Package: ddalpha
Type: Package
Title: Depth-Based Classification and Calculation of Data Depth
Version: 1.3.16
Date: 2024-09-29
Authors@R: c(person("Oleksii", "Pokotylo", role=c("aut", "cre"),
                                          email = "alexey.pokotylo@gmail.com"),
             person("Pavlo", "Mozharovskyi", role=c("aut"),
                                          email = "pavlo.mozharovskyi@telecom-paris.fr"),
             person("Rainer", "Dyckerhoff", role=c("aut"),
                                          email = "rainer.dyckerhoff@statistik.uni-koeln.de"),
             person("Stanislav", "Nagy", role=c("aut"),
                                          email = "nagy@karlin.mff.cuni.cz"))
Depends: R (>= 2.10), stats, utils, graphics, grDevices, MASS, class,
        robustbase, sfsmisc, geometry
Imports: Rcpp (>= 0.11.0)
LinkingTo: BH, Rcpp
Description: Contains procedures for depth-based supervised learning, which are entirely non-parametric, in particular the DDalpha-procedure (Lange, Mosler and Mozharovskyi, 2014 <doi:10.1007/s00362-012-0488-4>). The training data sample is transformed by a statistical depth function to a compact low-dimensional space, where the final classification is done. It also offers an extension to functional data and routines for calculating certain notions of statistical depth functions. 50 multivariate and 5 functional classification problems are included. (Pokotylo, Mozharovskyi and Dyckerhoff, 2019 <doi:10.18637/jss.v091.i05>).
License: GPL-2
NeedsCompilation: yes
Packaged: 2024-09-29 13:54:41 UTC; alexis
Author: Oleksii Pokotylo [aut, cre],
  Pavlo Mozharovskyi [aut],
  Rainer Dyckerhoff [aut],
  Stanislav Nagy [aut]
Maintainer: Oleksii Pokotylo <alexey.pokotylo@gmail.com>
Repository: CRAN
Date/Publication: 2024-09-30 22:50:02 UTC
Built: R 4.3.3; aarch64-apple-darwin20; 2025-01-24 18:03:04 UTC; unix
Archs: ddalpha.so.dSYM

Package: ddalpha
Type: Package
Title: Depth-Based Classification and Calculation of Data Depth
Version: 1.3.16
Date: 2024-09-29
Authors@R: c(person("Oleksii", "Pokotylo", role=c("aut", "cre"),
                                          email = "alexey.pokotylo@gmail.com"),
             person("Pavlo", "Mozharovskyi", role=c("aut"),
                                          email = "pavlo.mozharovskyi@telecom-paris.fr"),
             person("Rainer", "Dyckerhoff", role=c("aut"),
                                          email = "rainer.dyckerhoff@statistik.uni-koeln.de"),
             person("Stanislav", "Nagy", role=c("aut"),
                                          email = "nagy@karlin.mff.cuni.cz"))
Depends: R (>= 2.10), stats, utils, graphics, grDevices, MASS, class,
        robustbase, sfsmisc, geometry
Imports: Rcpp (>= 0.11.0)
LinkingTo: BH, Rcpp
Description: Contains procedures for depth-based supervised learning, which are entirely non-parametric, in particular the DDalpha-procedure (Lange, Mosler and Mozharovskyi, 2014 <doi:10.1007/s00362-012-0488-4>). The training data sample is transformed by a statistical depth function to a compact low-dimensional space, where the final classification is done. It also offers an extension to functional data and routines for calculating certain notions of statistical depth functions. 50 multivariate and 5 functional classification problems are included. (Pokotylo, Mozharovskyi and Dyckerhoff, 2019 <doi:10.18637/jss.v091.i05>).
License: GPL-2
NeedsCompilation: yes
Packaged: 2024-09-29 13:54:41 UTC; alexis
Author: Oleksii Pokotylo [aut, cre],
  Pavlo Mozharovskyi [aut],
  Rainer Dyckerhoff [aut],
  Stanislav Nagy [aut]
Maintainer: Oleksii Pokotylo <alexey.pokotylo@gmail.com>
Repository: CRAN
Date/Publication: 2024-09-30 22:50:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 02:25:29 UTC; unix
Archs: ddalpha.so.dSYM

export(ddalpha.train, ddalpha.classify,  
       ddalpha.test, ddalpha.getErrorRateCV, ddalpha.getErrorRatePart,
       ddalphaf.train, ddalphaf.classify, 
       ddalphaf.test, ddalphaf.getErrorRateCV, ddalphaf.getErrorRatePart,
       compclassf.train, compclassf.classify,
       depth., depth.space.,
       depth.zonoid, depth.halfspace, depth.Mahalanobis, depth.projection, depth.spatial, depth.simplicial, depth.simplicialVolume, depth.potential, depth.qhpeeling, depth.L2, depth.betaSkeleton,
       depth.space.zonoid, depth.space.halfspace, depth.space.Mahalanobis, depth.space.projection, depth.space.spatial, depth.space.simplicial, depth.space.simplicialVolume, depth.space.potential,
       is.in.convex, 
       dknn.train, dknn.classify, dknn.classify.trained,
       resetPar, depth.graph, draw.ddplot, depth.contours, depth.contours.ddalpha,
       getdata,
       dataf.geneexp, dataf.growth, dataf.medflies, dataf.tecator,
       dataf.sim.1.CFF07, dataf.sim.2.CFF07,
       plot.functional, lines.functional, points.functional,
       dataf,
       depthf.simplicialBand
       )
export(FKS, shape.fd.analysis, shape.fd.outliers, 
       depthf.,
       depthf.BD, depthf.ABD, depthf.fd1, depthf.fd2,
       depthf.HR, depthf.hM, depthf.hM2, depthf.RP1, depthf.RP2, derivatives.est, 
       L2metric, Cmetric, depth.sample, infimalRank, dataf.population, dataf.population2010,
       dataf2rawfd, rawfd2dataf
       )
useDynLib(ddalpha, .registration = TRUE)
S3method(predict, ddalpha)
S3method(predict, ddalphaf)
S3method(predict, compclassf)
S3method(plot, ddalpha)
S3method(plot, ddalphaf)
S3method(plot, functional)
S3method(lines, functional)
S3method(points, functional)
S3method(print, ddalpha)
S3method(print, ddalphaf)
S3method(summary, ddalpha)
S3method(summary, ddalphaf)
S3method(print, ddalpha.pattern)
S3method(print, ddalpha.alpha)
S3method(print, ddalpha.polynomial)
S3method(print, ddalpha.knnlm)
S3method(print, ddalpha.maxD)
import(stats)
import(utils)
import(graphics)
import(grDevices)
import(MASS)
import(class)
import(robustbase)
import(Rcpp)
import(geometry)

import(sfsmisc)

.depthf <- function(fname, funcargs, ...){
  f <- try(match.fun(fname), silent = T)
  if (is.function(f)){
    args = list(...)
    fcnArgs <- names(formals(f))
    fcnArgs <- unlist(fcnArgs, use.names=FALSE)
    keep <- intersect(names(args), fcnArgs)
    unused <- setdiff(names(args), fcnArgs)
    args <- args[keep]
    
    args <- c(args, funcargs)
    res <- do.call(fname, args=args)
    
    if(length(unused)>0)
      warning("Unused by '", fname, "' arguments: ", paste(unused, collapse = ', '))
    
    #res <- f(x, data, ...)
    
    return(res)
  } else {
    warning("There is no depth function ", fname)
  }
}

depthf. <- function(datafA, datafB, notion = c("ABD", "BD", "fd1", "fd2", "hM", "hM2", "HR", "RP1", "RP2"), ...){
  
  if(is.null(notion))
    stop("Parameter 'notion' must be set")
  t <- notion
  try(t <- match.arg(notion), silent = T)
  
  fname = paste0("depthf.", t)
  funcargs = list(datafA = datafA, datafB = datafB)

  return(.depthf(fname, funcargs, ...))
}

#depthf.space. <- function(dataf, cardinalities, notion = c("ABD", "BD", "fd1", "fd2", "hM", "hM2", "HR", "RP1", "RP2"), ...){
#  
#  if(is.null(notion))
#    stop("Parameter 'notion' must be set")
#  t <- notion
#  try(t <- match.arg(notion), silent = T)
#  
#  # try to find a depth
#  fname = paste0("depth.space.", t)
#  funcargs = list(cardinalities = cardinalities, data = data)
#  return(.depth(fname, funcargs, ...))
#}




# d =  depth(data$train, data$train, exact = T)

################################################################################
# File:             depth.contours.r
# Created by:       Oleksii Pokotylo
# First published:  03.07.2015
# Last revised:     13.11.2015
# 
# Visualization of the depth contours
################################################################################


gcolors = c("red", "blue", "green", "orange", "violet")

depth.contours.ddalpha <- function(ddalpha, main = "", xlab="", ylab = "", drawplot = T, frequency=100, levels = 10, drawsep = T, ...){
  
  if(class(ddalpha)!="ddalpha")
    stop("Not a 'ddalpha' classifier")
  
  if (ddalpha$dimension != 2)
  {
    warning ("The contours may be drawn only for 2 dimensional datasets")
    return(0)
  }
  
  if (ddalpha$needtransform == 1){
    data = ddalpha$patterns[[1]]$transformer(ddalpha$patterns[[1]]$points, inv = T)
    for (i in 1:length(ddalpha$patterns))
      data = rbind(data, ddalpha$patterns[[i]]$transformer(ddalpha$patterns[[i]]$points, inv = T))
  }else{
    data = ddalpha$patterns[[1]]$points
    for (i in 1:length(ddalpha$patterns))
      data = rbind(data, ddalpha$patterns[[i]]$points)
  }
  classes = rep(gcolors[1], ddalpha$patterns[[1]]$cardinality)
  for (i in 1:length(ddalpha$patterns))
    classes = c(classes, rep(gcolors[i], ddalpha$patterns[[i]]$cardinality))
  margins = c(min(data[,1]), max(data[,1]), min(data[,2]), max(data[,2]));  margins = margins + c(-0.1*(margins[2]-margins[1]), 0.1*(margins[2]-margins[1]), -0.1*(margins[4]-margins[3]), 0.1*(margins[4]-margins[3]))
  C = ncol(data)
  
  cr = 0
  
  if (drawplot)
    plot(data, col = classes, main = main, xlab=xlab, ylab = ylab, ...)
  if (!is.null(ddalpha$methodDepth))
  if (ddalpha$methodDepth == "Mahalanobis" && ddalpha$mahEstimate == "moment"){
    # Mahalanobis depth
    for (i in seq(length(ddalpha$patterns))){
      mahalanobisRegions(ddalpha$patterns[[i]]$points, levels, col = gcolors[i])
    }
  } else 
    if (ddalpha$methodDepth == "Mahalanobis" && ddalpha$mahEstimate == "MCD"){
      # Mahalanobis depth mit MCD
      for (i in seq(length(ddalpha$patterns))){
        mahalanobisMCDRegions(ddalpha$patterns[[i]]$points, ddalpha$mahParMcd, levels, col = gcolors[i])
      }
    } else 
#       if (ddalpha$methodDepth == "zonoid"){
#         if (require(WMTregions))
#           for (i in seq(length(ddalpha$patterns))){
#             wmtRegions(ddalpha$patterns[[i]]$points, depths = c(1:9/10), trtype = "zonoid", col = gcolors[i])
#           }
#         else{
#           contourRegions(ddalpha, margins = margins, depths = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1), frequency=frequency)      
#         }
#       } else 
#         if (ddalpha$methodDepth == "halfspace"){
#           library(depth)
#           for (i in seq(length(ddalpha$patterns))){
#             locationRegions(ddalpha$patterns[[i]]$points, depths = c(1:9/10), col = gcolors[i])
#           }    
#         }  else
          # no formal depth regions
        {
          cr = (contourRegions(ddalpha, margins = margins, depths = levels, frequency=frequency))
        }   
  
  lwd = 2
  if(drawsep){
    gx <- seq(min(ddalpha$raw[, 1]), max(ddalpha$raw[, 1]), length = frequency)
    gy <- seq(min(ddalpha$raw[, 2]), max(ddalpha$raw[, 2]), length = frequency)
    y <- as.matrix(expand.grid(gx, gy))   
    
    depthcontours = ddalpha.classify(ddalpha = ddalpha, objects = y)
    depthcontours = as.numeric(unlist(depthcontours))
    contour(gx, gy, matrix(depthcontours, nrow=length(gx), ncol = length(gy)), add = TRUE, levels = unique(depthcontours)+0.5, drawlabels = FALSE, col = "black", lwd = lwd)
  }
  
  invisible(cr)
}


depth.contours <- function(data, depth, main = "", xlab="", ylab = "", drawplot = T, frequency=100, levels = 10, col = "red", ...){
  
  if(!(is.matrix(data)||is.data.frame(data)))
    stop("Data is not a matrix or a data frame")
  
  if (ncol(data) != 2)
  {
    stop("The contours may be drawn only for 2 dimensional datasets")
  }
  
  if (drawplot){
    cc = tryCatchCapture(
      plot(data, col = col, main = main, xlab=xlab, ylab = ylab, ...)
      , err = F) # catch all warnings about unused params
    wrns = sort(unique(cc$warnings))
    unused = gsub("\"(.+)\".+", "\\1", wrns)
    if(length(unused)>0)
      warning("Unused by 'plot' arguments: ", paste(unused, collapse = ', '))
  }
  margins = c(min(data[,1]), max(data[,1]), min(data[,2]), max(data[,2]));  margins = margins + c(-0.1*(margins[2]-margins[1]), 0.1*(margins[2]-margins[1]), -0.1*(margins[4]-margins[3]), 0.1*(margins[4]-margins[3]))
  
  if (depth == "Mahalanobis"){
    # Mahalanobis depth
    mahalanobisRegions(data, levels, col = col, ...)
  } else 
#     if (depth == "zonoid"){
#       if (require(WMTregions))
#         wmtRegions(data, depths = c(1:9/10), trtype = "zonoid", col = col)
#       else{
#         contourRegionsData(data, depth = depth, margins = margins, depths = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1), frequency=frequency, col = col, ...)      
#       }
#     } else 
#       if (depth == "halfspace"){
#         library(depth)
#         locationRegions(data, depths = c(1:9/10), col = col)  
#       }  else 
        # no formal depth regions
      {
        invisible(contourRegionsData(data, depth = depth, margins = margins, depths = levels, frequency=frequency, col = col, ...))
      }   
  
  invisible(0)
}


# wmtRegions <- function(x, depths = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9), trtype = "zonoid", col = "black"){
#   fname <- "Cloud.dat"
#   #trname <- "TRegion_vertices.dat"
#   trname <- "tmp_vrtheap.dat"
#   fdir = getwd()
#   for (i in 1:length(depths)){
#     ffullname = as.character(paste(fdir, "/", fname, sep = ""))
#     write(trtype, ffullname)
#     write(depths[i], ffullname, append = TRUE)
#     write(ncol(x), ffullname, append = TRUE)
#     write(nrow(x), ffullname, append = TRUE)
#     write(array(t(x), dim = c(nrow(x), ncol(x))), ffullname, ncolumns = ncol(x), append = TRUE)
#     WMTR()
#     cat("Calculated for i = ", i, "\n")
#     wmtreg <- read.table(as.character(paste(fdir, "/", trname, sep = "")), sep=" ")
#     unlink(as.character(paste(fdir, "/", trname, sep = "")))
#     wmtreg <- as.matrix(wmtreg)
#     numribs <- nrow(wmtreg)/2
#     for (i in 1:numribs){
#       lines(rbind(wmtreg[i*2 - 1,], wmtreg[i*2,]), col = col)
#     }    
#   }
# }
# 
# locationRegions <- function(x, depths = NULL, col = "black"){
#   numLearn <- nrow(x)
#   vert = isodepth(x, c(2:(numLearn - 1)), output = T, dpth = as.integer(depths*numLearn/2))
#   for (verticles in vert){
#     if(is.null(verticles) || nrow(verticles)<1) next
#     verticles = rbind(verticles, verticles[1,])
#     lines(verticles, col = col)
#   }
# }

mahalanobisRegions <- function(x, depths = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1), col = "black", ...){
  if(is.na(depths) || is.null(depths))
    depths = 10
  if(length(depths) == 1 && depths > 1)
    depths = seq(0.0001, 1, length.out = depths)
  
  # PM (2018-06-22)
  if (("mah.estimate" %in% names(list(...))) && 
      toupper(list(...)[["mah.estimate"]]) == "MCD"){
    if (("mah.parMcd" %in% toupper(names(list(...)))) &&
        (list(...)[["mah.parMcd"]] >= 0.5) &&
        (list(...)[["mah.parMcd"]] <= 1)){
      parMcd <- list(...)[["mah.parMcd"]]
    }else{
      parMcd <- 0.75
    }
    est <- covMcd(x, alpha = parMcd)
    mu <- est$center
    sigma.inv <- solve(est$cov)
  }else{
    c <- c(0,0)
    for (i in 1:nrow(x)){
      c <- c + x[i,]
    }
    mu <- c / nrow(x)
    sigma.inv <- solve(cov(x))
  }
  for (i in 1:length(depths)){
    ellipsem(mu = mu, amat = sigma.inv, c2 = 1/depths[i] - 1, showcentre = F, col = col)
  }
}

mahalanobisMCDRegions <- function(x, alpha = 1/2, depths = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1), col = "black"){
  
  if(is.na(depths) || is.null(depths))
    depths = 10
  if(length(depths) == 1 && depths > 1)
    depths = seq(0.0001, 1, length.out = depths)
  
  #  library(robustbase)
  estimate <- covMcd(x, alpha = alpha)
  mu <- estimate$center
  sigma.inv <- solve(estimate$cov)
  for (i in 1:length(depths)){
    ellipsem(mu = mu, amat = sigma.inv, c2 = 1/depths[i] - 1, showcentre = F, col = col)
  }
}

contourRegions <- function(ddalpha, margins = NULL, depths, frequency=100){
  
  if (is.null(margins)){
    if (ddalpha$needtransform == 1)
      data = rbind(ddalpha$patterns[[1]]$transformer(ddalpha$patterns[[1]]$points, inv = T), ddalpha$patterns[[2]]$transformer(ddalpha$patterns[[2]]$points, inv = T))
    else
      data = rbind(ddalpha$patterns[[1]]$points, ddalpha$patterns[[2]]$points)
    margins = c(min(data[,1]), max(data[,1]), min(data[,2]), max(data[,2]));  margins = margins + c(-0.1*(margins[2]-margins[1]), 0.1*(margins[2]-margins[1]), -0.1*(margins[4]-margins[3]), 0.1*(margins[4]-margins[3]))
  }
  
  gx <- seq(margins[1], margins[2], length=frequency)
  gy <- seq(margins[3], margins[4], length=frequency)
  y <- as.matrix(expand.grid(gx, gy))    
  
  depthcontours <- .ddalpha.count.depths(ddalpha, y)  
  
  if(is.na(depths) || is.null(depths))
    depths = 10
  
  if(length(depths) == 1 && depths > 1)
    depths = seq(0, max(depthcontours), length.out = depths)
  
  for (i in seq(length(ddalpha$patterns))){
    contour(gx, gy, matrix(depthcontours[,i], nrow=length(gx), ncol=length(gy)), add=TRUE, levels=depths*max(depthcontours), drawlabels=FALSE, col = gcolors[i])
  }
  
  invisible(list(gx = gx, gy = gy, depthcontours = depthcontours))
}

contourRegionsData <- function(data, depth, margins = NULL, depths = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1), frequency=100, col = "red", ...){
  
  if (is.null(depth) || depth == "none"){
    invisible (0);}
  else{
  # df = switch(depth,
  #             "zonoid" = function(x, data, ...) depth.zonoid(x, data) ,
  #             "halfspace" = depth.halfspace,
  #             "simplicialVolume" = depth.simplicialVolume,
  #             "simplicial" = depth.simplicial,
  #             "Mahalanobis" = function(x, X) (.Mahalanobis_depth(x, colMeans(X), solve(cov(X)))),
  #             "projection" = depth.projection,
  #             "spatial" = depth.spatial
  # )
  df = function(x, data, ...) 
        ddalpha::depth.(x, data, notion = depth, ...)
  
  if (is.null(margins)){
    margins = c(min(data[,1]), max(data[,1]), min(data[,2]), max(data[,2]));  margins = margins + c(-0.1*(margins[2]-margins[1]), 0.1*(margins[2]-margins[1]), -0.1*(margins[4]-margins[3]), 0.1*(margins[4]-margins[3]))
  }
  
  gx <- seq(margins[1], margins[2], length=frequency)
  gy <- seq(margins[3], margins[4], length=frequency)
  y <- as.matrix(expand.grid(gx, gy))    
  
  depthcontours <- df(y,data,...)  
  
  if(is.na(depths) || is.null(depths))
    depths = 10
  if(length(depths) == 1 && depths > 1)
    depths = seq(0.0001, max(depthcontours), length.out = depths)
  
  contour(gx, gy, matrix(depthcontours, nrow=length(gx), ncol=length(gy)), add=TRUE, levels=depths, drawlabels=FALSE, col = col)
  
  invisible(list(gx = gx, gy = gy, depthcontours = depthcontours))
  }
}


ellipsem <- function (mu, amat, c2, npoints = 100, showcentre = T, col, ...){
  if (all(dim(amat) == c(2, 2))) {
    eamat <- eigen(amat)
    hlen <- sqrt(c2/eamat$val)
    theta <- angle(eamat$vec[1, 1], eamat$vec[2, 1])
    ellipse(hlen[1], hlen[2], theta, mu[1], mu[2], npoints = npoints, col = col, ...)
    if (showcentre)
      points(mu[1], mu[2], pch = 3)
  }
  invisible()
}

ellipse <- function (hlaxa = 1, hlaxb = 1, theta = 0, xc = 0, yc = 0, newplot = F, npoints = 100, ...){
  a <- seq(0, 2 * pi, length = npoints + 1)
  x <- hlaxa * cos(a)
  y <- hlaxb * sin(a)
  alpha <- angle(x, y)
  rad <- sqrt(x^2 + y^2)
  xp <- rad * cos(alpha + theta) + as.double(xc)
  yp <- rad * sin(alpha + theta) + as.double(yc)
  if (newplot)
    plot(xp, yp, type = "l", ...)
  else lines(xp, yp, ...)
  invisible()
}

angle <- function (x, y)
{
  angle2 <- function(xy) {
    x <- xy[1]
    y <- xy[2]
    if (x > 0) {
      atan(y/x)
    }
    else {
      if (x < 0 & y != 0) {
        atan(y/x) + sign(y) * pi
      }
      else {
        if (x < 0 & y == 0) {
          pi
        }
        else {
          if (y != 0) {
            (sign(y) * pi)/2
          }
          else {
            NA
          }
        }
      }
    }
  }
  apply(cbind(x, y), 1, angle2)
}

is.numeric_data.frame <- function(x){
  if (is.data.frame(x) && all(sapply(x,base::is.numeric)))
    return (T)
  return (F)
}

is.numeric <- function(x){
  if (base::is.numeric(x))
    return (T)
  if (is.data.frame(x) && all(sapply(x,base::is.numeric)))
    return (T)
  return (F)
}

resetPar <- function() {
  dev.new()
  op <- par(no.readonly = TRUE)
  dev.off()
  op
}

is.sorted <- function(x)  {
  return(!is.unsorted(x))
}

tryCatchCapture <- function(expr, warn = T, err = T) {
  val <- NULL
  myWarnings <- NULL
  wHandler <- function(w) {
    myWarnings <<- c(myWarnings, w$message)
    invokeRestart("muffleWarning")
  }
  myError <- NULL
  eHandler <- function(e) {
    myError <<- e$message
    NULL
  }
  if(warn && err){
    val <- tryCatch(withCallingHandlers(expr, warning = wHandler), error = eHandler)
    return(list(value = val, warnings = myWarnings, error=myError))
  }
  if(warn){
    val <- tryCatch(withCallingHandlers(expr, warning = wHandler))
    return(list(value = val, warnings = myWarnings))
  }
  if(err){
    val <- tryCatch(expr, error = eHandler)
    return(list(value = val, error=myError))
  }
  val <- expr
  return(list(value = val))
}

dataf <- function(fdatas, labels) 
{
  .split_data_matrix <-function(fdata) {
    funcs = split(fdata$data, row(fdata$data))
    dataf = lapply(funcs, function(func) list(args = fdata$argvals, vals = func))
  }
  
  if (inherits(fdatas, "fdata")) {
    if (nrow(fdatas$data) != length(labels))
      stop("the length of 'labels' must correspond to the number of functions in 'fdatas'")
    if (!is.null(fdatas$fdata2d) && fdatas$fdata2d)
      stop("fdata2d = TRUE is not supported")
    
    dataf.labels = as.list(labels)
    dataf.dataf = .split_data_matrix(fdatas)
    res = list(
      dataf = dataf.dataf,
      labels = dataf.labels,
      name = fdatas$names$main,
      args = fdatas$names$xlab,
      vals = fdatas$names$ylab
    )
    class(res) <- "functional"
    return (res)
  }
  
  
  if (length(fdatas) != length(labels))
    stop("'fdatas' and 'labels' must be vectors of the same length")

  dataf.dataf = list()
  dataf.labels = c()
  for (i in range(1:length(fdatas))) {
    fdata = fdatas[[i]]
    lab = labels[[i]]
    
    if (class(fdata) != "fdata")
      stop("elements of 'fdatas' must be of the 'fdata' class")
    if (!is.null(fdata$fdata2d) && fdata$fdata2d)
      stop("fdata2d = TRUE is not supported")
  
    dataf.labels = c(dataf.labels, rep(lab, nrow(fdata$data)))
    
    dataf.dataf = c(dataf.dataf, .split_data_matrix(fdata))
  }  
  
  res = list(
    dataf = as.list(dataf.dataf),
    labels = dataf.labels,
    # just take the names from the first data set
    name = fdatas[[1]]$names$main,
    args = fdatas[[1]]$names$xlab,
    vals = fdatas[[1]]$names$ylab
  )
  class(res) <- "functional"
  return (res)
  
}

mah.moment <- function(x){
  mu <- colMeans(x)
  
  scale.eig <- eigen(cov(x))
  B <- scale.eig$vectors %*% diag(sqrt(scale.eig$values)) 
  B_inv <- solve(B)
  return (list(mu = as.numeric(mu), b = B_inv, s = cov(x)))
}

mah.mcd <- function(x, alpha = 1/2){
  #library(robustbase)
  estimate <- covMcd(x, alpha = alpha)
  mu <- estimate$center
  
  scale.eig <- eigen(estimate$cov)
  B <- scale.eig$vectors %*% diag(sqrt(scale.eig$values)) 
  B_inv <- solve(B)
  return (list(mu = as.numeric(mu), b = B_inv, s = estimate$cov))
}

mah.transform <- function (x, mu, B_inv, inv = F)  {
  if (inv)
    return (t(solve(B_inv) %*% (t(x))+mu))
  return (t(B_inv %*% (t(x)-mu)))
}

mah.transform.back <- function (x, mu, B_inv)  {
  return (t(solve(B_inv) %*% (t(x))+mu))
}

MahMomentTransformer <- function(mu, b){
  f <- function(points, inv = F){
    return(mah.transform(points, mu, b, inv))
  }
  environment(f) <- new.env()
  environment(f)$mu = mu
  environment(f)$b = b
  return (f)
}

# mahalanobisRegionsX <- function(x, depths = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1), col = "black"){
#   c <- c(0,0)
#   for (i in 1:nrow(x)){
#     c <- c + x[i,]
#   }
#   mu <- c / nrow(x)
#   
#   mahalanobisRegions(mu, cov(x), col)
# }
# 
# mahalanobisRegions <- function(mu, s, depths = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1), col = "black"){
#   sigma.inv <- solve(s)
#   for (i in 1:length(depths)){
#     ellipsem(mu = mu, amat = sigma.inv, c2 = 1/depths[i] - 1, showcentre = F, col = col)
#   }
# }

depth.Mahalanobis <- function(x, data, mah.estimate = "moment", mah.parMcd = 0.75){
  if (mah.estimate == "moment") s = solve(cov(data))
  else if (mah.estimate == "MCD") s = solve(covMcd(data, alpha = mah.parMcd)$cov)
  else stop ("Wrong parameter 'estimate'")
  
  if (is.matrix(x) || is.data.frame(x)) {
    nx = nrow(x)
    if (ncol(x) != ncol(data))
      stop("Wrong dimension of x")
  }
  else if (is.vector(x)) {
    nx = 1
    if (length(x) != ncol(data))
      stop("Wrong dimension of x")
  } else stop("Wrong type of x")
  
#  if (mah.estimate == "MCD"){     # MCD in c++ is much slower
    depths <- .Mahalanobis_depth(x, center = colMeans(data), sigma = s)
    return (depths)
#  }
  
  points <- as.vector(t(data))
  objects <- as.vector(t(x))
  depths <- .C("MahalanobisDepth", 
           as.double(points), 
           as.double(objects), 
           as.integer(nrow(data)), 
           as.integer(nx), 
           as.integer(ncol(data)), 
           as.double(mah.parMcd),
           depths=double(nx))$depths
  
  return (depths)
}

depth.space.Mahalanobis <- function(data, cardinalities, mah.estimate = "moment", mah.parMcd = 0.75){
  if (is.data.frame(data))
    data <- data.matrix(data)
  if (!is.numeric(data)
      || !is.matrix(data) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.vector(cardinalities, mode = "numeric") 
      || is.na(min(cardinalities)) 
      || sum(.is.wholenumber(cardinalities)) != length(cardinalities) 
      || min(cardinalities) <= 0 
      || sum(cardinalities) != nrow(data)){
    stop("Argument \"cardinalities\" should be a vector of cardinalities of the classes in \"data\" ")
  }
  if (sum(cardinalities < ncol(data) + 1) != 0){
    stop("Not in all classes sufficiently enough objetcs")
  }
  
  depth.space <- NULL
  for (i in 1:length(cardinalities)){
    pattern <- data[(1 + sum(cardinalities[0:(i - 1)])):sum(cardinalities[1:i]),]
    pattern.depths <- depth.Mahalanobis (data, pattern, mah.estimate, mah.parMcd)
    depth.space <- cbind(depth.space, pattern.depths, deparse.level = 0)
  }
  
  return (depth.space)
}

################################################################################
# File:             depth.projection.r
# Created by:       Pavlo Mozharovskyi
# First published:  28.02.2013
# Last revised:     13.11.2015
# 
# Computation of the projection data depth.
################################################################################
# Conrains R-codes written by Subhajit Dutta,
# taken from http://www.isical.ac.in/~tijahbus/GAM/r_wilcox.txt.
################################################################################

depth.projection <- function(x, data, method = "random", num.directions = 1000, seed = 0){
  if (seed!=0) set.seed(seed)
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.matrix(x) 
      && is.vector(x)){
    x <- matrix(x, nrow=1)
  }
  if (method == "random"){
    dt <- as.vector(t(data))
    z <- as.vector(t(x))
    m <- nrow(x)
    d <- ncol(data)
    n <- nrow(data)
    q <- 1
    k <- num.directions
    newDirs <- 1
    rez <- .C("ProjectionDepth", as.double(dt), 
              as.double(z), 
              as.integer(m), 
              as.integer(d), 
              as.integer(n), 
              as.integer(q), 
              dirs=double(k*d), 
              prjs=double(k*n), 
              as.integer(k), 
              as.integer(1), 
              as.integer(seed),
              dps=double(m*q))
    return (rez$dps)
  }
  if (method == "linearize"){
    depths <- .zdepth(data, x)
    return (1/(1 + depths))
  }
}

depth.space.projection <- function(data, cardinalities, method = "random", num.directions = 1000, seed = 0){
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.vector(cardinalities, mode = "numeric") 
      || is.na(min(cardinalities)) 
      || sum(.is.wholenumber(cardinalities)) != length(cardinalities) 
      || min(cardinalities) <= 0 
      || sum(cardinalities) != nrow(data)){
    stop("Argument \"cardinalities\" should be a vector of cardinalities of the classes in \"data\" ")
  }
  if (sum(cardinalities < ncol(data) + 1) != 0){
    stop("Not in all classes sufficiently enough objetcs")
  }
  
  depth.space <- NULL
  for (i in 1:length(cardinalities)){
    pattern <- data[(1 + sum(cardinalities[0:(i - 1)])):sum(cardinalities[1:i]),]
    pattern.depths <- depth.projection (data, pattern, method, num.directions, seed)
    depth.space <- cbind(depth.space, pattern.depths, deparse.level = 0)
  }
  
  return (depth.space)
}


################################################################################
# R-codes of this function written by Subhajit Dutta,
# taken from http://www.isical.ac.in/~tijahbus/GAM/r_wilcox.txt.
################################################################################
.zdepth<-function(m,pts=m,zloc=median,zscale=mad){
  # 
  # Compute depth of points as in Zuo, Annals, 2003
  #
  if(!is.matrix(m))stop("argument m should be a matrix")
  if(!is.matrix(pts))stop("argument pts should be a matrix")
  if(ncol(m)!=ncol(pts))stop("Number of columns for m and pts are not equal")
  np<-ncol(m)
  val<-NA
  for(i in 1:nrow(pts)){
    pval<-pts[i,]
    START<-rep(1,np)/sqrt(np)
    temp<-.nelderv2(m,np,FN=.zdepth.sub,START=START,zloc=zloc,zscale=zscale,pts=pval)
    temp<-temp/sqrt(sum(temp^2))
    y<-t(t(m)*temp)
    y<-apply(y,1,sum)
    ppro<-sum(pval*temp)
    val[i]<-abs(ppro-zloc(y))/zscale(y)
  }
  val
}

################################################################################
# R-codes of this function written by Subhajit Dutta,
# taken from http://www.isical.ac.in/~tijahbus/GAM/r_wilcox.txt.
################################################################################
.zdepth.sub<-function(x,theta,zloc=median,zscale=mad,pts=NA){
  theta<-theta/sqrt(sum(theta^2))
  temp<-t(t(x)*theta)
  ppro<-sum(t(t(pts)*theta))
  yhat<-apply(temp,1,sum)
  val<-0-abs(ppro-zloc(yhat))/zscale(yhat)
  val
}

################################################################################
# R-codes of this function written by Subhajit Dutta,
# taken from http://www.isical.ac.in/~tijahbus/GAM/r_wilcox.txt.
################################################################################
.nelderv2<-function(x,N,FN,START=c(rep(1,N)),STEP=c(rep(1,N)),
                   XMIN=c(rep(0,N)),XSEC=c(rep(0,N)),...){
  #     NELDER-MEAD method for minimzing a function
  #
  #     TAKEN FROM OLSSON, J QUALITY TECHNOLOGY, 1974, 6, 56.
  #
  #     x= n by p matrix containing data; it is used by 
  #        function to be minimized.
  #     N= number of parameters 
  #
  #     FN=the function to be minimized
  #     FORM: FN(x,theta), theta is vector containing
  #     values for N parameters.
  #
  #     START = starting values.
  #     STEP=initial step.
  #     This function returns the N values for theta that minimize FN
  #
  ICOUNT<-500
  REQMIN<-.0000001
  NN<-N+1
  P<-matrix(NA,nrow=N,ncol=NN)
  P[,NN]<-START
  PBAR<-NA
  RCOEFF<-1
  ECOEFF<-2
  CCOEFF<-.5
  KCOUNT<-ICOUNT
  ICOUNT<-0
  DABIT<-2.04067e-35
  BIGNUM<-1.e38
  KONVGE<-5
  XN<-N
  DN<-N
  Y<-rep(0,NN)
  Y[NN]<-FN(x,START,...)
  ICOUNT<-ICOUNT+1
  for(J in 1:N){
    DCHK<-START[J]
    START[J]<-DCHK+STEP[J]
    for(I in 1:N){
      P[I,J]<-START[I]
    }
    Y[J]<-FN(x,START,...)
    ICOUNT<-ICOUNT+1
    START[J]<-DCHK
  }
  I1000<-T
  while(I1000){
    YLO<-Y[1]
    YNEWLO<-YLO
    ILO<-1
    IHI<-1
    for(I in 2:NN){
      if(Y[I] <  YLO){
        YLO<-Y[I]
        ILO<-I}
      if(Y[I] > YNEWLO){
        YNEWLO<-Y[I]
        IHI<-I}
    }
    DCHK<-(YNEWLO+DABIT)/(YLO+DABIT)-1
    if(abs(DCHK) < REQMIN){
      I1000<-F
      next
    }
    KONVGE<-KONVGE-1
    if(KONVGE == 0){
      KONVGE<-5
      for(I in 1:N){
        COORD1<-P[I,1]
        COORD2<-COORD1
        for(J in 2:NN){
          if(P[I,J] < COORD1)COORD1<-P[I,J]
          if(P[I,J] > COORD2)COORD2<-P[I,J]
        }     # 2010 CONTINUE
        DCHK<-(COORD2+DABIT)/(COORD1+DABIT)-1
        if(abs(DCHK) > REQMIN)break  
      } 
    }       
    if(ICOUNT >= KCOUNT){
      I1000<-F
      next
    }
    for(I in 1:N){
      Z<-0.0
      Z<-sum(P[I,1:NN]) # 6
      Z<-Z-P[I,IHI]
      PBAR[I]<-Z/DN
    }
    PSTAR<-(1.+RCOEFF)*PBAR-RCOEFF*P[,IHI]
    YSTAR<-FN(x,PSTAR,...)
    ICOUNT<-ICOUNT+1
    if(YSTAR < YLO && ICOUNT >= KCOUNT){
      P[,IHI]<-PSTAR
      Y[IHI]<-YSTAR
      next 
    }
    IFLAG<-T   
    if(YSTAR < YLO){
      P2STAR<-ECOEFF*PSTAR+(1-ECOEFF)*PBAR
      Y2STAR<-FN(x,P2STAR,...)
      ICOUNT<-ICOUNT+1
      if(Y2STAR >= YSTAR){ 
        P[,IHI]<-PSTAR
        Y[IHI]<-YSTAR
        next #In essence, go to 19 which goes to 1000
      }
      IFLAG<-T
      while(YSTAR < Y[IHI]){
        P[,IHI]<-P2STAR
        Y[IHI]<-Y2STAR
        IFLAG<-F
        break  
        L<-sum(Y[1:NN] > YSTAR)
        if(L > 1){
          P[,IHI]<-PSTAR
          Y[IHI]<-YSTAR
          IFLAG<-T
          break 
        }
        if(L > 1)break # go to 19
        if(L != 0){
          P[1:N,IHI]<-PSTAR[1:N]
          Y[IHI]<-YSTAR
        }
        I1000<-F
        break
        if(ICOUNT >= KCOUNT){
          I1000<-F
          next
        }
        P2STAR[1:N]<-CCOEFF*P[1:N,IHI]+(1-CCOEFF)*PBAR[1:N]
        Y2STAR<-FN(x,P2STAR,...)
        ICOUNT<-ICOUNT+1
      }   # END WHILE 
    }
    if(IFLAG){
      for(J in 1:NN){
        P[,J]=(P[,J]+P[,ILO])*.5
        XMIN<-P[,J]
        Y[J]<-FN(x,XMIN,...)
      }
      ICOUNT<-ICOUNT+NN
      if(ICOUNT < KCOUNT)next
      I1000<-F
      next
    }
    P[1:N,IHI]<-PSTAR[1:N]
    Y[IHI]<-YSTAR
  } 
  for(J in 1:NN){
    XMIN[1:N]<-P[1:N,J]
  }
  Y[J]<-FN(x,XMIN,...)
  YNEWLO<-BIGNUM
  for(J in 1:NN){
    if (Y[J] < YNEWLO){
      YNEWLO<-Y[J]
      IBEST<-J
    }}
  Y[IBEST]<-BIGNUM
  YSEC<-BIGNUM
  for(J in 1:NN){
    if(Y[J] < YSEC){
      YSEC<-Y[J]
      ISEC<-J
    }}
  XMIN[1:N]<-P[1:N,IBEST]
  XSEC[1:N]<-P[1:N,ISEC]
  XMIN
}

################################################################################
# File: depth.graph.r
# Created by:       Oleksii Pokotylo
# First published:  01.10.2014
# Last revised:     01.10.2014
# 
# Builds the data depth surfaces for 2-dimensional data  
################################################################################

depth.graph <- function (data, depth_f = c("halfspace", "Mahalanobis", "projection", "simplicial", "simplicialVolume", "spatial", "zonoid", "none"), apoint = NULL
                           , main = depth_f
                           , xlim = c(min(data[,1]), max(data[,1])), ylim = c(min(data[,2]), max(data[,2])), zlim = c(0,max(z))
                           , xnum = 250, ynum = 250
                           , theta=15, phi=60, bold = F, ...){
  
  x1 <- seq(xlim[1], xlim[2], length = xnum)
  x2 <- seq(ylim[1], ylim[2], length = ynum)
  x1.step <- (x1[2]-x1[1])
  x2.step <- (x2[2]-x2[1])
  all.points <- as.matrix(expand.grid(x1, x2))
  all.depths <- rep(0, nrow(all.points))
  #library(depth)
  df = depth_f
  if (!is.function(depth_f)){
    depth_f = match.arg (depth_f)            
    df = switch(depth_f,
                "none" = function(x, X,...) (0),
                "zonoid" = depth.zonoid,
                "halfspace" = depth.halfspace,
                "simplicialVolume" = depth.simplicialVolume,
                "simplicial" = depth.simplicial,
                "Mahalanobis" = function(x, X,...) (.Mahalanobis_depth(x, colMeans(X), solve(cov(X)))),
                "projection" = depth.projection,
                "spatial" = depth.spatial
    )
    if (depth_f == "none") zlim = c(0,1)
  }
  
  all.depths = df(all.points, data[,1:2], ...)
  z <- matrix(all.depths, ncol=ynum, nrow=xnum, byrow=FALSE)
  
  z.red <- as.integer((data[,1]-x1[1])/x1.step+1) + as.integer((data[,2]-x2[1])/x2.step+1)*(xnum-1)
  if (bold)
    z.red <- c(z.red, 
               as.integer((data[,1]-x1[1])/x1.step+2) + as.integer((data[,2]-x2[1])/x2.step+1)*(xnum-1),
               as.integer((data[,1]-x1[1])/x1.step+1) + as.integer((data[,2]-x2[1])/x2.step+2)*(xnum-1),
               as.integer((data[,1]-x1[1])/x1.step+0) + as.integer((data[,2]-x2[1])/x2.step+1)*(xnum-1),
               as.integer((data[,1]-x1[1])/x1.step+1) + as.integer((data[,2]-x2[1])/x2.step+0)*(xnum-1)
               )
			   if (!is.null(apoint) && is.numeric(apoint) && length(apoint) == 2){
			       z.black <- as.integer((apoint[1]-x1[1])/x1.step+1) + as.integer((apoint[2]-x2[1])/x2.step+1)*(xnum-1)
			       if (bold){
			         z.black <- c(z.black, 
			                    as.integer((apoint[1]-x1[1])/x1.step+2) + as.integer((apoint[2]-x2[1])/x2.step+1)*(xnum-1),
			                    as.integer((apoint[1]-x1[1])/x1.step+1) + as.integer((apoint[2]-x2[1])/x2.step+2)*(xnum-1),
			                    as.integer((apoint[1]-x1[1])/x1.step+0) + as.integer((apoint[2]-x2[1])/x2.step+1)*(xnum-1),
			                    as.integer((apoint[1]-x1[1])/x1.step+1) + as.integer((apoint[2]-x2[1])/x2.step+0)*(xnum-1)
								)
			       }
			     }else{
			       z.black <- NA
			     }
  
  zfacet <- z[-1, -1] + z[-1, -ynum] + z[-xnum, -1] + z[-xnum, -ynum]
  z.indices.zero <- which(zfacet == 0)
  cols <- rep("gray", (xnum-1)*(ynum-1))
  cols <- replace(cols, z.indices.zero, ifelse (depth_f == "none", NA,"lightblue"))
  cols <- replace(cols, z.red, "red")
  cols <- replace(cols, z.black, "black")
  
  par(bg = "white")
  persp(x1, x2, z, xlim=xlim, ylim=ylim, zlim=zlim, r = 10, theta=theta, phi=phi, 
        col=cols, main = main,
        ltheta=55, shade=0.55, ticktype="detailed", 
        xlab="x", ylab="y", zlab="D(x|X)", border=NA, box=FALSE, ...)
}

ddalphaf.test <- function(learn, learnlabels, test, testlabels, disc.type = c("LS", "comp"), ...){
  ops <- options(warn = -1)
  on.exit(options(ops))
  
  disc.type <- match.arg(disc.type)
  
  ftrain = switch(disc.type,
              "LS" = ddalphaf.train,
              "comp" = compclassf.train
  )
  fclassify = switch(disc.type,
                  "LS" = ddalphaf.classify,
                  "comp" = compclassf.classify
  )
  
  
  tryCatch({
    time <- system.time(
      ddalpha <- ftrain(learn, learnlabels, ...)
    )
    cc = fclassify(objectsf = test,ddalphaf = ddalpha)
    if (is.numeric(testlabels[[1]])){
      if(is.factor(cc[[1]]) || is.character(cc[[1]])){
        cc <- unlist(lapply(cc, as.character))
        cc[cc == "Ignored"] <- NA
      }
      equal = (cc == testlabels)
    } else {
      cc <- unlist(lapply(cc, as.character)) 
      equal = (cc == as.character(testlabels))      
    }
    
    if(!(T %in% equal) && !(F %in% equal)) 
    { return(NA)}
    error = sum(!equal,na.rm = T)/(sum(!equal,na.rm = T)+sum(equal,na.rm = T))
    return(list(error = error, correct = sum(equal,na.rm = T), incorrect = sum(!equal,na.rm = T), 
                total = length(cc)-sum(is.na(equal)), ignored = sum(is.na(equal)), n = length(cc),
                time = time[1])) 
  }
  #  tryCatch({}
  , error = function(e) {
    print ("ERROR T")
    print (e)
  }, finally = {          
  })
  return (NA)
}


ddalphaf.getErrorRateCV <- function(dataf, labels, numchunks = 10, disc.type = c("LS", "comp"),  ...){
  n = length(dataf)
  numchunks = min(n, numchunks)
  chunksize = ceiling(n/numchunks)
  
  sample = seq(from = 1, by = numchunks, length.out = chunksize)   
  
  errors = 0
  total = 0
  times = c()
  
  for (i in 1:numchunks){
    sample = sample[sample<=n]
    learn = dataf[-sample]
    test  = dataf[sample]
    learnlabels = labels[-sample]
    testlabels  = labels[sample]
    
    el = ddalphaf.test(learn, learnlabels, test, testlabels, disc.type, ...)
    if(is.list(el)){
      errors = errors + el$incorrect
      total = total + el$total
      times = c(times,el$time)
    }
    
    sample = sample+1
  }
  
  return (list(errors = errors/total, time = mean(times), time_sd = sd(times)))  
}


ddalphaf.getErrorRatePart <- function(dataf, labels, size = 0.3, times = 10, disc.type = c("LS", "comp"),  ...){
  
  if (!is.numeric(size) || size <=0 || size >= length(dataf)) stop("Wrong size of excluded sequences")
  
  if(size < 1)
    size = max(1, size*length(dataf)) # at least 1 point
  
  size = as.integer(size)
  
  indexes = 1:length(dataf)
  
  errors = c()
  total = 0
  time = c()
  
  for (i in 1:times){
    samp = sample(indexes, size)
    learn = dataf[-samp]
    test  = dataf[samp]
    learnlabels = labels[-samp]
    testlabels  = labels[samp]
    
    el = ddalphaf.test(learn, learnlabels, test, testlabels, disc.type, ...)
    if(is.list(el)){
      errors = c(errors,el$incorrect/el$total)
      time = c(time,el$time)
    }
  }
  
  return (list(errors = mean(errors), errors_sd = sd(errors), errors_vec = errors, time = mean(time), time_sd = sd(time)))  
}

################################################################################
# File:             dknn.r
# Created by:       Oleksii Pokotylo
# First published:  
# Last revised:     
# 
# Contains the realization of the Depth-based KNN classifier of Paindaveine and Van Bever (2015).
################################################################################

dknn.train <- function(data, kMax = -1, depth = "halfspace", seed = 0){

  dpth = switch (depth,
                 "halfspace" = 1,
                 "Mahalanobis" = 2,
                 "simplicial" = 3, 0)
  if(dpth == 0)
    stop("Wrong depth: ", depth)
  
  n = nrow(data)
  chunkNumber = n #10
  dimension = ncol(data)-1
  labs <- data[,dimension+1]
  labels <- integer(length(labs))
  
  uniquelab = unique(labs)
  cardinalities = unlist(lapply(uniquelab, function(l)sum(labs == l)))
  uniquelab = uniquelab[order(cardinalities)]
  cardinalities = cardinalities[order(cardinalities)]
  for (i in seq_along(uniquelab)){
    labels[labs == uniquelab[i]] <- i
  }
  if(length(uniquelab)<2)
    stop("There is only one class")
  
  if (!is.numeric(kMax) 
      || is.na(kMax) 
      || length(kMax) != 1 
      || !.is.wholenumber(kMax) 
      || !(kMax >= 1 
           && kMax <= (cardinalities[1]+rev(cardinalities)[1]) 
           || kMax == -1)){
    warning("In treatment number ", i, ": Argument \"k\" not specified 
            correctly. Defaults are applied")
    kMax <- - 1
  }
  if(kMax == -1)    kMax <- n/2

  kMax <- min(kMax, n - 1)
  kMax <- max(kMax, 2)
  
  points <- as.vector(t(data[,1:dimension]))
  k <- as.integer(.C("DKnnLearnCv", 
          as.double(points), 
          as.integer(labels),
          as.integer(n), 
          as.integer(dimension), 
          as.integer(kMax), 
          as.integer(dpth),
          k = integer(1),
          as.integer(chunkNumber),
          as.integer(seed))$k)
  
  dknn <- list(data = data, 
               n = n, 
               dimension = dimension,
               labels = labels,
               uniquelab = uniquelab,
               methodSeparator = "Dknn",
               k = k,
               depth = dpth,#depth,
               seed = seed)
  
  return (dknn)
}

dknn.classify.trained <- function(objects, dknn){
  
  # Correct input data
  if(is.data.frame(objects))
    objects = as.matrix(objects)
  if (!is.matrix(objects)){
    objects <- matrix(objects, nrow=1)
  }
  if(ncol(objects)!=dknn$dimension)
    stop("Parameter 'objects' has wrong dimension")
  
  points <- as.vector(t(dknn$data[,1:dknn$dimension]))
  obj <- as.vector(t(objects))
  output <- .C("DKnnClassify", 
          as.double(obj),
          as.integer(nrow(objects)),
          as.double(points), 
          as.integer(dknn$labels),
          as.integer(dknn$n), 
          as.integer(dknn$dimension), 
          as.integer(dknn$k), 
          as.integer(dknn$depth),
          as.integer(dknn$seed),
          output=integer(nrow(objects)))$output

  results = dknn$uniquelab[output]
  return (results)
}

dknn.classify <- function(objects, data, k, depth = "halfspace", seed = 0){
  
  n = nrow(data)
  dimension = ncol(data)-1
  labs <- data[,dimension+1]
  labels <- integer(length(labs))
  
  uniquelab = unique(labs)
  cardinalities = unlist(lapply(uniquelab, function(l)sum(labs == l)))
  uniquelab = uniquelab[order(cardinalities)]
  cardinalities = cardinalities[order(cardinalities)]
  for (i in seq_along(uniquelab)){
    labels[labs == uniquelab[i]] <- i
  }
  if(length(uniquelab)<2)
    stop("There is only one class")
  
  if (!is.numeric(k) 
      || is.na(k) 
      || length(k) != 1 
      || !.is.wholenumber(k) 
      || k < 1
      || k > nrow(data)){
    stop("Argument \"k\" not specified correctly.")
  }
  
  dpth = switch (depth,
                 "halfspace" = 1,
                 "Mahalanobis" = 2,
                 "simplicial" = 3, 0)
  if(dpth == 0)
    stop("Wrong depth: ", depth)
  
  dknn <- list(data = data, 
               n = n, 
               dimension = dimension,
               labels = labels,
               uniquelab = uniquelab,
               methodSeparator = "Dknn",
               k = k,
               depth = dpth,#depth,
               seed = seed)
  
  return (dknn.classify.trained(objects, dknn))
}

depth.spatial.local <- function(x, data, mah.estimate = "moment", mah.parMcd = 0.75, kernel.bandwidth = 1){
  if (!is.matrix(x) 
      && is.vector(x)){
    x <- matrix(x, nrow=1)
  }
  mean <- colMeans(data)
  if(mah.estimate == "none"){
    lambda = diag(ncol(data))
  } else {
    if(mah.estimate == "moment"){
      cov <- cov(data)
    } else if(mah.estimate == "MCD"){
      cov <- covMcd(data, mah.parMcd)$cov
    } else {stop("Wrong parameter 'mah.estimate'")}
    cov.eig <- eigen(cov)
    B <- cov.eig$vectors %*% diag(sqrt(cov.eig$values))
    lambda <- solve(B)
  }
  
  if (kernel.bandwidth>1){
    k <- function(x){
      return(sqrt(2*pi)^(-ncol(x))*exp(-rowSums(x^2)/2/kernel.bandwidth))
    }
  }else{  #1/kernel.bandwidth^d
    k <- function(x){
      return((sqrt(2*pi)*kernel.bandwidth)^(-ncol(x))*exp(-rowSums(x^2)/2/kernel.bandwidth))
    }
  }  

  depths <- apply(x, 1, function(x){  
            n = nrow(data)
            t1 = t(lambda %*% (x - t(data)))
            t1 <- t1[which(rowSums(t1) != 0),]
            
            return (sum(k(t1))/n - sqrt(sum((colSums( k(t1)*( t1/sqrt(rowSums(t1^2)) ) )/n)^2)))
          }  
        )
  return (depths)
}

depth.spatial <- function(x, data, mah.estimate = "moment", mah.parMcd = 0.75){
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if(is.data.frame(data))
    data = data.matrix(data)
  if (!is.matrix(x)){
    if(is.vector(x))
      x <- matrix(x, nrow=1)
    if(is.data.frame(x))
      x = data.matrix(x)
  }
  mean <- colMeans(data)
  if(mah.estimate == "none"){
    lambda = diag(ncol(data))
  } else {
    if(mah.estimate == "moment"){
      cov <- cov(data)
    } else if(mah.estimate == "MCD"){
      cov <- covMcd(data, mah.parMcd)$cov
    } else {stop("Wrong parameter 'mah.estimate'")}
    if(sum(is.na(cov)) == 0){
      cov.eig <- eigen(cov)
      B <- cov.eig$vectors %*% diag(sqrt(cov.eig$values))
      lambda <- solve(B)
    } else{
      lambda = diag(ncol(data))
    }
  }
  
  depths <- rep(-1, nrow(x))
  for (i in 1:nrow(x)){
    tmp1 <- t(lambda %*% (x[i,] - t(data)))
    tmp1 <- tmp1[which(rowSums(tmp1) != 0),]
    tmp2 <- 1/sqrt(rowSums(tmp1^2))
    depths[i] <- 1 - sqrt(sum((colSums(tmp2*tmp1)/nrow(data))^2))
  }
  return (depths)
}

depth.space.spatial <- function(data, cardinalities, mah.estimate = "moment", mah.parMcd = 0.75){
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.vector(cardinalities, mode = "numeric") 
      || is.na(min(cardinalities)) 
      || sum(.is.wholenumber(cardinalities)) != length(cardinalities) 
      || min(cardinalities) <= 0 
      || sum(cardinalities) != nrow(data)){
    stop("Argument \"cardinalities\" should be a vector of cardinalities of the classes in \"data\" ")
  }
  if (sum(cardinalities < ncol(data) + 1) != 0){
    stop("Not in all classes sufficiently enough objetcs")
  }
  
  depth.space <- NULL
  for (i in 1:length(cardinalities)){
    pattern <- data[(1 + sum(cardinalities[0:(i - 1)])):sum(cardinalities[1:i]),]
    pattern.depths <- depth.spatial (data, pattern, mah.estimate, mah.parMcd)
    depth.space <- cbind(depth.space, pattern.depths, deparse.level = 0)
  }
  
  return (depth.space)
}

################################################################################
# File:             depth.zonoid.r
# Created by:       Pavlo Mozharovskyi
# First published:  28.02.2013
# Last revised:     15.05.2013
# 
# Computation of the zonoid data depth.
################################################################################

depth.zonoid <- function(x, data, seed = 0){
  if (seed!=0) set.seed(seed)
  if (!is.matrix(x) 
      && is.vector(x)){
    x <- matrix(x, nrow=1)
  }
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.numeric(x)){
    stop("Argument \"x\" should be numeric")
  }
  
  if (ncol(x) != ncol(data)){
    stop("Dimensions of the arguments \"x\" and \"data\" should coincide")
  }
  if (ncol(data) + 1 > nrow(data)){ #?
    stop("To few data points")
  }
  
  points <- as.vector(t(data))
  objects <- as.vector(t(x))
  ds <- .C("ZDepth", 
           as.double(points), 
           as.double(objects), 
           as.integer(nrow(data)), 
           as.integer(nrow(x)), 
           as.integer(ncol(data)), 
           as.integer(seed),
           depths=double(nrow(x)))$depths
  
  return (ds)
}

################################################################################
# File:             ddalpha.classify.r
# Created by:       Pavlo Mozharovskyi
# First published:  28.02.2013
# Last revised:     15.05.2013
# 
# Contains the classification function of the DDalpha-classifier.
# 
# For a description of the algorithm, see:
#   Lange, T., Mosler, K. and Mozharovskyi, P. (2012). Fast nonparametric 
#     classification based on data depth. Statistical Papers.
#   Mozharovskyi, P., Mosler, K. and Lange, T. (2013). Classifying real-world 
#     data with the DDalpha-procedure. Mimeo.
################################################################################

ddalpha.classify <- function(ddalpha, 
                             objects,
                             subset,
                             outsider.method = NULL, 
                             use.convex = NULL){
  # Checks
  if (!is.matrix(objects) && !is.data.frame(objects)){
    objects <- matrix(objects, nrow=1)
  }
  if (!(is.matrix(objects) && is.numeric(objects)
        || is.data.frame(objects) && prod(sapply(objects, is.numeric)))){
    warning("Argument \"objects\" has unacceptable format. Classification can not be performed!!!")
    return (NULL)
  } 
  
  # convert using formula
  if(!is.null(ddalpha$classif.formula)){
    objects = model.frame(ddalpha$classif.formula, data = objects)
  }
  
  if(!missing(subset))
    objects = objects[subset,]
  
  if (ncol(objects) != ddalpha$dimension){
    warning("Dimension of the objects to be classified does not correspond to the dimension of the trained classifier. Classification can not be performed!!!")
    return (NULL)
  }
  
  if (ddalpha$methodSeparator == "Dknn")
    return(dknn.classify.trained(objects, ddalpha))
  
#  if (!is.character(outsider.method) 
#      || length(outsider.method) != 1){
#    warning("Argument \"outsidet.method\" not specified correctly. Outsiders will be ignored!!!")
#    outsider.method <- NULL
#  }
  if (is.null(use.convex)){
    use.convex <- ddalpha$useConvex
  }
  depths <- matrix(nrow=0, ncol=ddalpha$numPatterns) #?
  freePoints <- matrix(nrow=0, ncol=ncol(objects)) #?
  
  if (is.null(ddalpha$methodDepth)){ #use only outsiders treatment
    classifiableIndices <- c()
    resultsDepths <- list()
    freePoints <- objects
  } 
  else {
  # Define points that can be classified by the DD-Alpha and the outsiders
  if (use.convex){
    points <- ddalpha$patterns[[1]]$points
    cardinalities <- c(ddalpha$patterns[[1]]$cardinality)
    for (i in 2:ddalpha$numPatterns){
      points <- rbind(points, ddalpha$patterns[[i]]$points)
      cardinalities <- c(cardinalities, ddalpha$patterns[[i]]$cardinality)
    }
    classifiable <- .are_classifiable(objects, points, cardinalities)
    classifiableIndices <- which(classifiable == 1)
    if (length(classifiableIndices) == 0){
      depths <- matrix(nrow=0, ncol=ddalpha$numPatterns)
      freePoints <- objects
    }else{
      depths <- .ddalpha.count.depths(ddalpha, objects[classifiableIndices,])
      
      freePoints <- matrix(objects[-classifiableIndices,,drop=F], 
                           nrow=nrow(objects)-length(classifiableIndices))
    }
  }else{
    ifelse(ddalpha$methodDepth == "ddplot",
      depths <- objects,
      depths <- .ddalpha.count.depths(ddalpha, objects) )

    classifiableIndices <- c()
    for (i in 1:nrow(depths)){
      if (sum(depths[i,]) > 0){
        classifiableIndices <- c(classifiableIndices, i)
      }
    }
    if (length(classifiableIndices) == 0){
      depths <- matrix(nrow=0, ncol=ddalpha$numPatterns)
      freePoints <- objects
    }else{
      depths <- suppressWarnings( as.matrix(depths[classifiableIndices,,drop=F], 
                       nrow=length(classifiableIndices), ncol=ddalpha$numPatterns))
      freePoints <- # objects[-classifiableIndices,,drop=F]# 
                suppressWarnings( as.matrix(objects[-classifiableIndices,,drop=F], 
                           nrow=nrow(objects)-length(classifiableIndices), 
                           ncol=ncol(objects)))
    }
  }
  
  # Classify with the pure DD classifiers
  resultsDepths <- list()
  if (nrow(depths) > 0){

    fname = paste0(".", ddalpha$methodSeparator, "_classify")
    classify <- .getFunction(fname)

    resultsDepths1 <- list()
    if (ddalpha$methodSeparatorBinary){
      #### Binary classifiers
      
      votes <- matrix(rep(0, nrow(depths)*ddalpha$numPatterns), nrow=nrow(depths), ncol=ddalpha$numPatterns)
      for (i in 1:ddalpha$numClassifiers){
        xAxis <- ddalpha$classifiers[[i]]$index1
        yAxis <- ddalpha$classifiers[[i]]$index2
        
        result <- classify(ddalpha, ddalpha$classifiers[[i]], depths)
        
        for (obj in 1:nrow(depths)){
          if (result[obj] > 0){
            votes[obj,xAxis] <- votes[obj,xAxis] + 1
          }else{
            votes[obj,yAxis] <- votes[obj,yAxis] + 1
          }
        }
      }
      for (i in 1:nrow(depths)){
        resultsDepths[[i]] <- ddalpha$patterns[[which.max(votes[i,])]]$name
      }
    } else {
      #### Multiclass classifiers
      
      indexes <- classify(ddalpha, ddalpha$classifiers[[1]], depths)
      
      for (i in 1:nrow(depths)){
        resultsDepths[[i]] <- ddalpha$patterns[[indexes[i]]]$name
      }
    }

  }
  } # end if(!is.null(ddalpha$methodDepth))
  
  # Classify Outsiders
  resultsOutsiders <- as.list(rep("Ignored", nrow(freePoints)))
  freePoints
  if (is.null(outsider.method) && length(ddalpha$methodsOutsider) == 1)
    outsider.method = ddalpha$methodsOutsider[[1]]$name
  if (length(resultsOutsiders) > 0 && !is.null(outsider.method)){
    for (i in 1:length(ddalpha$methodsOutsider)){
      if (toupper(ddalpha$methodsOutsider[[i]]$name) == toupper(outsider.method)){
        resultsOutsiders <- .ddalpha.classify.outsiders(freePoints, ddalpha, ddalpha$methodsOutsider[[i]])
        break
      }
    }
  }
  
  # Merge classifiable and outsiders
  if (length(resultsOutsiders) == 0)
    results <- resultsDepths
  else if(length(resultsDepths) == 0)
    results <- resultsOutsiders
  else{
    
    if(is.factor(resultsOutsiders[[1]]) && !is.factor(resultsDepths[[1]]))
      resultsOutsiders = lapply(resultsOutsiders, as.character)
    
    if(is.numeric(resultsDepths[[1]]) && !is.numeric(resultsOutsiders[[1]]))
      resultsOutsiders = as.numeric(resultsOutsiders)
    
    results <- list()
    counterDepths <- 1
    counterOutsiders <- 1
    for (i in 1:nrow(objects)){
      if (i %in% classifiableIndices){
        results[[i]] <- resultsDepths[[counterDepths]]
        counterDepths <- counterDepths + 1
      }else{
        results[[i]] <- resultsOutsiders[[counterOutsiders]]
        counterOutsiders <- counterOutsiders + 1
      }
    }
  }

  if (length(results) == 1) return(results[[1]])
                       else return (results)
}

predict.ddalpha <- function(object, 
                            objects, 
                            subset,
                            outsider.method = NULL, 
                            use.convex = NULL, ...){
  return(ddalpha.classify(object, objects, subset, outsider.method, use.convex))
}

.ddalpha.learn.polynomial <- function(ddalpha){
  # Separating (calculating extensions and normals)
  counter <- 1
  # Determining multi-class behaviour
  if (ddalpha$methodAggregation == "majority"){
    for (i in 1:(ddalpha$numPatterns - 1)){
      for (j in (i + 1):ddalpha$numPatterns){
        # Creating a classifier
        polynomial <- .polynomial_learn_C(ddalpha$maxDegree, 
                                   rbind(ddalpha$patterns[[i]]$depths, 
                                         ddalpha$patterns[[j]]$depths), 
                                   ddalpha$patterns[[i]]$cardinality, 
                                   ddalpha$patterns[[j]]$cardinality, 
                                   ddalpha$numChunks, ddalpha$seed)
 # DEBUG       
 if (F){
        print(polynomial$coefficients)
        print(GetEmpiricalRisk (polynomial$coefficients,
                          rbind(ddalpha$patterns[[i]]$depths, ddalpha$patterns[[j]]$depths), 
                          ddalpha$patterns[[i]]$cardinality, 
                          ddalpha$patterns[[j]]$cardinality))
      }  
        # Adding the classifier to the list of classifiers
        ddalpha$classifiers[[counter]] <- 
          list(
            index          = counter,
            index1         = i,
            index2         = j,
            polynomial     = polynomial$coefficients,
            degree         = polynomial$degree,
            axis           = polynomial$axis)
        
        counter <- counter + 1
      }
    }
    ddalpha$numClassifiers <- counter - 1
  }
  if (ddalpha$methodAggregation == "sequent"){
    for (i in 1:ddalpha$numPatterns){
      anotherClass <- NULL
      for (j in 1:ddalpha$numPatterns){
        if (j != i){
          anotherClass <- rbind(anotherClass, ddalpha$patterns[[j]]$depths)
        }
      }
      polynomial <- .polynomial_learn_C(ddalpha$maxDegree, rbind(ddalpha$patterns[[i]]$depths, anotherClass), 
                                        ddalpha$patterns[[i]]$cardinality, nrow(anotherClass), ddalpha$numChunks, ddalpha$seed)

      # Adding the classifier to the list of classifiers
      ddalpha$classifiers[[i]] <- 
        list(index          = counter, 
             index1         = i, 
             index2         = -1, 
             polynomial     = polynomial$coefficients,
             degree         = polynomial$degree,
             axis           = polynomial$axis)
    }
    ddalpha$numClassifiers <- ddalpha$numPatterns
  }
  
  return (ddalpha)
}

################################################################################
# Functions for intermediate calculations are presented below
################################################################################

.polynomial_learn_C <- function(maxDegree, data, numClass1, numClass2, numChunks, seed){
  points <- as.vector(t(data))
  numPoints <- numClass1 + numClass2
  dimension <- ncol(data)
  cardinalities <- c(numClass1, numClass2)
  upToPower <- maxDegree
  minFeatures <- 2
  maxExtDimension <- (factorial(dimension + maxDegree) / (factorial(dimension)*factorial(maxDegree))) - 1;
  

  res <- .C("PolynomialLearnCV", 
          as.double(points), 
          as.integer(numPoints), 
          as.integer(dimension), 
          as.integer(cardinalities),  
          as.integer(upToPower), 
          as.integer(numChunks), 
          as.integer(seed),
          degree = integer(1),
          axis = integer(1),
          polynomial=double(upToPower))
  
  degree <- res$degree
  axis <- res$axis
  polynomial <- res$polynomial[1:degree]
  
  return(list(coefficients = polynomial, axis = axis, degree = degree))
}

GetNumsErrors <- function(polynomial, depths, numClass1, numClass2){
  # Calculates the number of classification error for two classes on the
  # basis of given depths
  # 
  # Args:
  #   polynomial: Polynomial as a vector of coefficients starting with the 
  #               first degree (a0 = 0 always)
  #   depths:     nx2 matrix of depths, where each column contains the depths
  #               against the corresponding class
  #   numClass1:  Number of points belonging to the first class
  #   numClass2:  Number of points belonging to the second class
  # Returns:
  #   Vector containing number of errors of the points from the firts and 
  #   the second class
  degree <- length(polynomial)
  numErrors1 <- 0
  if(numClass1 != 0){
    for(i in 1:numClass1){      
      val <- depths[i,1]
      res <- 0
      for(j in 1:degree){res <- res + polynomial[j]*val^j}
      if(depths[i,2] > res){
        numErrors1 <- numErrors1 + 1
      }
    }
  }
  numErrors2 <- 0
  if(numClass2 != 0){
    for(i in (numClass1 + 1):(numClass1 + numClass2)){
      val <- depths[i,1]
      res <- 0
      for(j in 1:degree){res <- res + polynomial[j]*val^j}
      if(depths[i,2] < res){
        numErrors2 <- numErrors2 + 1
      }
    }
  }
  return(c(numErrors1, numErrors2))
}

GetEmpiricalRiskSmoothed <- function(polynomial, depths, numClass1, numClass2){
  res = (colSums(sapply(depths[,1], '^', (1:length(polynomial)))*polynomial) - depths[,2])*c(rep(-1, numClass1), rep(1, numClass2))
  risk = sum(1/(1 + exp(-100*(res))))
  return (risk/(numClass1 + numClass2))
}

GetEmpiricalRisk <- function(polynomial, depths, numClass1, numClass2){
  # Calculates the empirical risk for two classes on the basis of given depths
  # 
  # Args:
  #   polynomial: Polynomial as a vector of coefficients starting with the 
  #               first degree (a0 = 0 always)
  #   depths:     nx2 matrix of depths, where each column contains the depths
  #               against the corresponding class
  #   numClass1:  Number of points belonging to the first class
  #   numClass2:  Number of points belonging to the second class
  # Returns:
  #   Empirical risk
  risk1 <- 0
  degree <- length(polynomial)
  for(i in 1:numClass1){
    val <- depths[i,1]
    res <- 0
    for(j in 1:degree){res <- res + polynomial[j]*val^j}
    if(depths[i,2] > res){
      risk1 <- risk1 + 1
    }
  }
  risk2 <- 0
  for(i in (numClass1 + 1):(numClass1 + numClass2)){
    val <- depths[i,1]
    res <- 0
    for(j in 1:degree){res <- res + polynomial[j]*val^j}
    if(depths[i,2] < res){
      risk2 <- risk2 + 1
    }
  }
  risk <- (risk1 + risk2)/(numClass1 + numClass2)
  return(risk)
}

################################################################################
# File:             depth.halfspace.r
# Created by:       Pavlo Mozharovskyi
# First published:  28.02.2013
# Last revised:     13.11.2015
# 
# Computation of the Tukey data depth.
################################################################################

.parse_HSD_pars <- function(exact, method){
  if(missing(exact) && missing(method))
    return(0)
  
  if(!missing(exact)){
    if(exact == F){
      if (missing(method))
        return(0)
      else if (method != 0 && method != "Sunif.1D")
        stop("Wrong combination of 'exact' and 'method' parameters.")
    }
    else{
      if (missing(method))
        return(1)           # default exact
      else
        if (!(method %in% 1:3 || method %in% c("recursive","plane","line")))
          stop("Wrong combination of 'exact' and 'method' parameters.")
    }
  }
  
  if (!(method %in% 0:3 || method %in% c("Sunif.1D","recursive","plane","line")))
    stop("Wrong parameter 'method'.")
      
  if (is.character(method))
    method =  switch (method, random = 0, recursive = 1, plane = 2, line = 3)
  
  return(method)
}

depth.halfspace <- function(x, data, exact, method, num.directions = 1000, seed = 0){
  if (!is.matrix(x) 
      && is.vector(x)){
    x <- matrix(x, nrow=1)
  }
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.numeric(x)){
    stop("Argument \"x\" should be numeric")
  }
  if (ncol(x) != ncol(data)){
    stop("Dimensions of the arguments \"x\" and \"data\" should coincide")
  }
  if (ncol(data) + 1 > nrow(data)){ #?
    stop("To few data points")
  }

  points <- as.vector(t(data))
  objects <- as.vector(t(x))
  
  method = .parse_HSD_pars(exact, method)
  
  if (method == 0)
  if (!is.numeric(num.directions) 
      || is.na(num.directions) 
      || length(num.directions) != 1 
      || !.is.wholenumber(num.directions) 
      || !(num.directions > 1 && num.directions < 10000000)){
    numDirections <- 1000
    warning("Argument \"num.directions\" not specified correctly. 1000 is used as a default value")
  }else{
    numDirections <- num.directions
  }
  
  if (method == 0){
    c <- as.vector(nrow(data))
    k <- numDirections
    ds <- .C("HDepth", 
             as.double(points), 
             as.double(objects), 
             as.integer(nrow(x)), 
             as.integer(ncol(data)), 
             as.integer(c), 
             as.integer(1), 
             dirs=double(k*ncol(data)), 
             prjs=double(k*nrow(data)), 
             as.integer(k), 
             as.integer(1), # use the same directions and projections
             as.integer(seed),
             depths=double(nrow(x)))$depths
  } else 
    if (method %in% 1:3){
      ds <- .C("HDepthEx", 
               as.double(points), 
               as.double(objects), 
               as.integer(nrow(data)),
               as.integer(nrow(x)),  
               as.integer(ncol(data)), 
               as.integer(method), 
               depths=double(nrow(x)))$depths  
    }
  else 
    stop("wrong choise of the algorithm, method = ", method)
  
  return (ds)
}




depth.space.halfspace <- function(data, cardinalities, exact, method, num.directions = 1000, seed = 0){
  if (seed != 0) set.seed(seed)
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.vector(cardinalities, mode = "numeric") 
      || is.na(min(cardinalities)) 
      || sum(.is.wholenumber(cardinalities)) != length(cardinalities) 
      || min(cardinalities) <= 0 
      || sum(cardinalities) != nrow(data)){
    stop("Argument \"cardinalities\" should be a vector of cardinalities of the classes in \"data\" ")
  }
  if (sum(cardinalities < ncol(data) + 1) != 0){
    stop("Not in all classes sufficiently enough objetcs")
  }
  if (!is.numeric(num.directions) 
      || is.na(num.directions) 
      || length(num.directions) != 1 
      || !.is.wholenumber(num.directions) 
      || !(num.directions > 1 && num.directions < 10000000)){
    numDirections <- 1000
    warning("Argument \"num.directions\" not specified correctly. 1000 is used as a default value")
  }else{
    numDirections <- num.directions
  }
  
  x <- as.vector(t(data))
  c <- as.vector(cardinalities)
  
  method = .parse_HSD_pars(exact, method)
  
  if (method == 0)
    if (!is.numeric(num.directions) 
        || is.na(num.directions) 
        || length(num.directions) != 1 
        || !.is.wholenumber(num.directions) 
        || !(num.directions > 1 && num.directions < 10000000)){
      numDirections <- 1000
      warning("Argument \"num.directions\" not specified correctly. 1000 is used as a default value")
    }else{
      numDirections <- num.directions
    }
  
  if (method == 0){
    k <- numDirections
    rez <- .C("HDSpace", 
              as.double(x), 
              as.integer(ncol(data)), 
              as.integer(c), 
              as.integer(length(cardinalities)), 
              as.integer(k), 
              as.integer(1), 
              as.integer(seed),
              dspc=double(nrow(data)*length(cardinalities)), 
              dirs=double(k*ncol(data)), 
              prjs=double(k*nrow(data)))
    depth.space <- matrix(rez$dspc, nrow=nrow(data), ncol=length(cardinalities), byrow=TRUE)
  }else 
    if (method %in% 1:3){
      ds <- .C("HDepthSpaceEx", 
               as.double(x), 
               as.double(x), 
               as.integer(c), 
               as.integer(length(cardinalities)), 
               as.integer(nrow(data)),  
               as.integer(ncol(data)), 
               as.integer(method), 
               depths=double(nrow(data)*length(cardinalities)))$depths  
      
      depth.space <- matrix(ds, nrow=nrow(data), ncol=length(cardinalities), byrow=F)
    }
  else 
    stop("wrong choise of the algorithm, method = ", method)
  
  return (depth.space)
}

depthf.simplicialBand <- function(objectsf, dataf, modified = TRUE, J = NULL, 
                                  range = NULL, d = 101){
  # Calculate the simplicial band depth
  # Follows the article:
  # Lopez-Pintado, Sun, Lin, Genton (2014). 
  # "Simplicial band depth for multivariate data", 
  # Advances in Data Analysis and Classification, 8(3), 321-338.
  # Args:
  #   objectsf: functoins for which the depth should be computed; a list 
  #             containing lists (functions) of two vectors of equal length, 
  #             named "args" and "vals": arguments sorted in ascending order 
  #             and corresponding them values respectively.
  #   dataf:    data sample of functoins w.r.t. which the depth should be 
  #             computed; structure as for "objectsf".
  #   modified: whether modified simplicial band depth should be computed; 
  #             logical, TRUE by default.
  #   J:        how many functions to consider in each tuple of the 
  #             U-statistics; integer, d+1 by default.
  #   range:    The common range of the domain where the functions of objectsf 
  #             and dataf are observed. Vector of length 2 with the left and 
  #             the right end of the interval. Must contain all arguments given 
  #             in objectsf and dataf.
  #   d:        Grid size to which all the functional data are transformed. For 
  #             depth computation, all functional observations are first 
  #             transformed into vectors of their functional values of 
  #             length d corresponding to equi-spaced points in the domain 
  #             given by the interval range. Functional values in these points 
  #             are reconstructed using linear interpolation, and extrapolation.
  # Returns:
  #   A vector of depths of each of "objectsf" w.r.t. "dataf".
  
  # Check input data for consistency:
  if (length(objectsf) < 1){
    stop("Number of functions for which the depth should be computed is < 1.")
  }
  #if (J < 2){
  #  stop("Impossible to calculate depth with 'J' < 2.")
  #}
  p <- ifelse(is.null(dim(objectsf[[1]]$vals)), 1, dim(objectsf[[1]]$vals)[2])
  J <- p + 1
  if (length(dataf) < J){
    stop("Number of functions w.r.t. which the depth should be computed 
         is < dimension + 1.")
  }
  m <- length(objectsf)
  n <- length(dataf)
  objArgs <- unique(unlist(lapply(objectsf, function(x){return(x$args)})))
  datArgs <- unique(unlist(lapply(dataf, function(x){return(x$args)})))
  if (length(objArgs) != length(datArgs) || 
      sum(objArgs == datArgs) != length(objArgs)){
    stop("Not the same arguments for 'objectsf' and 'dataf'.")
  }
  l <- length(objArgs)
  numObjArgs <- unlist(lapply(objectsf, function(x){return(length(x$args))}))
  if (sum(numObjArgs == length(objArgs)) != m){
    stop("Not the same arguments for all functions in 'objectsf'.")
  }
  numDatArgs <- unlist(lapply(dataf, function(x){return(length(x$args))}))
  if (sum(numDatArgs == length(datArgs)) != n){
    stop("Not the same arguments for all functions in 'dataf'.")
  }
  if (p == 1){
    numObjVals <- unlist(lapply(objectsf, function(x){return(length(x$vals))}))
    numDatVals <- unlist(lapply(dataf, function(x){return(length(x$vals))}))
  }else{
    numObjVals <- unlist(lapply(objectsf, function(x){return(nrow(x$vals))}))
    numDatVals <- unlist(lapply(dataf, function(x){return(nrow(x$vals))}))
  }
  if (sum(numObjVals == numObjArgs) != m){
    stop("Number of arguments and values for (some) functions in 'objectsf' differ.")
  }
  if (sum(numDatVals == numDatArgs) != n){
    stop("Number of arguments and values for (some) functions in 'dataf' differ.")
  }
  # Interpolate the data and prepare it for the C++ transfer
  A <- dataf2rawfd(objectsf, range = range, d = d)
  B <- dataf2rawfd(dataf, range = range, d = d)
  At <- apply(A, 1, function(x) t(x))
  Bt <- apply(B, 1, function(x) t(x))
  fArgs <- approx(dataf[[1]]$args, n = d)$x
  # fArgs <- objectsf[[1]]$args
  #fObjVals <- unlist(lapply(A, function(x){return(t(x$vals))}))
  #fDatVals <- unlist(lapply(B, function(x){return(t(x$vals))}))
  # Call the C++ routine
  #print(as.double(fObjVals))
  #print(as.double(fDatVals))
  #print(as.integer(m))
  #print(as.integer(n))
  #print(as.integer(l))
  #print(as.integer(d))
  ds <- .C("SimplicialBandDepthF", 
           as.double(At),
           as.double(Bt), 
           as.double(fArgs), 
           as.integer(m), 
           as.integer(n), 
           as.integer(d), 
           as.integer(p), 
           as.integer(modified), 
           as.integer(J), 
           depths = double(m))$depths
  return(ds)
}

################
# R call for the F procedures
# functional depth computation
# Stanislav Nagy
# nagy@karlin.mff.cuni.cz
# 09/10/2018
#
# Stanislav Nagy, Irene Gijbels & Daniel Hlubinka (2017) Depth-Based Recognition of Shape Outlying Functions, Journal of Computational and Graphical Statistics, 26:4, 883-893, DOI: 10.1080/10618600.2017.1336445
# Stanislav Nagy, Frederic Ferraty (2018) Data Depth for Noisy Random Functions. Under review.
################

if(FALSE){
  # for roxygenize only
  unlink("depth.fd",recursive=TRUE)
  library(roxygen2)
  library(devtools)
  package.skeleton("depth.fd",code_files="depth.fd.R")
  roxygenize("depth.fd")
}

#' @useDynLib depth.fd
#' @export FKS
#' @export shape.fd.analysis
#' @export shape.fd.outliers
#' @export depthf.BD
#' @export depthf.ABD
#' @export depthf.fd1
#' @export depthf.fd2
#' @export depthf.HR
#' @export depthf.hM
#' @export depthf.hM2
#' @export depthf.RP1
#' @export depthf.RP2
#' @export derivatives.est
#' @export L2metric
#' @export Cmetric
#' @export depth.sample
#' @export infimalRank
#' @export dataf2rawfd
#' @export rawfd2dataf

#' @title Transform a \code{dataf} object to raw functional data
#' 
#' @description
#' From a (possibly multivariate) functional data object \code{dataf} constructs an array of the functional values
#' evaluated at an equi-distant grid of points.
#' 
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords functional
#'
#' @param dataf Functions to be transformed, represented by a (possibly multivariate) \code{dataf} object of their arguments
#' and functional values. \code{m} stands for the number of functions. The grid of observation points for the 
#' functions in \code{dataf} may not be the same.
#'
#' @param range The common range of the domain where the fucntions \code{dataf} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{dataf}. If the range is not provided, the smallest interval in which all the arguments from the data functions
#' are contained is chosen as the domain.
#' 
#' @param d Grid size to which all the functional data are transformed. All functional observations are 
#' transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation, see Nagy et al. (2016).
#'
#' @return If the functional data are univariate (scalar-valued), a matrix of size \code{m*d} is given, with each row
#' corresponding to one function. If the functional data are \code{k}-variate with k>1, an array of size \code{m*d*k}
#' of the functional values is given.
#' 
#' @examples
#' ## transform a matrix into a functinal data set and back
#' n = 5
#' d = 21
#' X = matrix(rnorm(n*d),ncol=d)
#' R = rawfd2dataf(X,range=c(0,1))
#' R2 = dataf2rawfd(R,range=c(0,1),d=d)
#' all.equal(X,R2)
#' 
#' ## transform a functional dataset into a raw matrix of functional values
#' dataf = dataf.population()$dataf
#' dataf2rawfd(dataf,range=c(1950,2015),d=66)
#' 
#' ## transform an array into a multivariate functional data set and back
#' k = 3
#' X = array(rnorm(n*d*k),dim=c(n,d,k))
#' R = rawfd2dataf(X,range=c(-1,1))
#' dataf2rawfd(R,range=c(-1,1),d=50)
#'
#' @seealso \code{\link{rawfd2dataf}}
#' @seealso \code{\link{depthf.fd1}}
#' @seealso \code{\link{depthf.fd2}}

# M = dataf2rawfd(dataf.growth()$dataf)

dataf2rawfd = function(dataf, range = NULL, d = 101){
  # transform dataf format for functional data to a raw matrix
  # of functional values evaluated at a grid common to all functions
  # range: range of the common grid, if not specified the range of all the functions
  # d: no of discretized points in the grid, equidistant
  # approximation procedure follows Nagy et al. (2016, JMVA)
  
  # Check "dataf"
  if (!is.list(dataf))
    stop("Argument 'dataf' must be a list")
  
  if(is.vector(dataf[[1]]$vals)){
    mv = FALSE
    for (df in dataf) 
      if (!(is.list(df) && length(df) == 2 &&
            !is.null(df$args) && !is.null(df$vals) &&
            is.vector(df$args) && is.vector(df$vals) &&
            is.numeric(df$args) && is.numeric(df$vals) &&
            length(df$args) == length(df$vals) &&
            is.sorted(df$args)))
        stop("Argument 'dataf' must be a list containing lists (functions) 
             of two vectors of equal length, named 'args' and 'vals': 
             arguments sorted in ascending order and corresponding them 
             values respectively") 
  }
  
  if(is.matrix(dataf[[1]]$vals)){
    mv = TRUE
    for (df in dataf)
      if (!(is.list(df) && length(df) == 2 &&
            !is.null(df$args) && !is.null(df$vals) &&
            is.vector(df$args) && is.matrix(df$vals) &&
            is.numeric(df$args) && is.numeric(df$vals) &&
            length(df$args) == nrow(df$vals) &&
            is.sorted(df$args)))
        stop("Argument 'dataf' must be a list containing lists (functions) 
             of a vector named 'args' and a matrix named 'vals'. The arguments
             of 'args' must be sorted in ascending order. To each element of 'args'
             the corresponding row in 'vals' represents 
             the functional values at this point") 
  }
  
  # range construction
  rng = numeric(0)
  for (df in dataf)	rng = range(c(rng,df$args)) # common range of all the data
  if(!is.null(range)){ 
    if(!(length(range) == 2 && is.numeric(range) && 
         range[1]<=rng[1] && range[2]>=rng[2])) 
      stop("Argument 'range' must be a numeric vector of two components
           that defines the range of the domain of functional data. All
           functional data must have 'args' vales inside this domain.")
  } else range = rng
  if(!(range[1]<range[2])) stop("Argument 'range' must define a non-degenerate interval.")	
  
  t = seq(range[1],range[2],length=d)
  n = length(dataf)
  if(!mv){
    X = matrix(nrow=n,ncol=d)
    # functional data interpolation / extrapolation
    for(i in 1:n){
      ni = length(dataf[[i]]$args)
      X[i,] = approx(dataf[[i]]$args,dataf[[i]]$vals,t)$y
      X[i,t<=dataf[[i]]$args[1]] = dataf[[i]]$vals[1]
      X[i,t>=dataf[[i]]$args[ni]] = dataf[[i]]$vals[ni]
    }
  } else {
    k = ncol(dataf[[1]]$vals)
    X = array(dim=c(n,d,k))
    # functional data interpolation / extrapolation
    for(i in 1:n){
      ni = length(dataf[[i]]$args)
      for(j in 1:k) X[i,,j] = approx(dataf[[i]]$args,dataf[[i]]$vals[,j],t)$y
      X[i,t<=dataf[[i]]$args[1],] = dataf[[i]]$vals[1,]
      X[i,t>=dataf[[i]]$args[ni],] = dataf[[i]]$vals[ni,]
    }
  }
  return(X)
  }

#' @title Transform raw functional data to a \code{dataf} object
#' 
#' @description
#' Constructs a (possibly multivariate) functional data object given by an array of its functional values
#' evaluated at an equi-distant grid of points, and transforms it into a \code{dataf} object more suitable 
#' for work in the \code{ddalpha} package.
#' 
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords functional
#'
#' @param X Either a matrix of size \code{n*d}, or an array of dimension \code{n*d*k} of functional values. Here \code{n}
#' stands for the number of functions, \code{d} is the number of equi-distant points in the domain where the functional
#' values are evaluated, and if applicable, \code{k} is the dimensionality of the (vector-valued) functional data.
#'
#' @param range A vector of size two that represents the endpoints of the common domain of all functions \code{X}.
#'
#' @return A (possibly multivariate) \code{dataf} object corresponding to the functional data \code{X} evaluated at an
#' equi-distant grid of points.
#' 
#' @examples
#' ## transform a matrix into a functinal data set
#' n = 5
#' d = 21
#' X = matrix(rnorm(n*d),ncol=d)
#' rawfd2dataf(X,range=c(0,1))
#' 
#' ## transform an array into a multivariate functional data set
#' k = 3
#' X = array(rnorm(n*d*k),dim=c(n,d,k))
#' rawfd2dataf(X,range=c(-1,1))
#'
#' @seealso \code{\link{dataf2rawfd}}
#' @seealso \code{\link{depthf.fd1}}
#' @seealso \code{\link{depthf.fd2}}

rawfd2dataf = function(X, range){
  # transform a raw array of functional data values 
  # to a dataf format, where the domain is assumed to be
  # an equidistant grid in the interval given by range
  
  if(is.vector(X)) X = matrix(X,nrow=1) # if X is a single vector, it is considered to be a matrix of one scalar function
  
  # Check "rawfd"
  if (!is.array(X))
    stop("Argument 'X' must be an array (multivariate functional data)
         or a matix (univariate functional data)")
  mv = !is.matrix(X)
  
  # range construction
  if(!(range[1]<range[2])) stop("Argument 'range' must define a non-degenerate interval.")	
  n = dim(X)[1]
  d = dim(X)[2]
  if(mv) k = dim(X)[3]
  t = seq(range[1],range[2],length=d)
  dataf = list()
  for(i in 1:n){
    if(mv) df = list(args = t, vals = X[i,,]) else df = list(args = t, vals = X[i,])
    dataf[[i]] = df
  }
  return(dataf)
}

#' @title Fast kernel smoothing
#'
#' @description
#' Produces a kernel smoothed version of a function based on
#' the vectors given in the input. Bandwidth is selected using cross-validation.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords smoothing kernel functional
#'
#' @details
#' A vector of the same length as \code{Tout}
#' corresponding to the values of the
#' function produced using kernel smoothing, is provided. Bandwidth is selected using the
#' \code{K}-fold cross-validation of randomly shuffled input values.
#'
#' @param dataf A set of functional data given by a \code{dataf} object that are to be smoothed.
#'
#' @param Tout vector of values in the domain of the functions at which the
#' resulting smoothed function is evaluated
#' 
#' @param kernel Kernel used for smoothing. Admissible values are \code{uniform}, 
#' \code{triangular}, \code{Epanechnikov}, \code{biweight}, \code{triweight} and \code{Gaussian}.
#' By default, \code{uniform} is used.
#' 
#' @param m Number of points in the grid for choosing the cross-validated bandwidth.
#' 
#' @param K Performs \code{K}-fold cross-validation based on randomly shuffled data.
#'
#' @return A \code{dataf} object corresponding to \code{Tout} of smoothed functional values.
#' 
#' @examples
#' d = 10
#' T = sort(runif(d))
#' X = T^2+ rnorm(d,sd=.1)
#' Tout = seq(0,1,length=101)
#' 
#' plot(T,X)
#' dataf = list(list(args=T,vals=X))
#' data.sm = FKS(dataf,Tout,kernel="Epan")
#' lines(data.sm[[1]]$args,data.sm[[1]]$vals,col=2)
#' 
#' datafs = structure(list(dataf=dataf,labels=1:length(dataf)),class="functional")
#' plot(datafs)
#' points(T,X)
#' data.sms = structure(list(dataf=data.sm,labels=1:length(data.sm)),class="functional")
#' plot(data.sms)
#' 
#' n = 6
#' dataf = list()
#' for(i in 1:n) dataf[[i]] = list(args = T<-sort(runif(d)), vals = T^2 + rnorm(d,sd=.1))
#' data.sm = FKS(dataf,Tout,kernel="triweight")
#' data.sms = structure(list(dataf=data.sm,labels=1:length(data.sm)),class="functional")
#' plot(data.sms)

FKS = function(dataf,Tout,kernel=c("uniform","triangular","Epanechnikov","biweight","triweight","Gaussian"),m=51,K=20){
  # kernel smoothing
  # produces directly the kernel smoother with CV value of h
  # Hmax 	the maximal H for CV
  # m		no of elements of H for CV
  # K		K-fold CV for bandwidth selection
  #
  # Args:
  #   dataf:  list containing lists (functions) of two vectors of equal length, 
  #           named "args" and "vals": arguments sorted in ascending order and 
  #           corresponding them values respectively
  #   labels: output labels of the functinal observations
  
  # Check "dataf"
  if (!is.list(dataf))
    stop("Argument 'dataf' must be a list")
  for (df in dataf)
    if (!(is.list(df) && length(df) == 2 &&
          !is.null(df$args) && !is.null(df$vals) &&
          is.vector(df$args) && is.vector(df$vals) &&
          is.numeric(df$args) && is.numeric(df$vals) &&
          length(df$args) == length(df$vals) &&
          is.sorted(df$args)))
      stop("Argument 'dataf' must be a list containing lists (functions) 
           of two vectors of equal length, named 'args' and 'vals': 
           arguments sorted in ascending order and corresponding them 
           values respectively")  
  
  krn = c("uniform","triangular","Epanechnikov","biweight","triweight","Gaussian")
  kernel = match.arg(kernel)
  kernI = which(kernel==krn)
  for(df in dataf) if(sum(is.na(df$args))>0) stop("NA values in the functional args vector are not allowed")
  # produce the output structure
  sm.dataf = list()
  Tout = sort(Tout)
  for(j in 1:length(dataf)){
    T = dataf[[j]]$args
    X = dataf[[j]]$vals
    ST = dataf[[j]]$args
    eps = 10^(-6)
    Hmin = max(c(ST[1], (1 - ST[length(ST)]), max(ST[-1] - ST[-length(ST)])/2)) + eps
    Hmax = max(ST[length(ST)]-ST[1],(max(Tout)-min(Tout)))
    if (Hmin >= Hmax) Hmin = Hmax/10 # in the extreme case when Hmin > Hmax take Hmin very small
    H = seq(Hmin,Hmax,length=m)
    #
    n = length(ST)
    nR = max(1,floor(n/K))
    if(nR==1) K = n
    TRE = rep(NA,nR*K)
    XRE = rep(NA,nR*K)
    TNRE = rep(NA,K*(n-nR))
    XNRE = rep(NA,K*(n-nR))
    SH = sample(n,n)			# random shuffle of the points for CV
    for(i in 1:K){
      S = SH[(i-1)*nR+(1:nR)] 
      TRE[(i-1)*nR+(1:nR)] = T[S]
      XRE[(i-1)*nR+(1:nR)] = X[S]
      TNRE[(i-1)*(n-nR)+(1:(n-nR))] = T[-S]
      XNRE[(i-1)*(n-nR)+(1:(n-nR))] = X[-S]
    }
    Res = .Fortran("CVKERNSM",
                   as.double(T),
                   as.double(X),
                   as.double(Tout),
                   as.integer(length(T)),
                   as.integer(length(Tout)),
                   as.double(H),
                   as.integer(length(H)),
                   as.integer(kernI),
                   as.double(TRE),
                   as.double(XRE),
                   as.double(TNRE),
                   as.double(XNRE),
                   as.integer(nR),		
                   as.integer(K),
                   as.double(rep(0,length(Tout)))
    )
    Res[Res[[15]]>10^6] = Inf
    sm.dataf[[j]] = list(args = Tout, vals = Res[[15]])
  }
  return(sm.dataf)
}

#' @title Fast depth computation for univariate and bivariate random samples
#'
#' @description
#' Faster implementation of the halfspace and the simplicial depth. Computes the depth 
#' of a whole random sample of a univariate or a bivariate data in one run.	
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth
#'
#' @details
#' The function returns vectors of sample halfspace and simplicial depth values.
#'
#' @param A Univariate or bivariate points whose depth is computed, represented by a matrix of 
#' size \code{m*2}. \code{m} stands for the number of points, \code{d} is 1 for univariate and 2 
#' for bivariate data.
#'
#' @param B Random sample points with respect to which the depth of \code{A} is computed. 
#' \code{B} is represented by a matrix of size \code{n*2}, where \code{n} is the sample size.
#'
#' @return Vector of length \code{m} of depth halfspace depth values is returned.
#' 
#' @examples
#' n = 100
#' m = 150
#' A = matrix(rnorm(2*n),ncol=2)
#' B = matrix(rnorm(2*m),ncol=2)
#' depth.sample(A,B)
#' system.time(D1<-depth.halfspace(A,B))
#' system.time(D2<-depth.sample(A,B))
#' max(D1-D2$Half)
#'
#' A = rnorm(100)
#' B = rnorm(150)
#' depth.sample(A,B)
#' # depth.halfspace(matrix(A,ncol=1),matrix(B,ncol=1))
#'
#' @seealso \code{\link{depth.halfspace}}
#' @seealso \code{\link{depth.simplicial}}

depth.sample = function(A,B){
  # bivariate halfspace depth
  # A points whose depth I compute, M*2 matrix
  # B points wrt whose the depth is computed, N*2 matrix
  if(is.null(dim(A))){ # for univariate data
    A1 = as.vector(A)
    B1 = as.vector(B)	
    m = length(A)
    n = length(B)
    FD = .Fortran("dpth1",	
                  as.numeric(A1),		#	A1
                  as.numeric(B1),		#	B1
                  as.integer(m),		#	m
                  as.integer(n),		#	n
                  sdep=as.numeric(rep(-1,m)),
                  hdep=as.numeric(rep(-1,m))
    )
    return(list(Simpl = FD$sdep, Half = FD$hdep))
  } else {
    A1 = as.vector(A[,1])
    A2 = as.vector(A[,2])
    B1 = as.vector(B[,1])
    B2 = as.vector(B[,2])
    m = dim(A)[1]
    n = dim(B)[1]
    if((dim(B)[2]!=2)|(dim(A)[2]!=2)) stop("Computation for two dimensions only")
    FD = .Fortran("dpth2",	
                  as.numeric(A1),		#	A1
                  as.numeric(A2),		#	A2
                  as.numeric(B1),		#	B1
                  as.numeric(B2),		#	B2
                  as.integer(m),		#	m
                  as.integer(n),		#	n
                  sdep=as.numeric(rep(-1,m)),
                  hdep=as.numeric(rep(-1,m))
    )
    return(list(Simpl = FD$sdep, Half = FD$hdep))
  }
}

#' @title Univariate integrated and infimal depth for functional data
#'
#' @description
#' Usual, and order extended integrated and infimal depths for real-valued functional data based on the
#' halfspace and simplicial depth.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth functional
#'
#' @details
#' The function returns vectors of sample integrated and infimal depth values.
#'
#' @param datafA Functions whose depth is computed, represented by a \code{dataf} object of their arguments
#'  and functional values. \code{m} stands for the number of functions.
#'
#' @param datafB Random sample functions with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a \code{dataf} object of their arguments
#'  and functional values. \code{n} is the sample size. 
#' The grid of observation points for the 
#' functions \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation, see Nagy et al. (2016).
#'
#' @param order The order of the order extended integrated and infimal depths.
#' By default, this is set to \code{1}, meaning that the usual univariate depths of 
#' the functional values are computed. For \code{order=2} or \code{3}, the second
#' and the third order extended integrated and infimal depths are computed, 
#' respectively. 
#'
#' @param approx Number of approximations used in the computation of the order extended depth
#' for \code{order} greater than \code{1}. For \code{order=2}, the default
#' value is set to \code{0}, meaning that the depth is computed at all possible \code{d^order}
#' combinations of the points in the domain. For \code{order=3}, 
#' the default value is set to \code{101}. When \code{approx} is a positive integer, \code{approx}
#' points are randomly sampled in \code{[0,1]^order} and at these points the \code{order}-variate depths of the
#' corresponding functional values are computed.
#'
#' @return Four vectors of length \code{m} of depth values are returned:
#'	\itemize{
#'	\item \code{Simpl_FD} the integrated depth based on the simplicial depth,
#'	\item \code{Half_FD} the integrated depth based on the halfspace depth,
#'	\item \code{Simpl_ID} the infimal depth based on the simplicial depth,
#'	\item \code{Half_ID} the infimal depth based on the halfspace depth.
#'	}
#' In addition, two vectors of length \code{m} of the relative area of smallest depth values is returned:
#'	\itemize{
#'	\item \code{Simpl_IA} the proportions of points at which the depth \code{Simpl_ID} was attained,
#'	\item \code{Half_IA} the proportions of points at which the depth \code{Half_ID} was attained.
#'	}
#' The values \code{Simpl_IA} and \code{Half_IA} are always in the interval [0,1]. 
#' They introduce ranking also among functions having the same
#' infimal depth value - if two functions have the same infimal depth, the one with larger infimal area
#' \code{IA} is said to be less central. 	
#' For \code{order=2} and \code{m=1}, two additional matrices of pointwise depths are also returned:
#'	\itemize{
#'    \item \code{PSD} the matrix of size \code{d*d} containing the computed 
#'    pointwise bivariate simplicial depths used for the computation of \code{Simpl_FD} and \code{Simpl_ID},
#'    \item \code{PHD} the matrix of size \code{d*d} containing the computed 
#'    pointwise bivariate halfspace depths used for the computation of \code{Half_FD} and \code{Half_ID}.
#'	}
#' For \code{order=3}, only \code{Half_FD} and \code{Half_ID} are provided.
#' 
#' @references Nagy, S., Gijbels, I. and Hlubinka, D. (2016). 
#' Weak convergence of discretely observed functional data with applications. 
#' \emph{Journal of Multivariate Analysis}, \bold{146}, 46--62.
#' 
#' @references Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
#' Depth-based recognition of shape outlying functions. 
#' \emph{Journal of Computational and Graphical Statistics}, \bold{26:4}, 883--893.
#'
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#' depthf.fd1(datafA,datafB)
#' depthf.fd1(datafA,datafB,order=2)
#' depthf.fd1(datafA,datafB,order=3,approx=51)
#'
#' @seealso \code{\link{depthf.fd2}}, \code{\link{infimalRank}}

depthf.fd1 = function(datafA,datafB,range=NULL,d=101,order=1,approx=0){
  # univariate integrated depth
  # A functions whose depth I compute, M*D matrix
  # B functions wrt whose the depth is computed, N*D matrix
  # both 1dimensional, n*d, n nr of functions, d dimensionality
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)
  
  if(order==1){
    A1 = as.vector(A)
    B1 = as.vector(B)
    d = dim(A)[2]
    m = dim(A)[1]
    n = dim(B)[1]
    if(dim(B)[2]!=d) stop("dimension mismatch")
    FD = .Fortran("funD1",	
                  as.numeric(A1),		#	A
                  as.numeric(B1),		#	B
                  as.integer(m),		#	m
                  as.integer(n),		#	n
                  as.integer(d),		#	d
                  funsdep=as.numeric(rep(-1,m)),	
                  funhdep=as.numeric(rep(-1,m)),
                  fIsdep =as.numeric(rep(-1,m)),
                  fIhdep =as.numeric(rep(-1,m)),
                  IAsdep =as.integer(rep(-1,m)),
                  IAhdep =as.integer(rep(-1,m))
    )
    return(list(	Simpl_FD = FD$funsdep, Half_FD = FD$funhdep,
                 Simpl_ID = FD$fIsdep, Half_ID = FD$fIhdep, 
                 Simpl_IA = FD$IAsdep/d, Half_IA = FD$IAhdep/d ))
  }
  if(order==2) return(DiffDepth(A,B,approx))
  if(order==3) return(DiffDepth3D(A,B,approx))
}

#' @title Fast computation of the \eqn{L^2} metric for sets of functional data
#'
#' @description
#' Returns the matrix of \eqn{L^2} distances between two sets of functional data.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords metric functional
#'
#' @details
#' For two sets of functional data of sizes \code{m} and \code{n}
#' represented by matrices of their functional values on the common domain {1,...,d}, 
#' this function returns the symmetric matrix of size \code{m*n} whose entry in the
#' \code{i}-th row and \code{j}-th column is the approximated \eqn{L^2} distance of the 
#' \code{i}-th function from the first set, and the \code{j}-th function from the second set.
#' This function is utilized in the computation of the h-mode depth.
#'
#' @param A Functions of the first set, represented by a matrix of their functional values of 
#' size \code{m*d}. \code{m} stands for the number of functions, \code{d}
#' is the number of the equi-distant points {1,...,d} in the domain of the data [1,d] at which the functional
#' values of the \code{m} functions are evaluated.
#'
#' @param B Functions of the second set, represented by a matrix of their functional values of 
#' size \code{n*d}. \code{n} stands for the number of functions, \code{d}
#' is the number of the equi-distant points {1,...,d} in the domain of the data [1,d] at which the functional
#' values of the \code{n} functions are evaluated. The grid of observation points for the 
#' functions \code{A} and \code{B} must be the same.
#'
#' @return A symmetric matrix of the distances of the functions of size \code{m*n}.
#' 
#' @examples
#' datapop = dataf2rawfd(dataf.population()$dataf,range=c(1950,2015),d=66)
#' A = datapop[1:20,]
#' B = datapop[21:50,]
#' L2metric(A,B)
#'
#' @seealso \code{\link{depthf.hM}}
#' @seealso \code{\link{dataf2rawfd}}

L2metric = function(A,B){
  # computes fast approximation of L2 distance between fctions A and B
  M = .Fortran("metrl2",
               as.numeric(A),
               as.numeric(B),
               as.integer(m<-nrow(A)),
               as.integer(n<-nrow(B)),
               as.integer(d<-length(as.numeric(A))/nrow(A)),
               m = as.numeric(rep(-1,m*n)))$m
  return(M = matrix(M,nrow=m))
}

depthf.M = function(A,B,q=.2){
  # h-mode depth for the L2 metric
  mdist = L2metric(B,B)
  mdist2 = L2metric(A,B)
  hq2 = quantile(mdist[mdist>0],probs=q,type=1)	# probs=.2 as in Cuevas et al. (2007)
  return(rowSums(dnorm(mdist2/hq2)))
}

#' @title Fast computation of the uniform metric for sets of functional data
#'
#' @description
#' Returns the matrix of \eqn{C} (uniform) distances between two sets of functional data.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords metric functional
#'
#' @details
#' For two sets of functional data of sizes \code{m} and \code{n}
#' represented by matrices of their functional values, 
#' this function returns the symmetric matrix of size \code{m*n} whose entry in the
#' \code{i}-th row and \code{j}-th column is the approximated \eqn{C} (uniform) distance of the 
#' \code{i}-th function from the first set, and the \code{j}-th function from the second set.
#' This function is utilized in the computation of the h-mode depth.
#'
#' @param A Functions of the first set, represented by a matrix of their functional values of 
#' size \code{m*d}. \code{m} stands for the number of functions, \code{d}
#' is the number of the equi-distant points in the domain of the data at which the functional
#' values of the \code{m} functions are evaluated.
#'
#' @param B Functions of the second set, represented by a matrix of their functional values of 
#' size \code{n*d}. \code{n} stands for the number of functions, \code{d}
#' is the number of the equi-distant points in the domain of the data at which the functional
#' values of the \code{n} functions are evaluated. The grid of observation points for the 
#' functions \code{A} and \code{B} must be the same.
#'
#' @return A symmetric matrix of the distances of the functions of size \code{m*n}.
#' 
#' @examples
#' datapop = dataf2rawfd(dataf.population()$dataf,range=c(1950,2015),d=66)
#' A = datapop[1:20,]
#' B = datapop[21:50,]
#' Cmetric(A,B)
#'
#' @seealso \code{\link{depthf.hM}}
#' @seealso \code{\link{dataf2rawfd}}

Cmetric = function(A,B){
  # computes fast approximation of \eqn{C} distance between fctions A and B
  M = .Fortran("metrC",
               as.numeric(A),
               as.numeric(B),
               as.integer(m<-nrow(A)),
               as.integer(n<-nrow(B)),
               as.integer(d<-length(as.numeric(A))/nrow(A)),
               m = as.numeric(rep(-1,m*n)))$m
  return(M = matrix(M,nrow=m))
}

depthf.MC = function(A,B,q=.2){
  # h-mode depth for the \eqn{C} metric
  mdist = Cmetric(B,B)
  mdist2 = Cmetric(A,B)
  hq2 = quantile(mdist[mdist>0],probs=q,type=1)	# probs=.2 as in Cuevas et al. (2007)
  return(rowSums(dnorm(mdist2/hq2)))
}

#' @title h-mode depth for functional data
#'
#' @description
#' The h-mode depth of functional real-valued data.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth functional
#'
#' @details
#' The function returns the vectors of the sample h-mode depth values. The kernel 
#' used in the evaluation is the standard Gaussian kernel, the bandwidth value is chosen
#' as a quantile of the non-zero distances between the random sample curves.
#'
#' @param datafA Functions whose depth is computed, represented by a \code{dataf} object of their arguments
#'  and functional values. \code{m} stands for the number of functions.
#'
#' @param datafB Random sample functions with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a \code{dataf} object of their arguments
#'  and functional values. \code{n} is the sample size. 
#' The grid of observation points for the 
#' functions \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @param norm The norm used for the computation of the depth. Two possible 
#' choices are implemented: \code{C} for the uniform norm of continuous functions, 
#' and \code{L2} for the \eqn{L^2} norm of integrable functions.
#'
#' @param q The quantile used to determine the value of the bandwidth \eqn{h}
#' in the computation of the h-mode depth. \eqn{h} is taken as the \code{q}-quantile of
#' all non-zero distances between the functions \code{B}. By default, this value is set
#' to \code{q=0.2}, in accordance with the choice of Cuevas et al. (2007).
#'
#' @return A vector of length \code{m} of the h-mode depth values.
#' 
#' @references Cuevas, A., Febrero, M. and Fraiman, R.  (2007).
#' Robust estimation and classification for functional data via projection-based depth notions. 
#' \emph{Computational Statistics} \bold{22} (3), 481--496.
#' 
#' @references Nagy, S., Gijbels, I. and Hlubinka, D. (2016). 
#' Weak convergence of discretely observed functional data with applications. 
#' \emph{Journal of Multivariate Analysis}, \bold{146}, 46--62.
#'
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#' depthf.hM(datafA,datafB)
#' depthf.hM(datafA,datafB,norm="L2")

depthf.hM = function(datafA,datafB,range=NULL,d=101, norm = c("C","L2"), q=.2){
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)
  
  norm = match.arg(norm)
  switch(norm,
         C = depthf.MC(A,B,q),
         L2 = depthf.M(A,B,q)
  )
}

AdjBDKernL = function(b,v,J=3){
  # v is a matrix m x eval
  m = dim(v)[1]                                   # sample size
  eval = length(b)                                
  com = combn(m,J)
  poc = dim(com)[2]
  s = .Fortran("AdjLP",
               as.integer(eval),
               as.integer(J),
               as.integer(m),
               as.integer(poc),
               as.integer(as.vector(com)),
               as.numeric(b),
               as.numeric(v),
               dj = as.numeric(1))$dj
  return(s)
}

AdjBDL = function(b,v,J=3,K=1){
  eval = length(b)
  m = dim(v)[1]                                   # sample size
  sam = sample.int(m,m)                           # shuffle
  if (K==0) K=1                                   # make sure K!=0
  nk = m%/%K                                      # subsample size
  nlast = m %% K+nk                               # last subsample size
  Dpom = rep(NA,K)	   					# subsample depth
  if (K>1){
    for (k in 1:(K-1)) 
      if (eval>1) Dpom[k] = AdjBDKernL(b,v[sam[((k-1)*nk+1):(k*nk)],],J)
      else  Dpom[k] = AdjBDKernL(b,t(as.matrix(v[sam[((k-1)*nk+1):(k*nk)],])),J)
  }
  {if (nlast>0){	
    if (eval>1) Dpom[K]= AdjBDKernL(b,v[sam[((K-1)*nk+1):m],],J)
    else  Dpom[K]= AdjBDKernL(b,t(as.matrix(v[sam[((K-1)*nk+1):m],])),J)
  }
    else (K=K-1)}
  return(mean(Dpom[1:K]))
}

AdjBDsampleL = function(A,B,J=3,K=1){
  m = dim(A)[1]
  hlb = rep(NA,m)
  for (i in 1:m) hlb[i] = AdjBDL(A[i,],B,J,K)
  return(hlb)
}

AdjBDKernC = function(b,v,J=3){
  # v ma rozmery m x eval
  m = dim(v)[1]						# sample size
  eval = length(b)					
  com = combn(m,J)	
  poc = dim(com)[2]
  s = .Fortran("AdjC",
               as.integer(eval),
               as.integer(J),
               as.integer(m),
               as.integer(poc),
               as.integer(as.vector(com)),
               as.numeric(b),
               as.numeric(v),
               dj = as.numeric(1))$dj
  return(s)
}

AdjBDC = function(b,v,J=3,K=1){
  # v is a matrix m x eval
  eval = length(b)
  m = dim(v)[1]                                   # sample size
  sam = sample.int(m,m)                           # shuffle
  if (K==0) K=1                                   # make sure K!=0
  nk = m%/%K                                      # subsample size
  nlast = m %% K+nk                               # last subsample size
  Dpom = rep(NA,K)	   					# subsample depth
  if (K>1){
    for (k in 1:(K-1)) 
      if (eval>1) Dpom[k] = AdjBDKernC(b,v[sam[((k-1)*nk+1):(k*nk)],],J)
      else  Dpom[k] = AdjBDKernC(b,t(as.matrix(v[sam[((k-1)*nk+1):(k*nk)],])),J)
  }
  {if (nlast>0){
    if (eval>1) Dpom[K]= AdjBDKernC(b,v[sam[((K-1)*nk+1):m],],J)
    else  Dpom[K]= AdjBDKernC(b,t(as.matrix(v[sam[((K-1)*nk+1):m],])),J)
  }
    else (K=K-1)}
  return(mean(Dpom[1:K]))
}

AdjBDsampleC = function(A,B,J=3,K=1){
  m = dim(A)[1]
  hlb = rep(NA,m)
  for (i in 1:m) hlb[i] = AdjBDC(A[i,],B,J,K)
  return(hlb)
}

#' @title Adjusted band depth for functional data
#'
#' @description
#' The adjusted band depth 
#' of functional real-valued data based on either the
#' \eqn{C} (uniform) norm, or on the \eqn{L^2} norm of functions.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth functional
#'
#' @details
#' The function returns the vector of the sample adjusted band depth values. The kernel 
#' used in the evaluation is the function \eqn{K(u) = exp(-u)}.
#'
#' @param datafA Functions whose depth is computed, represented by a \code{dataf} object of their arguments
#'  and functional values. \code{m} stands for the number of functions.
#'
#' @param datafB Random sample functions with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a \code{dataf} object of their arguments
#'  and functional values. \code{n} is the sample size. 
#' The grid of observation points for the 
#' functions \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation, see Nagy et al. (2016).
#'
#' @param norm The norm used for the computation of the depth. Two possible 
#' choices are implemented: \code{C} for the uniform norm of continuous functions, 
#' and \code{L2} for the \eqn{L^2} norm of integrable functions. 
#'
#' @param J The order of the adjusted band depth, that is the maximal number of functions
#' taken in a band. Acceptable values are \code{2}, \code{3},... By default this value is set to \code{2}. 
#' Note that this is NOT the order as
#' defined in the order-extended version of adjusted band depths in Nagy et al. (2016), used
#' for the detection of shape outlying curves.
#' 
#' @param K Number of sub-samples of the functions from \code{B} taken to speed up the
#' computation. By default, sub-sampling is not performed. Values of \code{K} larger than \code{1}
#' result in an approximation of the adjusted band depth.
#'
#' @return A vectors of length \code{m} of the adjusted band depths.
#' 
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#' depthf.ABD(datafA,datafB)
#' depthf.ABD(datafA,datafB,norm="L2")
#'
#' @seealso \code{\link{depthf.BD}}
#'
#' @references Gijbels, I., Nagy, S. (2015).
#' Consistency of non-integrated depths for functional data.
#' \emph{Journal of Multivariate Analysis} \bold{140}, 259--282.
#' 
#' @references Nagy, S., Gijbels, I. and Hlubinka, D. (2016). 
#' Weak convergence of discretely observed functional data with applications. 
#' \emph{Journal of Multivariate Analysis}, \bold{146}, 46--62.
#' 
#' @references Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
#' Depth-based recognition of shape outlying functions. 
#' \emph{Journal of Computational and Graphical Statistics}, \bold{26:4}, 883--893.

depthf.ABD = function(datafA,datafB,range=NULL,d=101, norm = c("C","L2"), J=2, K=1){
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)
  
  norm = match.arg(norm)
  switch(norm,
         C = AdjBDsampleC(A,B,J,K),
         L2 = AdjBDsampleL(A,B,J,K)
  )
}

DiffDepth = function(A,B,approx=0){
  # A functions whose depth I compute
  # B functions wrt whose the depth is computed
  # both 1dimensional, n*d, n nr of functions, d dimensionality
  # approx is number of approximations to be used, 0 for full computation
  
  d = dim(A)[2]
  m = dim(A)[1]
  n = dim(B)[1]
  if(dim(B)[2]!=d) stop("dimension mismatch")
  Av = as.vector(A)
  Bv = as.vector(B)
  if (approx>0){
    C = combn(1:d,2)		# all couples 
    if (approx>=ncol(C)){ 
      ind.sample = 1:ncol(C) 
      approx = ncol(C) } else 
        ind.sample = sample.int(ncol(C),approx)
      RN = matrix(C[,ind.sample],nrow=2)		#	RN = random numbers of size 2*REP from 1:d
  } else RN = 0		
  FD = .Fortran("DiffD",	
                as.numeric(Av),		#	A
                as.numeric(Bv),		#	B
                as.integer(m),		#	m
                as.integer(n),		#	n
                as.integer(d),		#	d
                as.integer(approx),	#	REP
                as.integer(RN),		#	RN
                funsdep=as.numeric(rep(-1,m)),	
                funhdep=as.numeric(rep(-1,m)),
                funsdepm=as.numeric(rep(-1,m)),	
                funhdepm=as.numeric(rep(-1,m)),
                Psdep = as.numeric(rep(-1,d*d)),
                Phdep = as.numeric(rep(-1,d*d)),
                IAsdep =as.integer(rep(-1,m)),
                IAhdep =as.integer(rep(-1,m)))
  if(approx==0){
    S_IA = FD$IAsdep/(d^2)
    H_IA = FD$IAhdep/(d^2)
  } else {
    S_IA = FD$IAsdep/(approx)
    H_IA = FD$IAhdep/(approx)		
  }
  if (m>1) return(list(	Simpl_FD = FD$funsdep, Half_FD = FD$funhdep,
                        Simpl_ID = FD$funsdepm, Half_ID = FD$funhdepm,
                        Simpl_IA = S_IA, Half_IA = H_IA))
  if ((m==1)&(approx==0)){
    PSD = matrix(FD$Psdep,ncol=d)
    PSD = pmax(PSD,0)
    PSD = PSD + t(PSD)
    diag(PSD) = NA
    PHD = matrix(FD$Phdep,ncol=d)
    PHD = pmax(PHD,0)
    PHD = PHD + t(PHD)
    diag(PHD) = NA
    return(list(	Simpl_FD = FD$funsdep, Half_FD = FD$funhdep,
                 Simpl_ID = FD$funsdepm, Half_ID = FD$funhdepm,
                 PSD = PSD, PHD = PHD,
                 Simpl_IA = S_IA, Half_IA = H_IA))
  }
  if ((m==1)&(approx>0)){
    PSD = matrix(FD$Psdep,ncol=d)
    for(i in 1:approx) PSD[RN[2*i-1],RN[2*i]]=PSD[RN[2*i],RN[2*i-1]]	# symmetrize
    PSD[PSD==-1] = NA
    PHD = matrix(FD$Phdep,ncol=d)
    for(i in 1:approx) PHD[RN[2*i-1],RN[2*i]]=PHD[RN[2*i],RN[2*i-1]]	# symmetrize
    PHD[PHD==-1] = NA
    return(list(	Simpl_FD = FD$funsdep, Half_FD = FD$funhdep, 
                 Simpl_ID = FD$funsdepm, Half_ID = FD$funhdepm,
                 PSD = PSD, PHD = PHD,
                 Simpl_IA = S_IA, Half_IA = H_IA))
  }
}

DiffDepth3D = function(A,B,approx=101){
  
  d = dim(A)[2]
  m = dim(A)[1]
  n = dim(B)[1]
  if(dim(B)[2]!=d) stop("dimension mismatch")
  if(approx==0) approx=101
  DT = matrix(nrow=m,ncol=approx)
  for(a in 1:approx){
    I = sample.int(d,3)
    DT[,a] = depth.halfspace(A[,I],B[,I],exact=TRUE)  
  }
  D = apply(DT,1,mean)	# integrated depth
  DI = apply(DT,1,min)	# infimal depth
  IA = apply(DT,1,function(x) sum(x==min(x)))/approx	# infimal area
  return(list(Half_FD=D,Half_ID=DI,Half_IA=IA))
}

#' @title Bivariate h-mode depth for functional data based on the \eqn{L^2} metric
#'
#' @description
#' The h-mode depth 
#' of functional bivariate data (that is, data of the form \eqn{X:[a,b] \to R^2},
#' or \eqn{X:[a,b] \to R} and the derivative of \eqn{X}) based on the
#' \eqn{L^2} metric of functions.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth functional derivatives
#'
#' @details
#' The function returns the vectors of sample h-mode depth values. The kernel 
#' used in the evaluation is the standard Gaussian kernel, the bandwidth value is chosen
#' as a quantile of the non-zero distances between the random sample curves.
#'
#' @param datafA Bivariate functions whose depth is computed, represented by a multivariate \code{dataf} object of 
#' their arguments (vector), and a matrix with two columns of the corresponding bivariate functional values. 
#' \code{m} stands for the number of functions.
#'
#' @param datafB Bivariate random sample functions with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a multivariate \code{dataf} object of their arguments
#'  (vector), and a matrix with two columns of the corresponding bivariate functional values.
#' \code{n} is the sample size. The grid of observation points for the 
#' functions \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @param q The quantile used to determine the value of the bandwidth \eqn{h}
#' in the computation of the h-mode depth. \eqn{h} is taken as the \code{q}-quantile of
#' all non-zero distances between the functions \code{B}. By default, this value is set
#' to \code{q=0.2}, in accordance with the choice of Cuevas et al. (2007).
#'
#' @return Three vectors of length \code{m} of h-mode depth values are returned:
#'	\itemize{
#'	\item \code{hM} the unscaled h-mode depth,
#'	\item \code{hM_norm} the h-mode depth \code{hM} linearly transformed so that its range is [0,1],
#'	\item \code{hM_norm2} the h-mode depth \code{FD} linearly transformed by a transformation such that 
#'    the range of the h-mode depth of \code{B} with respect to \code{B} is [0,1]. This depth may give negative values.
#'	}
#'
#' @references Cuevas, A., Febrero, M. and Fraiman, R.  (2007). 
#' Robust estimation and classification for functional data via projection-based depth notions. 
#' \emph{Computational Statistics} \bold{22} (3), 481--496.
#' 
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#'
#' datafA2 = derivatives.est(datafA,deriv=c(0,1))
#' datafB2 = derivatives.est(datafB,deriv=c(0,1))
#' 
#' depthf.hM2(datafA2,datafB2)
#'
#' depthf.hM2(datafA2,datafB2)$hM
#' # depthf.hM2(cbind(A2[,,1],A2[,,2]),cbind(B2[,,1],B2[,,2]))$hM
#' # the two expressions above should give the same result
#'
#' @seealso \code{\link{depthf.hM}}

depthf.hM2 = function(datafA,datafB,range=NULL,d=101,q=.2){
  # h-Mode depth Cuevas_etal2007
  # A functions whose depth is computed :	either M*D matrix (M functions, D points per function), 
  #							or M*D*2 array (2 derivative levels), 
  # B functions of random sample :		as A
  # q :							quantile, bandwidth value in the resulting kernel estimate
  #							computes the same as depthf.M (with cbind-ed 2 levels if derivatives are involved)
  #							depthf.M(cbind(A[,,1],A[,,2]),cbind(B[,,1],B[,,2]))
  #							depthf.M2(A,B)$FD
  #							depthf.M2(cbind(A[,,1],A[,,2]),cbind(B[,,1],B[,,2]))$FD
  # 							all should give the same result
  
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)
  
  # q=.2 Cuevas et al, 0.15 default v usc.fda
  M = .Fortran(	"funMD",
                as.numeric(A),
                as.numeric(B),
                as.integer(m<-nrow(A)),
                as.integer(n<-nrow(B)),
                as.integer(d<-length(as.numeric(A))/nrow(A)),
                as.numeric(q),
                md = as.numeric(rep(-1,m))
  )$md
  #	because of scaling in depth.mode() in fda.usc
  M.ori = .Fortran(	"funMD",
                    as.numeric(B),
                    as.numeric(B),
                    as.integer(n),
                    as.integer(n),
                    as.integer(d),
                    as.numeric(q),
                    md = as.numeric(rep(-1,n))
  )$md
  Mn = (M-min(M))/(max(M)-min(M))
  Mn2 = (M-min(M.ori))/(max(M.ori)-min(M.ori))
  return(list(hM = M, hM_norm = Mn, hM_norm2 = Mn2))
}

#' @title Univariate random projection depths for functional data
#'
#' @description
#' Random projection depth and random functional depth for functional data.	
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth functional
#'
#' @details
#' The function returns the vectors of sample random projection, and random functional depth values. 
#' The random projection depth described in Cuevas et al. (2007) is based on the average univariate depth
#' of one-dimensional projections of functional data. The projections are taken randomly as a sample of standard
#' normal \code{d}-dimensional random variables, where \code{d} stands for the dimensionality of the discretized
#' functional data. 
#'
#' The random functional depth (also called random Tukey depth, or random halfspace depth) is described in
#' Cuesta-Albertos and Nieto-Reyes (2008). The functional data are projected into the real line in random 
#' directions as for the random projection depths. Afterwards, an approximation of the halfspace (Tukey) depth
#' based on this limited number of univariate projections is assessed. 
#'
#' @param datafA Functions whose depth is computed, represented by a \code{dataf} object of their arguments
#'  and functional values. \code{m} stands for the number of functions.
#'
#' @param datafB Random sample functions with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a \code{dataf} object of their arguments
#'  and functional values. \code{n} is the sample size. 
#' The grid of observation points for the 
#' functions \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @param nproj Number of projections taken in the computation of the random projection depth. By default taken
#' to be \code{51}.
#'
#' @param nproj2 Number of projections taken in the computation of the random functional depth. By default taken
#' to be \code{5}. \code{nproj2} should be much smaller than \code{d}, the dimensionality of the discretized 
#' functional data.
#'
#' @return Three vectors of depth values of length \code{m} are returned:
#'	\itemize{
#'	\item \code{Simpl_FD} the random projection depth based on the univariate simplicial depth,
#'	\item \code{Half_FD} the random projection depth based on the univariate halfspace depth,
#'	\item \code{RHalf_FD} the random halfspace depth.
#'	}
#'
#' @references Cuevas, A., Febrero, M. and Fraiman, R.  (2007).
#' Robust estimation and classification for functional data via projection-based depth notions, 
#' \emph{Computational Statistics} \bold{22} (3), 481--496.
#'
#' @references Cuesta-Albertos, J.A. and Nieto-Reyes, A. (2008).
#'  The random Tukey depth. 
#' \emph{Computational Statistics & Data Analysis} \bold{52} (11), 4979--4988.
#' 
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#'
#' depthf.RP1(datafA,datafB)
#'
#' @seealso \code{\link{depthf.RP2}}

depthf.RP1 = function(datafA,datafB,range=NULL,d=101,nproj=50,nproj2=5){
  # simple 1D projection depth
  # nproj nr of projections taken
  #
  #	SFD : projection \eqn{L^2} -> R -> mean of univariate Simplicial Depth (nproj projections)
  #	HFD : projection \eqn{L^2} -> R -> mean of univariate Halfspace Depth (nproj projections)
  #	RFD : projection \eqn{L^2} -> R -> infimum of univariate Halfspace Depths (nproj2 projections)
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)  
  
  A1 = as.vector(A)
  B1 = as.vector(B)
  d = dim(A)[2]
  m = dim(A)[1]
  n = dim(B)[1]
  V = rnorm(nproj*d)
  if(dim(B)[2]!=d) stop("dimension mismatch")
  FD = .Fortran("funRPD1",	
                as.numeric(A1),		#	A
                as.numeric(B1),		#	B
                as.integer(m),		#	m
                as.integer(n),		#	n
                as.integer(d),		#	d
                as.integer(nproj),	#	nproj
                as.integer(nproj2),	#	nproj2
                as.numeric(V),		#	projections
                funsdep=as.numeric(rep(-1,m)),	
                funhdep=as.numeric(rep(-1,m)),
                funrdep=as.numeric(rep(-1,m)))
  return(list(Simpl_FD = FD$funsdep, Half_FD = FD$funhdep, RHalf_FD = FD$funrdep))
}

#' @title Bivariate integrated and infimal depth for functional data
#'
#' @description
#' Integrated and infimal depths 
#' of functional bivariate data (that is, data of the form \eqn{X:[a,b] \to R^2},
#' or \eqn{X:[a,b] \to R} and the derivative of \eqn{X}) based on the
#' bivariate halfspace and simplicial depths.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth functional derivatives
#'
#' @details
#' The function returns the vectors of sample integrated and infimal depth values.
#'
#' @param datafA Bivariate functions whose depth is computed, represented by a multivariate \code{dataf} object of 
#' their arguments (vector), and a matrix with two columns of the corresponding bivariate functional values. 
#' \code{m} stands for the number of functions.
#'
#' @param datafB Bivariate random sample functions with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a multivariate \code{dataf} object of their arguments
#'  (vector), and a matrix with two columns of the corresponding bivariate functional values.
#' \code{n} is the sample size. The grid of observation points for the 
#' functions \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @return Four vectors of length \code{m} are returned:
#'	\itemize{
#'	\item \code{Simpl_FD} the integrated depth based on the bivariate simplicial depth,
#'	\item \code{Half_FD} the integrated depth based on the bivariate halfspace depth,
#'	\item \code{Simpl_ID} the infimal depth based on the bivariate simplicial depth,
#'	\item \code{Half_ID} the infimal depth based on the bivariate halfspace depth.
#'	}
#' In addition, two vectors of length \code{m} of the relative area of smallest depth values is returned:
#'	\itemize{
#'	\item \code{Simpl_IA} the proportions of points at which the depth \code{Simpl_ID} was attained,
#'	\item \code{Half_IA} the proportions of points at which the depth \code{Half_ID} was attained.
#'	}
#' The values \code{Simpl_IA} and \code{Half_IA} are always in the interval [0,1]. 
#' They introduce ranking also among functions having the same
#' infimal depth value - if two functions have the same infimal depth, the one with larger infimal area
#' \code{IA} is said to be less central. 	
#'
#' @references Hlubinka, D., Gijbels, I., Omelka, M. and Nagy, S. (2015). 
#' Integrated data depth for smooth functions and its application in supervised classification. 
#' \emph{Computational Statistics}, \bold{30} (4), 1011--1031.
#'
#' @references Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
#' Depth-based recognition of shape outlying functions. 
#' \emph{Journal of Computational and Graphical Statistics}, \bold{26:4}, 883--893.
#'  
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#'
#' dataf2A = derivatives.est(datafA,deriv=c(0,1))
#' dataf2B = derivatives.est(datafB,deriv=c(0,1))
#' depthf.fd2(dataf2A,dataf2B)
#'
#' @seealso \code{\link{depthf.fd1}}, \code{\link{infimalRank}}

depthf.fd2 = function(datafA,datafB,range=NULL,d=101){
  # A functions whose depth I compute
  # B functions wrt whose the depth is computed
  # both 2dimensional, n*d*2, n nr of functions, d dimensionality
  # now provides also infimal depth (inf_D2)
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)
  
  A1 = as.vector(A[,,1])
  A2 = as.vector(A[,,2])
  B1 = as.vector(B[,,1])
  B2 = as.vector(B[,,2])
  d = dim(A)[2]
  m = dim(A)[1]
  n = dim(B)[1]
  if(dim(B)[2]!=d) stop("dimension mismatch")
  FD = .Fortran("funD2",	
                as.numeric(A1),		#	A1
                as.numeric(A2),		#	A2
                as.numeric(B1),		#	B1
                as.numeric(B2),		#	B2
                as.integer(m),		#	m
                as.integer(n),		#	n
                as.integer(d),		#	d
                funsdep=as.numeric(rep(-1,m)),	
                funhdep=as.numeric(rep(-1,m)),
                fIsdep =as.numeric(rep(-1,m)),
                fIhdep =as.numeric(rep(-1,m)),
                IAsdep =as.integer(rep(-1,m)),
                IAhdep =as.integer(rep(-1,m))
  )
  return(list(	Simpl_FD = FD$funsdep, Half_FD = FD$funhdep,
               Simpl_ID = FD$fIsdep,	Half_ID = FD$fIhdep,
               Simpl_IA = FD$IAsdep/d, Half_IA = FD$IAhdep/d ))
}

#' @title Bivariate random projection depths for functional data
#'
#' @description
#' Double random projection depths of functional bivariate data (that is, data of the form \eqn{X:[a,b] \to R^2},
#' or \eqn{X:[a,b] \to R} and the derivative of \eqn{X}).	
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth functional derivatives
#'
#' @details
#' The function returns the vectors of sample double random projection depth values. 
#' The double random projection depths are described in Cuevas et al. (2007). They are of two types: RP2 type, and
#' RPD type. Both types of depths are based on bivariate projections of the bivariate functional data. 
#' These projections are taken randomly as a sample of standard
#' normal \code{d}-dimensional random variables, where \code{d} stands for the dimensionality of the internally 
#' represented discretized
#' functional data. For RP2 type depths, the average bivariate depth of the projected quantities is assessed.
#' For RPD type depths, further univariate projections of these bivariate projected quantities are evaluated, and
#' based on these final univariate quantities, the average univariate depth is computed.
#' 
#' @param datafA Bivariate functions whose depth is computed, represented by a multivariate \code{dataf} object of 
#' their arguments (vector), and a matrix with two columns of the corresponding bivariate functional values. 
#' \code{m} stands for the number of functions.
#'
#' @param datafB Bivariate random sample functions with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a multivariate \code{dataf} object of their arguments
#'  (vector), and a matrix with two columns of the corresponding bivariate functional values.
#' \code{n} is the sample size. The grid of observation points for the 
#' functions \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @param nproj Number of projections taken in the computation of the double random projection depth. By default taken
#' to be \code{51}.
#'
#' @return Five vectors of length \code{m} are returned:
#'	\itemize{
#'	\item \code{Simpl_FD} the double random projection depth RP2 based on the bivariate simplicial depth,
#'	\item \code{Half_FD} the double random projection depth RP2 based on the bivariate halfspace depth,
#'	\item \code{hM_FD} the double random projection depth RP2 based on the bivariate h-mode depth,
#'	\item \code{Simpl_DD} the double random projection depth RPD based on the univariate simplicial depth,
#'	\item \code{Half_DD} the random projection depth RPD based on the univariate halfspace depth,
#'	}
#'
#' @references Cuevas, A., Febrero, M. and Fraiman, R.  (2007).
#' Robust estimation and classification for functional data via projection-based depth notions.
#' \emph{Computational Statistics} \bold{22} (3), 481--496.
#' 
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#'
#' dataf2A = derivatives.est(datafA,deriv=c(0,1))
#' dataf2B = derivatives.est(datafB,deriv=c(0,1))
#' depthf.RP2(dataf2A,dataf2B)
#'
#'
#' @seealso \code{\link{depthf.RP1}}

depthf.RP2 = function(datafA,datafB,range=NULL,d=101,nproj=51){
  # double projection depth
  # nproj nr of projections taken
  #
  #	SFD :	(\eqn{L^2})^2 -> R^2 -> 2D Mean Simplicial Depth
  #	HFD : (\eqn{L^2})^2 -> R^2 -> 2D Mean Halfspace Depth
  #	MFD : (\eqn{L^2})^2 -> R^2 -> 2D Mean h-Mode Depth
  #	SDD : (\eqn{L^2})^2 -> R^2 -> R^1 -> Mean Simplicial Depth
  #	HDD : (\eqn{L^2})^2 -> R^2 -> R^1 -> Mean Halfspace Depth
  #
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)
  
  q = .2	# quantile for modal depth
  A1 = as.vector(A[,,1])
  A2 = as.vector(A[,,2])
  B1 = as.vector(B[,,1])
  B2 = as.vector(B[,,2])
  d = dim(A)[2]
  m = dim(A)[1]
  n = dim(B)[1]
  V = rnorm(nproj*d+nproj*2)
  if(dim(B)[2]!=d) stop("dimension mismatch")
  FD = .Fortran("funRPD2",	
                as.numeric(A1),		#	A1
                as.numeric(A2),		#	A2
                as.numeric(B1),		#	B1
                as.numeric(B2),		#	B2
                as.integer(m),		#	m
                as.integer(n),		#	n
                as.integer(d),		#	d
                as.integer(nproj),	#	nproj
                as.numeric(V),		#	projections
                as.numeric(q),		# 	q
                funsdep=as.numeric(rep(-1,m)),	
                funhdep=as.numeric(rep(-1,m)),
                funmdep=as.numeric(rep(-1,m)),
                funsddep=as.numeric(rep(-1,m)),
                funhddep=as.numeric(rep(-1,m)))
  return(list(Simpl_FD = FD$funsdep, Half_FD = FD$funhdep, hM_FD = FD$funmdep, Simpl_DD = FD$funsddep, Half_DD = FD$funhddep))
}

#' @title Half-region depth for functional data
#'
#' @description
#' The half-region depth 
#' for functional real-valued data.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth functional
#'
#' @details
#' The function returns the vector of the sample half-region depth values.
#'
#' @param datafA Functions whose depth is computed, represented by a \code{dataf} object of their arguments
#'  and functional values. \code{m} stands for the number of functions.
#'
#' @param datafB Random sample functions with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a \code{dataf} object of their arguments
#'  and functional values. \code{n} is the sample size. 
#' The grid of observation points for the 
#' functions \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @return A vector of length \code{m} of the half-region depth values.
#'
#' @references Lopez-Pintado, S. and Romo, J. (2011).
#' A half-region depth for functional data.
#' \emph{Computational Statistics & Data Analysis} \bold{55} (4), 1679--1695.
#' 
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#' depthf.HR(datafA,datafB)

depthf.HR = function(datafA,datafB,range=NULL,d=101){
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)
  
  d = dim(A)[2]
  m = dim(A)[1]
  n = dim(B)[1]
  if(dim(B)[2]!=d) stop("dimension mismatch")
  FD = .Fortran("HRD",	
                as.numeric(A),		#	A
                as.numeric(B),		#	B
                as.integer(m),		#	m
                as.integer(n),		#	n
                as.integer(d),		#	d
                FD=as.numeric(rep(-1,m)))
  return(FD$FD)
}

#' @title Band depth for functional data
#'
#' @description
#' The (unadjusted) band depth 
#' for functional real-valued data of order \code{J=2}.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth functional
#'
#' @details
#' The function returns the vector of the sample (unadjusted) band depth values.
#'
#' @param datafA Functions whose depth is computed, represented by a \code{dataf} object of their arguments
#'  and functional values. \code{m} stands for the number of functions.
#'
#' @param datafB Random sample functions with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a \code{dataf} object of their arguments
#'  and functional values. \code{n} is the sample size. 
#' The grid of observation points for the 
#' functions \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @return A vector of length \code{m} of the band depth values.
#' 
#' @references Lopez-Pintado, S. and Romo, J. (2009), On the concept of depth for functional data,
#' \emph{J. Amer. Statist. Assoc.} \bold{104} (486), 718 - 734.
#'
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#' depthf.BD(datafA,datafB)
#'
#' @seealso \code{\link{depthf.ABD}}, \code{\link{depthf.fd1}}

depthf.BD = function(datafA,datafB,range=NULL,d=101){
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)  
  
  d = dim(A)[2]
  m = dim(A)[1]
  n = dim(B)[1]
  if(dim(B)[2]!=d) stop("dimension mismatch")
  FD = .Fortran("BD",	
                as.numeric(A),		#	A
                as.numeric(B),		#	B
                as.integer(m),		#	m
                as.integer(n),		#	n
                as.integer(d),		#	d
                FD=as.numeric(rep(-1,m))
  )
  return(FD$FD)
}

#' @title Adjusted ranking of functional data based on the infimal depth
#'
#' @description
#' Returns a vector of adjusted depth-based ranks for infimal depth for functional data.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords rank depth functional
#'
#' @details
#' Infimal depths for functional data tend to give to many functional observations the same 
#' value of depth. Using this function, the data whose depth is the same is ranked according
#' to the infimal area indicator. This indicator is provided in functions \code{depthf.fd1} along
#' the value of the infimal depth. 
#'
#' @param ID The vector of infimal depths of the curves of length \code{n}.
#'
#' @param IA The vector of the infimal areas corresponding to the infimal depths from \code{ID}
#' of length \code{n}.
#'
#' @param ties.method Parameter for breaking ties in infimal area index. By default \code{max}, see 
#' \code{rank}.
#' 
#' @return A vector of length \code{n}. Low depth values mean high ranks, i.e. potential outlyingess. 
#' If some of the infimal depths are identical, the ranking of these functions is made according to the 
#' values of the infimal area. There, higher infimal area index means higher rank, i.e. non-centrality.
#' 
#' @references Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
#' Depth-based recognition of shape outlying functions. 
#' \emph{Journal of Computational and Graphical Statistics}, \bold{26:4}, 883--893. 
#' 
#' @examples
#' datafA = dataf.population()$dataf[1:20]
#' datafB = dataf.population()$dataf[21:50]
#' D = depthf.fd1(datafA,datafB)
#' infimalRank(D$Half_ID,D$Half_IA) 
#' 
#' ID = c(0,1,0,0,0,1,1)
#' IA = c(2,3,1,0,2,4,1)
#' infimalRank(ID,IA)


infimalRank = function(ID,IA,ties.method="max"){
  # finds the adjusted rank for appropriate for the infimal depth for functional data
  # ID is the vector of infimal depths
  # IA is the vector of the corresponding infimal areas
  # returns a vector of ranks
  n = length(ID)
  if(length(IA)!=n) stop("Lengths of the vectors differ")
  U = sort(unique(ID),decreasing=TRUE)
  R = rep(NA,n)
  cR = 0						# currently assigned rank	
  for(u in U){
    Iu = (ID==u)
    R[Iu] = cR+rank(IA[Iu],ties.method=ties.method)
    cR = sum(!is.na(R))
  }
  return(R)
}

#' @title Estimation of the first two derivatives for functional data
#'
#' @description
#' Returns the estimated values of derivatives of functional data.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords derivatives kernel functional
#' 
#' @details
#' If the input \code{dataf} is a functional random sample of size \code{m}, 
#' the function returns a \code{dataf} object of \code{nd}-dimensional functional data, where 
#' in the elements of the vector-valued functional data represent the estimated values of the 
#' derivatives of \code{dataf}. All derivatives are evaluated at an equi-distant grid of \code{d}
#' points in the domain given by \code{range}. \code{nd} here stands for \code{1}, \code{2} or \code{3}, 
#' depending on how many derivatives of \code{dataf} are
#' requested to be computed. For the estimation, functions \code{D1ss} and \code{D2ss} from the package
#' \code{sfsmisc} are utilized. 
#'
#' @param dataf Functional dataset, represented by a \code{dataf} object of their arguments
#'  and functional values. \code{m} stands for the number of functions.
#'  
#' @param range The common range of the domain where the fucntions \code{dataf} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{dataf}.
#' 
#' @param d Grid size to which all the functional data are transformed. For computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @param spar If provided, this parameter is passed to functions \code{D1ss} and \code{D2ss} from package \code{sfsmisc}
#' as the value of the smoothing spline parameter in order to numerically approximate
#' the derivatives of \code{dataf}.
#' 
#' @param deriv A vector composed of \code{0}, \code{1}, and \code{2} of the demanded 
#' functional values / derivatives of the functions in the rows of \code{dataf}.
#' \code{0} stands for the functional values, \code{1} for the first derivatives, 
#' \code{2} for the second derivatives.
#' 
#' @return A multivariate \code{dataf} object of the functional values and / or the derivatives of \code{dataf}. 
#' The dimensionality of the vector-valued functional data is \code{nd}. The arguments of the data are all equal to 
#' an equi-distant grid of \code{d} points in the domain given by \code{range}. \code{nd} is the demanded number 
#' of derivatives at the output, i.e. the length of the vector \code{deriv}.
#' 
#' @seealso \code{\link[sfsmisc]{D1ss}} in package sfsmisc
#' @seealso \code{\link[sfsmisc]{D2ss}} in package sfsmisc
#' 
#' @examples
#' dataf = dataf.population()$dataf
#' derivatives.est(dataf,deriv=c(0,1,2))

derivatives.est = function(dataf,range=NULL,d=101,spar=NULL,deriv=c(0,1)){
  
  X = dataf2rawfd(dataf, range = range, d = d)
  derd = length(deriv)
  XK = array(dim=c(dim(X),derd))
  # range construction
  rng = numeric(0)
  for (df in dataf)	rng = range(c(rng,df$args)) # common range of all the data
  if(!is.null(range)){ 
    if(!(length(range) == 2 && is.numeric(range) && 
         range[1]<=rng[1] && range[2]>=rng[2])) 
      stop("Argument 'range' must be a numeric vector of two components
           that defines the range of the domain of functional data. All
           functional data must have 'args' vales inside this domain.")
  } else range = rng
  if(!(range[1]<range[2])) stop("Argument 'range' must define a non-degenerate interval.")
  
  t = seq(range[1],range[2],length=d)
  n = nrow(X)	
  pb = txtProgressBar(min = 0, max = n, style = 3)	
  for (nd in 1:derd){
    if (deriv[nd]==0) XK[,,nd] = X else
      for(i in 1:n){
        if (deriv[nd]==1){
          if(is.null(spar))	XK[i,,nd] = D1ss(t,X[i,]) else
            XK[i,,nd] = D1ss(t,X[i,],spl.spar=spar)
        }
        if (deriv[nd]==2){
          if(is.null(spar))	XK[i,,nd] = D2ss(t,X[i,])$y else
            XK[i,,nd] = D2ss(t,X[i,],spl.spar=spar)$y
        }
        setTxtProgressBar(pb, i)
      }
  }
  close(pb)
  return(rawfd2dataf(XK,range=range))
  }

#' @title Diagnostic plot for first and second order integrated and infimal depths
#'
#' @description
#' Produce the diagnostic plot based on the fist or second order extended integrated / infimal depths.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth shape outlier plot functional
#'
#' @details
#' Plots a diagnostic plot of pointwise univariate (or bivariate) depths for all possible points (or couples of points) from the domain of the 
#' functional data. From such a plot it is possible to infer into the first order (or second order) properties of a single function \emph{x} with respect 
#' to the given set of functional data. For \code{order=1}, the integral of the displayed function is the integrated depth of \emph{x}, 
#' the smallest value of the function is the infimal depth of \emph{x}. 
#' For \code{order=2}, the bivariate integral of the displayed surface gives the second order extended 
#' integrated depth of \emph{x}, the infimum of this bivariate function gives the second order infimal depth of \emph{x}. 
#' For details see Nagy et al. (2016) and \code{\link{depthf.fd1}}.
#'
#' @param datafA A single function whose depth is computed, represented by a 
#' \code{dataf} object of arguments and functional values.
#'
#' @param datafB Functional dataset with respect to which the depth of \code{datafA} is computed. 
#' \code{datafB} is represented by a \code{dataf} object of arguments and functional values. 
#' \code{n} stands for the number of functions. The grid of observation points for the 
#' functions in \code{datafA} and \code{datafB} may not be the same.
#' 
#' @param range The common range of the domain where the fucntions \code{datafA} and \code{datafB} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{datafA} and \code{datafB}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @param order The order of the depth to be used in the plot, for \code{order=1} produces
#' the plot of univariate marginal depth of \code{A} and \code{nfun} functions from \code{B} 
#' over the domain of the functions. For \code{order=2} produces the bivariate contour plot 
#' of the bivariate depths of \code{A} at couples of points from the domain.
#'
#' @param method The depth that is used in the diagnostic plot. possible values are \code{halfspace} for 
#' the halfspace depth, or \code{simplicial} for the simplicial depth. 
#'
#' @param approx For \code{order=2}, the number of approximations used in the computation of the order extended depth. By default
#' this is set to \code{0}, meaning that the depth is computed at all possible \code{d^2}
#' combinations of the points in the domain. When set to a positive integer, \code{approx}
#' bivariate points are randomly sampled in unit square, and at these points the bivariate depths of the
#' corresponding functional values are computed.
#'
#' @param title The title of the diagnostic plot.
#'
#' @param nfun For \code{order=1}, the number of functions from \code{B} whose coordinate-wise
#' univariate depths of functional values should be displayes with the depth of \code{A}.
#' The depth of \code{A} is displayed in solid red line, the depths of the functions from \code{B}
#' in dashed black.
#'
#' @param plot Logical: should the function by plotted?
#' 
#' @return For \code{order=1} two depth values, and two vectors of pointwise depths:
#'	\itemize{
#'	\item \code{Simpl_FD} the first order integrated depth based on the simplicial depth,
#'	\item \code{Half_FD} the first order integrated depth based on the halfspace depth,
#'	\item \code{Simpl_ID} the first order infimal depth based on the simplicial depth,
#'	\item \code{Half_ID} the first order infimal depth based on the halfspace depth,
#'    \item \code{PSD} the vector of length \code{d} containing the computed 
#'    pointwise univariate simplicial depths used for the computation of \code{Simpl_FD} and \code{Simpl_ID},
#'    \item \code{PHD} the vector of length \code{d} containing the computed 
#'    pointwise univariate halfspace depths used for the computation of \code{Half_FD} and \code{Half_ID}.
#'	}
#'    In addition, the first order integrated / infimal depth diagnostic plot of the function \code{A} with respect to
#'    the random sample given by the functions corresponding to the rows of the matrix \code{B} is produced.
#'
#'    For \code{order=2} four depth values, and two matrices of pointwise depths:
#'	\itemize{
#'	\item \code{Simpl_FD} the second order integrated depth based on the simplicial depth,
#'	\item \code{Half_FD} the second order integrated depth based on the halfspace depth,
#'	\item \code{Simpl_ID} the second order infimal depth based on the simplicial depth,
#'	\item \code{Half_ID} the second order infimal depth based on the halfspace depth,
#'    \item \code{PSD} the matrix of size \code{d*d} containing the computed 
#'    pointwise bivariate simplicial depths used for the computation of \code{Simpl_FD} and \code{Simpl_ID},
#'    \item \code{PHD} the matrix of size \code{d*d} containing the computed 
#'    pointwise bivariate halfspace depths used for the computation of \code{Half_FD} and \code{Half_ID}.
#'	}
#'    In addition, the second order integrated / infimal depth diagnostic plot of the function \code{A} with respect to
#'    the random sample given by the functions corresponding to the rows of the matrix \code{B} is produced.
#' 
#' @references Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
#' Depth-based recognition of shape outlying functions. 
#' \emph{Journal of Computational and Graphical Statistics}, \bold{26:4}, 883--893.
#'
#' @examples
#' datafA = dataf.population()$dataf[1]
#' dataf = dataf.population()$dataf[2:20]
#' shape.fd.analysis(datafA,dataf,order=1)
#' shape.fd.analysis(datafA,dataf,order=2,approx=0)
#'
#' @seealso \code{\link{depthf.fd1}}

shape.fd.analysis = function(datafA,datafB,range=NULL,d=101,order=1,method=c("halfspace","simplicial"),approx=0,title="",nfun=10,plot=TRUE){
  
  A = dataf2rawfd(datafA, range = range, d = d)
  B = dataf2rawfd(datafB, range = range, d = d)  
  
  if (nrow(A)>1) stop("works only for a single function A (matrix 1*d)")
  method = match.arg(method)
  #
  # diagonal
  d = ncol(A)
  HDP = matrix(nrow=nrow(B)+1,ncol=d)
  SDP = HDP
  for(i in 1:d){
    FL = ecdf(B[,i])
    FU = ecdf(-B[,i])
    HDP[,i] = pmin(FL(rbind(B,A)[,i]),FU(-rbind(B,A)[,i]))
    SDP[,i] = FL(rbind(B,A)[,i])*FU(-rbind(B,A)[,i])*nrow(B)^2/choose(nrow(B),2)
  }
  if(order==2){
    DD = DiffDepth(A,B,approx)
    if (method=="halfspace") diag(DD$PHD) = HDP[nrow(B)+1,] else diag(DD$PSD) = SDP[nrow(B)+1,]
    if (plot){
      if (method=="halfspace") filled.contour(DD$PHD,main=title,color.palette= function(x)rev(heat.colors(x))) else
        filled.contour(DD$PSD,main=title,color.palette = function(x)rev(heat.colors(x)))
    }
    return(DD)
  }
  if(order==1){
    nfun = min(nrow(B),nfun)
    t = seq(0,1,length=ncol(B))
    DD = list(Simpl_FD = mean(SDP[nrow(B)+1,]), Half_FD = mean(HDP[nrow(B)+1,]), 
              Simpl_ID = min(SDP[nrow(B)+1,]), Half_ID = min(HDP[nrow(B)+1,]), 
              PHD = HDP[nrow(B)+1,], PSD = SDP[nrow(B)+1,]) 
    if (plot) { if (method=="halfspace"){
      plot(rep(t,nrow(B)+1),HDP,type="n",main=title,ylim = c(0,max(HDP)),ann=FALSE)
      for(i in 1:nfun) lines(t,HDP[i,],lwd=.75,lty=2)
      lines(t,HDP[nrow(B)+1,],col=2,lwd=3,lty=1)
    } else {
      plot(rep(t,nrow(B)+1),SDP,type="n",main=title,ylim = c(0,max(SDP)),ann=FALSE)
      for(i in 1:nfun) lines(t,SDP[i,],lwd=.75,lty=2)
      lines(t,SDP[nrow(B)+1,],col=2,lwd=3,lty=1)
    }
    }	
    return(DD)
  }
}

#' @title Functional depth-based shape outlier detection
#'
#' @description
#' Detects functional outliers of first three orders, based on the order extended integrated depth for functional data.		
#'
#' @author Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
#' @keywords depth
#' @keywords outlier
#' @keywords functional
#'
#' @details
#' Using the procedure described in Nagy et al. (2016), the function uses the order extended integrated depths for functions, 
#' see \code{\link{depthf.fd1}} and \code{\link{shape.fd.analysis}}, to perform informal functional shape outlier detection. 
#' Outliers of the first order (horizontal shift outliers) are found as the functions with \code{q} \% of smallest (first order)
#' integrated depth values. Second and third order outliers (shape outliers) are found using the extension of the boxplot method
#' for depths as described in the paper Nagy et al. (2016).
#'
#' @param dataf Functional dataset, represented by a \code{dataf} object of their arguments
#'  and functional values. \code{n} stands for the number of functions.
#'  
#' @param range The common range of the domain where the fucntions \code{dataf} are observed.
#' Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
#' \code{dataf}.
#' 
#' @param d Grid size to which all the functional data are transformed. For depth computation, 
#' all functional observations are first transformed into vectors of their functional values of length \code{d}
#' corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
#' points are reconstructed using linear interpolation, and extrapolation.
#'
#' @param q The quantile presenting a threshold for the first order outlier detection. Functions with first order integrated depth
#' smaller than the \code{q} quantile of this sample of depths are flagged as potential outliers. If set to \code{NULL}, the
#' the outliers are detected from the first order integrated depth after the log-transformation, as for higher order outliers.
#'
#' @param method The depth that is used in the diagnostic plot. possible values are \code{halfspace} for 
#' the halfspace depth, or \code{simplicial} for the simplicial depth. 
#'
#' @param approx For the computation of the third order integrated depth,
#' the number of approximations used in the computation of the order extended depth. By default
#' this is set to \code{100}, meaning that \code{100}
#' trivariate points are randomly sampled in unit cube, and at these points the trivariate depths of the
#' corresponding functional values. May be set to \code{0} to compute the depth at all possible \code{d^3}
#' combinations of the points in the domain. This choice may result in very slow computation, see also \code{\link{depthf.fd1}}. 
#'
#' @param print If the rows of \code{X} are named, \code{print=TRUE} enables a graphical output when the names of the outlying curves
#' are displayed.
#'
#' @param plotpairs If set to \code{TRUE}, the scatter plot of the computed depths for orders \code{1}, \code{2} and \code{3} is
#' is displayed. Here, the depths corresponding to the flagged outliers are plotted in colour.
#'
#' @param max.order Maximal order of shape outlyingness to be computed, can be set to \code{1}, \code{2}, or \code{3}.
#'
#' @param exclude.out Logical variable; exclude the detected lower order outliers in the flagging process? By default \code{TRUE}.
#'
#' @param output Output method, can be set to \code{matrix} for a matrix with logical entries (\code{TRUE} for outliers), or \code{list} for 
#' a list of outliers.
#' 
#' @param identifiers A vector of names for the data observations. Facilitates identification of outlyig functions.
#' 
#' @return A matrix of logical values of size \code{n*4}, where \code{n} is the sample size. In the first three rows indicators of outlyingness
#' of the corresponding functions for orders \code{1}, \code{2} and \code{3} are given, in the fourth row the indicator of outlyingness
#' with respect to the comparison of the first, and third order depths is given. That is, the fist row corresponds to the first order outliers, 
#' the second row to the second order outliers, and the last two rows formally to the third order outliers. Please consult Nagy et al. (2016)
#' to interpret the notion of shape outlyingness.
#' 
#' @references Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
#' Depth-based recognition of shape outlying functions. 
#' \emph{Journal of Computational and Graphical Statistics}, \bold{26:4}, 883--893.
#'
#' @examples
#' n = 30
#' dataf = dataf.population()$dataf[1:n]
#' shape.fd.outliers(dataf,print=TRUE,plotpairs=TRUE,
#' identifiers=unlist(dataf.population()$identifier)[1:n])
#'
#' @seealso \code{\link{depthf.fd1}}, \code{\link{shape.fd.analysis}}

shape.fd.outliers = function(dataf,range=NULL,d=101,q=.05,method=c("halfspace","simplicial"),
                             approx=100,print=FALSE,plotpairs=FALSE,max.order=3,
                             exclude.out = TRUE, output=c("matrix","list"), identifiers = NULL){
  
  X = dataf
  
  method = match.arg(method)
  output = match.arg(output)
  if(is.null(identifiers) | length(identifiers)!=length(dataf)){
    print = FALSE
    warning("Inconsistent identifiers, print is set to FALSE")
  }
  if((max.order>3)|(max.order<1)){
    max.order=3
    warning("Maximal order set to 3")
  }
  if (max.order==3){
    if(approx<50){
      approx=50
      warning("Too small approx value, approximation set to 50")
    }
  }
  n = length(X)									# set up the depths
  D = matrix(nrow=max.order,ncol=n)
  if(method=="halfspace"){
    D[1,] = depthf.fd1(X,X)$Half_FD
    if(max.order>1) D[2,] = depthf.fd1(X,X,range=range,d=d,order=2)$Half_FD
    if(max.order>2) D[3,] = depthf.fd1(X,X,range=range,d=d,order=3,approx=approx)$Half_FD
  }
  if(method=="simplicial"){
    D[1,] = depthf.fd1(X,X)$Simpl_FD
    if(max.order>1) D[2,] = depthf.fd1(X,X,range=range,d=d,order=2)$Simpl_FD
    if(max.order>2) D[3,] = depthf.fd1(X,X,range=range,d=d,order=3,approx=approx)$Simpl_FD
  }
  D = apply(D,1:2,function(x) max(0,x-1/n))
  if(max.order==1) out.rows = 1
  if(max.order==2) out.rows = 2
  if(max.order==3) out.rows = 4
  O = matrix(nrow=out.rows,ncol=ncol(D))					# compute the outliers
  if (print){
    colnames(O) = identifiers
    print("first order outliers: ")
  }
  if(!is.null(q)) O[1,]<-D[1,]<quantile(D[1,],q)
  if(is.null(q)){                  # for q null perform the log-transform
    S = -log(D[1,]/1)
    S[D[1,]==0] = Inf
    B = boxplot(S,plot=FALSE)
    O[1,]<-((S>B$stats[5,])|(S==Inf))            
  } 	
  if (print) print(which(O[1,]))
  if(max.order>1) for(i in 1:(max.order-1)){
    if (print){
      if (i==1) print("second order outliers: ")
      if (i==2) print("third order outliers: ")
    }
    S = -log(D[i+1,]/D[i,])
    S[D[i+1,]==0] = Inf
    B = boxplot(S,plot=FALSE)
    if(exclude.out) O[i+1,]<-(((S>B$stats[5,])|(S==Inf))&(O[i,]==FALSE)&(O[1,]==FALSE))
    if(!exclude.out) O[i+1,]<-((S>B$stats[5,])|(S==Inf))
    if (print) print(which(O[i+1,]))
  }
  if(max.order==3){
    if (print) print("1/3 outliers")
    S = -log(D[3,]/D[1,])
    S[D[3,]==0] = Inf
    B = boxplot(S,plot=FALSE)
    if(exclude.out) O[4,]<-(((S>B$stats[5,])|(S==Inf))&(O[3,]==FALSE)&(O[2,]==FALSE)&(O[1,]==FALSE))
    if(!exclude.out) O[4,]<-((S>B$stats[5,])|(S==Inf))
    if (print) print(which(O[4,]))
  }
  if(max.order==1) rownames(O) = "1st"
  if(max.order==2) rownames(O) = c("1st","2nd")
  if(max.order==3){
    rownames(O) = c("1st","2nd","3rd","1/3rd")
    if (plotpairs) DpairsPlot(D,O)
  }
  if(output=="matrix") return(O)
  if(output=="list") return(apply(O,1,which))
}

DpairsPlot = function(DB2,O,sp.index=NULL){
  # plots a 2x2 scatter of all the pairs of order extended integrated depth values
  # for orders 1, 2, and 3 as presented in the Nagy et al. (2016), Figure 5
  cexaxis = 1.3
  cexlab = 1.5
  n = ncol(O)
  col = rep("grey",n)
  colout1 = "olivedrab3"
  colout2 = "navy"
  colout3 = "darkorange"
  colout = c(colout1,colout2,colout3)
  col[O[1,]] = colout1
  col[O[2,]] = colout2
  col[O[3,]] = colout3
  col[O[4,]] = colout3
  pch = rep(16,n)
  pch[O[1,]] = 1
  pch[O[2,]] = 2
  pch[O[3,]] = 18
  pch[O[4,]] = 18
  cx = 1.5
  cex = rep(1,n)
  cex[O[1,]] = cx
  cex[O[2,]] = cx
  cex[O[3,]] = cx
  cex[O[4,]] = cx
  OI = (colSums(O)>0)	# outlier indicator
  
  op<-par(cex.axis=cexaxis,cex.lab=cexlab,mfrow = c(2, 2),     # 2x2 layout
          oma = c(3, 3.5, 0.45, 0.45), # two rows of text at the outer left and bottom margin
          mar = c(1, 1, 0, 0), # space for one row of text at ticks and to separate plots
          mgp = c(2, 1, 0),    # axis label at 2 rows distance, tick labels at 1 row
          xpd = NA)            # allow content to protrude into outer margin (and beyond)
  plot(c(0,M<-max(DB2)),c(0,max(DB2)),type="n",ann=FALSE,axes=FALSE,frame=TRUE)
  axis(2)
  title(ylab=expression(FD[2]),line=2.5)
  points(DB2[1,],DB2[2,],col=col,pch=pch,cex=cex,lwd=1.35*cex)
  for(i in 1:n) if(OI[i]) points(DB2[1,i],DB2[2,i],col=col[i],pch=pch[i],cex=cex[i],lwd=1.35*cex[i])
  if (!is.null(sp.index)) points(DB2[1,sp.index],DB2[2,sp.index],pch=16,col="orange")
  segments(0,0,M,M,lty=3,lwd=2)
  plot(c(0,1),c(0,1),type="n",ann=FALSE,axes=FALSE)
  legend(0,.7,c("1st order outliers","2nd order outliers","3rd order outliers"),col=colout,pch=c(1,2,18),cex=cx,bty="n")
  plot(c(0,max(DB2)),c(0,max(DB2)),type="n",ann=FALSE)
  axis(2)
  title(ylab=expression(FD[3]^A),line=2.5)
  axis(1)
  title(xlab=expression(FD[1]),line=3)
  points(DB2[1,],DB2[3,],col=col,pch=pch,cex=cex,lwd=1.35*cex)
  for(i in 1:n) if(OI[i]) points(DB2[1,i],DB2[3,i],col=col[i],pch=pch[i],cex=cex[i],lwd=1.35*cex[i])
  if (!is.null(sp.index)) points(DB2[1,sp.index],DB2[3,sp.index],pch=16,col="orange")	
  segments(0,0,M,M,lty=3,lwd=2)
  plot(c(0,max(DB2)),c(0,max(DB2)),type="n",ann=FALSE,axes=FALSE,frame=TRUE)
  axis(1)
  title(xlab=expression(FD[2]),line=3)
  points(DB2[2,],DB2[3,],col=col,pch=pch,cex=cex,lwd=1.35*cex)
  for(i in 1:n) if(OI[i]) points(DB2[2,i],DB2[3,i],col=col[i],pch=pch[i],cex=cex[i],lwd=1.35*cex[i])
  if (!is.null(sp.index)) points(DB2[2,sp.index],DB2[3,sp.index],pch=16,col="orange")
  segments(0,0,M,M,lty=3,lwd=2)
  par(op)
}

################################################################################
# File:             depth.betaSkeleton.r
# Created by:       Pavlo Mozharovskyi
# First published:  08.03.2018
# Last revised:     08.03.2018
# 
# Computation of the beta-skeleton depth.
################################################################################

.betaSkeleton_validate <- function(ddalpha, beta = 2, distance = "Lp", Lp.p = 2, 
                          mah.estimate = "moment", mah.parMcd = 0.75, ...) {
  # only validate and stop if anything is wrong
  if (!is.numeric(beta) || length(beta) > 1 || beta < 1){
    stop("Argument \"beta\" should be a real value >= 1")
  }
  if (!is.numeric(Lp.p) || length(Lp.p) > 1 || Lp.p < 0){
    stop("Argument \"Lp.p\" should be a real value >= 0")
  }
  if (toupper(distance) == "MAHALANOBIS"){
    if( !(toupper(mah.estimate) %in% c("NONE", "MOMENT", "MCD"))){
        stop("Wrong argument \"mah.estimate\", should be one of \"moment\", \"MCD\", \"none\"")
    }
  } else if (toupper(distance) != "LP") {
      stop("Argument \"distance\" should be either \"Lp\" or \"Mahalanobis\"")
  }
  
  
  return (list(beta = beta, distance = distance, Lp.p = Lp.p, 
              mah.estimate = mah.estimate, mah.parMcd = mah.parMcd))
}

.betaSkeleton_learn <- function(ddalpha) {
  #1. Calculate statistics based on data
  
  if (toupper(ddalpha$distance) == "MAHALANOBIS"){
    ddalpha$distance.code <- 5L
    if(toupper(ddalpha$mah.estimate) == "NONE"){
      for (i in 1:ddalpha$numPatterns){
        ddalpha$patterns[[i]]$sigma <- diag(ddalpha$dimension)
      }
    } else {
      for (i in 1:ddalpha$numPatterns){
        if(toupper(ddalpha$mah.estimate) == "MOMENT"){
          cov <- cov(as.matrix(ddalpha$patterns[[i]]$points))
        } else if(toupper(ddalpha$mah.estimate) == "MCD"){
          cov <- covMcd(as.matrix(ddalpha$patterns[[i]]$points), ddalpha$mah.parMcd)$cov
        } else {stop("Wrong argument \"mah.estimate\", should be one of \"moment\", \"MCD\", \"none\"")}
        if(sum(is.na(cov)) == 0){
          ddalpha$patterns[[i]]$sigma <- solve(cov)
        } else{
          ddalpha$patterns[[i]]$sigma <- diag(ddalpha$dimension)
          warning("Covariance estimate not found for pattern ", ddalpha$patterns[[i]]$name, ", no affine-invariance-adjustment")
        }
      }
    }
  }else{
    for (i in 1:ddalpha$numPatterns){
      ddalpha$patterns[[i]]$sigma <- 0;
    }
    if (toupper(ddalpha$distance) == "LP"){
      ddalpha$distance.code <- 4L
      if (ddalpha$Lp.p == 1){
        ddalpha$distance.code <- 1L
      }
      if (ddalpha$Lp.p == 2){
        ddalpha$distance.code <- 2L
      }
      if (is.infinite(ddalpha$Lp.p) && ddalpha$Lp.p > 0){
        ddalpha$distance.code <- 3L
      }
    } else {stop("Argument \"distance\" should be either \"Lp\" or \"Mahalanobis\"")}
  }
  
  #2. Calculate depths for each pattern
  for (i in 1:ddalpha$numPatterns){
    ddalpha$patterns[[i]]$depths = .betaSkeleton_depths(ddalpha, ddalpha$patterns[[i]]$points)
  }
  
  return (ddalpha)
}

.betaSkeleton_depths <- function(ddalpha, objects){
  depths <- NULL
  for (i in 1:ddalpha$numPatterns){
    pattern <- ddalpha$patterns[[i]]$points
    x <- as.vector(t(pattern))
    y <- as.vector(t(objects))

    ds <- .C("BetaSkeletonDepth", 
             as.double(x), 
             as.double(y), 
             as.integer(nrow(pattern)), 
             as.integer(nrow(objects)), 
             as.integer(ncol(pattern)), 
             as.double(ddalpha$beta), 
             as.integer(ddalpha$distance.code), 
             as.double(ddalpha$Lp.p), 
             as.double(ddalpha$patterns[[i]]$sigma), 
             depths=double(nrow(objects)))$depths
    
    depths <- cbind(depths, ds)
  }
  return (depths)
}

depth.betaSkeleton <- function(x, data, beta = 2, distance = "Lp", Lp.p = 2, 
                               mah.estimate = "moment", mah.parMcd = 0.75){
  if (!is.matrix(x) 
      && is.vector(x)){
    x <- matrix(x, nrow=1)
  }
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!(is.matrix(x) && is.numeric(x)
        || is.data.frame(x) && prod(sapply(x, is.numeric))) ){
    stop("Argument \"x\" should be numeric")
  }
  if (ncol(x) != ncol(data)){
    stop("Dimensions of the arguments \"x\" and \"data\" should coincide")
  }
  if (nrow(data) < 2){
    stop("Too few data points")
  }
  if (!is.numeric(beta) || length(beta) > 1 || beta < 1){
    stop("Argument \"beta\" should be a real value >= 1")
  }
  if (!is.numeric(Lp.p) || length(Lp.p) > 1 || Lp.p < 0){
    stop("Argument \"Lp.p\" should be a real value >= 0")
  }
  if (toupper(distance) == "MAHALANOBIS"){
    distance.code <- 5L
    if(toupper(mah.estimate) == "NONE"){
      sigma = diag(ncol(data))
    } else {
      if(toupper(mah.estimate) == "MOMENT"){
        cov <- cov(data)
      } else if(toupper(mah.estimate) == "MCD"){
        cov <- covMcd(data, mah.parMcd)$cov
      } else {stop("Wrong argument \"mah.estimate\", should be one of \"moment\", \"MCD\", \"none\"")}
      if(sum(is.na(cov)) == 0){
        sigma <- solve(cov)
      } else{
        sigma = diag(ncol(data))
        warning("Covariance estimate not found, no affine-invariance-adjustment")
      }
    }
  }else{
    sigma = 0;
    if (toupper(distance) == "LP"){
      distance.code <- 4L
      if (Lp.p == 1){
        distance.code <- 1L
      }
      if (Lp.p == 2){
        distance.code <- 2L
      }
      if (is.infinite(Lp.p) && Lp.p > 0){
        distance.code <- 3L
      }
    } else {stop("Argument \"distance\" should be either \"Lp\" or \"Mahalanobis\"")}
  }

  points <- as.vector(t(data))
  objects <- as.vector(t(x))
  ds <- .C("BetaSkeletonDepth", 
           as.double(points), 
           as.double(objects), 
           as.integer(nrow(data)), 
           as.integer(nrow(x)), 
           as.integer(ncol(data)), 
           as.double(beta), 
           as.integer(distance.code), 
           as.double(Lp.p), 
           as.double(sigma), 
           depths=double(nrow(x)))$depths
  
  return (ds)
}

################################################################################
# File:             depth.simplicialVolume.r
# Created by:       Oleksii Pokotylo, Pavlo Mozharovskyi
# First published:  15.06.2015
# Last revised:     20.02.2019
# 
# Computation of the simplicial volume data depth.
################################################################################

.longtoint <- function(k){
  limit = 2000000000
  k1 = as.integer(k/limit)
  k2 = k - k1*limit
  return(c(k1, k2))
}

depth.simplicialVolume <- function(x, data, exact = F, k = 0.05, mah.estimate = "moment", mah.parMcd = 0.75, seed = 0){
  if (seed!=0) set.seed(seed)
  if (!is.matrix(x) 
      && is.vector(x)){
    x <- matrix(x, nrow=1)
  }
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.numeric(x)){
    stop("Argument \"x\" should be numeric")
  }
  if (ncol(x) != ncol(data)){
    stop("Dimensions of the arguments \"x\" and \"data\" should coincide")
  }
  if (ncol(data) + 1 > nrow(data)){ #?
    stop("To few data points")
  }
  if (!exact) if (k <= 0) stop("k must be positive")
      else if (k < 1) k = choose(nrow(data), ncol(data))*k
  if(toupper(mah.estimate) == "NONE"){
    useCov <- 0
    covEst <- diag(ncol(data))
  } else if(toupper(mah.estimate) == "MOMENT"){
    useCov <- 1
    covEst <- cov(data)
  } else if(toupper(mah.estimate) == "MCD"){
    useCov <- 2
    covEst <- covMcd(data, mah.parMcd)$cov
  } else {stop("Wrong argument \"mah.estimate\", should be one of \"moment\", \"MCD\", \"none\"")}
      
  points <- as.vector(t(data))
  objects <- as.vector(t(x))
  ds <- .C("OjaDepth", 
           as.double(points), 
           as.double(objects), 
           as.integer(nrow(data)), 
           as.integer(nrow(x)), 
           as.integer(ncol(data)), 
           as.integer(seed),
           as.integer(exact),
           as.integer(.longtoint(k)),
           as.integer(useCov),
           as.double(as.vector(t(covEst))),
           depths=double(nrow(x)))$depths
  
  return (ds)
}



depth.space.simplicialVolume <- function(data, cardinalities, exact = F, k = 0.05, mah.estimate = "moment", mah.parMcd = 0.75, seed = 0){
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.vector(cardinalities, mode = "numeric") 
      || is.na(min(cardinalities)) 
      || sum(.is.wholenumber(cardinalities)) != length(cardinalities) 
      || min(cardinalities) <= 0 
      || sum(cardinalities) != nrow(data)){
    stop("Argument \"cardinalities\" should be a vector of cardinalities of the classes in \"data\" ")
  }
  if (sum(cardinalities < ncol(data) + 1) != 0){
    stop("Not in all classes sufficiently enough objetcs")
  }
  if (toupper(mah.estimate) == "NONE"){
    useCov <- 0
  } else if (toupper(mah.estimate) == "MOMENT"){
    useCov <- 1
  } else if (toupper(mah.estimate) == "MCD"){
    useCov <- 2
  } else {stop("Wrong argument \"mah.estimate\", should be one of \"moment\", \"MCD\", \"none\"")}
  
  depth.space <- NULL
  for (i in 1:length(cardinalities)){
    pattern <- data[(1 + sum(cardinalities[0:(i - 1)])):sum(cardinalities[1:i]),]
    pattern.depths <- depth.simplicialVolume(data, pattern, exact, k, mah.estimate, mah.parMcd, seed)
    depth.space <- cbind(depth.space, pattern.depths, deparse.level = 0)
  }
  
  return (depth.space)
}

qda.train <- function(data){
  new.frm <- data.frame(data)
  z <- qda(formula=as.formula(paste("X", ncol(data), " ~ .", sep="")), data = new.frm)#, prior = rep(1, dimension)/dimension)
  return (z)
}

qda.classify <- function(objects, qda){
  z <- predict(qda, data.frame(objects))$class
  return (z)
}

ddalphaf.train <- function(dataf, labels, subset,
                           adc.args = list(instance = "avr", 
                                           numFcn = -1, 
                                           numDer = -1), 
                           classifier.type = c("ddalpha", "maxdepth", "knnaff", "lda", "qda"), 
                           cv.complete = FALSE, 
                           maxNumIntervals = min(25, ceiling(length(dataf[[1]]$args)/2)),
                           seed = 0,
                           ...){
  # Trains the functional DDalpha-classifier
  # Args:
  #   dataf:  list containing lists (functions) of two vectors of equal length, 
  #           named "args" and "vals": arguments sorted in ascending order and 
  #           corresponding them values respectively
  #   labels: output labels of the functinal observations
  #   other arguments: TODO
  # Returns:
  #   Functional DDalpha-classifier
  
  # Check "dataf"
  if (!is.list(dataf))
    stop("Argument 'dataf' must be a list")
  for (df in dataf)
    if (!(is.list(df) && length(df) == 2 &&
            !is.null(df$args) && !is.null(df$vals) &&
            is.vector(df$args) && is.vector(df$vals) &&
            is.numeric(df$args) && is.numeric(df$vals) &&
            length(df$args) == length(df$vals) &&
            is.sorted(df$args)))
      stop("Argument 'dataf' must be a list containing lists (functions) of two vectors of equal length, named 'args' and 'vals': arguments sorted in ascending order and corresponding them values respectively")
  
  if(!missing(subset)) {
    dataf = dataf[subset]
    labels = labels[subset]
  }
  
  # Check "labels"
  if (!(length(dataf)==length(labels) && length(unique(labels))>=2))
    stop("Argument 'labels' has wrong format")
  
  # Check classifier.type
  classifier.type = match.arg(classifier.type)
  
  # Check "adc.method"
  adc.method = 'equalCover'
  
  if (seed != 0) set.seed(seed)
  
  # Check "adc.args"
  if(!is.null(names(adc.args))){
    if (!(adc.args$instance %in% c("val", "avr") && 
        ((adc.args$numFcn >= 0 &&  adc.args$numDer >= 0 && (adc.args$numFcn + adc.args$numDer >= 2)) ||
         (adc.args$numFcn == -1 &&  adc.args$numDer == -1))))
      stop("Argument 'adc.args' has wrong format")
  } else {
    if(!is.list(adc.args))
      stop("Argument 'adc.args' has wrong format")
    for(.adc.args in adc.args){
        if (!(.adc.args$instance %in% c("val", "avr") && 
            ((.adc.args$numFcn >= 0 &&  .adc.args$numDer >= 0 && (.adc.args$numFcn + .adc.args$numDer >= 2)) ||
             (.adc.args$numFcn == -1 &&  .adc.args$numDer == -1))))
        stop("Argument 'adc.args' has wrong format")
    }
  }
  
  # CV
  if (is.null(names(adc.args)) || adc.args$numFcn == -1 && adc.args$numDer == -1){
    if (cv.complete){
      res <- getBestSpaceCV(dataf, labels, adc.method, adc.args, 
                            classifier.type, num.chunks=10, numMax = maxNumIntervals, ...)
    }else{
      res <- getBestSpace(dataf, labels, adc.method, adc.args,
                          classifier.type, num.chunks=10, numMax = maxNumIntervals, ...)
    }
    the.args <- res$args
    num.cv <- res$num.cv
  }else{
    the.args <- adc.args
    num.cv <- 0
  }
  # Pointize
  points <- GetPoints(dataf, labels, adc.method, the.args)
  
  if(any(!is.finite(points$data))) {
    warning("infinite or missing values in 'points$data'")
    return (NA)
  }
  
  # Apply chosen classifier to train the data
  if (classifier.type == "ddalpha"){
    classifier <- ddalpha.train(points$data, seed = seed, ...)
  }
  if (classifier.type == "maxdepth"){
    classifier <- ddalpha.train(points$data, separator = "maxD", seed = seed, ...)
  }
  if (classifier.type == "knnaff"){
    classifier <- knnaff.train(points$data, i = 0, ...)
  }
  if (classifier.type == "lda"){
    classifier <- lda.train(points$data, ...)
  }
  if (classifier.type == "qda"){
    classifier <- qda.train(points$data, ...)
  }
  # Create the eventual output structure
  ddalphaf <- 
    list(dataf = points$dataf, 
      #labels_orig = labels,
      labels = points$labels, 
      adc.method = adc.method, 
      adc.args = the.args, 
      adc.num.cv = num.cv, 
      adc.transmat = points$adc.transmat, 
      data = points$data, 
      classifier.type = classifier.type, 
      classifier = classifier)
  class(ddalphaf) <- "ddalphaf"
  
  return (ddalphaf)
}

ddalphaf.classify <- function(ddalphaf, objectsf, subset, ...){
  # Classifies functions
  # Args:
  #   objectsf: sample to classify, a list containing lists (functions) of 
  #             two vectors of equal length, named "args" and "vals": 
  #             arguments sorted in ascending order and corresponding them 
  #             values respectively
  #   ddalphaf: functional DDalpha-classifier
  # Returns:
  #   List of labels assigned to the functions from "objectsf"

  # Check "objectsf"
  if (!is.list(objectsf))
    stop("Argument 'objectsf' must be a list")
  if (!is.null(objectsf$args)){
    objectsf = list(objectsf) # there was a single element 
  }
  
  if(!missing(subset)) {
    objectsf = objectsf[subset]
  }
  
  for (df in objectsf)
    if (!(is.list(df) && length(df) == 2 &&
            !is.null(df$args) && !is.null(df$vals) &&
            is.vector(df$args) && is.vector(df$vals) &&
            is.numeric(df$args) && is.numeric(df$vals) &&
            length(df$args) == length(df$vals) &&
            is.sorted(df$args)))
      stop("Argument 'objectsf' must be a list containing lists (functions) of two vectors of equal length, named 'args' and 'vals': arguments sorted in ascending order and corresponding them values respectively")

  # Prepare to multivariate classification
  objectsf.equalized <- equalize(objectsf)
  if (ddalphaf$adc.method == "equalCover"){
    if (ddalphaf$adc.args$instance == "val"){
      input <- getValGrid(objectsf.equalized, 
                          ddalphaf$adc.args$numFcn, ddalphaf$adc.args$numDer)
    }
    if (ddalphaf$adc.args$instance == "avr"){
      input <- getAvrGrid(objectsf.equalized, 
                          ddalphaf$adc.args$numFcn, ddalphaf$adc.args$numDer)
    }
    if (!is.null(ddalphaf$adc.transmat)){
      input <- input%*%ddalphaf$adc.transmat
    }
  }
  # Classify and assign class labels
  if (ddalphaf$classifier.type == "ddalpha" || ddalphaf$classifier.type == "maxdepth"){
    output <- ddalpha.classify(objects = input, ddalpha = ddalphaf$classifier, ...)
  }
  if (ddalphaf$classifier.type == "knnaff"){
    output <- knnaff.classify(input, ddalphaf$classifier, ...)
  }
  if (ddalphaf$classifier.type == "lda"){
    output <- lda.classify(input, ddalphaf$classifier, ...)
  }
  if (ddalphaf$classifier.type == "qda"){
    output <- qda.classify(input, ddalphaf$classifier, ...)
  }
  classes <- list()
  for (i in 1:length(output)){
#    if (is.numeric(output[[i]])){
      classes[[i]] <- ddalphaf$labels[[ output[[i]] ]]
#    }else{
#      classes[[i]] <- output[[i]]
#    }
  }

  return (classes)
}

predict.ddalphaf <- function(object, objectsf, subset, ...){
  return(ddalphaf.classify(object, objectsf, subset, ...))
}

is.in.convexf <- function(objectsf, dataf, cardinalities, 
                          adc.method = "equalCover", 
                          adc.args = list(instance = "val", 
                                          numFcn = 5, 
                                          numDer = 5), seed = 0){
  # Checks if the function(s) lie(s) inside convex hulls of the 
  # functions from the sample in the projection space
  # Args:
  #   objectsf:      list containing lists (functions) of two vectors of equal 
  #                  length, named "args" and "vals": arguments sorted in 
  #                  ascending order and corresponding them values 
  #                  respectively. These functions are supposed to be checked 
  #                  for 'outsiderness'
  #   dataf:         list containing lists (functions) of two vectors of equal 
  #                  length, named "args" and "vals": arguments sorted in 
  #                  ascending order and corresponding them values respectively
  #   cardinalities: cardinalities of the classes in "dataf"
  #   other arguments: TODO
  # Returns:
  #   Vector with 1s for those lying inside of at least one of the convex hulls 
  #   of the classes and 0s for those lying beyond them
  
  # Data-consistency checks
  # TODO
  
  # Project "objectsf" into a multivariate space
  objectsf.equalized <- equalize(objectsf)
  if (adc.method == "equalCover"){
    if (adc.args$instance == "val"){
      objects <- getValGrid(objectsf.equalized, adc.args$numFcn, adc.args$numDer)
    }
    if (adc.args$instance == "avr"){
      objects <- getAvrGrid(objectsf.equalized, adc.args$numFcn, adc.args$numDer)
    }
  }
  # Project "dataf" into a multivariate space
  dataf.equalized <- equalize(dataf)
  if (adc.method == "equalCover"){
    if (adc.args$instance == "val"){
      data <- getValGrid(dataf.equalized, adc.args$numFcn, adc.args$numDer)
    }
    if (adc.args$instance == "avr"){
      data <- getAvrGrid(dataf.equalized, adc.args$numFcn, adc.args$numDer)
    }    
  }
  in.convex <- is.in.convex(objects, data, cardinalities, seed)
  
  return (in.convex)
}

print.ddalphaf <- function(x, ...){
  cat("ddalphaf:\n")
  cat("\t num.functions = ", length(x$dataf), 
  ", num.patterns = ", length(unique(x$labels)), "\n", sep="")
#  cat("\t adc.method", x$adc.method, "\"\n", sep="")
  cat("\t adc: \"", x$adc.args$instance, "; numFcn:", x$adc.args$numFcn, "; numDer:", x$adc.args$numDer, "\"\n", sep="")
  cat("\t adc.num.cv \"", x$adc.num.cv, "\"\n", sep="")
  cat("\t adc.transmat \"", x$adc.transmat, "\"\n", sep="")
  cat("\t classifier.type \"", x$classifier.type, "\"\n", sep="")
  cat("classifier: ") 
  print(x$classifier)
}

summary.ddalphaf <- function(object, ...) 
  print.ddalphaf(object, ...)
################################################################################
# Functions below are used for intermediate computations                       #
################################################################################

equalize <- function(dataf){
  # 1. Adjusts the data to have equal (the largest) argument interval
  # 2. Calclates - numerically - derivative
  # Args:
  #   dataf: list containing lists (functions) of two vectors 
  #          of equal length, named "args" and "vals": arguments 
  #          sorted in ascending order and corresponding them 
  #          values respectively
  # Returns:
  #   The list of lists of the same structure, 'equalized', 
  #   contating derivatives as 3rd vector named "der1"
  
  # Check whether every function contains fields "args" and "vals", 
  # whether they are numerical and of equal length, have no NAs or 
  # ties and are sorted in ascending order
  # TODO
  
  # 1.
  # Get argument bounds
  min <- Inf
  max <- -Inf
  for (i in 1:length(dataf)){
    if (dataf[[i]]$args[1] < min){min <- dataf[[i]]$args[1]}
    if (dataf[[i]]$args[length(dataf[[i]]$args)] > max){
      max <- dataf[[i]]$args[length(dataf[[i]]$args)]
    }
  }
  # and apply them to equalize functions timely
  for (i in 1:length(dataf)){
    if (dataf[[i]]$args[1] > min){
      dataf[[i]]$args <- c(min, dataf[[i]]$args)
      dataf[[i]]$vals <- c(dataf[[i]]$vals[1], dataf[[i]]$vals)
    }
    if (dataf[[i]]$args[length(dataf[[i]]$args)] < max){
      dataf[[i]]$args <- c(dataf[[i]]$args, max)
      dataf[[i]]$vals <- c(dataf[[i]]$vals, 
                           dataf[[i]]$vals[length(dataf[[i]]$vals)])
    }
    # Computational trick - add "-1" to the "left"
    #dataf[[i]]$args <- c(min - 1, dataf[[i]]$args)
    #dataf[[i]]$vals <- c(dataf[[i]]$vals[1], dataf[[i]]$vals)
  }
  
  # 2.
  for (i in 1:length(dataf)){
    dataf[[i]] <- derive(dataf[[i]])
  }
  
  return (dataf)
}

derive <- function(fcn){
  # Adds 1st derivative to the function: a vector named "der1"
  # Args:
  #   fcn: function, a list of two vectors of equal length, named "args" 
  #   and "vals": arguments sorted in ascending order and corresponding 
  #   them values respectively
  # Returns:
  #   The list of of the same structure contating derivative as 3rd 
  #   vector named "der1"
  
  fcn$der1 <- rep(0, length(fcn$args))
  fcn$der1[1] = 0
  for (i in 2:length(fcn$der1)){
    fcn$der1[i] <- (fcn$vals[i] - fcn$vals[i - 1])/
      (fcn$args[i] - fcn$args[i - 1])
  }
  
  return (fcn)
}

getValue <- function(fcn, arg, fcnInstance){
  # Gets the value of the function or its derivative for the given argument 
  # value
  # Args:
  #   fcn:         function, a list of vectors of equal length, named "args" 
  #                (arguments), "vals" (function values) [and it's 
  #                derivatives of order "I", named derI]; arguments (and 
  #                corresponding values) sorted in ascending order
  #   arg:         argument value at which the function (derivative) value 
  #                is to be taken
  #   fcnInstance: inctance to be evaluated; "vals" for the function values, 
  #                "der1" for the first derivative
  # Returns:
  #   Value of the function (derivative)
  
  # Check "arg", evt. "fcnInstance"
  # TODO
  
  # Find corresponding interval
  index <- 2
  while (arg > fcn$args[index]){
    index <- index + 1
  }
  # Get the function value(s) by linear approximation
  if (fcnInstance == "vals"){
    value <- fcn$vals[index - 1] + 
      (fcn$vals[index] - fcn$vals[index - 1])*
      ((arg - fcn$args[index - 1])/(fcn$args[index] - fcn$args[index - 1]))
  }
  if (fcnInstance == "der1"){
    value <- fcn$der1[index]
  }
  
  return (value)
}

getAvrValue <- function(fcn, argFrom, argTo, fcnInstance){
  # Gets the average value of the function or its derivative on the given 
  # interval
  # Args:
  #   fcn:         function, a list of vectors of equal length, named "args" 
  #                (arguments), "vals" (function values) [and it's 
  #                derivatives of order "I", named derI]; arguments (and 
  #                corresponding values) sorted in ascending order
  #   argFrom:     argument value from which the function (derivative) value 
  #                is to be averaged
  #   argTo:       argument value to which the function (derivative) value 
  #                is to be averaged
  #   fcnInstance: inctance to be evaluated; "vals" for the function values, 
  #                "der1" for the first derivative
  # Returns:
  #   Average value of the function (derivative) on the interval 
  #   (argFrom, argTo)
  
  # Check "argFrom" and "argTo", evt. "fcnInstance"
  # TODO
  
  # Define 'from' and 'to' interval
  indexFrom <- 2
  while (argFrom > fcn$args[indexFrom]){
    indexFrom <- indexFrom + 1
  }
  indexTo <- 2
  while (argTo > fcn$args[indexTo]){
    indexTo <- indexTo + 1
  }
  average <- 0
  valTo <- getValue(fcn, argTo, fcnInstance)
  # Integrate
  curArgFrom <- argFrom
  if (fcnInstance == "vals"){
    valFrom <- getValue(fcn, curArgFrom, "vals")
    while(indexFrom < indexTo){
      average <- average + (valFrom + fcn$vals[indexFrom])*
        (fcn$args[indexFrom] - curArgFrom)/2
      valFrom <- fcn$vals[indexFrom]
      curArgFrom <- fcn$args[indexFrom]
      indexFrom <- indexFrom + 1
    }
    average <- average + (valFrom + valTo)*(argTo - curArgFrom)/2
  }
  if (fcnInstance == "der1"){
    while(indexFrom < indexTo){
      average <- average + (fcn$der1[indexFrom])*
        (fcn$args[indexFrom] - curArgFrom)
      curArgFrom <- fcn$args[indexFrom]
      indexFrom <- indexFrom + 1
    }
    average <- average + valTo*(argTo - curArgFrom)
  }
  average <- average/(argTo - argFrom)
  
  return (average)
}

getValGrid <- function(dataf, numFcn, numDer){
  # Represents a function sample as a multidimensional (d="numFcn"+"numDer") 
  # one averaging for that each function and it derivative on "numFcn" 
  # (resp. "numDer") equal nonoverlapping covering intervals
  # Args:
  #   dataf:  list containing lists (functions) of vectors of equal length, 
  #           first two named "args" and "vals" are arguments sorted in 
  #           ascending order and having same bounds for all functions and 
  #           corresponding them values respectively
  #   numFcn: number of function intervals
  #   numDer: number of first-derivative intervals
  # Returns:
  #   Matrix - a multidimensional presentation of the functional sample
  
  # Get argument bounds ("dataf" is equalized)
  min <- dataf[[1]]$args[1]
  max <- dataf[[1]]$args[length(dataf[[1]]$args)]
  # Get argument grid
  args <- dataf[[1]]$args
  argsFcn <- min + 0:numFcn*(max - min)/(numFcn - 1)
  argsDer <- min + 0:numDer*(max - min)/(numDer - 1)
  # Get function/derivative grid
  fcnGrid <- matrix(nrow = length(dataf), ncol = numFcn)
  derGrid <- matrix(nrow = length(dataf), ncol = numDer)
  if (numFcn > 0){
    for (i in 1:length(dataf)){
      # Set merging algorithm (Novikoff, Piter)
      cArgs <- 1
      cArgsFcn <- 1
      fcnGrid[i,1] <- dataf[[i]]$vals[1]
      while (cArgsFcn != numFcn){
#        print(argsFcn)
#        print(fcnGrid[i,])
#        cat(cArgs, " and ", cArgsFcn, "\n")
#        cat(args[cArgs + 1], " and ", argsFcn[cArgsFcn + 1], "\n")
        if (args[cArgs + 1] < argsFcn[cArgsFcn + 1]){
          cArgs <- cArgs + 1
        }else{
          nextArg <- argsFcn[cArgsFcn + 1]
          fcnGrid[i,cArgsFcn + 1] <- dataf[[i]]$vals[cArgs] + (nextArg - args[cArgs])*dataf[[i]]$der1[cArgs + 1]
          if (args[cArgs + 1] == argsFcn[cArgsFcn + 1]){
            cArgs <- cArgs + 1
          }
          cArgsFcn <- cArgsFcn + 1
        }
      }
    }
  }
  if (numDer > 0){
    for (i in 1:length(dataf)){
      # Again, set merging algorithm (Novikoff, Piter)
      cArgs <- 1
      cArgsDer <- 1
      derGrid[1] <- dataf[[i]]$ders[2]
      while (cArgsDer != numDer){
#        print(argsDer)
#        print(derGrid[i,])
#        cat(cArgs, " and ", cArgsDer, "\n")
#        cat(args[cArgs + 1], " and ", argsDer[cArgsDer + 1], "\n")
        if (args[cArgs + 1] < argsDer[cArgsDer + 1]){
          cArgs <- cArgs + 1
        }else{
          derGrid[i,cArgsDer + 1] <- dataf[[i]]$der1[cArgs + 1]
          if (args[cArgs + 1] == argsDer[cArgsDer + 1]){
            cArgs <- cArgs + 1
          }
          cArgsDer <- cArgsDer + 1
        }
      }
    }
  }
  mvX <- cbind(fcnGrid, derGrid)
  return (mvX)
}

getAvrGrid <- function(dataf, numFcn, numDer){
  # Represents a function sample as a multidimensional (d="numFcn"+"numDer") 
  # one averaging for that each function and it derivative on "numFcn" 
  # (resp. "numDer") equal nonoverlapping covering intervals
  # Args:
  #   dataf:  list containing lists (functions) of vectors of equal length, 
  #           first two named "args" and "vals" are arguments sorted in 
  #           ascending order and having same bounds for all functions and 
  #           corresponding them values respectively
  #   numFcn: number of function intervals
  #   numDer: number of first-derivative intervals
  # Returns:
  #   Matrix - a multidimensional presentation of the functional sample
  
  # Get argument bounds ("dataf" is equalized)
  min <- dataf[[1]]$args[1]
  max <- dataf[[1]]$args[length(dataf[[1]]$args)]
  # Get argument grid
  args <- dataf[[1]]$args
  argsFcn <- min + 0:numFcn*(max - min)/numFcn
  argsDer <- min + 0:numDer*(max - min)/numDer
  # Get function/derivative grid
  fcnGrid <- matrix(nrow = length(dataf), ncol = numFcn)
  derGrid <- matrix(nrow = length(dataf), ncol = numDer)
  if (numFcn > 0){
    for (i in 1:length(dataf)){
      # Set merging algorithm (Novikoff, Piter)
      cArgs <- 1
      cArgsFcn <- 1
      curArg <- min
      curFcn <- dataf[[i]]$vals[1]
      curAvr <- 0
      while (cArgsFcn != numFcn + 1){
#        print(argsFcn)
#        print(fcnGrid[i,])
#        cat(cArgs, " and ", cArgsFcn, "\n")
#        cat(args[cArgs + 1], " and ", argsFcn[cArgsFcn + 1], "\n")
        if (args[cArgs + 1] < argsFcn[cArgsFcn + 1]){
          nextArg <- args[cArgs + 1]
          nextFcn <- dataf[[i]]$vals[cArgs + 1]
          curAvr <- curAvr + (nextArg - curArg)*(nextFcn + curFcn)/2
          cArgs <- cArgs + 1
        }else{
          nextArg <- argsFcn[cArgsFcn + 1]
          nextFcn <- dataf[[i]]$vals[cArgs] + (nextArg - args[cArgs])*dataf[[i]]$der1[cArgs + 1]
          fcnGrid[i,cArgsFcn] <- curAvr + (nextArg - curArg)*(nextFcn + curFcn)/2
          curAvr <- 0
          if (args[cArgs + 1] == argsFcn[cArgsFcn + 1]){
            cArgs <- cArgs + 1
          }
          cArgsFcn <- cArgsFcn + 1
        }
        curArg <- nextArg
        curFcn <- nextFcn
      }
    }
  }
  fcnGrid <- fcnGrid/(argsFcn[2] - argsFcn[1])
  if (numDer > 0){
    for (i in 1:length(dataf)){
      # Again, set merging algorithm (Novikoff, Piter)
      cArgs <- 1
      cArgsDer <- 1
      curArg <- min
      curAvr <- 0
      while (cArgsDer != numDer + 1){
#        print(argsDer)
#        print(derGrid[i,])
#        cat(cArgs, " and ", cArgsDer, "\n")
#        cat(args[cArgs + 1], " and ", argsDer[cArgsDer + 1], "\n")
        if (args[cArgs + 1] < argsDer[cArgsDer + 1]){
          nextArg <- args[cArgs + 1]
          curAvr <- curAvr + (nextArg - curArg)*dataf[[i]]$der1[cArgs + 1]
          cArgs <- cArgs + 1
        }else{
          nextArg <- argsDer[cArgsDer + 1]
          derGrid[i,cArgsDer] <- curAvr + (nextArg - curArg)*dataf[[i]]$der1[cArgs + 1]
          curAvr <- 0
          if (args[cArgs + 1] == argsDer[cArgsDer + 1]){
            cArgs <- cArgs + 1
          }
          cArgsDer <- cArgsDer + 1
        }
        curArg <- nextArg
      }
    }
  }
  derGrid <- derGrid/(argsDer[2] - argsDer[1])
  mvX <- cbind(fcnGrid, derGrid)
  return (mvX)
}

getVapnikBound <- function(points, dim = NULL){
  n <- nrow(points)
  d <- ncol(points) - 1
  lda <- lda.train(points)
  result <- lda.classify(points[,1:d], lda)
  empRisk <- sum(result != points[,d + 1])/n
  # Calculate the deviation from the empirical risk
  nu <- 1/n
  C <- 0
  for (k in 0:d){
    C <- C + 2*choose(n - 1, k)
  }
  epsilon <- sqrt( (log(C) - log(nu)) / (2*n) )
  
  return (empRisk + epsilon)
}

GetPoints <- function(dataf, labels, adc.method, adc.args){
  # Numerize labels
  names <- unique(labels)
  output <- rep(0, length(labels))
  for (i in 1:length(labels)){
    for (j in 1:length(names)){
      if (labels[[i]] == names[[j]]){
        output[i] = j
        break
      }
    }
  }
  # Pointize data
  dataf.equalized <- equalize(dataf)
  if (adc.method == "equalCover"){
    if (adc.args$instance == "val"){
      input <- getValGrid(dataf.equalized, adc.args$numFcn, adc.args$numDer)
    }
    if (adc.args$instance == "avr"){
      input <- getAvrGrid(dataf.equalized, adc.args$numFcn, adc.args$numDer)
    }
    # Reduce dimension if needed
    princomps <- NULL
    newDim <- ncol(input)
    for (i in 1:length(names)){
      classi <- input[output == i,1:ncol(input)]
      
      if(any(!is.finite(classi))) {
        warning("infinite or missing values in 'classi'")
        next
      }
      
      princompsi <- prcomp(x=classi, tol=sqrt(.Machine$double.eps))
      #print(princompsi$sdev)
      newDimi <- sum(princompsi$sdev > sqrt(.Machine$double.eps))
      if (newDimi < newDim){
        newDim <- newDimi
        princomps <- princompsi
      }
    }
    #print(newDim)
    transmat <- NULL
    if (newDim < ncol(input)){
      transmat <- matrix(as.vector(princomps$rotation[,1:newDim]), ncol=newDim)
      input <- input%*%transmat
    }
    # Combine data
    data <- cbind(input, output, deparse.level=0)
  }
  
  return (list(data = data, 
               dataf = dataf.equalized, 
               labels = names, 
               adc.transmat = transmat))
}

GetPointsAll <- function(dataf, labels, adc.method = "equalCover", 
                         adc.args = list(instance = "avr", 
                                         numFcn = -1, 
                                         numDer = -1), numMax){
  # Numerize labels
  names <- unique(labels)
  output <- rep(0, length(labels))
  for (i in 1:length(labels)){
    for (j in 1:length(names)){
      if (labels[[i]] == names[[j]]){
        output[i] = j
        break
      }
    }
  }
  # Prepare values
  min <- dataf[[1]]$args[1]
  max <- dataf[[1]]$args[length(dataf[[1]]$args)]
  args <- dataf[[1]]$args
  fcnsAll <- list("")
  dersAll <- list("")
  dataf.equalized <- equalize(dataf)
  # Generate all one-type-argument ("fcn" or "der") for all num(Fcn,Der)-values
  for (numFcn in 1:numMax){
    numDer <- numFcn
    argsFcn <- min + 0:numFcn*(max - min)/numFcn
    argsDer <- min + 0:numDer*(max - min)/numDer
    # Get function/derivative grid
    fcnGrid <- matrix(nrow = length(dataf.equalized), ncol = numFcn)
    derGrid <- matrix(nrow = length(dataf.equalized), ncol = numDer)
    for (i in 1:length(dataf.equalized)){
      # Set merging algorithm (Novikoff, Piter)
      cArgs <- 1
      cArgsFcn <- 1
      curArg <- min
      curFcn <- dataf.equalized[[i]]$vals[1]
      curAvr <- 0
      while (cArgsFcn != numFcn + 1){
        if (args[cArgs + 1] < argsFcn[cArgsFcn + 1]){
          nextArg <- args[cArgs + 1]
          nextFcn <- dataf.equalized[[i]]$vals[cArgs + 1]
          curAvr <- curAvr + (nextArg - curArg)*(nextFcn + curFcn)/2
          cArgs <- cArgs + 1
        }else{
          nextArg <- argsFcn[cArgsFcn + 1]
          nextFcn <- dataf.equalized[[i]]$vals[cArgs] + (nextArg - args[cArgs])*dataf.equalized[[i]]$der1[cArgs + 1]
          fcnGrid[i,cArgsFcn] <- curAvr + (nextArg - curArg)*(nextFcn + curFcn)/2
          curAvr <- 0
          if (args[cArgs + 1] == argsFcn[cArgsFcn + 1]){
            cArgs <- cArgs + 1
          }
          cArgsFcn <- cArgsFcn + 1
        }
        curArg <- nextArg
        curFcn <- nextFcn
      }
    }
    fcnsAll[[numFcn]] <- fcnGrid/(argsFcn[2] - argsFcn[1])
    for (i in 1:length(dataf.equalized)){
      # Again, set merging algorithm (Novikoff, Piter)
      cArgs <- 1
      cArgsDer <- 1
      curArg <- min
      curAvr <- 0
      while (cArgsDer != numDer + 1){
        if (args[cArgs + 1] < argsDer[cArgsDer + 1]){
          nextArg <- args[cArgs + 1]
          curAvr <- curAvr + (nextArg - curArg)*dataf.equalized[[i]]$der1[cArgs + 1]
          cArgs <- cArgs + 1
        }else{
          nextArg <- argsDer[cArgsDer + 1]
          derGrid[i,cArgsDer] <- curAvr + (nextArg - curArg)*dataf.equalized[[i]]$der1[cArgs + 1]
          curAvr <- 0
          if (args[cArgs + 1] == argsDer[cArgsDer + 1]){
            cArgs <- cArgs + 1
          }
          cArgsDer <- cArgsDer + 1
        }
        curArg <- nextArg
      }
    }
    dersAll[[numDer]] <- derGrid/(argsDer[2] - argsDer[1])
  }
  pointsAll <- list("")
  counter <- 1
  # Construct the spaces, reducing dimension if needed
  for (dim in 2:numMax){
    for(nDer in 0:dim){
      nFcn <- dim - nDer
      tmp.args <- list(instance=adc.args$instance, numFcn=nFcn, numDer=nDer)
      if (nFcn == 0){
        input <- dersAll[[nDer]]
      }else{
        if (nDer == 0){
          input <- fcnsAll[[nFcn]]
        }else{
          input <- cbind(fcnsAll[[nFcn]], dersAll[[nDer]])
        }
      }
      # Reduce dimension if needed
      princomps <- NULL
      newDim <- ncol(input)
      for (i in 1:length(names)){
        classi <- input[output == i,1:ncol(input)]
        princompsi <- prcomp(x=classi, tol=sqrt(.Machine$double.eps))
        newDimi <- sum(princompsi$sdev > sqrt(.Machine$double.eps))
        if (newDimi < newDim){
          newDim <- newDimi
          princomps <- princompsi
        }
      }
      transmat <- NULL
      if (newDim < ncol(input)){
        transmat <- matrix(as.vector(princomps$rotation[,1:newDim]), ncol=newDim)
        input <- input%*%transmat
      }
      # Combine data
      tmp.points <- cbind(input, output, deparse.level=0)
      # Save to the list
      pointsAll[[counter]] <- list(data = tmp.points, 
                                   adc.args = tmp.args, 
                                   adc.transmat = transmat)
      counter <- counter + 1
    }
  }
  return (pointsAll)
}

# adc.args is a list of args
GetPointsArgs <- function(dataf, labels, adc.method = "equalCover", 
                         adc.args){
  pointsAll = list()
  counter = 1
  for(tmp.args in adc.args){
    dat = GetPoints(dataf, labels, adc.method, tmp.args)
    pointsAll[[counter]] <- list(data = dat$data, 
                                 adc.args = tmp.args, 
                                 adc.transmat = dat$adc.transmat)
    counter <- counter + 1
  
  }
  return (pointsAll)
}

getBestSpace <- function(dataf, labels, adc.method = "equalCover", 
                         adc.args = list(instance = "avr", 
                                         numFcn = -1, 
                                         numDer = -1), 
                         classifier.type = "ddalpha", 
                         num.chunks = 10, numMax,
                         ...){
  # First, get Vapnik bounds for all plausible spaces
  if(!is.null(names(adc.args))){
    pointsAll <- GetPointsAll(dataf, labels, adc.method, adc.args, numMax)
    numTries <- numMax * (numMax + 1) / 2 + numMax + 1 - 3  
  } else {
    pointsAll <- GetPointsArgs(dataf, labels, adc.method, adc.args)  
    numTries <- length(adc.args)
  }
  Vapnik.bounds <- rep(Inf, numTries)
  curTry <- 1
  for (i in 1:length(pointsAll)){
    tmp.args <- pointsAll[[i]]$adc.args
    tmp.points <- pointsAll[[i]]$data
    Vapnik.bounds[curTry] <- getVapnikBound(tmp.points, ncol(tmp.points) - 1)
    curTry <- curTry + 1
  }
  # Second, get 5 best (i.e. lowest) Vapnik bounds
  #   etalons <- sort(Vapnik.bounds)[1:5]
  #   best.indices <- c()
  #   while (length(best.indices) < 5 && length(etalons) > 0){
  #     best.indices <- c(best.indices, which(Vapnik.bounds == etalons[1]))
  #     etalons <- etalons[-1]
  #   }
  #   best.indices <- best.indices[1:5]
  best.indices <- which(Vapnik.bounds %in% sort(Vapnik.bounds)[1:min(5, numTries)])
  # Third, cross-validate over these best spaces
  errors <- rep(0, length(best.indices))
  for (i in 1:length(best.indices)){
    tmp.args <- pointsAll[[best.indices[i]]]$adc.args
    points.all <- pointsAll[[best.indices[i]]]$data
    d <- ncol(points.all) - 1
    # Actually CV
    num.points <- nrow(points.all)
    indices.off <- num.chunks*(0:(ceiling(num.points/num.chunks) - 1))
    for (j in 1:num.chunks){
      # Determine points to be taken off
      take.off <- (indices.off + j)[(indices.off + j) <= num.points]
      # Apply chosen classifier
      if (classifier.type == "ddalpha"){
        classifier <- ddalpha.train(points.all[-take.off,], ...)
        results <- ddalpha.classify(objects = points.all[take.off,1:d], ddalpha = classifier)
      }
      if (classifier.type == "maxdepth"){
        classifier <- ddalpha.train(points.all[-take.off,], separator = "maxD", ...)
        results <- ddalpha.classify(objects = points.all[take.off,1:d], ddalpha = classifier)
      }
      if (classifier.type == "knnaff"){
        classifier <- knnaff.train(points.all[-take.off,], i = i, ...)
        results <- knnaff.classify(points.all[take.off,1:d], classifier)
      }
      if (classifier.type == "lda"){
        classifier <- lda.train(points.all[-take.off,], ...)
        results <- lda.classify(points.all[take.off,1:d], classifier)
      }
      if (classifier.type == "qda"){
        classifier <- qda.train(points.all[-take.off,], ...)
        results <- qda.classify(points.all[take.off,1:d], classifier)
      }
      # Collect errors
      errors[i] <- errors[i] + sum(
        unlist(results) != points.all
        [take.off,d + 1])
    }
  }
  best.i <- which.min(errors)
  new.args <- pointsAll[[best.indices[best.i]]]$adc.args
  return (list(args = new.args, num.cv = length(best.indices)))
}

getBestSpaceCV <- function(dataf, labels, adc.method = "equalCover", 
                           adc.args = list(instance = "val", 
                                           numFcn = -1, 
                                           numDer = -1), 
                           classifier.type = "ddalpha", 
                           num.chunks = 10, numMax,
                         ...){
  if(!is.null(names(adc.args))){
    pointsAll <- GetPointsAll(dataf, labels, adc.method, adc.args, numMax)  
    numTries <- numMax * (numMax + 1) / 2 + numMax + 1 - 3
  } else {
    pointsAll <- GetPointsArgs(dataf, labels, adc.method, adc.args)  
    numTries <- length(adc.args)
  }
  curTry <- 1
  errors <- rep(0, length(pointsAll))
  for (i in 1:length(pointsAll)){
    points.all <- pointsAll[[i]]$data
    d <- ncol(points.all) - 1
    # Actually CV
    num.points <- nrow(points.all)
    indices.off <- num.chunks*(0:(ceiling(num.points/num.chunks) - 1))
    for (j in 1:num.chunks){
      # Determine points to be taken off
      take.off <- (indices.off + j)[(indices.off + j) <= num.points]
      # Apply chosen classifier
      if (classifier.type == "ddalpha"){
        classifier <- ddalpha.train(points.all[-take.off,], ...)
        results <- ddalpha.classify(objects = points.all[take.off,1:d], ddalpha = classifier)
      }
      if (classifier.type == "maxdepth"){
        classifier <- ddalpha.train(points.all[-take.off,], separator = "maxD", ...)
        results <- ddalpha.classify(objects = points.all[take.off,1:d], ddalpha = classifier)
      }
      if (classifier.type == "knnaff"){
        classifier <- knnaff.train(points.all[-take.off,], i = i, ...)
        results <- knnaff.classify(points.all[take.off,1:d], classifier)
      }
      if (classifier.type == "lda"){
        classifier <- lda.train(points.all[-take.off,], ...)
        results <- lda.classify(points.all[take.off,1:d], classifier)
      }
      if (classifier.type == "qda"){
        classifier <- qda.train(points.all[-take.off,], ...)
        results <- qda.classify(points.all[take.off,1:d], classifier)
      }
      # Collect errors
      errors[i] <- errors[i] + sum(
        unlist(results) != points.all
        [take.off,d + 1])
    }
  }
  best.i <- which.min(errors)
  new.args <- pointsAll[[best.i]]$adc.args
  return (list(args = new.args, num.cv = length(pointsAll)))
}

################################################################################
# File:             ddalpha-internal.r
# Created by:       Pavlo Mozharovskyi, Oleksii Pokotylo
# First published:  28.02.2013
# Last revised:     20.02.2019
# 
# Contains the internal functions of the DDalpha-classifier.
# 
# For a description of the algorithm, see:
#   Lange, T., Mosler, K. and Mozharovskyi, P. (2014). Fast nonparametric 
#     classification based on data depth. Statistical Papers.
#   Mozharovskyi, P., Mosler, K. and Lange, T. (2015). Classifying real-world 
#     data with the DDalpha-procedure. Mimeo.
################################################################################

.ddalpha.create.structure <- function(formula, data, subset, ...){
  
  # if the user calls ddalpha(data, <other named parameters>, ...)
  # for backward compatibility
  if(!missing(formula) && missing(data) && (is.data.frame(formula) || is.matrix(formula))){
    data = formula
    formula = NULL
  }
  
  # formula is present
  if(!missing(formula) && !is.null(formula)){
    
    needed.frame <- 1 # PM(2018-06-22)

    cl <- match.call(call = sys.call(sys.parent(n = needed.frame)))
    mf <- match.call(expand.dots = FALSE, call = sys.call(sys.parent(n = needed.frame)))
    m <- match(c("formula", "data", "subset"#, "weights", "na.action", "offset"
                 ), names(mf), 0L)
    mf <- mf[c(1L, m)]
    mf$drop.unused.levels <- TRUE
    mf[[1L]] <- quote(stats::model.frame)
    mf <- eval(mf, parent.frame())
    
    clres = colnames(mf)
    cat("Selected columns: ", paste(clres, collapse = ",   "), "\n")
    classif.formula = delete.response(terms(mf))
    
    data = cbind(mf[,-1,drop=F], mf[,1,drop=F])
    
  #  y = model.response(mf)
  #  mm = model.matrix(formula, data = mf)
  #  mm = as.data.frame(mm[,-1,drop=F])
    
  #  colnames(mm) <- paste0("P_",1:ncol(mm))
  #  data = cbind(mm, y)
    
    #colnames(mf) <- paste0("P_",1:ncol(mm))
                           
  } else {  # no formula
    # Check for data consistency
    if (!(is.matrix(data) && is.numeric(data)
          || is.data.frame(data) && prod(sapply(data[,-ncol(data)], is.numeric)))){
      stop("Argument data has unacceptable format. Classifier can not be trained!!!")
    }
    cl = NULL
    classif.formula = NULL
    clres = colnames(data)
    
    if(!is.data.frame(data))
      data = as.data.frame(data)
    if(!missing(subset))
      data = data[subset,]
  }
  
  names(data)[ncol(data)] <- "CLASS"
  
  # Elemantary statistics
  dimension <- ncol(data) - 1
  numOfPoints <- nrow(data)
  classNames <- unique(data[,dimension + 1])
  numOfClasses <- length(classNames)

  if(!is.data.frame(data))
    data = as.data.frame(data)
  names(data)[ncol(data)] <- "CLASS"
  
  # Creating overall structure
  ddalpha <- list(
         call = cl,
         colnames = clres, # colnames after formula is applied
         classif.formula = classif.formula, # for classification
         raw = data, # after formula is applied
         dimension = dimension, 
         numPatterns = numOfClasses, 
         numPoints = numOfPoints, 
         patterns = list(), 
         needtransform = 0,    # 0 - no transform, 1 - transform new points before classification, all classes the same, 2 - transform differently w.r.t. to classes.
         classifiers = list(), 
#         numClassifiers = 0, 
#         methodDepth = "halfspace", 
#         methodSeparator = "alpha", 
#         methodAggregation = "majority", 
#         methodsOutsider = NULL, 
#         numDirections = 1000, 
#         directions = NULL, 
#         projections = NULL, 
         sameDirections = TRUE, 
         useConvex = FALSE, 
         maxDegree = 3, 
         numChunks = numOfPoints, 
#         knnrange = NULL,
#         mahEstimate = "moment", 
#         mahParMcd = 0.75, 
#         mahPriors = NULL, 
#         knnK = 1, 
#         knnX = NULL, 
#         knnY = NULL, 
#         knnD = NULL, 
         treatments = c("LDA", "KNNAFF", "KNN", "DEPTH.MAHALANOBIS", "RANDEQUAL", "RANDPROP", "IGNORE"))

  # Ordering patterns according to their cardinalities
  classCardinalities <- rep(0, numOfClasses)
  for (i in 1:numOfClasses){
    classCardinalities[i] <- nrow(data[data[,dimension + 1] == classNames[i],])
  }
  # Creating pattern templates
  patterns <- as.list("")
  for (i in 1:numOfClasses){
    maxCarIndex <- which.max(classCardinalities)
    # Creating a single template
    ddalpha$patterns[[i]] <- list(
      index = i, 
      points = data[data[,dimension + 1] == classNames[maxCarIndex],1:dimension], 
      name = classNames[maxCarIndex], 
      cardinality = classCardinalities[maxCarIndex], 
      depths = matrix(rep(0, numOfClasses*classCardinalities[maxCarIndex]), nrow = classCardinalities[maxCarIndex], ncol = numOfClasses), 
      votes = 0#, 
      #         center = 0, 
      #         cov = 0, 
      #         sigma = 0, 
      #         centerMcd = 0, 
      #         covMcd = 0, 
      #         sigmaMcd = 0
    )    

    # Adding pattern template to the list of patterns
    class(ddalpha$patterns[[i]])<-"ddalpha.pattern"
    # Deleting processed pattern
    classCardinalities[maxCarIndex] <- -1    
  }

  return (ddalpha)
}

.check.depth.exists <- function(depth) {
  fname = paste0(".", depth, "_validate")
  f <- try(match.fun(fname), silent = T)
  if (!is.function(f))
    warning(paste0("No validator function: ", fname))
  fname = paste0(".", depth, "_learn")
  
  f <- (match.fun(fname))
  if (!is.function(f))
    stop(paste0("No function: ", fname))
  fname = paste0(".", depth, "_depths")
  f <- (match.fun(fname))
  if (!is.function(f))
    stop(paste0("No function: ", fname))
}

.ddalpha.learn.depth <- function(ddalpha){
  
  # try to find a custom depth
  fname = paste0(".", ddalpha$methodDepth, "_learn")
  f <- try(match.fun(fname), silent = T)
  if (is.function(f)){
    ddalpha = f(ddalpha)
    return(ddalpha)
  }
  
  # If it's the random Tukey depth, compute it first
  if (ddalpha$methodDepth == "halfspace"){
    dSpaceStructure <- .halfspace_space(ddalpha)
    ddalpha$directions <- dSpaceStructure$directions
    ddalpha$projections <- dSpaceStructure$projections
    tmpDSpace <- dSpaceStructure$dspace
  }
  if (ddalpha$methodDepth == "projection"){
    dSpaceStructure <- .projection_space(ddalpha)
    tmpDSpace <- dSpaceStructure$dspace
    if (ddalpha$dmethod == "random"){
      ddalpha$directions <- dSpaceStructure$directions
      ddalpha$projections <- dSpaceStructure$projections
    }
  }
  
  classBegin = 1
  if (ddalpha$methodDepth == "potential"){
    tmpDSpace <- .ddalpha.count.depths(ddalpha, NULL)
  }
  
  # Calculating depths in each pattern
  for (i in 1:ddalpha$numPatterns){
    if (   ddalpha$methodDepth == "halfspace" 
        || ddalpha$methodDepth == "projection"
        || ddalpha$methodDepth == "potential"){
      # Random depth is already calculated, just distribute
      ddalpha$patterns[[i]]$depths <- tmpDSpace[classBegin:(classBegin+ddalpha$patterns[[i]]$cardinality-1),]
      classBegin = classBegin+ddalpha$patterns[[i]]$cardinality
    } else
    if (ddalpha$methodDepth == "zonoid"){
      # Calculate depths for the class w.r.t all classes, saying to which of the classes the chunk belongs
      ddalpha$patterns[[i]]$depths <- .zonoid_depths(ddalpha, ddalpha$patterns[[i]]$points, i)
    } else
    if (ddalpha$methodDepth == "Mahalanobis"){
      ddalpha$patterns[[i]]$depths <- .Mahalanobis_depths(ddalpha, ddalpha$patterns[[i]]$points)
    } else
    if (ddalpha$methodDepth == "spatial"){
      # Calculate depths for the class w.r.t all classes, saying to which of the classes the chunk belongs
      ddalpha$patterns[[i]]$depths <- .spatial_depths(ddalpha, ddalpha$patterns[[i]]$points)
    } else
    if (ddalpha$methodDepth == "spatialLocal"){
      # Calculate depths for the class w.r.t all classes, saying to which of the classes the chunk belongs
      ddalpha$patterns[[i]]$depths <- .spatialLocal_depths(ddalpha, ddalpha$patterns[[i]]$points)
    } else
    if (ddalpha$methodDepth == "simplicial"){
      # Calculate depths for the class w.r.t all classes, saying to which of the classes the chunk belongs
      ddalpha$patterns[[i]]$depths <- .simplicial_depths(ddalpha, ddalpha$patterns[[i]]$points)
    } else
    if (ddalpha$methodDepth == "simplicialVolume"){
      # Calculate depths for the class w.r.t all classes, saying to which of the classes the chunk belongs
      ddalpha$patterns[[i]]$depths <- .simplicialVolume_depths(ddalpha, ddalpha$patterns[[i]]$points)
    } else
      stop("Unknown depth ", ddalpha$methodDepth)
  }

  return (ddalpha)
}

.getFunction <- function(fname) {
  f <- try(match.fun(fname), silent = T)
  if (!is.function(f))
    stop("Wrong or absent function: ", fname)
  f
}

.ddalpha.learn.binary <- function(ddalpha){
  
  fname = paste0(".", ddalpha$methodSeparator, "_learn")
  learn <- try(match.fun(fname), silent = T)
  if (!is.function(learn))
    stop("Wrong or absent function: ", fname)
  
  # Separating (calculating extensions and normals)
  counter <- 1
  # Determining multi-class behaviour
  if (ddalpha$methodAggregation == "majority"){
    for (i in 1:(ddalpha$numPatterns - 1)){
      for (j in (i + 1):ddalpha$numPatterns){
        # Creating a classifier
        classifier <- learn(ddalpha, i, j,
                            ddalpha$patterns[[i]]$depths, 
                            ddalpha$patterns[[j]]$depths)
        
        classifier$index          = counter
        classifier$index1         = i
        classifier$index2         = j
        
        if(inherits(classifier, "list"))
          class(classifier) <- paste0("ddalpha.", ddalpha$methodSeparator)
        else
          class(classifier) <- c(class(classifier), paste0("ddalpha.", ddalpha$methodSeparator))
        
        # Adding the classifier to the list of classifiers
        ddalpha$classifiers[[counter]] <- classifier
        
        counter <- counter + 1
      }
    }
    ddalpha$numClassifiers <- counter - 1
  }
  if (ddalpha$methodAggregation == "sequent"){
    for (i in 1:ddalpha$numPatterns){
      anotherClass <- NULL
      for (j in 1:ddalpha$numPatterns){
        if (j != i){
          anotherClass <- rbind(anotherClass, ddalpha$patterns[[j]]$depths)
        }
      }
      classifier <- learn(ddalpha, i, -i,
                          ddalpha$patterns[[i]]$depths, 
                          anotherClass)
      
      classifier$index          = counter
      classifier$index1         = i
      classifier$index2         = -i
      
      if(inherits(classifier, "list"))
        class(classifier) <- paste0("ddalpha.", ddalpha$methodSeparator)
      else
        class(classifier) <- c(class(classifier), paste0("ddalpha.", ddalpha$methodSeparator))
      
      # Adding the classifier to the list of classifiers
      ddalpha$classifiers[[i]] <- classifier
    }
    ddalpha$numClassifiers <- ddalpha$numPatterns
  }
  
  return (ddalpha)
}

.alpha_learn <- function(ddalpha, index1, index2, depths1, depths2){
  points <- as.vector(t(rbind(depths1, depths2)))
  numClass1 <- nrow(depths1)
  numClass2 <- nrow(depths2)
  numPoints <- numClass1 + numClass2
  dimension <- ncol(depths1)
  cardinalities <- c(numClass1, numClass2)
  upToPower <- ddalpha$maxDegree
  minFeatures <- 2
  maxExtDimension <- (factorial(dimension + upToPower) / (factorial(dimension)*factorial(upToPower))) - 1;
  
  p <- .C("AlphaLearnCV", as.double(points), as.integer(numPoints), as.integer(dimension), as.integer(cardinalities),  as.integer(upToPower), as.integer(ddalpha$numChunks), as.integer(minFeatures), as.integer(ddalpha$debug), portrait=double(maxExtDimension + 1))$portrait
  
  degree <- p[1];
  extDimension <- (factorial(dimension + degree) / (factorial(dimension)*factorial(degree))) - 1;
  d <- 0
  for (i in 2:(extDimension + 1)){
    if (p[i] != 0){
      d <- d + 1
    }
  }
  
  return(list(
    hyperplane     = p,
    degree         = degree,
    dimProperties  = length(p) - 1,
    dimFeatures    = d
  ))
}

.alpha_classify <- function(ddalpha, classifier, depths){
  toClassify <- as.double(as.vector(t(depths)))
  m = as.integer(nrow(depths))
  q = as.integer(ncol(depths))
  result <- .C("AlphaClassify", 
               toClassify, 
               m, q, 
               as.integer(classifier$degree), 
               as.double(classifier$hyperplane), 
               output=integer(m))$output
  return(result)
}

.polynomial_learn <- function(ddalpha, index1, index2, depths1, depths2){
  
  polynomial <- .polynomial_learn_C(ddalpha$maxDegree, 
                                    rbind(depths1[,c(index1, index2)], depths2[,c(index1, index2)]),
                                    nrow(depths1), nrow(depths2), 
                                    ddalpha$numChunks, ddalpha$seed)
  return(list(
    polynomial     = polynomial$coefficients,
    degree         = polynomial$degree,
    axis           = polynomial$axis))
}

.polynomial_classify <- function(ddalpha, classifier, depths){

  x = ifelse(classifier$axis == 0, classifier$index1, classifier$index2)
  y = ifelse(classifier$axis == 0, classifier$index2, classifier$index1)
  
  result <- (- depths[,y])
  for (obj in 1:nrow(depths)){        
    val <- depths[obj,x]
    for(j in 1:classifier$degree){
      result[obj] <- result[obj] + classifier$polynomial[j]*val^j
    }
  }
  return(result)
}


.ddalpha.learn.multiclass <- function(ddalpha){
  
  fname = paste0(".", ddalpha$methodSeparator, "_learn")
  learn <- try(match.fun(fname), silent = T)
  if (!is.function(learn))
    stop("Wrong or absent function: ", fname)
  
  classifier <- learn(ddalpha)
  if(inherits(classifier, "list"))
    class(classifier) <- paste0("ddalpha.", ddalpha$methodSeparator)
  else
    class(classifier) <- c(class(classifier), paste0("ddalpha.", ddalpha$methodSeparator))
      
  ddalpha$classifiers[[1]] <- classifier
  ddalpha$numClassifiers <- 1
  
  return (ddalpha)
}

#.ddalpha.learn.knnlm 
.knnlm_learn <- function(ddalpha){
  
  x <- NULL
  y <- NULL
  for (i in 1:ddalpha$numPatterns){
    x <- rbind(x, ddalpha$patterns[[i]]$depths)
    y <- c(y, rep(i - 1, ddalpha$patterns[[i]]$cardinality))
  }
  x <- as.vector(t(x))
  y <- as.vector(y)

  k <- .C("KnnLearnJK", 
          as.double(x), 
          as.integer(y), 
          as.integer(ddalpha$numPoints), 
          as.integer(ddalpha$numPatterns), 
          as.integer(ddalpha$knnrange), 
          as.integer(2), 
          k=integer(1))$k

  return (list(knnK = k,
               knnX = x,
               knnY = y))
}

.knnlm_classify <- function(ddalpha, classifier, depths){
  z <- as.vector(t(depths))
  output <- .C("KnnClassify", 
               as.double(z), 
               as.integer(nrow(depths)), 
               as.double(classifier$knnX), 
               as.integer(classifier$knnY), 
               as.integer(ddalpha$numPoints), 
               as.integer(ddalpha$numPatterns), 
               as.integer(classifier$knnK), 
               as.integer(2), 
               output=integer(nrow(depths)))$output
  
  return(output+1)
}

.maxD_learn <- function(ddalpha) return(list())

.maxD_classify <- function(ddalpha, classifier, depths) apply(depths, 1, which.max)

.ddalpha.learn.outsiders <- function(ddalpha, methodsOutsider = "LDA", settingsOutsider = NULL){
  # Refine treatments
  if (is.null(settingsOutsider)){
    ddalpha$methodsOutsider <- .parse.methods(methodsOutsider)
  }else{
    ddalpha$methodsOutsider <- .parse.settings(ddalpha, settingsOutsider)
  }
  # Train treatments
  treatments = list(LDA = .lda_learn, QDA = .qda_learn, 
                    KNN = .knn_learn, KNNAff = .knnAff_learn, depth.Mahalanobis = .mah_learn,
                    Ignore = NA, Mark = NA, RandEqual = NA, RandProp = NA)
  for (i in 1:length(ddalpha$methodsOutsider)){
    .treatment = treatments[[ddalpha$methodsOutsider[[i]]$method]]
    if(is.null(.treatment)) stop("Unknown outsiders treatment method ", ddalpha$methodsOutsider[[i]]$method)
    if(!is.function(.treatment)) next; # need no training
    ddalpha$methodsOutsider[[i]] <- .treatment(ddalpha, ddalpha$methodsOutsider[[i]])
  }
  
  return(ddalpha)
}

.ddalpha.count.depths <- function(ddalpha, objects, ...){
  fname = paste0(".", ddalpha$methodDepth, "_depths")
  f <- (match.fun(fname))
  if (!is.function(f))
    stop(paste0("Wrong function for ", ddalpha$methodDepth))
  
  # Count for all data
  if (is.null(objects))  
    for (i in 1:ddalpha$numPatterns){
      objects <- rbind(objects, ddalpha$patterns[[i]]$points)
    }
  else ## if needtransform == 1 the data is already scaled
  # Transform the data once
  if (ddalpha$needtransform == 1){
    objects <- ddalpha$patterns[[1]]$transformer(objects)
  } 
  
  # Calculate depths for all classes together
  if (ddalpha$needtransform != 2){
    res <- f(ddalpha, objects, ...)
    return(res) 
  }
  # Calculate depths w.r.t. each class
  else {
    d <- NULL
    # w.r.t. each class
    for (cls in 1:ddalpha$numPatterns){      
      depth <- f(ddalpha, objects, class = cls, ...)
      
      d = cbind(d, depth)
    }
    
    return(d)    
  }
}

.ddalpha.classify.outsiders<- function (objects, ddalpha, settings){
  if (settings$method == "Ignore"){
    return (.ignore_classify(nrow(objects)))
  }
  if (settings$method == "Mark"){
    return (.right_classify(nrow(objects)))
  }
  if (settings$method == "RandEqual"){
    return (.randequal_classify(nrow(objects), ddalpha))
  }
  if (settings$method == "RandProp"){
    return (.randprop_classify(nrow(objects), ddalpha, settings))
  }
  
  treatments = list(LDA = .lda_classify, QDA = .qda_classify, 
                    KNN = .knn_classify, KNNAff = .knnAff_classify, 
                    depth.Mahalanobis = .mah_classify)

  .treatment = treatments[[settings$method]]
  if(is.null(.treatment)) stop("Unknown outsiders treatment method ", settings$method)
  classified = .treatment(objects, ddalpha, settings)
  return(classified)
}

################################################################################
# Functions used for intermediate calculations and checks are presented below
################################################################################

.are_classifiable <- function(objects, points, cardinalities){
  convexes <- .count_convexes(objects, points, cardinalities)
  return (ifelse(rowSums(convexes)>0,1,0))
}

.count_convexes <- function(objects, points, cardinalities, seed = 0){
  if (is.na(seed)) seed = 0
  x <- as.vector(t(points))
  dimension <- ncol(points)
  numClasses <- length(cardinalities)
  o <- as.vector(t(objects))
  numObjects <- nrow(objects)
  result <- .C("IsInConvexes", 
               as.double(x), 
               as.integer(dimension), 
               as.integer(cardinalities), 
               as.integer(numClasses), 
               as.double(o), 
               as.integer(numObjects), 
               as.integer(seed), 
               isInConvexes=integer(numObjects*numClasses))$isInConvexes
  result <- matrix(result, byrow = T, ncol = numClasses)
  return (result)
}

.halfspace_space <- function(ddalpha){
  points <- NULL
  cardinalities <- NULL
  for (i in 1:ddalpha$numPatterns){
    points <- rbind(points, ddalpha$patterns[[i]]$points)
    cardinalities <- c(cardinalities, ddalpha$patterns[[i]]$cardinality)
  }
  x <- as.vector(t(points))
  c <- as.vector(cardinalities)
  k <- ddalpha$numDirections

  method = ddalpha$dmethod
  
  if (method == 0){
    if (ddalpha$sameDirections){
      rez <- .C("HDSpace", as.double(x), as.integer(ncol(points)), as.integer(c), as.integer(ddalpha$numPatterns), as.integer(k), as.integer(1), as.integer(ddalpha$seed), dspc=double(nrow(points)*ddalpha$numPatterns), dirs=double(k*ncol(points)), prjs=double(k*nrow(points)))
      return (list(dspace=matrix(rez$dspc, nrow=nrow(points), ncol=ddalpha$numPatterns, byrow=TRUE), directions=rez$dirs, projections=rez$prjs))
    }else{
      rez <- .C("HDSpace", as.double(x), as.integer(ncol(points)), as.integer(c), as.integer(ddalpha$numPatterns), as.integer(k), as.integer(0), as.integer(ddalpha$seed), dspc=double(nrow(points)*ddalpha$numPatterns), dirs=double(k*ncol(points)), prjs=double(k*nrow(points)))
      return (list(dspace=matrix(rez$dspc, nrow=nrow(points), ncol=ddalpha$numPatterns, byrow=TRUE), directions=0, projections=0))
    }
  } else 
    if (method %in% 1:3){
      
      ds <- .C("HDepthSpaceEx", 
               as.double(x), 
               as.double(x), 
               as.integer(c), 
               as.integer(length(cardinalities)), 
               as.integer(nrow(points)),  
               as.integer(ncol(points)), 
               as.integer(method), 
               depths=double(nrow(points)*length(cardinalities)))$depths  
      
      return (list(dspace=matrix(ds, nrow=nrow(points), ncol=ddalpha$numPatterns, byrow=F)))
    }
  else 
    stop("wrong choise of the algorithm, method = ", method)
}

.halfspace_depths <- function(ddalpha, objects){
  points <- NULL
  cardinalities <- NULL
  for (i in 1:ddalpha$numPatterns){
    points <- rbind(points, ddalpha$patterns[[i]]$points)
    cardinalities <- c(cardinalities, ddalpha$patterns[[i]]$cardinality)
  }
  x <- as.vector(t(points))
  y <- as.vector(t(objects))
  c <- as.vector(cardinalities)
  k <- ddalpha$numDirections
  method <- ddalpha$dmethod
  if (method == 0){
    if (ddalpha$sameDirections){
      result <- .C("HDepth", 
                   as.double(x), 
                   as.double(y), 
                   as.integer(nrow(objects)), 
                   as.integer(ncol(points)), 
                   as.integer(c), 
                   as.integer(ddalpha$numPatterns), 
                   as.double(ddalpha$directions), 
                   as.double(ddalpha$projections), 
                   as.integer(k), 
                   as.integer(1), 
                   as.integer(ddalpha$seed),
                   depths=double(ddalpha$numPatterns*nrow(objects)))
    }else{
      result <- .C("HDepth", 
                   as.double(x), 
                   as.double(y), 
                   as.integer(nrow(objects)), 
                   as.integer(ncol(points)), 
                   as.integer(c), 
                   as.integer(ddalpha$numPatterns), 
                   dirs=double(k*ncol(points)), 
                   prjs=double(k*nrow(points)),
                   as.integer(k), 
                   as.integer(0), 
                   as.integer(ddalpha$seed),
                   depths=double(ddalpha$numPatterns*nrow(objects)))
    }
  }
  else
    if (method %in% 1:3){
      
      ds <- .C("HDepthSpaceEx", 
               as.double(x), 
               as.double(y), 
               as.integer(c), 
               as.integer(length(cardinalities)), 
               as.integer(nrow(objects)),  
               as.integer(ncol(points)), 
               as.integer(method), 
               depths=double(nrow(objects)*length(cardinalities)))$depths  
      
      depths <- matrix(ds, nrow=nrow(objects), ncol=length(cardinalities), byrow=F)
      return (depths)
  }
  else 
    stop("wrong choise of the algorithm, method = ", method)
  return (matrix(result$depths, nrow=nrow(objects), ncol=ddalpha$numPatterns, byrow=TRUE))
}

.zonoid_depths <- function(ddalpha, objects, ownPattern = 0){
  depths <- NULL
  for (i in 1:ddalpha$numPatterns){
    pattern <- ddalpha$patterns[[i]]$points
    x <- as.vector(t(pattern))
    y <- as.vector(t(objects))
    ds <- .C("ZDepth", as.double(x), as.double(y), as.integer(nrow(pattern)), as.integer(nrow(objects)), 
             as.integer(ncol(pattern)), as.integer(ddalpha$seed), depths=double(nrow(objects)))$depths
    if (i == ownPattern){
      ds <- replace(ds, which(ds < 1/nrow(pattern) - sqrt(.Machine$double.eps)), 1/nrow(pattern))
    }else{
      ds <- replace(ds, which(ds < 1/nrow(pattern) - sqrt(.Machine$double.eps)), 0)
    }
    depths <- cbind(depths, ds)
  }
  return (depths)
}

.estimate_moments <- function(ddalpha){
  for (i in 1:ddalpha$numPatterns){
    if (ddalpha$mahEstimate == "moment"){
      
      ddalpha$patterns[[i]]$center <- colMeans(ddalpha$patterns[[i]]$points)
      ddalpha$patterns[[i]]$cov    <- cov(ddalpha$patterns[[i]]$points)
      try(
        ddalpha$patterns[[i]]$sigma  <- solve(ddalpha$patterns[[i]]$cov)
      )        
    }
    if (ddalpha$mahEstimate == "MCD"){
      try(
        estimate <- covMcd(ddalpha$patterns[[i]]$points, ddalpha$mahParMcd)
      )
      try(
        ddalpha$patterns[[i]]$center <- estimate$center
      )
      try(
        ddalpha$patterns[[i]]$cov    <- estimate$cov
      )
      try(
        ddalpha$patterns[[i]]$sigma  <- solve(estimate$cov)
      )
    }
  }
  return(ddalpha)
}

.Mahalanobis_learn <- function(ddalpha){

  if (is.null(ddalpha$mahPriors)){
    ddalpha$mahPriors <- c()
    for (i in 1:ddalpha$numPatterns){
      ddalpha$mahPriors[i] <- ddalpha$patterns[[i]]$cardinality/ddalpha$numPoints
    }
  }
  
  ddalpha <- .estimate_moments(ddalpha)
  
  for (i in 1:ddalpha$numPatterns){
    ddalpha$patterns[[i]]$depths <- .Mahalanobis_depths(ddalpha, ddalpha$patterns[[i]]$points)
  }
  
  return(ddalpha)
}

.Mahalanobis_depths <- function(ddalpha, objects){
  depths <- NULL
  for (j in 1:ddalpha$numPatterns){
    depths <- cbind(depths, .Mahalanobis_depth (objects, center = ddalpha$patterns[[j]]$center, sigma = ddalpha$patterns[[j]]$sigma))
  }
  return (depths)
}

.Mahalanobis_depth <- function(points, center = colMeans(points), sigma = solve(cov(points))){
  if (is.data.frame(points))
    points <- as.matrix(points, drop = F)
  if(is.vector(points))
    points <- t(as.matrix(points, drop = F))
  if (!is.matrix(points))
    stop("Wrong format of 'points'")
  
  i = 1; step = 200
  d <- NULL
  while (i<=nrow(points)){
    tmp1 <- t(t(points[i:min(i+step, nrow(points)),, drop = F]) - center)

    dd <- diag(tmp1 %*% sigma %*% t(tmp1))
    d <- c(d,1/(1 + dd))
    i = i+1+step
  }

   # d <- 1/(1 + (points - center) %*% sigma %*% t(points - center))
  return (d)
}

.projection_space <- function(ddalpha){
  points <- NULL
  cardinalities <- NULL
  for (i in 1:ddalpha$numPatterns){
    points <- rbind(points, ddalpha$patterns[[i]]$points)
    cardinalities <- c(cardinalities, ddalpha$patterns[[i]]$cardinality)
  }
  if (ddalpha$dmethod == "random"){
    x <- as.vector(t(points))
    y <- as.vector(t(points))
    c <- as.vector(cardinalities)
    k <- ddalpha$numDirections
    result <- .C("ProjectionDepth", 
                 as.double(x), 
                 as.double(y), 
                 as.integer(nrow(points)), 
                 as.integer(ncol(points)), 
                 as.integer(c), 
                 as.integer(ddalpha$numPatterns), 
                 dirs=double(k*ncol(points)), 
                 prjs=double(k*nrow(points)), 
                 as.integer(k), 
                 as.integer(1), 
                 as.integer(ddalpha$seed),
                 dspc=double(ddalpha$numPatterns*nrow(points)))
    return (list(dspace=matrix(result$dspc, nrow=nrow(points), 
                               ncol=ddalpha$numPatterns, byrow=TRUE), 
                 directions=result$dirs, projections=result$prjs))
  }
  if (ddalpha$dmethod == "linearize"){
    depths <- NULL
    for (i in 1:ddalpha$numPatterns){
      ds <- .zdepth(ddalpha$patterns[[i]]$points, points)
      depths <- cbind(depths, ds)
    }
    return (list(dspace=depths))
  }
}

.projection_depths <- function(ddalpha, objects){
  points <- NULL
  cardinalities <- NULL
  for (i in 1:ddalpha$numPatterns){
    points <- rbind(points, ddalpha$patterns[[i]]$points)
    cardinalities <- c(cardinalities, ddalpha$patterns[[i]]$cardinality)
  }
  if (ddalpha$dmethod == "random"){
    x <- as.vector(t(points))
    y <- as.vector(t(objects))
    c <- as.vector(cardinalities)
    k <- ddalpha$numDirections
    result <- .C("ProjectionDepth", 
                 as.double(x), 
                 as.double(y), 
                 as.integer(nrow(objects)), 
                 as.integer(ncol(points)), 
                 as.integer(c), 
                 as.integer(ddalpha$numPatterns), 
                 as.double(ddalpha$directions), 
                 as.double(ddalpha$projections), 
                 as.integer(k), 
                 as.integer(0), 
                 as.integer(ddalpha$seed),
                 depths=double(ddalpha$numPatterns*nrow(objects)))
    return (matrix(result$depths, nrow=nrow(objects), ncol=ddalpha$numPatterns, byrow=TRUE))
  }
  if (ddalpha$dmethod == "linearize"){
    depths <- NULL
    for (i in 1:ddalpha$numPatterns){
      ds <- .zdepth(ddalpha$patterns[[i]]$points, objects)
      depths <- cbind(depths, ds)
    }
    return (depths)
  }
}

.simplicialVolume_depths <- function(ddalpha, objects){
  if (is.data.frame(objects))
    objects = as.matrix(objects)
  depths <- NULL
  for (i in 1:ddalpha$numPatterns){
    pattern <- ddalpha$patterns[[i]]$points
    
    points <- as.vector(t(pattern))
    x <- as.vector(t(objects))
    if (ddalpha$d_useCov == 0){
      covEst <- diag(ncol(pattern))
    } else if (ddalpha$d_useCov == 1){
      covEst <- cov(pattern)
    } else if (ddalpha$d_useCov == 2){
      covEst <- covMcd(pattern, ddalpha$d_parMcd)$cov
    }
    ds <- .C("OjaDepth", 
             as.double(points), 
             as.double(x), 
             as.integer(nrow(pattern)), 
             as.integer(nrow(objects)), 
             as.integer(ncol(pattern)), 
             as.integer(ddalpha$seed),
             as.integer(ddalpha$d_exact),
             as.integer(.longtoint(ddalpha$d_k)),
             as.integer(ddalpha$d_useCov),
             as.double(as.vector(t(covEst))),
             depths=double(nrow(objects)))$depths
    depths <- cbind(depths, ds, deparse.level = 0)
  }  
  return (depths) 
}

.simplicial_depths <- function(ddalpha, objects){
  if (is.data.frame(objects))
    objects = as.matrix(objects)
  depths <- NULL
  for (i in 1:ddalpha$numPatterns){
    pattern <- ddalpha$patterns[[i]]$points
    
    points <- as.vector(t(pattern))
    x <- as.vector(t(objects))
    ds <- .C("SimplicialDepth", 
             as.double(points), 
             as.double(x), 
             as.integer(nrow(pattern)), 
             as.integer(nrow(objects)), 
             as.integer(ncol(pattern)), 
             as.integer(ddalpha$seed),
             as.integer(ddalpha$d_exact),
             as.integer(.longtoint(ddalpha$d_k)),           
             depths=double(nrow(objects)))$depths
    depths <- cbind(depths, ds, deparse.level = 0)
  }  
  return (depths) 
}



.spatial_learn <- function(ddalpha){
  if (ddalpha$mahEstimate == "none"){
    for (i in 1:ddalpha$numPatterns){
        ddalpha$patterns[[i]]$center <- colMeans(ddalpha$patterns[[i]]$points)
        ddalpha$patterns[[i]]$cov <- NA
    }
  }else{
    ddalpha <- .estimate_moments(ddalpha)
  }
  
  for (i in 1:ddalpha$numPatterns){
    ddalpha$patterns[[i]]$depths <- .spatial_depths(ddalpha, ddalpha$patterns[[i]]$points)
  }
  
  return(ddalpha)
}

.spatialLocal_learn <- function(ddalpha){
  
  ddalpha <- .estimate_moments(ddalpha)
  
  for (i in 1:ddalpha$numPatterns){
    ddalpha$patterns[[i]]$depths <- .spatialLocal_depths(ddalpha, ddalpha$patterns[[i]]$points)
  }
  
  return(ddalpha)
}

.spatial_depths <- function(ddalpha, objects){
  if (is.data.frame(objects))
    objects = as.matrix(objects)
  depths <- NULL
  for (i in 1:ddalpha$numPatterns){
    pattern <- ddalpha$patterns[[i]]$points
    mean <- ddalpha$patterns[[i]]$center
    cov <- ddalpha$patterns[[i]]$cov
    #suppressWarnings(
    if(sum(is.na(cov)) == 0){
      cov.eig <- eigen(cov)
      B <- cov.eig$vectors %*% diag(sqrt(cov.eig$values))
      lambda <- solve(B)
    } else{
      lambda = diag(ncol(pattern))
    }#)
    ds <- rep(-1, nrow(objects))
    for (i in 1:nrow(objects)){
      tmp1 <- t(lambda %*% (objects[i,] - t(pattern)))
      tmp1 <- tmp1[which(rowSums(tmp1) != 0),]
      tmp2 <- 1/sqrt(rowSums(tmp1^2))
      ds[i] <- 1 - sqrt(sum((colSums(tmp2*tmp1)/nrow(pattern))^2))
    }
    depths <- cbind(depths, ds)
  }
  return (depths)
}

.spatialLocal_depths <- function(ddalpha, objects){
  depths <- NULL
  for (i in 1:ddalpha$numPatterns){
    depths <- cbind(depths, depth.spatial.local(objects, ddalpha$patterns[[i]]$points, ddalpha$kernel.bandwidth[i]))
  }
  return (depths)
}

.NONE_depths <- function(ddalpha, objects){
  depths <- matrix(0, ncol = ddalpha$numPatterns, nrow = nrow(objects))
  
  return (depths)
}
#==========================================================

.alpha_learnOLD <- function(maxDegree, data, numClass1, numClass2, numChunks, debug = F){
  points <- as.vector(t(data))
  numPoints <- numClass1 + numClass2
  dimension <- ncol(data)
  cardinalities <- c(numClass1, numClass2)
  upToPower <- maxDegree
  minFeatures <- 2
  maxExtDimension <- (factorial(dimension + maxDegree) / (factorial(dimension)*factorial(maxDegree))) - 1;
  
  p <- .C("AlphaLearnCV", as.double(points), as.integer(numPoints), as.integer(dimension), as.integer(cardinalities),  as.integer(upToPower), as.integer(numChunks), as.integer(minFeatures), as.integer(debug), portrait=double(maxExtDimension + 1))$portrait
  
  degree <- p[1];
  extDimension <- (factorial(dimension + degree) / (factorial(dimension)*factorial(degree))) - 1;
  d <- 0
  for (i in 2:(extDimension + 1)){
    if (p[i] != 0){
      d <- d + 1
    }
  }
  return(list(por=p,dim=d,deg=degree))
}

.parse.methods <- function(methods){
  if (!is.vector(methods) || length(methods) <= 0){return(list())}
  methods.refined <- unique(toupper(methods))
  
  treatments.settings <- list()
  counter <- 1
  for (i in 1:length(methods.refined)){
    supported <- FALSE
    if (methods.refined[i] == "LDA"){
      supported <- TRUE
      treatment.settings <- structure(
        list(name = "LDA", 
             method = "LDA", 
             priors = NULL, 
             lda = NULL), 
        .Names = c("name", "method", "priors", "lda"))
    }
    if (methods.refined[i] == "QDA"){
      supported <- TRUE
      treatment.settings <- structure(
        list(name = "QDA", 
             method = "QDA", 
             priors = NULL, 
             qda = NULL))
    }
    if (methods.refined[i] == "KNNAFF"){
      supported <- TRUE
      treatment.settings <- structure(
        list(name = "KNNAff", 
             method = "KNNAff", 
             knnAff.methodAggregation = "majority", 
             knnAff.range = -1, 
             knnAff.k = -1, 
             knnAff.classifiers = NULL), 
        .Names = c("name", "method", "knnAff.methodAggregation", "knnAff.range", "knnAff.k", "knnAff.classifiers"))
    }
    if (methods.refined[i] == "KNN"){
      supported <- TRUE
      treatment.settings <- structure(
        list(name = "KNN", 
             method = "KNN", 
             knn.range = -1, 
             knn.k = -1, 
             knn.train = NULL, 
             knn.cl = NULL), 
        .Names = c("name", "method", "knn.range", "knn.k", "knn.train", "knn.cl"))
    }
    if (methods.refined[i] == "DEPTH.MAHALANOBIS"){
      supported <- TRUE
      treatment.settings <- structure(
        list(name = "depth.Mahalanobis", 
             method = "depth.Mahalanobis", 
             mah.estimate = "moment", 
             priors = NULL, 
             mah.classes = NULL, 
             mcd.alpha = 0.5), 
        .Names = c("name", "method", "mah.estimate", "priors", "mah.classes", "mcd.alpha"))
    }
    if (methods.refined[i] == "RANDEQUAL"){
      supported <- TRUE
      treatment.settings <- structure(
        list(name = "RandEqual", 
             method = "RandEqual"), 
        .Names = c("name", "method"))
    }
    if (methods.refined[i] == "RANDPROP"){
      supported <- TRUE
      treatment.settings <- structure(
        list(name = "RandProp", 
             method = "RandProp", 
             priors = NULL), 
        .Names = c("name", "method", "priors"))
    }
    if (methods.refined[i] == "IGNORE"){
      supported <- TRUE
      treatment.settings <- structure(
        list(name = "Ignore", 
             method = "Ignore"), 
        .Names = c("name", "method"))
    }
    if (supported){
      treatments.settings[[counter]] <- treatment.settings
      counter <- counter + 1
    }
  }
  return(treatments.settings)
}

.parse.settings <- function(ddalpha, settings){
  if (!is.list(settings) || length(settings) <= 0){return(list())}
  treatments.names <- c()
  
  treatments.settings <- list()
  counter <- 1
  for (i in 1:length(settings)){
    supported <- FALSE
    if (!is.list(settings[[i]]) 
        || is.null(settings[[i]]$name) 
        || is.null(settings[[i]]$method)){
      warning("In treatment number ", i, ": The treatment has unacceptable format. The treatment will be ignored")
      next
    }
    if (!is.character(settings[[i]]$method) 
        || length(settings[[i]]$method) != 1 
        || !(toupper(settings[[i]]$method) %in% ddalpha$treatments)){
      warning("In treatment number ", i, ": The method name of the treatment is not acceptable. The treatment will be ignored")
      next
    }
    if (!is.character(settings[[i]]$name) 
        || length(settings[[i]]$name) != 1){
      warning("In treatment number ", i, ": The name of the treatment is not acceptable. The treatment will be ignored")
      next
    }
    if (settings[[i]]$name %in% treatments.names){
      warning("In treatment number ", i, ": Treatment with the name ", settings[[i]]$name, " already exists. The treatment will be ignored")
      next
    }else{
      treatments.names <- c(treatments.names, settings[[i]]$name)
    }
    if (toupper(settings[[i]]$method) == "LDA"){
      tmp.name <- settings[[i]]$name
      if (!is.vector(settings[[i]]$priors, mode = "double") 
          || is.na(min(settings[[i]]$priors)) 
          || length(settings[[i]]$priors) != ddalpha$numPatterns 
          || min(settings[[i]]$priors) <= 0 
          || max(settings[[i]]$priors) <= 0){
        warning("In treatment number ", i, ": Argument \"priors\" not specified correctly. Defaults in the form of class portions are applied")
        tmp.priors <- NULL
      }else{
        tmp.priors <- settings[[i]]$priors/sum(settings[[i]]$priors)
      }
      treatment.settings <- structure(
        list(name = tmp.name, 
             method = "LDA", 
             priors = tmp.priors, 
             lda = NULL), 
        .Names = c("name", "method", "priors", "lda"))
      supported <- TRUE
    }
    if (toupper(settings[[i]]$method) == "KNNAFF"){
      tmp.name <- settings[[i]]$name
      if (!is.character(settings[[i]]$knnAff.methodAggregation) 
          || length(settings[[i]]$knnAff.methodAggregation) != 1 
          || !(settings[[i]]$knnAff.methodAggregation %in% c("majority", "sequent"))){
        warning("In treatment number ", i, ": Argument \"knnAff.methodAggregation\" not specified correctly. \"majority\" is used as a default value")
        tmp.knnAff.methodAggregation <- "majority"
      }else{
        tmp.knnAff.methodAggregation <- settings[[i]]$knnAff.methodAggregation
      }
      if (!is.numeric(settings[[i]]$knnAff.range) 
          || is.na(settings[[i]]$knnAff.range) 
          || length(settings[[i]]$knnAff.range) != 1 
          || !.is.wholenumber(settings[[i]]$knnAff.range) 
          || !(settings[[i]]$knnAff.range >= 2 
               && settings[[i]]$knnAff.range <= (ddalpha$patterns[[ddalpha$numPatterns]]$cardinality + ddalpha$patterns[[ddalpha$numPatterns - 1]]$cardinality - 1) 
               || settings[[i]]$knnAff.range == -1)){
        warning("In treatment number ", i, ": Argument \"knnAff.range\" not specified correctly. Defaults are applied")
        tmp.knnAff.range <- -1
      }else{
        tmp.knnAff.range <- settings[[i]]$knnAff.range
      }
      if (!is.numeric(settings[[i]]$knnAff.k) 
          || is.na(settings[[i]]$knnAff.k) 
          || length(settings[[i]]$knnAff.k) != 1 
          || !.is.wholenumber(settings[[i]]$knnAff.k) 
          || !(settings[[i]]$knnAff.k >= 1 
               && settings[[i]]$knnAff.k <= (ddalpha$patterns[[ddalpha$numPatterns]]$cardinality + ddalpha$patterns[[ddalpha$numPatterns - 1]]$cardinality) 
               || settings[[i]]$knnAff.k == -1)){
        warning("In treatment number ", i, ": Argument \"knnAff.k\" not specified correctly. Defaults are applied")
        tmp.knnAff.k <- -1
      }else{
        tmp.knnAff.k <- settings[[i]]$knnAff.k
      }
      treatment.settings <- structure(
        list(name = tmp.name, 
             method = "KNNAff", 
             knnAff.methodAggregation = tmp.knnAff.methodAggregation, 
             knnAff.range = tmp.knnAff.range, 
             knnAff.k = tmp.knnAff.k, 
             knnAff.classifiers = NULL), 
        .Names = c("name", "method", "knnAff.methodAggregation", "knnAff.range", "knnAff.k", "knnAff.classifiers"))
      supported <- TRUE
    }
    if (toupper(settings[[i]]$method) == "KNN"){
      tmp.name <- settings[[i]]$name
      if (!is.numeric(settings[[i]]$knn.range) 
          || is.na(settings[[i]]$knn.range) 
          || length(settings[[i]]$knn.range) != 1 
          || !.is.wholenumber(settings[[i]]$knn.range) 
          || !(settings[[i]]$knn.range >= 2 
               && settings[[i]]$knn.range <= (ddalpha$numPoints - 1) 
               || settings[[i]]$knn.range == -1)){
        warning("In treatment number ", i, ": Argument \"knn.range\" not specified correctly. Defaults are applied")
        tmp.knn.range <- -1
      }else{
        tmp.knn.range <- settings[[i]]$knn.range
      }
      if (!is.numeric(settings[[i]]$knn.k) 
          || is.na(settings[[i]]$knn.k) 
          || length(settings[[i]]$knn.k) != 1 
          || !.is.wholenumber(settings[[i]]$knn.k) 
          || !(settings[[i]]$knn.k >= 1 
               && settings[[i]]$knn.k <= (ddalpha$numPoints) 
               || settings[[i]]$knn.k == -1)){
        warning("In treatment number ", i, ": Argument \"knn.k\" not specified correctly. Defaults are applied")
        tmp.knn.k <- -1
      }else{
        tmp.knn.k <- settings[[i]]$knn.k
      }
      treatment.settings <- structure(
        list(name = tmp.name, 
             method = "KNN", 
             knn.range = tmp.knn.range, 
             knn.k = tmp.knn.k, 
             knn.train = NULL, 
             knn.cl = NULL), 
        .Names = c("name", "method", "knn.range", "knn.k", "knn.train", "knn.cl"))
      supported <- TRUE
    }
    if (toupper(settings[[i]]$method) == "DEPTH.MAHALANOBIS"){
      tmp.name <- settings[[i]]$name
      if (!is.character(settings[[i]]$mah.estimate) 
          || length(settings[[i]]$mah.estimate) != 1 
          || !(settings[[i]]$mah.estimate %in% c("moment", "MCD"))){
        warning("In treatment number ", i, ": Argument \"mah.estimate\" not specified correctly. \"moment\" is used as a default value")
        tmp.mah.estimate <- "moment"
      }else{
        tmp.mah.estimate <- settings[[i]]$mah.estimate
      }
      if (!is.vector(settings[[i]]$priors, mode = "double") 
          || is.na(min(settings[[i]]$priors)) 
          || length(settings[[i]]$priors) != ddalpha$numPatterns 
          || min(settings[[i]]$priors) <= 0 
          || max(settings[[i]]$priors) <= 0){
        warning("In treatment number ", i, ": Argument \"priors\" not specified correctly. Defaults in the form of class portions are applied")
        tmp.priors <- NULL
      }else{
        tmp.priors <- settings[[i]]$priors/sum(settings[[i]]$priors)
      }
      if (!is.vector(settings[[i]]$mcd.alpha, mode = "double") 
          || is.na(min(settings[[i]]$mcd.alpha)) 
          || length(settings[[i]]$mcd.alpha) != 1 
          || settings[[i]]$mcd.alpha < 0.5 
          || settings[[i]]$mcd.alpha > 1){
        if (tmp.mah.estimate == "MCD"){
          warning("In treatment number ", i, ": Argument \"mcd.alpha\" not specified correctly. 0.75 is used as a default value")
        }
        tmp.mcd.alpha <- 0.75
      }else{
        tmp.mcd.alpha <- settings[[i]]$mcd.alpha
      }
      treatment.settings <- structure(
        list(name = tmp.name, 
             method = "depth.Mahalanobis", 
             mah.estimate = tmp.mah.estimate, 
             priors = tmp.priors, 
             mah.classes = NULL, 
             mcd.alpha = tmp.mcd.alpha), 
        .Names = c("name", "method", "mah.estimate", "priors", "mah.classes", "mcd.alpha"))
      supported <- TRUE
    }
    if (toupper(settings[[i]]$method) == "RANDEQUAL"){
      tmp.name <- settings[[i]]$name
      treatment.settings <- structure(
        list(name = tmp.name, 
             method = "RandEqual"), 
        .Names = c("name", "method"))
      supported <- TRUE
    }
    if (toupper(settings[[i]]$method) == "RANDPROP"){
      tmp.name <- settings[[i]]$name
      if (!is.vector(settings[[i]]$priors, mode = "double") 
          || is.na(min(settings[[i]]$priors)) 
          || length(settings[[i]]$priors) != ddalpha$numPatterns 
          || min(settings[[i]]$priors) <= 0 
          || max(settings[[i]]$priors) <= 0){
        warning("In treatment number ", i, ": Argument \"priors\" not specified correctly. Defaults in the form of class portions are applied")
        tmp.priors <- NULL
      }else{
        tmp.priors <- settings[[i]]$priors/sum(settings[[i]]$priors)
      }
      treatment.settings <- structure(
        list(name = tmp.name, 
             method = "RandProp", 
             priors = NULL), 
        .Names = c("name", "method", "priors"))
      supported <- TRUE
    }
    if (toupper(settings[[i]]$method) == "IGNORE"){
      tmp.name <- settings[[i]]$name
      treatment.settings <- structure(
        list(name = tmp.name, 
             method = "Ignore"), 
        .Names = c("name", "method"))
      supported <- TRUE
    }
    if (supported){
      treatments.settings[[counter]] <- treatment.settings
      counter <- counter + 1
    }
  }

  return (treatments.settings)
}

.lda_learn <- function(ddalpha, settings){
  settings$lda <- MASS::lda(formula=as.formula("CLASS ~ ."), data=ddalpha$raw, priors=settings$priors)
  settings$priors <- settings$lda$prior
  return (settings)
}

.lda_classify <- function(objects, ddalpha, settings){
  if (!is.data.frame(objects)) objects = as.data.frame(objects)
  names(objects) <- names(ddalpha$raw)[1:ncol(objects)]
  return (as.list(predict(settings$lda, objects)$class))
}

.qda_learn <- function(ddalpha, settings){
  settings$qda <- MASS::qda(formula=as.formula("CLASS ~ ."), data=ddalpha$raw, priors=settings$priors)
  settings$priors <- settings$qda$prior
  return (settings)
}

.qda_classify <- function(objects, ddalpha, settings){
  if (!is.data.frame(objects)) objects = as.data.frame(objects)
  names(objects) <- names(ddalpha$raw)[1:ncol(objects)]
  return (as.list(predict(settings$qda, objects)$class))
}

.knnAff_learn <- function(ddalpha, settings){
  counter <- 1
  # Determining multi-class behaviour
  if (settings$knnAff.methodAggregation == "majority"){
    for (i in 1:(ddalpha$numPatterns - 1)){
      for (j in (i + 1):ddalpha$numPatterns){
        # Creating a classifier
        classifier.index          <- counter
        classifier.index1         <- i
        classifier.index2         <- j
        classifier.points         <- as.double(t(rbind(ddalpha$patterns[[i]]$points, ddalpha$patterns[[j]]$points)))
        classifier.cardinalities  <- as.integer(c(ddalpha$patterns[[i]]$cardinality, ddalpha$patterns[[j]]$cardinality))
        if (settings$knnAff.k < 1 || settings$knnAff.k > (ddalpha$patterns[[i]]$cardinality + ddalpha$patterns[[j]]$cardinality - 1))
        {
          if (settings$knnAff.range < 2 || settings$knnAff.range > (ddalpha$patterns[[i]]$cardinality + ddalpha$patterns[[j]]$cardinality - 1)){
            maxk <- 10*( (ddalpha$numPoints)^(1/ddalpha$dimension) ) + 1
          }else{
            maxk <- settings$knnAff.range
          }
          maxk <- min(maxk, ddalpha$patterns[[i]]$cardinality + ddalpha$patterns[[j]]$cardinality - 1)
          maxk <- max(maxk, 2)
          classifier.range <- maxk
          classifier.k <- as.integer(.C("KnnAffInvLearnJK", 
                             classifier.points, 
                             as.integer(ddalpha$dimension), 
                             classifier.cardinalities, 
                             as.integer(maxk), 
                             k=integer(1))$k)
        }else{
          classifier.range <- settings$knnAff.range
          classifier.k <- as.integer(settings$knnAff.k)
        }
        # Adding the classifier to the list of classifiers
        settings$knnAff.classifiers[[counter]] <- 
          list(index = classifier.index, 
               index1 = classifier.index1, 
               index2 = classifier.index2, 
               points = classifier.points, 
               cardinalities = classifier.cardinalities, 
               k = classifier.k, 
               range = classifier.range)
        counter <- counter + 1
      }
    }
  }
  if (settings$knnAff.methodAggregation == "sequent"){
    for (i in 1:ddalpha$numPatterns){
      anotherClass <- NULL
      for (j in 1:ddalpha$numPatterns){
        if (j != i){
          anotherClass <- rbind(anotherClass, ddalpha$patterns[[j]]$points)
        }
      }
      classifier.index          <- counter
      classifier.index1         <- i
      classifier.index2         <- -1
      classifier.points         <- as.double(t(rbind(ddalpha$patterns[[i]]$points, anotherClass)))
      classifier.cardinalities  <- as.integer(c(ddalpha$patterns[[i]]$cardinality, nrow(anotherClass)))
      if (settings$knnAff.k < 1 || settings$knnAff.k > ddalpha$numPoints)
      {
        if (settings$knnAff.range < 2 || settings$knnAff.range > (ddalpha$numPoints - 1)){
          maxk <- 10*( (ddalpha$numPoints)^(1/ddalpha$dimension) ) + 1
        }else{
          maxk <- settings$knnAff.range
        }
        maxk <- min(maxk, ddalpha$numPoints - 1)
        maxk <- max(maxk, 2)
        classifier.range <- maxk
        classifier.k <- as.integer(.C("KnnAffInvLearnJK", 
                           classifier.points, 
                           as.integer(ddalpha$dimension), 
                           classifier.cardinalities, 
                           as.integer(maxk), 
                           k=integer(1))$k)
      }else{
        classifier.range <- settings$knnAff.range
        classifier.k <- as.integer(settings$knnAff.k)
      }
      # Adding the classifier to the list of classifiers
      settings$knnAff.classifiers[[counter]] <- 
        list(index = classifier.index, 
             index1 = classifier.index1, 
             index2 = classifier.index2, 
             points = classifier.points, 
             cardinalities = classifier.cardinalities, 
             k = classifier.k, 
             range = classifier.range)
      counter <- counter + 1
    }
  }
  return (settings)
}

.knnAff_classify <- function(objects, ddalpha, settings){
  # Correct input data
  if (!is.matrix(objects)){
    objects <- matrix(objects, nrow=1)
  }
  # Initialization of the vote array
  votes <- matrix(rep(0, nrow(objects)*ddalpha$numPatterns), nrow=nrow(objects), ncol=ddalpha$numPatterns)
  for (i in 1:length(settings$knnAff.classifiers)){
    res <- .C("KnnAffInvClassify", 
              as.double(t(objects)), 
              as.integer(nrow(objects)), 
              settings$knnAff.classifiers[[i]]$points, 
              as.integer(ddalpha$dimension), 
              settings$knnAff.classifiers[[i]]$cardinalities, 
              settings$knnAff.classifiers[[i]]$k, 
              output=integer(nrow(objects)))$output
    for (j in 1:nrow(objects)){
      if (res[j] == 0){
        votes[j,settings$knnAff.classifiers[[i]]$index1] <- votes[j,settings$knnAff.classifiers[[i]]$index1] + 1
      }else{
        votes[j,settings$knnAff.classifiers[[i]]$index2] <- votes[j,settings$knnAff.classifiers[[i]]$index2] + 1
      }
    }
  }
  # Collect results
  results <- list()
  for (i in 1:nrow(objects)){
    results[[i]] <- ddalpha$patterns[[which.max(votes[i,])]]$name
  }
  return (results)
}

.knn_learn <- function(ddalpha, settings){
  settings$knn.train <- ddalpha$raw[,1:ddalpha$dimension]
  settings$knn.cl <- ddalpha$raw[,ddalpha$dimension + 1]
  if (settings$knn.k < 1 || settings$knn.k > ddalpha$numPoints){
    if (settings$knn.range < 1 || settings$knn.range > ddalpha$numPoints - 1){
      settings$knn.range <- 10*( (ddalpha$numPoints)^(1/ddalpha$dimension) ) + 1
      settings$knn.range <- min(settings$knn.range, ddalpha$numPoints - 1)
      settings$knn.range <- max(settings$knn.range, 2)
    }
    cv.err <- c()
    ks <- 1:settings$knn.range
    for (i in ks){
      newpre <- as.vector(class::knn.cv(settings$knn.train, settings$knn.cl, k = i))
      cv.err <- c(cv.err, sum(settings$knn.cl != newpre))
    }
    settings$knn.k <- ks[which.min(cv.err)] 
  }
  return (settings)
}

.knn_classify <- function(objects, ddalpha, settings){
  return (class::knn(settings$knn.train, objects, settings$knn.cl, settings$knn.k))
}

.mah_learn <- function(ddalpha, settings){
  settings$mah.classes <- list()
  if (is.null(settings$priors)){
    settings$priors <- c()
    for (i in 1:ddalpha$numPatterns){
      settings$priors[i] <- ddalpha$patterns[[i]]$cardinality/ddalpha$numPoints
    }
  }
  for (i in 1:ddalpha$numPatterns){
    class.prior <- NULL
    class.mean <- NULL
    class.cov <- NULL
    class.sigma <- NULL
    class.prior <- settings$priors[i]
    if (settings$mah.estimate == "moment"){
      class.mean <- colMeans(ddalpha$patterns[[i]]$points)
      class.cov <- cov(ddalpha$patterns[[i]]$points)
      class.sigma <- solve(class.cov)
    }
    if (settings$mah.estimate == "MCD"){
      estimate <- robustbase::covMcd(ddalpha$patterns[[i]]$points, alpha=settings$mcd.alpha)
      class.mean <- estimate$center
      class.cov <- estimate$cov
      class.sigma <- solve(class.cov)
    }
    settings$mah.classes[[i]] <- structure(
      list(index = i, 
           prior = class.prior, 
           mean = class.mean, 
           cov = class.cov, 
           sigma = class.sigma), 
      .Names = c("index", "prior", "mean", "cov", "sigma"))
  }
  return (settings)
}

.mah_classify <- function(objects, ddalpha, settings){
  # Correct input data
  if (!is.matrix(objects)){
    objects <- matrix(objects, nrow=1)
  }
  # Initialization of the vote array
  votes <- matrix(rep(0, nrow(objects)*ddalpha$numPatterns), nrow=nrow(objects), ncol=ddalpha$numPatterns)
  for (i in 1:nrow(objects)){
    for (j in 1:length(settings$mah.classes)){
      votes[i,j] <- settings$mah.classes[[j]]$prior*.Mahalanobis_depth(objects[i,], settings$mah.classes[[j]]$mean, settings$mah.classes[[j]]$sigma)
    }
  }
  # Collect results
  results <- list()
  for (i in 1:nrow(objects)){
    results[[i]] <- ddalpha$patterns[[which.max(votes[i,])]]$name
  }
  return (results)
}

.ignore_classify <- function(nobjects){
  return (as.list(rep("Ignored", nobjects)))
}

.right_classify <- function(nobjects){
  return (as.list(rep("Outsider", nobjects)))
}

.randequal_classify <- function(nobjects, ddalpha){
  results <- list()
  for (i in 1:nobjects){
    results[[i]] <- ddalpha$patterns[[sample(1:ddalpha$numPatterns,1)]]$name
  }
  return (results)
}

.randprop_classify <- function(nobjects, ddalpha, settings){
  priors <- settings$priors
  if (is.null(priors)){
    priors <- c()
    for (i in 1:ddalpha$numPatterns){
      priors[i] <- ddalpha$patterns[[i]]$cardinality/ddalpha$numPoints
    }
  }
  results <- list()
  for (i in 1:nobjects){
    results[[i]] <- ddalpha$patterns[[sample(1:ddalpha$numPatterns,1,prob = priors)]]$name
  }
  return (results)
}

# Function is taken from the R-documentation, "Examples" to the function "is.integer"
.is.wholenumber <- function(x, tol = .Machine$double.eps^0.5)  abs(x - round(x)) < tol

summary.ddalpha <- function(object, printseparators = T, ...){
  x = object
  if(!is.null(x$call)){
      cat("Call: ", format(x$call), "\nResulting table: ", paste(x$colnames, collapse = ",   "),"\n")
  }
  
  cat("ddalpha:\n")
  cat("\t num.points = ", x$numPoints, 
      ", dimension = ", x$dimension, 
      ", num.patterns = ", x$numPatterns, "\n", sep="")
  cat("\t depth = \"", x$methodDepth, "\"\n", sep="")
  cat("\t separator = \"", x$methodSeparator, "\"\n", sep="")
  cat("\t aggregation.method = \"", x$methodAggregation, "\"\n", sep="")
  if (is.numeric(x$numChunks)) cat("\t num.chunks =", x$numChunks, "\n")
  if (is.numeric(x$numDirections)) cat("\t num.directions =", x$numDirections, "\n")
  cat("\t use.convex =", x$useConvex, "\n")
  if (is.numeric(x$maxDegree)) cat("\t max.degree =", x$maxDegree, "\n")
  cat("patterns:\n")
  for (i in 1:length(x$patterns)){
    cat("\t ");print(x$patterns[[i]])
  }
  cat("num.classifiers = ", x$numClassifiers, "\n")
  if(x$numClassifiers == 1 || printseparators)
    for (i in 1:x$numClassifiers){
      print(x$classifiers[[i]], prefix = "\t ")
    }
  cat("outsider.methods:\n")
  if (is.null(x$methodsOutsider)){
    cat ("\t Absent\n", sep="")
  }else{
    for (i in 1:length(x$methodsOutsider)){
      cat ("\t \"",x$methodsOutsider[[i]]$name, "\":\n", sep="")
      cat("\t\t method = \"", x$methodsOutsider[[i]]$method, "\"\n", sep="")
      if (   x$methodsOutsider[[i]]$method == "LDA" 
          || x$methodsOutsider[[i]]$method == "RandProp" 
          || x$methodsOutsider[[i]]$method == "depth.Mahalanobis"){
        cat("\t\t priors =", x$methodsOutsider[[i]]$priors, "\n")
      }
      if (x$methodsOutsider[[i]]$method == "KNNAff"){
        cat("\t\t aggregation.method = \"", 
            x$methodsOutsider[[i]]$knnAff.methodAggregation, "\"\n", sep="")
        for (j in 1:length(x$methodsOutsider[[i]]$knnAff.classifiers)){
          cat("\t\t k.range = ", format(paste("1:", 
                                              x$methodsOutsider[[i]]$knnAff.classifiers[[j]]$range, sep=""), 
                                        justify="right", width=10), sep="")
        }
        cat("\n")
        for (j in 1:length(x$methodsOutsider[[i]]$knnAff.classifiers)){
          cat("\t\t k       = ", format( 
            x$methodsOutsider[[i]]$knnAff.classifiers[[j]]$k, 
            justify="right", width=10), sep="")
        }
        cat("\n")
      }
      if (x$methodsOutsider[[i]]$method == "KNN"){
        cat("\t\t k.range = 1:", x$methodsOutsider[[i]]$knn.range, "\n", sep="")
        cat("\t\t k =", x$methodsOutsider[[i]]$knn.k, "\n")
      }
      if (x$methodsOutsider[[i]]$method == "depth.Mahalanobis"){
        cat("\t\t estimate = \"", x$methodsOutsider[[i]]$mah.estimate, "\"\n", sep="")
        if (x$methodsOutsider[[i]]$mah.estimate == "MCD"){
          cat("\t\t mcd.alpha = ", x$methodsOutsider[[i]]$mcd.alpha, "\n")
        }
      }
    }
  }
  invisible(x)
}

print.ddalpha <- function(x, printseparators = F, ...){
  if(!is.null(x$call)){
    cat("Call: ", format(x$call), "\nResulting table: ", paste(x$colnames, collapse = ",   "),"\n")
  }

  cat("depth = \"", x$methodDepth, "\"\n", sep="")
  cat("separator = \"", x$methodSeparator, "\"\n", sep="")
  cat("aggregation.method = \"", x$methodAggregation, "\"\n", sep="")
  cat("patterns:\n")
  for (i in 1:length(x$patterns)){
    cat("\t ");print(x$patterns[[i]])
  }
  cat("num.classifiers = ", x$numClassifiers, "\n")
  if(x$numClassifiers == 1 || printseparators)
    for (i in 1:x$numClassifiers){
      print(x$classifiers[[i]], prefix = "\t ", full = F)
    }
  cat("outsider.methods:\n")
  if (is.null(x$methodsOutsider)){
    cat ("\t Absent\n", sep="")
  }else{
    for (i in 1:length(x$methodsOutsider)){
      cat("\t method = \"", x$methodsOutsider[[i]]$method, "\"\n", sep="")
      if (   x$methodsOutsider[[i]]$method == "LDA" 
             || x$methodsOutsider[[i]]$method == "RandProp" 
             || x$methodsOutsider[[i]]$method == "depth.Mahalanobis"){
        cat("\t priors =", x$methodsOutsider[[i]]$priors, "\n")
      }
      if (x$methodsOutsider[[i]]$method == "KNNAff"){
        cat("\t aggregation.method = \"", 
            x$methodsOutsider[[i]]$knnAff.methodAggregation, "\"\n", sep="")
        for (j in 1:length(x$methodsOutsider[[i]]$knnAff.classifiers)){
          cat("\t k.range = ", format(paste("1:", 
                                              x$methodsOutsider[[i]]$knnAff.classifiers[[j]]$range, sep=""), 
                                        justify="right", width=10), sep="")
        }
        cat("\n")
        for (j in 1:length(x$methodsOutsider[[i]]$knnAff.classifiers)){
          cat("\t k       = ", format( 
            x$methodsOutsider[[i]]$knnAff.classifiers[[j]]$k, 
            justify="right", width=10), sep="")
        }
        cat("\n")
      }
      if (x$methodsOutsider[[i]]$method == "KNN"){
        cat("\t k.range = 1:", x$methodsOutsider[[i]]$knn.range, "\n", sep="")
        cat("\t k =", x$methodsOutsider[[i]]$knn.k, "\n")
      }
      if (x$methodsOutsider[[i]]$method == "depth.Mahalanobis"){
        cat("\t estimate = \"", x$methodsOutsider[[i]]$mah.estimate, "\"\n", sep="")
        if (x$methodsOutsider[[i]]$mah.estimate == "MCD"){
          cat("\t mcd.alpha = ", x$methodsOutsider[[i]]$mcd.alpha, "\n")
        }
      }
    }
  }
  invisible(x)
}

print.ddalpha.pattern <- function(x, ...){
  cat("pattern[", x$index, "]:", sep="")
  cat("\t ", x$cardinality, " points, label = \"", x$name, "\"\n", sep="")
  invisible(x)
}

print.ddalpha.alpha <- function(x, prefix = "", full = T, ...){
  if(full){
    cat(prefix,x$index, ". alpha:\n", sep="")
    cat(prefix,"   degree: ", x$degree,"; axes: ", x$index1, ", ", x$index2, "\n", sep="")
    cat(prefix,"   hyperplane: ", paste(x$hyperplane, collapse = ", "), "\n", sep="")
  } else{
    cat(prefix, x$index, ". degree: ", x$degree,"; axes: ", x$index1, ", ", x$index2, "\n", sep="")
    cat(prefix,"   hyperplane: ", paste(x$hyperplane, collapse = ", "), "\n", sep="")
  }
  invisible(x)
}
print.ddalpha.polynomial <- function(x, prefix = "", full = T,...){
  if(full){
    cat(prefix,x$index, ". polynomial:\n", sep="")
    cat(prefix,"   degree: ", x$degree,"; axes: ", x$index1, ", ", x$index2, ifelse(x$axis!=0, ", invert", ""), "\n", sep="")
    cat(prefix,"   polynomial: ", paste(x$polynomial, collapse = ", "), "\n", sep="")
  } else {
    cat(prefix,x$index, ". degree: ", x$degree,"; axes: ", x$index1, ", ", x$index2, ifelse(x$axis!=0, ", invert", ""), "\n", sep="")
    cat(prefix,"   polynomial: ", paste(x$polynomial, collapse = ", "), "\n", sep="")
  }
  invisible(x)
}
print.ddalpha.knnlm <- function(x, prefix = "", full = T,...){
  if(full)
    cat(prefix,"knnlm:  k = ", x$knnK, "\n", sep="")
  else
    cat(prefix,"k = ", x$knnK, "\n", sep="")
  invisible(x)
}
print.ddalpha.maxD <- function(x, prefix = "", full = T,...){
  if(full)
    cat(prefix,"maximum depth separator\n")
  invisible(x)
}

# .ddalpha.learn.knnlm <- function(ddalpha){
#   
#   # Prepare outputs and distance matrix
#   y <- NULL
#   allPoints <- NULL
#   for (i in 1:ddalpha$numPatterns){
#     allPoints<- rbind(allPoints, ddalpha$patterns[[i]]$depths)
#     y <- c(y, rep(ddalpha$patterns[[i]]$name, ddalpha$patterns[[i]]$cardinality))
#   }
#   #  print(y)
#   dists <- knn.dist(allPoints, dist.meth="maximum")
#   #  plot(ddalpha$patterns[[1]]$depths, col = "red", xlim=c(0,1), ylim=c(0,1))
#   #  points(ddalpha$patterns[[2]]$depths, col = "blue", xlim=c(0,1), ylim=c(0,1))
#   # Cross-validate knn
#   cvErr  <- c()
#   allIndices <- 1:ddalpha$numPoints
#   krange <- 10*( (ddalpha$numPoints)^(1/ddalpha$numPatterns) ) + 1
#   krange <- min(krange, ceiling(ddalpha$numPoints/2))
#   if (ddalpha$numPatterns == 2){krange <- min(krange, 50)}
#   krange <- max(krange, 2)
#   #  cat("Range: 1:", krange, ".\n", sep="")
#   for (i in 1:krange){
#     curPreErr <- 0
#     for (j in 0:(ddalpha$numChunks - 1)){
#       testSel <- allIndices[allIndices%%ddalpha$numChunks == j]
#       test <- allIndices[testSel]
#       train <- allIndices[-test]
#       #      cat("1. Train: ", train, ", test: ", test, ".\n")
#       curPreErr <- curPreErr + sum(knn.predict(train, test, y, dists, k=i, agg.meth="majority", ties.meth="first") != y[test])
#     }
#     cvErr <- c(cvErr, curPreErr)
#     #    cat("i = ", i, "done.\n")
#     #    print(cvErr)
#   }
#   # Collect results
#   ddalpha$knnK <- (1:krange)[which.min(cvErr)]
#   ddalpha$knnX <- allPoints
#   ddalpha$knnY <- y
#   ddalpha$knnD <- dists
#   
#   #  print(ddalpha$knnK)
#   
#   return (ddalpha)
# }

.potentialKernelTypes <- c("EDKernel", "GKernel", "EKernel", "TriangleKernel", "VarGKernel")

.potential_depths <- function(ddalpha, objects, class = 0){
  
  if (ddalpha$ignoreself) stop("ignoreself not supported")  # todo ignore only if objects were NULL, @ what class do the points belong to? when seperately scaled
  
  # count all potential on the untransformed data
  if (class == 0) { 
    a = ddalpha$kernel.bandwidth
    data <- NULL
    cardinalities <- c()
    for (i in 1:ddalpha$numPatterns){
      data <- rbind(data, ddalpha$patterns[[i]]$points)
      cardinalities <- c(cardinalities, ddalpha$patterns[[i]]$cardinality)
    }
    if (is.null(objects))
      objects = data
  }
  # count potential w.r.t the given class
  else {
    a = ddalpha$kernel.bandwidth[class]
    data <- ddalpha$patterns[[class]]$transformer( ddalpha$patterns[[class]]$points )
    cardinalities <- c(ddalpha$patterns[[class]]$cardinality)      
    objects <- ddalpha$patterns[[class]]$transformer(objects)
    
    #   if (is.null(objects)){
    #     for (i in (1:ddalpha$numPatterns)){
    #        objects <- rbind(objects, ddalpha$patterns[[i]]$points)
    #     }
    #      objects <- ddalpha$patterns[[class]]$transformer(objects)
    #   } else {
    #     ignoreself = F
    #    }
  }
  
  kernelType = ddalpha$kernel
  if (is.character(kernelType))
    kernelType =  switch (kernelType, EDKernel = 1, GKernel = 2, EKernel = 3, TriangleKernel = 4, 1)
  
  points <- as.vector(t(data))
  numPoints <- sum(cardinalities)
  dimension <- ncol(data)
  points2 <- as.vector(t(objects))
  numPoints2 <- nrow(objects)
  classes <- length(cardinalities)
  
  depth <- .C("PotentialDepthsCount", 
              as.double(points), 
              as.integer(numPoints), 
              as.integer(dimension), 
              as.integer(classes),
              as.integer(cardinalities), 
              as.double(points2), 
              as.integer(numPoints2), 
              as.integer(kernelType), 
              as.double(a), 
              as.integer(ddalpha$ignoreself),
              depth=double(numPoints2*classes))$depth
  
  d = as.matrix(depth)
  dim(d)<-c(numPoints2,classes)
  
  return(d)
}

.potential_depths_wrt <- function(ddalpha, objects_){
  
  d <- NULL
  
  # w.r.t. each class
  for (cls in 1:ddalpha$numPatterns){
    
    data <- ddalpha$patterns[[cls]]$transformer( ddalpha$patterns[[cls]]$points )
    cardinalities <- c(ddalpha$patterns[[cls]]$cardinality)
    
    objects = objects_
    if (is.null(objects_))  # count for all data
      for (i in 1:ddalpha$numPatterns){
        objects <- rbind(objects, ddalpha$patterns[[i]]$points)
      }
    
    objects <- ddalpha$patterns[[cls]]$transformer(objects)
    
    kernelType = ddalpha$kernel
    if (is.character(kernelType))
      kernelType =  switch (kernelType, EDKernel = 1, GKernel = 2, EKernel = 3, TriangleKernel = 4, 1)
    
    points <- as.vector(t(data))
    numPoints <- sum(cardinalities)
    dimension <- ncol(data)
    points2 <- as.vector(t(objects))
    numPoints2 <- nrow(objects)
    classes <- length(cardinalities)
    
    depth <- .C("PotentialDepthsCount", 
                as.double(points), 
                as.integer(numPoints), 
                as.integer(dimension), 
                as.integer(classes),
                as.integer(cardinalities), 
                as.double(points2), 
                as.integer(numPoints2), 
                as.integer(kernelType), 
                as.double(ddalpha$kernel.bandwidth[cls]),
                as.integer(ddalpha$ignoreself),
                depth=double(numPoints2*classes))$depth
    
    d = cbind(d, depth)
  }
  
  return(d)
}

depth.potential <- function(x, data, pretransform = "1Mom", kernel = "GKernel", kernel.bandwidth = NULL, mah.parMcd = 0.75){
  if (!is.numeric(x)){
    stop("Argument \"x\" should be numeric")
  }
  if (!is.matrix(x)) {
    if(is.vector(x))
      x <- matrix(x, nrow=1)
    if(is.data.frame(x))
      x <- data.matrix(x)
  }
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if(is.data.frame(data))
    data <- data.matrix(data)
  if (ncol(x) != ncol(data)){
    stop("Dimensions of the arguments \"x\" and \"data\" should coincide")
  }
  if (ncol(data) + 1 > nrow(data)){ #?
    stop("To few data points")
  }
  
  if(!is.null(pretransform)){
    if (pretransform == "1Mom" || pretransform == "NMom")
      mm <- mah.moment(data)
    else if (pretransform == "1MCD" || pretransform == "NMCD")
      mm <- mah.mcd(data, mah.parMcd)              
    
    transformer <- MahMomentTransformer(mm$mu, mm$b)
    data <- transformer(data)
    x <- transformer(x)
  }
  
  kernelType = kernel
  if (is.character(kernelType))
    kernelType =  switch (kernelType, EDKernel = 1, GKernel = 2, EKernel = 3, TriangleKernel = 4, 1)
  
  if (is.null(kernel.bandwidth)) {    # use the rule of thumb
    kernel.bandwidth = nrow(data) ^ (-2/(ncol(data)+4))
  }
  else{
    if (length(kernel.bandwidth) != 1 || is.na(kernel.bandwidth) || kernel.bandwidth == 0)
      stop("Argument \"kernel.bandwidth\" has invaid format.")
  }
  
  points <- as.vector(t(data))
  numPoints <- nrow(data)
  dimension <- ncol(data)
  points2 <- as.vector(t(x))
  numPoints2 <- nrow(x)
  cardinalities = numPoints
  classes <- 1
  
  ignoreself = F
  
  depth <- .C("PotentialDepthsCount", 
              as.double(points), 
              as.integer(numPoints), 
              as.integer(dimension), 
              as.integer(classes),
              as.integer(cardinalities), 
              as.double(points2), 
              as.integer(numPoints2), 
              as.integer(kernelType), 
              as.double(kernel.bandwidth),
              as.integer(ignoreself),
              depth=double(numPoints2*classes))$depth
  
  return(depth)
}




depth.space.potential <- function(data, cardinalities, pretransform = "NMom", kernel = "GKernel", kernel.bandwidth = NULL, mah.parMcd = 0.75){
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if(is.data.frame(data))
    data <- data.matrix(data)
  if (!is.vector(cardinalities, mode = "numeric") 
      || is.na(min(cardinalities)) 
      || sum(.is.wholenumber(cardinalities)) != length(cardinalities) 
      || min(cardinalities) <= 0 
      || sum(cardinalities) != nrow(data)){
    stop("Argument \"cardinalities\" should be a vector of cardinalities of the classes in \"data\" ")
  }
  if (sum(cardinalities < ncol(data) + 1) != 0){
    stop("Not in all classes sufficiently enough objetcs")
  }
  
  d <- NULL
  
  needtransform = F
  
  if (pretransform == "1Mom" || pretransform == "1MCD"){
    if (pretransform == "1Mom")
      mm <- mah.moment(data)
    else   # "1MCD"
      mm <- mah.mcd(data, mah.parMcd)              
    
    transformer <- MahMomentTransformer(mm$mu, mm$b)
    data <- transformer(data)
    
    if (is.null(kernel.bandwidth)) {    # use the rule of thumb
      kernel.bandwidth = nrow(data) ^ (-2/(ncol(data)+4))
    }
    else{
      if (length(kernel.bandwidth) != 1 || is.na(kernel.bandwidth) || kernel.bandwidth == 0)
        stop("Argument \"kernel.bandwidth\" has invaid length, Zero or NA elements.")
    }
    
  } else if (pretransform == "NMom" || pretransform == "NMCD"){
    needtransform = T
    
    if (is.null(kernel.bandwidth)) {    # use the rule of thumb
      #separately calculated later
    }
    else{
      if (!is.numeric(kernel.bandwidth)
          ||!(is.vector(kernel.bandwidth) || is.list(kernel.bandwidth))){
        stop("Argument \"kernel.bandwidth\" has invaid format.")
      }
      if (length(kernel.bandwidth) == 1)
        kernel.bandwidth = rep(kernel.bandwidth, length(cardinalities))
      if (sum(!is.na(kernel.bandwidth)) != length(cardinalities) || sum(kernel.bandwidth != 0) != length(cardinalities)){
        stop("Argument \"kernel.bandwidth\" has invaid length, Zero or NA elements.")
      } 
    }
  }
  
  if(needtransform)
    # w.r.t. each class
    for (cls in 1:length(cardinalities)){
      pattern <- data[(1 + sum(cardinalities[0:(cls - 1)])):sum(cardinalities[1:cls]),]
      if(is.null(kernel.bandwidth)) band <- NULL
      else band <- kernel.bandwidth[cls]
      
      depth <- depth.potential(data, pattern, pretransform, kernel, band, mah.parMcd)
      d <- cbind(d, depth)
    }
  else # not w.r.t.
  {
    points <- as.vector(t(data))
    numPoints <- sum(cardinalities)
    dimension <- ncol(data)
    points2 <- as.vector(t(data))
    numPoints2 <- nrow(data)
    classes <- length(cardinalities)
    
    kernelType = kernel
    if (is.character(kernelType))
      kernelType =  switch (kernelType, EDKernel = 1, GKernel = 2, EKernel = 3, TriangleKernel = 4, 1)
    
    ignoreself = F
    
    depth <- .C("PotentialDepthsCount", 
                as.double(points), 
                as.integer(numPoints), 
                as.integer(dimension), 
                as.integer(classes),
                as.integer(cardinalities), 
                as.double(points2), 
                as.integer(numPoints2), 
                as.integer(kernelType), 
                as.double(kernel.bandwidth), 
                as.integer(ignoreself),
                depth=double(numPoints2*classes))$depth
    
    d = as.matrix(depth)
    dim(d)<-c(numPoints2,classes)
  }
  
  return(d)
}

c----------------------------------------------------------------------
c     Stanislav Nagy
c     nagy@karlin.mff.cuni.cz
c     27/06/2016
c
c     Fortran implementation of functional depths and kernel smoothing
c     - Nagy, Ferraty: Depth for Noisy Random Functions
c     - Nagy, Gijbels, Hlubinka: Depth-Based Recognition of Shape 
c           Outlying Functions
c
c----------------------------------------------------------------------

c----------------------------------------------------------------------
c     kernel smoothing of a set of functions
c----------------------------------------------------------------------

      SUBROUTINE KERNSM(T,X,G,M,N,H,K,R)
c     kernel smoothing of a single function
c     T vector of observation points (M)
c     X vector of observed values (M)
c     G grid to evaluate (N)
c     H bandwidth
c     K kernel code to use, see below
c     R resulting estimate (N)

      integer M,N,K
      integer I,J
      double precision T(M),X(M),G(N),R(N),H
      double precision DEN,P
      
      DO 10 I=1,N
            R(I)=0.0
            DEN = 0.0
            DO 5 J=1,M
                  CALL KERN((G(I)-T(J))/H,P,K)
                  R(I) = R(I) + X(J)*P
                  DEN = DEN + P
5           CONTINUE
            IF (DEN.GT.0) THEN
                  R(I) = R(I)/DEN
               ELSE
                  R(I) = 10**6
            ENDIF      
10    CONTINUE
      RETURN
      END
      
c----------------------------------------------------------------------

      SUBROUTINE KERN(T,R,K)
c     kernel functions
      
      double precision T,R
      INTEGER K
      
      IF(K.EQ.1) THEN
c     uniform kernel
            IF (ABS(T).LE.1.0) THEN
                  R = .5
                  ELSE
                  R = 0.0
            ENDIF
      ENDIF
      IF(K.EQ.2) THEN
c     triangular kernel
            IF (ABS(T).LE.1.0) THEN
                  R = 1-ABS(T)
                  ELSE
                  R = 0.0
            ENDIF
      ENDIF
      IF(K.EQ.3) THEN
c     Epanechnikov kernel
            IF (ABS(T).LE.1.0) THEN
                  R = 3.0/4.0*(1.0-T**2.0)
                  ELSE
                  R = 0.0
            ENDIF
      ENDIF
      IF(K.EQ.4) THEN
c     biweight kernel
            IF (ABS(T).LE.1.0) THEN
                  R = 15.0/16.0*(1.0-T**2.0)**2
                  ELSE
                  R = 0.0
            ENDIF
      ENDIF
      IF(K.EQ.5) THEN
c     triweight kernel
            IF (ABS(T).LE.1.0) THEN
                  R = 35.0/32.0*(1.0-T**2.0)**3
                  ELSE
                  R = 0.0
            ENDIF
      ENDIF
      IF(K.EQ.6) THEN
c     Gaussian kernel
            R = (2*4.D0*DATAN(1.D0))**(-0.5)*EXP(-0.5*T**2)
      ENDIF
      RETURN
      END
  
c----------------------------------------------------------------------
      
      SUBROUTINE CVKERNSM(T,X,G,M,N,H,NH,KER,TRE,XRE,TNRE,XNRE,NR,NT,R)
c     as in KERNSM, but with automated BW seletction
c     from the vector H(NH)
c     KER kernel code
c     TRE,XRE vectors from T,X for omission for the CV (NR*NT)
c     TNRE,XNRE supplementary vectors to TRE,NRE ((M-NR)*NT)
c     NR no of random elements for each H choice
c     NT no of random trials for each H choice
c     for each H we choose NR random points from X
c     compute the kernel smoother without these points
c     and evaluate the error. repeat NT times, then average
c     the errors to get the CV estimate

      integer M,N,NH,NR,NT,KER
      integer I,J,K,OM(NR),L
      double precision T(M),X(M),G(N),R(N),H(NH),TRE(NR*NT),XRE(NR*NT)
      double precision TNRE((M-NR)*NT),XNRE((M-NR)*NT)
      double precision TOM(NR),XOM(NR),TNOM(M-NR),XNOM(M-NR)
      double precision CV(NH),SR(NR),MCV
      
      DO 30 I=1,NH
            CV(I) = 0.0
            DO 20 J=1,NT
                  DO 5 K=1,NR
                        TOM(K) = TRE((J-1)*NR+K)
                        XOM(K) = XRE((J-1)*NR+K)
5                 CONTINUE
                  DO 6 K=1,(M-NR)
                        TNOM(K) = TNRE((J-1)*(M-NR)+K)
                        XNOM(K) = XNRE((J-1)*(M-NR)+K)
6                 CONTINUE
            CALL KERNSM(TNOM,XNOM,TOM,(M-NR),NR,H(I),KER,SR)
                  DO 10 K=1,NR
                  CV(I) = CV(I) + (XOM(K)-SR(K))**2
10                CONTINUE
20          CONTINUE            
30    CONTINUE
      K = 0
      MCV = CV(1)+1
      DO 40 I=1,NH
            IF (CV(I).LT.MCV) THEN
                  K = I
                  MCV = CV(I)
            ENDIF
40    CONTINUE
      CALL KERNSM(T,X,G,M,N,H(K),KER,R)
      RETURN
      END
      
c----------------------------------------------------------------------
c     fucntional data depth computation      
c----------------------------------------------------------------------

      SUBROUTINE funD1(A,B,M,N,D,funSDEP,funHDEP,fIsdep,fIhdep,IAsdep,
     +IAhdep)
c     univariate integrated depth
       
      INTEGER N,M,D
      INTEGER IAsdep(M),IAhdep(M)
      double precision A(M*D),B(N*D),funSDEP(M),funHDEP(M),
     +fIsdep(M),fIhdep(M) 
      double precision BH(N)
      double precision hSDEP,hHDEP
      INTEGER I,J,K

c     initializing
      DO 5 I=1,M
            funSDEP(I) = 0.0
            funHDEP(I) = 0.0
            fISDEP(I) = 2.0
            fIHDEP(I) = 2.0 
            IAsdep(I) = 0
            IAhdep(I) = 0
5     CONTINUE
c     essential loop, computing 1D depths at each point
      DO 30 I=1,D
            DO 10 K=1,N
                  BH(K) = B((I-1)*N+K)
10          CONTINUE
            DO 20 J=1,M
                  hSDEP = 0.0
                  hHDEP = 0.0
                  CALL fD1(A((I-1)*M+J),N,BH,hSDEP,hHDEP)
c     integrated depth evaluation
                  funSDEP(J) = funSDEP(J) + hSDEP
                  funHDEP(J) = funHDEP(J) + hHDEP
c     counting the area of the smallest depth for each function
                  IF (hSDEP.EQ.fISDEP(J)) THEN
                        IAsdep(J) = IAsdep(J)+1
                  ELSEIF (hSDEP.LT.fISDEP(J)) THEN 
                        IAsdep(J) = 1
                  ENDIF
                  IF (hHDEP.EQ.fIHDEP(J)) THEN
                        IAhdep(J) = IAhdep(J)+1
                  ELSEIF (hHDEP.LT.fIHDEP(J)) THEN 
                        IAhdep(J) = 1
                  ENDIF 
c     infimal depth evaluation                 
                  fISDEP(J) = min(fISDEP(J),hSDEP)
                  fIHDEP(J) = min(fIHDEP(J),hHDEP)
20          CONTINUE
30    CONTINUE
c     dividing the resulting depths by number of points d
      DO 40 I=1,M
            funSDEP(I) = funSDEP(I)/(D+0.0)
            funHDEP(I) = funHDEP(I)/(D+0.0)
40    CONTINUE
      RETURN
      END 
      
c----------------------------------------------------------------------

      SUBROUTINE fD1(U,N,X,SDEP,HDEP)
c     computes 1D simplicial and halfspace depth of U wrt X of size N
c     used in the univariate integrated depth

      double precision U,X(N),SDEP,HDEP
      INTEGER N,RB,RA,I
      
      RB = 0
      RA = 0
      DO 10 I=1,N
            IF (U .LE. X(I)) THEN 
                  RB = RB + 1
                  ENDIF
            IF (U .GE. X(I)) THEN 
                  RA = RA + 1
                  ENDIF
10    CONTINUE     
      HDEP = min(RA+0.0,RB+0.0)/(N+0.0)
      SDEP = (RA+0.0)*(RB+0.0)/(K(N,2)+0.0)
      RETURN
      END
      
c----------------------------------------------------------------------

      INTEGER FUNCTION K(M,J)
c     combination number (m choose j)

      integer m,j
      IF (M.LT.J) THEN
          K=0
      ELSE
          IF (J.EQ.1) K=M
          IF (J.EQ.2) K=(M*(M-1))/2
          IF (J.EQ.3) K=(M*(M-1)*(M-2))/6
      ENDIF
      RETURN
      END
      
c----------------------------------------------------------------------

      SUBROUTINE metrl2(A,B,M,N,D,METR)
c     computes a fast approximation of the L2 metric between A and B
c     supporting functions on regular grids only
c     used in the h-mode depth 
        
      INTEGER N,M,D
      double precision A(M*D),B(N*D),METR(M*N)
      INTEGER I,J,K

      DO 15 I=1,M
            DO 10 J=1,N
            METR((J-1)*M+I) = 0.0
                  DO 5 K=1,D
                  METR((J-1)*M+I) = METR((J-1)*M+I) + (A((K-1)*M+I)-
     +B((K-1)*N+J))**2
5                 CONTINUE
            METR((J-1)*M+I) = sqrt(METR((J-1)*M+I) - 
     +((A((0)*M+I)-B((0)*N+J))**2+(A((D-1)*M+I)-B((D-1)*N+J))**2)
     +/(2.0))
10          CONTINUE
15    CONTINUE
      RETURN
      END 
      
c----------------------------------------------------------------------

      SUBROUTINE metrC(A,B,M,N,D,METR)
c     computes a fast approximation of the C metric between A and B
c     supporting functions on regular grids only
c     used in the h-mode depth 
        
      INTEGER N,M,D
      double precision A(M*D),B(N*D),METR(M*N)
      INTEGER I,J,K

      DO 15 I=1,M
            DO 10 J=1,N
            METR((J-1)*M+I) = 0.0
                  DO 5 K=1,D
                  METR((J-1)*M+I) = MAX(METR((J-1)*M+I),A((K-1)*M+I)-
     +B((K-1)*N+J))
            METR((J-1)*M+I) = MAX(METR((J-1)*M+I),-A((K-1)*M+I)+
     +B((K-1)*N+J))
5                 CONTINUE
10          CONTINUE
15    CONTINUE
      RETURN
      END 

c----------------------------------------------------------------------

      double precision FUNCTION MAX(A,B)
c     maximum of two numbers A and B

      double precision A,B
      
      IF (A.LE.B) THEN
            MAX = B
      ELSE
            MAX = A
      ENDIF
      RETURN
      END 

      double precision FUNCTION AdjLPindicator(EVAL,J,B,V)
c     adjusted band depth core function, smoothing (exp(-u)), L2 metric
c     b is a vector of length eval
c     v is a matrix j*eval

      INTEGER EVAL, J
      double precision B(EVAL), V(J*EVAL)
      double precision MINI, MAXI, DIST, POWER
      INTEGER I, II

      DIST = 0.0
      POWER = 1.0
c     power in exp(-power*dist) for weighting
      DO 10 I=1,EVAL 
            MINI = V((I-1)*J+1)
            MAXI = V((I-1)*J+1)
            DO 5 II=1,J
            IF (MINI.GT.V((I-1)*J+II)) THEN
                  MINI = V((I-1)*J+II)
           ENDIF
            IF (MAXI.LT.V((I-1)*J+II)) THEN
                  MAXI = V((I-1)*J+II)
            ENDIF
5           CONTINUE
            IF  ((B(I).GE.MINI).AND.(B(I).LE.MAXI)) THEN
                  DIST = DIST + 0.0
            ELSE
                  IF(B(I).GT.MAXI) THEN
                  DIST = DIST + (B(I)-MAXI)**2
                  ENDIF
                  IF(B(I).LT.MINI) THEN
                  DIST = DIST + (MINI-B(I))**2
                  ENDIF
            ENDIF
10    CONTINUE
      DIST = DIST/(EVAL+0.0)
      DIST = EXP(-POWER*DIST)
      AdjLPindicator = DIST
      RETURN
      END

      SUBROUTINE AdjLP(EVAL,J,M,KOMB,COM,B,V,DJ)
c     adjusted band depth function, smoothing (exp(-u)), L2 metric
c     eval     no of time points for each functions
c     J        order of the depth
c     m        sample size
c     komb     m choose J number
c     com      vector of all the combinations of J elements from m
c     b        functional values of the x function, length eval
c     v        matrix of sample funct. values, dimension [m x eval]

      INTEGER EVAL,J,M,KOMB,COM(KOMB*J)
      double precision B(EVAL),V(M*EVAL),DJ
      double precision VPOM(J*EVAL)
      INTEGER C,I,II
      double precision ADJLPINDICATOR

      DJ = 0.0
      DO 10 C=1,KOMB
            DO 5 I=1,J
                  DO 3 II=1,EVAL
                        VPOM((II-1)*J+I) = V((II-1)*M+COM((C-1)*J+I))
3                 CONTINUE
5           CONTINUE 
            DJ = DJ + AdjLPindicator(EVAL,J,B,VPOM)     
10    CONTINUE
      DJ = DJ/(KOMB+0.0)
      RETURN
      END

c----------------------------------------------------------------------
      
      double precision FUNCTION AdjCindicator(EVAL,J,B,V)
c     adjusted band depth core function, smoothing (exp(-u)), C metric
c     b is a vector of length eval
c     v is a matrix j*eval

      INTEGER EVAL, J
      double precision B(EVAL), V(J*EVAL)
      double precision MINI, MAXI, DIST, POWER, MAX
      INTEGER I, II

      DIST = 0.0
      POWER = 1.0
c     power in exp(-power*dist) for weighting
      DO 10 I=1,EVAL 
            MINI = V((I-1)*J+1)
            MAXI = V((I-1)*J+1)
            DO 5 II=1,J
            IF (MINI.GT.V((I-1)*J+II)) THEN
                  MINI = V((I-1)*J+II)
           ENDIF
            IF (MAXI.LT.V((I-1)*J+II)) THEN
                  MAXI = V((I-1)*J+II)
            ENDIF
5           CONTINUE
            IF  ((B(I).GE.MINI).AND.(B(I).LE.MAXI)) THEN
                  DIST = DIST + 0.0
            ELSE
                  IF(B(I).GT.MAXI) THEN
                  DIST = MAX(DIST,(B(I)-MAXI))
                  ENDIF
                  IF(B(I).LT.MINI) THEN
                  DIST = MAX(DIST,(MINI-B(I)))
                  ENDIF
            ENDIF
10    CONTINUE
      DIST = EXP(-POWER*DIST)
      AdjCindicator = DIST
      RETURN
      END

      SUBROUTINE AdjC(EVAL,J,M,KOMB,COM,B,V,DJ)
c     adjusted band depth function, smoothing (exp(-u)), C metric
c     eval     no of time points for each functions
c     J        order of the depth
c     m        sample size
c     komb     m choose J number
c     com      vector of all the combinations of J elements from m
c     b        functional values of the x function, length eval
c     v        matrix of sample funct. values, dimension [m x eval]

      INTEGER EVAL,J,M,KOMB,COM(KOMB*J)
      double precision B(EVAL),V(M*EVAL),DJ
      double precision VPOM(J*EVAL)
      INTEGER C,I,II
      double precision ADJCINDICATOR

      DJ = 0.0
      DO 10 C=1,KOMB
            DO 5 I=1,J
                  DO 3 II=1,EVAL
                        VPOM((II-1)*J+I) = V((II-1)*M+COM((C-1)*J+I))
3                 CONTINUE
5           CONTINUE 
            DJ = DJ + AdjCindicator(EVAL,J,B,VPOM)     
10    CONTINUE
      DJ = DJ/(KOMB+0.0)
      RETURN
      END
      
c-----------------------------------
c     DiffDepth.f
c-----------------------------------

c-------------------------------------
c     first Difference Depth for 1D functions
c     for 1d functions, halfspace and simplicial depth
c-------------------------------------

      SUBROUTINE DiffD(A,B,M,N,D,REP,RN,funSDEP,funHDEP,funSDEPm,
     +funHDEPm,PSDEP,PHDEP,IAsdep,IAhdep)

c     using fdepth computes 2dimensional diff depth of functions in A
c     with respect to the functions in B
c     M     size of functions in A
c     N     size of functions in B
c     D     dimensionality of functions
c     REP   is the number of simulations to approximate the real value
c           of depth, if set to 0 full comutation is made
c     RN    random numbers, arrray of randoms from 1:D of size 2*REP
c     funSDEPm    DiffDepth when taking infimum of projected depths
c     funSDEP     DiffDepth when taking integral of projected depths
c     P.DEP pointwise depth, depth at each point of domain x domain
c     P.DEP is computed only if M=1, for simplicity
        
      INTEGER N,M,D,REP
      INTEGER IAsdep(M),IAhdep(M)
      double precision A(M*D),B(N*D),funSDEP(M),funHDEP(M),PSDEP(D*D),
     +PHDEP(D*D)
      double precision funSDEPm(M),funHDEPm(M),B1H(N),B2H(N)
      double precision hSDEP,hHDEP
      double precision hALPHA(N)
      INTEGER hF(N),I,J,K,RN(2*REP)

      DO 5 I=1,M
            funSDEP(I) = 0.0
            funHDEP(I) = 0.0
            funSDEPm(I) = 2.0
            funHDEPm(I) = 2.0
            IAsdep(I) = 0
            IAhdep(I) = 0
5     CONTINUE
c     initialization for easier recognition of unfilled elements
      IF (M.EQ.1) THEN 
            DO 8 I=1,(D*D)
                  PSDEP(I) = -1.0
                  PHDEP(I) = -1.0
8                 CONTINUE
            ENDIF
c     essential loop, computing 2D depths at each point
      IF (REP.EQ.0) THEN
c     if we compute the complete, non-approximated depth
      DO 30 I=1,D
            DO 25 J=(I+1),D
c     not taking into account diagonal elements, there is 0 depth
                  DO 10 K=1,N
                  B1H(K) = B((I-1)*N+K)
                  B2H(K) = B((J-1)*N+K)
10                CONTINUE
                  DO 20 L=1,M
                        hSDEP = 0.0
                        hHDEP = 0.0
                        hALPHA(1) = N+0.0
                        hF(1) = N
      CALL fD2(A((I-1)*M+L),A((J-1)*M+L),N,B1H,B2H,hALPHA,hF,
     +hSDEP,hHDEP)
                        funSDEP(L) = funSDEP(L) + hSDEP
                        funHDEP(L) = funHDEP(L) + hHDEP
c     counting the area of the smallest depth for each function
                        IF (hSDEP.EQ.funSDEPm(L)) THEN
                              IAsdep(L) = IAsdep(L)+1
                        ELSEIF (hSDEP.LT.funSDEPm(L)) THEN 
                              IAsdep(L) = 1
                        ENDIF
                        IF (hHDEP.EQ.funHDEPm(L)) THEN
                              IAhdep(L) = IAhdep(L)+1
                        ELSEIF (hHDEP.LT.funHDEPm(L)) THEN 
                              IAhdep(L) = 1
                        ENDIF 
                        funSDEPm(L) = min(funSDEPm(L),hSDEP)
                        funHDEPm(L) = min(funHDEPm(L),hHDEP)
                        IF (M.EQ.1) THEN
                              PSDEP((I-1)*D+J) = hSDEP
                              PHDEP((I-1)*D+J) = hHDEP
                              ENDIF
20                CONTINUE
25          CONTINUE
30    CONTINUE
c       dividing the resulting depths by number of points d
c     for infimal area, the result is *2+D because the diagonal 
c     has zero depth by default, and only the lower triangle 
c     of the matrix is actually computed
      DO 40 I=1,M
            funSDEP(I) = 2.0*funSDEP(I)/(D*(D-1.0))
c     diagonals are always zero, do not count into the sum 
c ((D+0.0)*(D+1.0)/2+0.0-D)
            funHDEP(I) = 2.0*funHDEP(I)/(D*(D-1.0))
c ((D+0.0)*(D+1.0)/2+0.0-D)
            IAhdep(I) = IAhdep(I)*2+D
            IAsdep(I) = IAsdep(I)*2+D
40    CONTINUE
      ELSE 
c     else of if(REP.EQ.0), that is if we approximate
      DO 70 I=1,REP
c     going through random elements
                  DO 50 K=1,N
                  B1H(K) = B((RN(2*I-1)-1)*N+K)
                  B2H(K) = B((RN(2*I)-1)*N+K)
50                CONTINUE
                  DO 60 L=1,M
                        hSDEP = 0.0
                        hHDEP = 0.0
                        hALPHA(1) = N+0.0
                        hF(1) = N
      CALL fD2(A((RN(2*I-1)-1)*M+L),A((RN(2*I)-1)*M+L),N,
     +B1H,B2H,hALPHA,hF,hSDEP,hHDEP)
                        funSDEP(L) = funSDEP(L) + hSDEP
                        funHDEP(L) = funHDEP(L) + hHDEP
c     counting the area of the smallest depth for each function
                        IF (hSDEP.EQ.funSDEPm(L)) THEN
                              IAsdep(L) = IAsdep(L)+1
                        ELSEIF (hSDEP.LT.funSDEPm(L)) THEN 
                              IAsdep(L) = 1
                        ENDIF
                        IF (hHDEP.EQ.funHDEPm(L)) THEN
                              IAhdep(L) = IAhdep(L)+1
                        ELSEIF (hHDEP.LT.funHDEPm(L)) THEN 
                              IAhdep(L) = 1
                        ENDIF 
                        funSDEPm(L) = min(funSDEPm(L),hSDEP)
                        funHDEPm(L) = min(funHDEPm(L),hHDEP)
                        IF (M.EQ.1) THEN
                              PSDEP((RN(2*I-1)-1)*D+RN(2*I)) = hSDEP
                              PHDEP((RN(2*I-1)-1)*D+RN(2*I)) = hHDEP
                              ENDIF
60                CONTINUE
70    CONTINUE
      DO 80 I=1,M
            funSDEP(I) = funSDEP(I)/(REP+0.0) 
            funHDEP(I) = funHDEP(I)/(REP+0.0)
80    CONTINUE      
      ENDIF
      RETURN
      END 
      
c----------------------------------------------------------------------

      SUBROUTINE fD2(U,V,N,X,Y,ALPHA,F,SDEP,HDEP)

c        Rousseuw, P.J., and Ruts, I. (1996), AS 307 : Bivariate location
c        depth, Applied Statistics (JRRSS-C), vol.45, 516-526

      double precision U,V,X(n),Y(n),ALPHA(n)
      double precision P,P2,EPS,D,XU,YU,ANGLE,ALPHK,BETAK,SDEP,HDEP
      INTEGER F(N),GI
      integer n,nums,numh,nt,i,nn,nu,ja,jb,nn2,nbad,nf,j,ki,k
      NUMS=0
      NUMH=0
      SDEP=0.0
      HDEP=0.0
      IF (N.LT.1) RETURN
      P=ACOS(-1.0)
      P2=P*2.0
      EPS=0.00000001
      NT=0
C
C  Construct the array ALPHA.
C
      DO 10 I=1,N
          D=SQRT((X(I)-U)*(X(I)-U)+(Y(I)-V)*(Y(I)-V))
          IF (D.LE.EPS) THEN
              NT=NT+1
          ELSE
              XU=(X(I)-U)/D
              YU=(Y(I)-V)/D
              IF (ABS(XU).GT.ABS(YU)) THEN
                  IF (X(I).GE.U) THEN
                      ALPHA(I-NT)=ASIN(YU)
                      IF(ALPHA(I-NT).LT.0.0) THEN
                          ALPHA(I-NT)=P2+ALPHA(I-NT)
                      ENDIF
                  ELSE
                      ALPHA(I-NT)=P-ASIN(YU)
                  ENDIF
              ELSE
                  IF (Y(I).GE.V) THEN
                      ALPHA(I-NT)=ACOS(XU)
                  ELSE
                      ALPHA(I-NT)=P2-ACOS(XU)
                  ENDIF
              ENDIF
              IF (ALPHA(I-NT).GE.(P2-EPS)) ALPHA(I-NT)=0.0
          ENDIF
  10  CONTINUE
      NN=N-NT
      IF (NN.LE.1) GOTO 60
C
C  Sort the array ALPHA.
C
      CALL SORT(ALPHA,NN)
C
C  Check whether theta=(U,V) lies outside the data cloud.
C
      ANGLE=ALPHA(1)-ALPHA(NN)+P2
      DO 20 I=2,NN
          ANGLE=MAX(ANGLE,(ALPHA(I)-ALPHA(I-1)))
  20  CONTINUE
      IF (ANGLE.GT.(P+EPS)) GOTO 60
C
C  Make smallest alpha equal to zero,
C  and compute NU = number of alpha < pi.
C
      ANGLE=ALPHA(1)
      NU=0
      DO 30 I=1,NN
          ALPHA(I)=ALPHA(I)-ANGLE
          IF (ALPHA(I).LT.(P-EPS)) NU=NU+1
  30  CONTINUE
      IF (NU.GE.NN) GOTO 60
C
C  Mergesort the alpha with their antipodal angles beta,
C  and at the same time update I, F(I), and NBAD.
C
      JA=1
      JB=1
      ALPHK=ALPHA(1)
      BETAK=ALPHA(NU+1)-P
      NN2=NN*2
      NBAD=0
      I=NU
      NF=NN
      DO 40 J=1,NN2
          IF ((ALPHK+EPS).LT.BETAK) THEN
              NF=NF+1
              IF (JA.LT.NN) THEN
                  JA=JA+1
                  ALPHK=ALPHA(JA)
              ELSE
                  ALPHK=P2+1.0
              ENDIF
          ELSE
              I=I+1
              IF (I.EQ.(NN+1)) THEN
                  I=1
                  NF=NF-NN
              ENDIF
              F(I)=NF
              NBAD=NBAD+K((NF-I),2)
              IF (JB.LT.NN) THEN
                  JB=JB+1
                  IF ((JB+NU).LE.NN) THEN
                      BETAK=ALPHA(JB+NU)-P
                  ELSE
                      BETAK=ALPHA(JB+NU-NN)+P
                  ENDIF
              ELSE
                  BETAK=P2+1.0
              ENDIF
          ENDIF
  40  CONTINUE
      NUMS=K(NN,3)-NBAD
C
C  Computation of NUMH for halfspace depth.
C
      GI=0
      JA=1
      ANGLE=ALPHA(1)
      NUMH=MIN(F(1),(NN-F(1)))
      DO 50 I=2,NN
          IF(ALPHA(I).LE.(ANGLE+EPS)) THEN
              JA=JA+1
          ELSE
              GI=GI+JA
              JA=1
              ANGLE=ALPHA(I)
          ENDIF
          KI=F(I)-GI
          NUMH=MIN(NUMH,MIN(KI,(NN-KI)))
   50 CONTINUE
C
C  Adjust for the number NT of data points equal to theta:
C
  60  NUMS=NUMS+K(NT,1)*K(NN,2)+K(NT,2)*K(NN,1)+K(NT,3)
      IF (N.GE.3) SDEP=(NUMS+0.0)/(K(N,3)+0.0)
      NUMH=NUMH+NT
      HDEP=(NUMH+0.0)/(N+0.0)
      RETURN
      END
c----------------------------------------------------------------------


c      INTEGER FUNCTION K(M,J)
c      integer m,j
c      IF (M.LT.J) THEN
c          K=0
c      ELSE
c          IF (J.EQ.1) K=M
c          IF (J.EQ.2) K=(M*(M-1))/2
c          IF (J.EQ.3) K=(M*(M-1)*(M-2))/6
c      ENDIF
c      RETURN
c      END

c----------------------------------------------------------------------

      SUBROUTINE SORT(B,N)
C  Sorts an array B (of length N<=1000) in O(NlogN) time.

      double precision B(N),x(n)
      integer q(n),i,n
      
      call indexx(n,b,q)
      
      do 10 i=1,n
        x(i)=b(i)
10    continue
 
      do 20 i=1,n
        b(i)=x(q(i))
20    continue
      end
c----------------------------------------------------------------------

      SUBROUTINE INDEXX(N,ARRIN,INDX)

      integer INDX(N)
      integer n,j,l,ir,indxt,i
      double precision arrin(n),q

      DO 11 J=1,N
        INDX(J)=J
11    CONTINUE
      L=N/2+1
      IR=N
10    CONTINUE
        IF(L.GT.1)THEN
          L=L-1
          INDXT=INDX(L)
          Q=ARRIN(INDXT)
        ELSE
          INDXT=INDX(IR)
          Q=ARRIN(INDXT)
          INDX(IR)=INDX(1)
          IR=IR-1
          IF(IR.EQ.1)THEN
            INDX(1)=INDXT
            RETURN
          ENDIF
        ENDIF
        I=L
        J=L+L
20      IF(J.LE.IR)THEN
          IF(J.LT.IR)THEN
            IF(ARRIN(INDX(J)).LT.ARRIN(INDX(J+1)))J=J+1
          ENDIF
          IF(Q.LT.ARRIN(INDX(J)))THEN
            INDX(I)=INDX(J)
            I=J
            J=J+J
          ELSE
            J=IR+1
          ENDIF
        GO TO 20
        ENDIF
        INDX(I)=INDXT
      GO TO 10
      END
      
cccccccccccccccccc
c
c     newdepth
c
cccccccccccccccccc

c----------------------------------------------------------------------
      SUBROUTINE funMD(A,B,M,N,D,Q,funDEP)
      
      INTEGER N,M,D,I,J      
      double precision A(M*D),B(N*D),METR1(N*N),METR2(N*M),funDEP(M)
      double precision Q,H,PI
c     Q is quantile, default 0.15, Cuevas 0.2     
      
      CALL metrl2(B,B,N,N,D,METR1)
      CALL metrl2(A,B,M,N,D,METR2)
      CALL SORT(METR1,N*N)
      H = METR1(INT(Q*(N*N+0.0)))
      PI=4.D0*DATAN(1.D0)
      DO 10 I=1,M*N
            METR2(I) = EXP(-(METR2(I)/H)**2/2.0)/SQRT(2.0*PI)
10    CONTINUE
      DO 20 I=1,M
            funDEP(I)=0.0
            DO 15 J=1,N
                  funDEP(I) = funDEP(I)+METR2((J-1)*M+I)
15          CONTINUE
20    CONTINUE      
      RETURN
      END
      

c----------------------------------------------------------------------
      SUBROUTINE funRPD1(A,B,M,N,D,NPROJ,NPROJ2,V,funSDEP,
     +funHDEP,funRDEP)
      
c     computes fast random projection depth for 1D functions using 
c     NPROJ simple iid gaussian processes projection and then 1D
c     simplicial or halfspace depth 
c     RDEP random halfspace depth (Cuesta-Albertos, Nieto-Reyes 2008)
c     NPROJ2 nr of projections for RDEP

      INTEGER M,N,D,NPROJ,NPROJ2
      double precision A(M*D),B(N*D),funSDEP(M),funHDEP(M),funRDEP(M)
      double precision V(D*NPROJ),AP,BP(N),VNORM
      INTEGER I,J,K
      double precision hSDEP,hHDEP

      DO 5 I=1,M
            funSDEP(I) = 0.0
            funHDEP(I) = 0.0
            funRDEP(I) = 2.0
5     CONTINUE      
c     main loop
      DO 40 K=1,NPROJ
c     generating a single gaussian processes
      VNORM = 0.0
      DO 10 I=1,D
             Vnorm = VNORM + V((K-1)*D+I)**2
10    CONTINUE
      VNORM = sqrt(VNORM - (V((K-1)*D+1)**2+V((K-1)*D+D)**2)/(2.0))
c     loop projecting j-th function of the random sample B on v
      DO 18 J=1,N
            BP(J) = 0.0
            DO 15 I=1,D
                  BP(J) = BP(J)+ B((I-1)*N+J)*V((K-1)*D+I)/VNORM
15          CONTINUE
            BP(J) = BP(J)/(D+0.0)
18    CONTINUE
c     loop with projecting j-th function of the random sample A on v
      DO 30 J=1,M
            AP = 0.0
c     compute projection of j-th function A
            DO 20 I=1,D
                  AP = AP + A((I-1)*M+J)*V((K-1)*D+I)/VNORM
20          CONTINUE
            AP = AP/(D+0.0)
            hSDEP = 0.0
            hHDEP = 0.0
            CALL fD1(AP,N,BP,hSDEP,hHDEP)
            funSDEP(J) = funSDEP(J) + hSDEP
            funHDEP(J) = funHDEP(J) + hHDEP
            IF (K .LE. NPROJ2) THEN 
                  funRDEP(J) = min(funRDEP(J),hHDEP)
                  ENDIF
30    CONTINUE
40    CONTINUE
c     averaging projection depth over all projections
      DO 50 I=1,M
            funSDEP(I) = funSDEP(I)/(NPROJ+0.0)
            funHDEP(I) = funHDEP(I)/(NPROJ+0.0)
50    CONTINUE
      RETURN
      END
      
c-------------------------------------
c Fraiman Muniz type integrated depths
c for 1 and 2d functions, halfspace and simplicial depth
c-------------------------------------

      SUBROUTINE funD2(A1,A2,B1,B2,M,N,D,funSDEP,funHDEP,
     +fIsdep,fIhdep,IAsdep,IAhdep)

c     using fdepth computes 2dimensional depth of functions in A1,A2
c     with respect to the functions in B1,B2
c     M size of functions in A
c     N size of functions in B
c     D dimensionality of functions
        
      INTEGER N,M,D
      INTEGER IAsdep(M),IAhdep(M)
      double precision A1(M*D),A2(M*D),B1(N*D),B2(N*D),funSDEP(M),
     +funHDEP(M),fIsdep(M),fIhdep(M) 
      double precision B1H(N),B2H(N)
      double precision hSDEP,hHDEP
      double precision hALPHA(N)
      INTEGER hF(N),I,J,K

      DO 5 I=1,M
            funSDEP(I) = 0.0
            funHDEP(I) = 0.0
            fISDEP(I) = 2.0
            fIHDEP(I) = 2.0 
            IAsdep(I) = 0
            IAhdep(I) = 0           
5     CONTINUE
c       essential loop, computing 2D depths at each point
      DO 30 I=1,D
            DO 10 K=1,N
                  B1H(K) = B1((I-1)*N+K)
                  B2H(K) = B2((I-1)*N+K)
10          CONTINUE
            DO 20 J=1,M
                  hSDEP = 0.0
                  hHDEP = 0.0
                  hALPHA(1) = N+0.0
                  hF(1) = N
      CALL fD2(A1((I-1)*M+J),A2((I-1)*M+J),N,B1H,B2H,hALPHA,hF,
     +hSDEP,hHDEP)
                  funSDEP(J) = funSDEP(J) + hSDEP
                  funHDEP(J) = funHDEP(J) + hHDEP
c     counting the area of the smallest depth for each function
                  IF (hSDEP.EQ.fISDEP(J)) THEN
                        IAsdep(J) = IAsdep(J)+1
                  ELSEIF (hSDEP.LT.fISDEP(J)) THEN 
                        IAsdep(J) = 1
                  ENDIF
                  IF (hHDEP.EQ.fIHDEP(J)) THEN
                        IAhdep(J) = IAhdep(J)+1
                  ELSEIF (hHDEP.LT.fIHDEP(J)) THEN 
                        IAhdep(J) = 1
                  ENDIF 
c     infimal depth evaluation
                  fISDEP(J) = min(fISDEP(J),hSDEP)
                  fIHDEP(J) = min(fIHDEP(J),hHDEP)                  
20          CONTINUE
30    CONTINUE
c       dividing the resulting depths by number of points d
      DO 40 I=1,M
            funSDEP(I) = funSDEP(I)/(D+0.0)
            funHDEP(I) = funHDEP(I)/(D+0.0)
40    CONTINUE
      RETURN
      END 
      
c----------------------------------------------------------------------
c     2D RANDOM PROJECTION DEPTH
c     PROJECTIONS ON WHITE NOISE
c     DEPENDS ON THE RANDOM NUMBER GENERATOR VERSION
c----------------------------------------------------------------------
      
      SUBROUTINE funRPD2(A1,A2,B1,B2,M,N,D,NPROJ,V,Q,funSDEP,
     +funHDEP,funMDEP,funSDDEP,funHDDEP)
      
c     computes fast random projection depth for 2D functions using 
c     NPROJ simple iid gaussian processes projection and then 2D
c     simplicial or halfspace depth 
c     Q is quantile used in MDEP
c     funDDEP is double random projection (2*D-dim)->(2-dim)->(1-dim)
c     and then appying Halfspace depth (equiv to quantile as Cuevas 2007)

      INTEGER M,N,D,NPROJ
      double precision A1(M*D),A2(M*D),B1(N*D),B2(N*D),funSDEP(M),
     +funHDEP(M),funMDEP(M),Q,funSDDEP(M),funHDDEP(M)
      double precision V(D*NPROJ+2*NPROJ),A1P,A2P,B1P(N),B2P(N),
     +VNORM,VNORMF,AP(2*M),BP(2*N),VF(2),AFP,BFP(N)
c     last 2*NPROJ of V are for the second random projection 2d->1d
c     VF(2) is the final projection
      INTEGER I,J,K
      double precision hSDEP,hHDEP,hMDEP(M),hHDDEP,hSDDEP
      double precision hALPHA(N)
      INTEGER hF(N)

      DO 5 I=1,M
            funSDEP(I) = 0.0
            funHDEP(I) = 0.0
            funMDEP(I) = 0.0
            funSDDEP(I) = 0.0
            funHDDEP(I) = 0.0
5     CONTINUE      
c     main loop
      DO 40 K=1,NPROJ
c     generating a single gaussian processes
      VNORM = 0.0
      VNORMF = sqrt(V(NPROJ*D+(K-1)*2+1)**2+V(NPROJ*D+(K-1)*2+2)**2)
      VF(1) = V(NPROJ*D+(K-1)*2+1)/VNORMF
      VF(2) = V(NPROJ*D+(K-1)*2+2)/VNORMF
c     final projection is now normed
      DO 10 I=1,D
             VNORM = VNORM + V((K-1)*D+I)**2
10    CONTINUE
      VNORM = sqrt(VNORM - (V((K-1)*D+1)**2+V((K-1)*D+D)**2)/(2.0))
c     loop projecting j-th function of the random sample B on v
      DO 18 J=1,N
            B1P(J) = 0.0
            B2P(J) = 0.0
            DO 15 I=1,D
                  B1P(J) = B1P(J)+ B1((I-1)*N+J)*V((K-1)*D+I)/VNORM
                  B2P(J) = B2P(J)+ B2((I-1)*N+J)*V((K-1)*D+I)/VNORM
15          CONTINUE
            BFP(J) = B1P(J)*VF(1)+B2P(J)*VF(2)
18    CONTINUE
      DO 19 I=1,N
            BP(I)=B1P(I)
            BP(N+I)=B2P(I)
19    CONTINUE
c     loop with projecting j-th function of the random sample A on v
      DO 30 J=1,M
            A1P = 0.0
            A2P = 0.0
c     compute projection of j-th function A
            DO 20 I=1,D
                  A1P = A1P + A1((I-1)*M+J)*V((K-1)*D+I)/VNORM
                  A2P = A2P + A2((I-1)*M+J)*V((K-1)*D+I)/VNORM
20          CONTINUE
            AP(J) = A1P
            AP(J+M) = A2P
            hSDEP = 0.0
            hHDEP = 0.0
            hMDEP(J) = 0.0
            hALPHA(1) = N+0.0
            hF(1) = N
            CALL fD2(A1P,A2P,N,B1P,B2P,hALPHA,hF,
     +hSDEP,hHDEP)
            funSDEP(J) = funSDEP(J) + hSDEP
            funHDEP(J) = funHDEP(J) + hHDEP
c     and now the second projection of A
            AFP = A1P*VF(1)+A2P*VF(2)
            hHDDEP = 0.0
            hSDDEP = 0.0
            CALL fD1(AFP,N,BFP,hSDDEP,hHDDEP)
            funSDDEP(J) = funSDDEP(J) + hSDDEP
            funHDDEP(J) = funHDDEP(J) + hHDDEP
30    CONTINUE
      CALL funMD(AP,BP,M,N,2,Q,hMDEP)
      DO 35 I=1,M
            funMDEP(I) = funMDEP(I) + hMDEP(I)
35    CONTINUE
40    CONTINUE
c     averaging projection depth over all projections
      DO 50 I=1,M
            funSDEP(I) = funSDEP(I)/(NPROJ+0.0)
            funHDEP(I) = funHDEP(I)/(NPROJ+0.0)
            funMDEP(I) = funMDEP(I)/(NPROJ+0.0)
            funSDDEP(I) = funSDDEP(I)/(NPROJ+0.0)
            funHDDEP(I) = funHDDEP(I)/(NPROJ+0.0)
50    CONTINUE
      RETURN
      END
      
c
c
c     halfregiondepth.f
c
c

      SUBROUTINE HRD(A,B,M,N,D,FD)
c     computes fast half-region depth of Lopez-Pintado and Romo 2011
      INTEGER M,N,D
      double precision A(M*D),B(N*D),FD(M)
      INTEGER I,J,K, U, L, UI, LI

      DO 30 I=1,M
            FD(I) = 0.0
            U = 0
            L = 0
            DO 20 J=1,N
                  UI = 0
                  LI = 0
                  K = 0
                  DO 10 WHILE ((K .LT. D) 
     +                  .AND. (UI .EQ. 0 .OR. LI .EQ. 0)) 
                        K = K + 1
                        IF(A((K-1)*M+I) .GT. B((K-1)*N+J)) UI = UI + 1
                        IF(A((K-1)*M+I) .LT. B((K-1)*N+J)) LI = LI + 1
10                CONTINUE
                  IF (UI .EQ. 0) U = U+1
                  IF (LI .EQ. 0) L = L+1
20          CONTINUE
            FD(I) = (MIN(U,L)+0.0)/(N+0.0)             
30    CONTINUE
      RETURN
      END 

c----------------------------------------------------------------------      
      SUBROUTINE BD(A,B,M,N,D,FD)
c     computes fast 2nd order band depth of Lopez-Pintado and Romo 2009
      INTEGER M,N,D
      double precision A(M*D),B(N*D),FD(M),L,U
      INTEGER I,J,JJ,K,W,WI

      DO 30 I=1,M
            FD(I) = 0.0
            W = 0
            DO 20 J=1,(N-1)
                  DO 15 JJ=(J+1),N
                        WI = 0
                        K = 0
                  DO 10 WHILE ( WI .GE. 0) 
c     -1 means TRUE, the function is in the band of J and JJ
c     -2 means FALSE, the function crosses
                        WI = WI + 1
                        K = K+1
                        L = MIN(B((K-1)*N+J),B((K-1)*N+JJ))
                        U = MAX(B((K-1)*N+J),B((K-1)*N+JJ))
                        IF (A((K-1)*M+I) .LT. L .OR. 
     +  A((K-1)*M+I) .GT. U) WI = -2
                        IF (WI .NE. -2 .AND. WI .EQ. D) WI = -1
10                CONTINUE
                  IF (WI .EQ. -1) W = W+1
15                CONTINUE
20          CONTINUE
            FD(I) = (W+0.0)/((N*(N-1))/2 + 0.0)          
30    CONTINUE
      RETURN
      END 
      
c-------------------------------------
c bivariate halfspace and simplicial depth
c-------------------------------------

      SUBROUTINE DPTH2(A1,A2,B1,B2,M,N,SDEP,HDEP)

c     computes 2dimensional depth of A1,A2
c     with respect to B1,B2
c     M size of A
c     N size of B
        
      INTEGER N,M
      double precision A1(M),A2(M),B1(N),B2(N),SDEP(M),HDEP(M)
      double precision hSDEP,hHDEP
      double precision hALPHA(N)
      INTEGER hF(N),I,J,K

      DO 5 I=1,M
            SDEP(I) = 0.0
            HDEP(I) = 0.0
5     CONTINUE
      DO 10 J=1,M
            hSDEP = 0.0
            hHDEP = 0.0
            hALPHA(1) = N+0.0
            hF(1) = N
      CALL fD2(A1(J),A2(J),N,B1,B2,hALPHA,hF,
     +hSDEP,hHDEP)
            SDEP(J) = hSDEP
            HDEP(J) = hHDEP
10    CONTINUE
      RETURN
      END 
      
c-------------------------------------
c univariate halfspace and simplicial depth
c-------------------------------------

      SUBROUTINE DPTH1(A1,B1,M,N,SDEP,HDEP)

c     computes 1dimensional depth of A1
c     with respect to B1
c     M size of A
c     N size of B
        
      INTEGER N,M
      double precision A1(M),B1(N),SDEP(M),HDEP(M)
      double precision hSDEP,hHDEP
      double precision hALPHA(N)
      INTEGER hF(N),I,J,K

      DO 5 I=1,M
            SDEP(I) = 0.0
            HDEP(I) = 0.0
5     CONTINUE
      DO 10 J=1,M
            hSDEP = 0.0
            hHDEP = 0.0
            hALPHA(1) = N+0.0
            hF(1) = N
      CALL fD1(A1(J),N,B1,hSDEP,hHDEP)
            SDEP(J) = hSDEP
            HDEP(J) = hHDEP
10    CONTINUE
      RETURN
      END 

dataf.sim.1.CFF07 <- function(numTrain = 100, numTest = 50, numDiscrets = 51, plot = FALSE){
  # Processes:
  # X(t) = m_0(t) + e(t), m_0(t) = 30*(1-t)*t^1.2
  # Y(t) = m_1(t) + e(t), m_1(t) = 30*(1-t)^1.2*t
  # e(t): Gaussian with mean = 0, cov(X(s), X(t)) = 0.2*exp(-abs(s - t)/0.3)
  
  t <- 0:(numDiscrets - 1)/(numDiscrets - 1)
  mean0 <- 30*(1-t)*t^1.2
  mean1 <- 30*(1-t)^1.2*t
  cov <- matrix(nrow=numDiscrets, ncol=numDiscrets)
  for (i in 1:numDiscrets){
    for (j in 1:numDiscrets){
      cov[i,j] <- 0.2*exp(-abs(t[i] - t[j])/0.3)
    }
  }
  
  X <- mvrnorm(n=numTrain+numTest, mu=mean0, Sigma=cov)
  Y <- mvrnorm(n=numTrain+numTest, mu=mean1, Sigma=cov)
  
  datafX <- list()
  datafY <- list()
  labelsX <- as.list(rep(0,numTrain+numTest))
  labelsY <- as.list(rep(1,numTrain+numTest))
  for (i in 1:(numTrain + numTest)){
    datafX[[i]] <- list(args = t, vals = X[i,])
    datafY[[i]] <- list(args = t, vals = Y[i,])
  }
  
  learn <- list(dataf = c(head(datafX, numTrain), head(datafY, numTrain)), 
                labels = c(head(labelsX, numTrain), head(labelsY, numTrain)))
  class(learn) = "functional"
  test <- list(dataf = c(tail(datafX, numTest), tail(datafY, numTest)), 
               labels = c(tail(labelsX, numTest), tail(labelsY, numTest)))
  class(test) = "functional"
  if (plot){
    plot(0, type="n", xlim=c(0,1), ylim=c(0, 9), 
         main=paste("Model 1 from CuevasFF07: ", 
                    "0 red (", sum(unlist(learn$labels) == 0), "), ", 
                    "1 blue (", sum(unlist(learn$labels) == 1), "), ", sep=""))
    grid()
    for (i in 1:length(learn$dataf)){
      if (learn$labels[[i]] == 0){
        lineColor <- "red"
        lineType <- 1
      }
      if (learn$labels[[i]] == 1){
        lineColor <- "blue"
        lineType <- 2
      }
      lines(learn$dataf[[i]]$args, learn$dataf[[i]]$vals, col=lineColor, lty=lineType)
    }
  }
  
  return (list(learn = learn, test = test))
}

dataf.sim.2.CFF07 <- function(numTrain = 100, numTest = 50, numDiscrets = 51, plot = FALSE){
  # Processes
  # X(t) = m_0(t) + e(t), m_0(t) = 30*(1-t)*t^2 + 0.5*abs(sin(20*pi*t))
  # Y(t) = smooth.spline with 8 knots
  # e(t): Gaussian with mean = 0, cov(X(s), X(t)) = 0.2*exp(-abs(s - t)/0.3)
  
  t <- 0:(numDiscrets - 1)/(numDiscrets - 1)
  mean0 <- 30*(1 - t)*t^2 + 0.5*abs(sin(20*pi*t))
  cov <- matrix(nrow=numDiscrets, ncol=numDiscrets)
  for (i in 1:numDiscrets){
    for (j in 1:numDiscrets){
      cov[i,j] <- 0.2*exp(-abs(t[i] - t[j])/0.3)
    }
  }
  
  X <- mvrnorm(n=numTrain+numTest, mu=mean0, Sigma=cov)
  Y <- NULL
  for (i in 1:nrow(X)){
    Y <- rbind(Y, smooth.spline(t, X[i,], nknots = 8)$y)
  }
  
  datafX <- list()
  datafY <- list()
  labelsX <- as.list(rep(0,numTrain+numTest))
  labelsY <- as.list(rep(1,numTrain+numTest))
  for (i in 1:(numTrain + numTest)){
    datafX[[i]] <- list(args = t, vals = X[i,])
    datafY[[i]] <- list(args = t, vals = Y[i,])
  }
  
  learn <- list(dataf = c(head(datafX, numTrain), head(datafY, numTrain)), 
                labels = c(head(labelsX, numTrain), head(labelsY, numTrain)))
  class(learn) = "functional"
  test <- list(dataf = c(tail(datafX, numTest), tail(datafY, numTest)), 
               labels = c(tail(labelsX, numTest), tail(labelsY, numTest)))
  class(test) = "functional"
  
  if (plot){
    plot(0, type="n", xlim=c(0,1), ylim=c(0, 7), 
         main=paste("Model 2 from CuevasFF07: ", 
                    "0 red (", sum(unlist(learn$labels) == 0), "), ", 
                    "1 blue (", sum(unlist(learn$labels) == 1), "), ", sep=""))
    grid()
    for (i in 1:length(learn$dataf)){
      if (learn$labels[[i]] == 0){
        lineColor <- "red"
        lineType <- 1
      }
      if (learn$labels[[i]] == 1){
        lineColor <- "blue"
        lineType <- 2
      }
      lines(learn$dataf[[i]]$args, learn$dataf[[i]]$vals, col=lineColor, lty=lineType)
    }
  }
  
  return (list(learn = learn, test = test))
}

ddalpha.test <- function(learn, test, ...){
  ops <- options(warn = -1)
  on.exit(options(ops))
  
  if(is.vector(test))
    test = t(as.matrix(test))
  
  tryCatch({
    time <- system.time(
      ddalpha <- ddalpha.train(learn, ...)
    )
    C = ddalpha$dimension+1
    cc = ddalpha.classify(objects = test[,-C], ddalpha = ddalpha)
    if (is.numeric(test[,C])){
      if(is.factor(cc[[1]]) || is.character(cc[[1]])){
        cc <- unlist(lapply(cc, as.character))
        cc[cc == "Ignored"] <- NA
      }
      equal = (cc == test[,C])
    } else {
      cc <- unlist(lapply(cc, as.numeric)) 
      equal = (cc == as.numeric(test[,C]))      
    }
    
    if(!(T %in% equal) && !(F %in% equal)) 
    { return(NA)}
    error = sum(!equal,na.rm = T)/(sum(!equal,na.rm = T)+sum(equal,na.rm = T))
    return(list(error = error, correct = sum(equal,na.rm = T), incorrect = sum(!equal,na.rm = T), 
                total = length(cc)-sum(is.na(equal)), ignored = sum(is.na(equal)), n = length(cc),
                time = time[1])) 
  }
  #  tryCatch({}
  , error = function(e) {
    print ("ERROR T")
    print (e)
  }, finally = {          
  })
  return (NA)
}


ddalpha.getErrorRateCV <- function(data, numchunks = 10,  ...){
  n = nrow(data)
  numchunks = min(n, numchunks)
  chunksize = ceiling(n/numchunks)
  
  sample = seq(from = 1, by = numchunks, length.out = chunksize)   
  
  errors = 0
  total = 0
  times = c()
  
  for (i in 1:numchunks){
    sample = sample[sample<=n]
    learn = data[-sample,,drop = F]
    test  = data[sample,,drop = F]
    
    el = ddalpha.test(learn, test, ...)
    if(is.list(el)){
      errors = errors + el$incorrect
      total = total + el$total
      times = c(times,el$time)
    }
    
    sample = sample+1
  }
  
  return (list(errors = errors/total, time = mean(times), time_sd = sd(times)))  
}


ddalpha.getErrorRatePart <- function(data, size = 0.3, times = 10,  ...){
  
  if (!is.numeric(size) || size <=0 || size >= nrow(data)) stop("Wrong size of excluded sequences")
  
  if(size < 1)
    size = max(1, size*nrow(data)) # at least 1 point
  
  size = as.integer(size)
  
  indexes = 1:nrow(data)
  
  errors = c()
  total = 0
  time = c()
  
  for (i in 1:times){
    samp = sample(indexes, size)
    learn = data[-samp,,drop = F]
    test  = data[samp,,drop = F]
    
    el = ddalpha.test(learn, test, ...)
    if(is.list(el)){
      errors = c(errors,el$incorrect/el$total)
      time = c(time,el$time)
    }
  }
  
  return (list(errors = mean(errors), errors_sd = sd(errors), errors_vec = errors, time = mean(time), time_sd = sd(time)))  
}

################################################################################
# File:             is.in.convex.r
# Created by:       Pavlo Mozharovskyi
# First published:  28.02.2013
# Last revised:     28.02.2013
# 
# Check if points lie in the convex hulls of the data clouds.
################################################################################

is.in.convex <- function(x, data, cardinalities, seed = 0){
  if (!is.numeric(data)
      || !is.matrix(data) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.vector(cardinalities, mode = "numeric") 
      || is.na(min(cardinalities)) 
      || sum(.is.wholenumber(cardinalities)) != length(cardinalities) 
      || min(cardinalities) <= 0 
      || sum(cardinalities) != nrow(data)){
    stop("Argument \"cardinalities\" should be a vector of cardinalities of the classes in \"data\" ")
  }
  if (sum(cardinalities < ncol(data) + 1) != 0){
    stop("Not in all classes sufficiently enough objetcs")
  }
  if (!is.matrix(x) 
       && is.vector(x)){
    x <- matrix(x, nrow=1)
  }
  if (!is.numeric(x)){
    stop("Argument \"x\" should be numeric")
  }
  if (ncol(x) != ncol(data)){
    stop("Dimensions of the arguments \"x\" and \"data\" should coincide")
  }
  
  is.in.convex <- .count_convexes(x, data, cardinalities, seed)
  
  return (is.in.convex)
}

draw.ddplot <- function(ddalpha, depth.space, cardinalities,
                        main = "DD plot", xlab = "C1", ylab = "C2", xlim, ylim, 
                        classes = c(1,2), colors = c("red", "blue", "green"), drawsep = T){
  
  if(!missing(ddalpha)){
    
    col = c()
    points = NULL
    for (c in classes){
    points = rbind(points, ddalpha$patterns[[c]]$depths[,classes])
    col = c(col, rep(colors[c], ddalpha$patterns[[c]]$cardinality))
    }
    
    if (class(xlab) != "expression" && class(ylab) != "expression" &&
        xlab == "C1" && ylab == "C2"){
      xlab = ddalpha$patterns[[1]]$name
      ylab = ddalpha$patterns[[2]]$name
    }
    
    if (missing(xlim)) xlim = c(0, max(points[,]))
    if (missing(ylim)) ylim = c(0, max(points[,]))
    
    plot(points, col = col, main = main, xlab = xlab, ylab = ylab, asp = T, xlim = xlim, ylim = ylim)
    
    if(drawsep && ddalpha$methodSeparator %in% c("alpha", "polynomial"))
    {
      gx <- seq(-0.1, 1.2*max(points[,1]), length=100)
      gy <- seq(0, 1.2*max(points[,2]), length=100)
      y <- as.matrix(expand.grid(x = gx, y = gy))   
      
      if (ddalpha$methodSeparator == "alpha") {
      ray = ddalpha$classifiers[[1]]$hyperplane[-1]
      
      funcs = list(function(x) x[1], function(x) x[2], function(x) x[1]^2, function(x) x[1]*x[2], function(x) x[2]^2, function(x) x[1]^3, function(x) x[1]^2*x[2], function(x) x[1]*x[2]^2, function(x) x[2]^3)
      
      depthcontours = apply(y, 1, function(xx) {
        res = 0
        for(i in 1:ddalpha$classifiers[[1]]$dimProperties)(res = res+funcs[[i]](xx)*ray[i])
        res
        })
      } else if (ddalpha$methodSeparator == "polynomial"){
        if (ddalpha$classifiers[[1]]$axis == 0){
          xAxis <- ddalpha$classifiers[[1]]$index1
          yAxis <- ddalpha$classifiers[[1]]$index2
        }else{
          xAxis <- ddalpha$classifiers[[1]]$index2
          yAxis <- ddalpha$classifiers[[1]]$index1
        }
        
       depthcontours = apply(y, 1, function(xx) {
          res = 0
        for(j in 1:ddalpha$classifiers[[1]]$degree){res <- res + ddalpha$classifiers[[1]]$polynomial[j]*xx[xAxis]^j}
          res = res-xx[yAxis]
        })
      }
      contour(gx, gy, matrix(depthcontours, nrow=length(gx), ncol=length(gy)), add=TRUE, levels=0, drawlabels=FALSE, col = "black")
    }
    
  } else if(!missing(depth.space)){
    
    col = c()
    for (c in classes){
      col = c(col, rep(colors[c], cardinalities[c]))
    }
    
    if (missing(xlim)) xlim = c(0, max(depth.space[,classes]))
    if (missing(ylim)) ylim = c(0, max(depth.space[,classes]))
    
    plot(depth.space[,classes], col = col, main = main, xlab = xlab, ylab = ylab, xlim = xlim, ylim = ylim)
    
  } else stop("Both 'ddalpha' and 'depth.space' area missing")
  
}

plot.ddalpha <- function(x, type = c("ddplot", "depth.contours"), ...){
  type = match.arg(type)
  if(type == "ddplot") 
    draw.ddplot(x, ...)
  if(type == "depth.contours")
    depth.contours.ddalpha(x, ...)
}

plot.ddalphaf <- function(x, type = c("functional.data", "ddplot", "depth.contours"), ...){
  type = match.arg(type)
  if(type == "functional.data")
    plot.functional(list(dataf = x$dataf, labels = lapply(x$data[,ncol(x$data)], function(o){x$labels[[o]]})), ...)
  
  if(class(x$classifier)!="ddalpha")
    stop(type, " is available only for the ddalpha classifier")
  if(type == "ddplot") 
    draw.ddplot(x$classifier, ...)
  if(type == "depth.contours")
    depth.contours.ddalpha(x$classifier, ...)
}
  
  

################################################################################
# File:             depth.space.zonoid.r
# Created by:       Pavlo Mozharovskyi
# First published:  28.02.2013
# Last revised:     15.05.2013
# 
# Computation of the depth space based on the zonoid data depth.
################################################################################

depth.space.zonoid <- function(data, cardinalities, seed = 0){
  if (seed!=0) set.seed(seed)
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.vector(cardinalities, mode = "numeric") 
      || is.na(min(cardinalities)) 
      || sum(.is.wholenumber(cardinalities)) != length(cardinalities) 
      || min(cardinalities) <= 0 
      || sum(cardinalities) != nrow(data)){
    stop("Argument \"cardinalities\" should be a vector of cardinalities of the classes in \"data\" ")
  }
  if (sum(cardinalities < ncol(data) + 1) != 0){
    stop("Not in all classes sufficiently enough objetcs")
  }
  
  dim <- ncol(data)
  depth.space <- NULL
  for (i in 1:length(cardinalities)){
    objects <- as.vector(t(data[(1 + sum(cardinalities[0:(i - 1)])):sum(cardinalities[1:i]),]))
    pattern.depths <- NULL
    for (j in 1:length(cardinalities)){
      pattern <- as.vector(t(data[(1 + sum(cardinalities[0:(j - 1)])):sum(cardinalities[1:j]),]))
      ds <- .C("ZDepth", 
               as.double(pattern), 
               as.double(objects), 
               as.integer(cardinalities[j]), 
               as.integer(cardinalities[i]), 
               as.integer(dim), 
               as.integer(seed),
               depths=double(cardinalities[i]))$depths
      if (j == i){
        ds <- replace(ds, which(ds < 1/cardinalities[j] - sqrt(.Machine$double.eps)), 1/cardinalities[j])
      }else{
        ds <- replace(ds, which(ds < 1/cardinalities[j] - sqrt(.Machine$double.eps)), 0)
      }
      pattern.depths <- cbind(pattern.depths, ds, deparse.level = 0)
    }
    depth.space <- rbind(depth.space, pattern.depths, deparse.level = 0)
  }

  return (depth.space)
}

################################################################################
# File:             ddalpha.train.r
# Created by:       Pavlo Mozharovskyi, Oleksii Pokotylo
# First published:  28.02.2013
# Last revised:     20.02.2019
# 
# Contains the training function of the DDalpha-classifier.
# 
# For a description of the algorithm, see:
#   Lange, T., Mosler, K. and Mozharovskyi, P. (2012). Fast nonparametric 
#     classification based on data depth. Statistical Papers.
#   Mozharovskyi, P., Mosler, K. and Lange, T. (2013). Classifying real-world 
#     data with the DDalpha-procedure. Mimeo.
################################################################################

ddalpha.train <- function(formula, data, subset,
                          depth = "halfspace", 
                          separator = "alpha", 
                          outsider.methods = "LDA", 
                          outsider.settings = NULL, 
                          aggregation.method = "majority",
                          pretransform = NULL,
                          use.convex = FALSE,
                          seed = 0,
                          
                          ...
                          
                          #                           # knn
                          #                           knnrange = NULL, 
                          #                           # alpha
                          #                           num.chunks = 10, 
                          #                           max.degree = 3, 
                          #                          
#                           # halfspace depth
                          #                           num.directions = 1000,                          
                          #                           # Mahalanobis depth
                          #                           mah.estimate = "moment", 
                          #                           mah.parMcd = 0.75, 
                          #                           mah.priors = NULL
                          
){
  
 
  # Raw data processing
  ddalpha <- .ddalpha.create.structure(formula, data, subset, ...) 

  # Check for data consistency #1 (moved to .ddalpha.create.structure)
  #if (!(is.matrix(data) && is.numeric(data)
  #    || is.data.frame(data) && prod(sapply(data[,-ncol(data)], is.numeric)))){
  #  stop("Argument data has unacceptable format. Classifier can not be trained!!!")
  #}
  
  # Check for data consistency #2
  if (ddalpha$numPatterns < 2){
    stop("Number of patterns is < 2. Classifier can not be trained!!!")
  }
  
  # TODO ddalpha$numPatterns should be restricted from above as well
  if (ddalpha$dimension < 2){
    stop("Data dimension is < 2. Classifier can not be trained!!!")
  }
  
  if(separator == "Dknn"){
    dknn = dknn.train(ddalpha$raw, depth = depth, seed = seed, ...)
    dknn$call = ddalpha$call
    dknn$colnames = ddalpha$colnames
    dknn$classif.formula = ddalpha$classif.formula
    return(dknn)
  }
  
  for (i in 1:ddalpha$numPatterns){
    if (ddalpha$patterns[[i]]$cardinality < ddalpha$dimension + 1){
      stop("At least in one patern number of the points < (dimension + 1). Classifier can not be trained!!!")
    }
  }
  if (seed != 0){
    set.seed(seed)
  }
  ddalpha$seed = seed
  
  ## Validating the properties
  depthsThatNeedNoScaling = c("zonoid", "halfspace", "Mahalanobis", "projection", "spatial", "spatialLocal", "simplicial", "simplicialVolume", "ddplot") # note: "spatialLocal" thansforms the data inside, by itself
  supportedDepths = c(depthsThatNeedNoScaling, "potential")
  if (is.null(depth) || toupper(depth) %in% c("", "NULL", "NONE")){
      ddalpha$methodDepth <- NULL
      warning("Argument \"depth\" specified as NULL.")    
  } else
  if (!is.character(depth)
      || length(depth) != 1){
    stop("Argument \"depth\" not specified correctly.")
  } else
  if(!(depth %in% supportedDepths)){
    .check.depth.exists(depth)
    ddalpha$methodDepth <- depth
  }else{
    ddalpha$methodDepth <- depth
  }
  if (!is.character(separator)
      || length(separator) != 1){
    stop("Argument \"separator\" not specified correctly.")
  } else
  if(!(separator %in% c("alpha", "polynomial", "knnlm", "maxD"))){
    fname = paste0(".", separator, "_validate")
    f <- try(match.fun(fname), silent = T)
    if (!is.function(f))
      warning(paste0("No validator function: ", fname, ". Cannot set 'methodSeparatorBinary'."))
    fname = paste0(".", separator, "_learn")
    f <- (match.fun(fname))
    if (!is.function(f))
      stop(paste0("No function: ", fname))
    fname = paste0(".", separator, "_classify")
    f <- (match.fun(fname))
    if (!is.function(f))
      stop(paste0("No function: ", fname))
    ddalpha$methodSeparator <- separator
  }else{
    ddalpha$methodSeparator <- separator
  }
  if (!is.character(aggregation.method)
      || length(aggregation.method) != 1
      || !(aggregation.method %in% c("majority", "sequent", "none"))){
    ddalpha$methodAggregation <- "majority"
    warning("Argument \"aggregation.method\" not specified correctly. \"majority\" is used as a default value")
  }else{
    ddalpha$methodAggregation <- aggregation.method
    
    ddalpha$methodSeparatorBinary = (aggregation.method !="none")
  }
  
  ddalpha$needtransform = 0
  if (!is.null(pretransform))
    if (ddalpha$methodDepth %in% depthsThatNeedNoScaling){
      warning("The used depth method is affine-invariant and pretransform doesn't influence the result. The data won't be transformed.")
    } else if (pretransform == "1Mom" || pretransform == "1MCD"){
      ddalpha$needtransform = 1
      
      if (pretransform == "1Mom")
        mm <- mah.moment(data[,-ncol(data)])
      else   # "1MCD"
        mm <- mah.mcd(data[,-ncol(data)], .mah.parMcd.fromDots(...))              
      
      for (i in 1:ddalpha$numPatterns){
        ddalpha$patterns[[i]]$transformer <- MahMomentTransformer(mm$mu, mm$b)
        ddalpha$patterns[[i]]$points <- ddalpha$patterns[[i]]$transformer(ddalpha$patterns[[i]]$points)
      }
    } else if (pretransform == "NMom" || pretransform == "NMCD"){
      ddalpha$needtransform = 2
      
      for (i in 1:ddalpha$numPatterns){
        if (pretransform == "NMom")
          mm <- mah.moment(ddalpha$patterns[[i]]$points)
        else   # "NMCD"
          mm <- mah.mcd(ddalpha$patterns[[i]]$points, .mah.parMcd.fromDots(...))
        
        ddalpha$patterns[[i]]$transformer <- MahMomentTransformer(mm$mu, mm$b)
      }
    } else {
      warning("Argument pretransform has unacceptable format. The data won't be transformed.")
    }  
  
  # appends ddalpha with the values from the given list (lst)
  ddalpha.append <- function(lst){ if(is.list(lst)) for(k in names(lst))  ddalpha[[k]] <<- lst[[k]] }
  # calls the validation method for the selected separator || depth
  # NO errors or warnings if the function doesn't exist!!
  # ddalpha is READONLY inside the validators
  validate <- function(method_name){
    f <- try(match.fun(paste0(".", method_name, "_validate")), silent = T)
    if (is.function(f)){
      lst  <- f(ddalpha, ...)
      ddalpha.append(lst) 
    }
  }
  
  ## Separator parameters validation
  
  if (!is.null(ddalpha$methodDepth))  
  validate(ddalpha$methodSeparator)
  
  ## Depth parameters validation
  
  if (!is.logical(use.convex) 
      || length(use.convex) != 1 
      || !(use.convex %in% c(TRUE, FALSE))){
    ddalpha$useConvex <- FALSE
    warning("Argument \"use.convex\" not specified correctly. FALSE is used as a default value")
  }else{
    ddalpha$useConvex <- use.convex
  }
  
  if (!is.null(ddalpha$methodDepth))
  validate(ddalpha$methodDepth)
  
  ## The learning procedures
  
  if (!is.null(ddalpha$methodDepth)){
    # Calculate depths
    if(ddalpha$methodDepth == "ddplot"){
      for(i in 1:ddalpha$numPatterns)
        ddalpha$patterns[[i]]$depths <- ddalpha$patterns[[i]]$points
    } else{
      ddalpha <- .ddalpha.learn.depth(ddalpha)
      if(is.null(ddalpha))
        stop("The depth method did not return the 'ddalpha' object.")
    }
    
    # Learn classification machine
    if (ddalpha$methodSeparatorBinary){
      ddalpha <- .ddalpha.learn.binary(ddalpha)
    } else {
      ddalpha <- .ddalpha.learn.multiclass(ddalpha)
    }
    if(is.null(ddalpha))
      stop("The separator method did not return the 'ddalpha' object.")
    
    # if (ddalpha$methodSeparator == "alpha"){
    #   ddalpha <- .ddalpha.learn.alpha(ddalpha)
    # } else
    # if (ddalpha$methodSeparator == "polynomial"){
    #   ddalpha <- .ddalpha.learn.polynomial(ddalpha)
    # } else
    # if (ddalpha$methodSeparator == "knnlm"){
    #   ddalpha <- .ddalpha.learn.knnlm(ddalpha)
    # } else
    #   stop("Define custom classifier")
  } else {
    ddalpha$numClassifiers = 0
  }
  
  # Learn outsider treatments if needed
  if (is.null(ddalpha$methodDepth) || !(ddalpha$methodDepth %in% 
          c("Mahalanobis", "projection", "spatial", "simplicialVolume", "potential"))){#, "simplicial" (may obtain too small values)
    ddalpha <- .ddalpha.learn.outsiders(ddalpha = ddalpha, 
                                        methodsOutsider = outsider.methods, 
                                        settingsOutsider = outsider.settings)
  }
  class(ddalpha) <- "ddalpha"
  
  return (ddalpha)
}

################################################################################
# Validation functions
################################################################################

.mah.parMcd.fromDots <- function(mah.parMcd = 0.75, ...) {
  return(mah.parMcd)
}

.alpha_validate  <- function(ddalpha, num.chunks = 10, max.degree = 3, debug = F,...){
  
  if (ddalpha$methodAggregation == "majority"){
    maxChunks <- ddalpha$patterns[[ddalpha$numPatterns]]$cardinality + ddalpha$patterns[[ddalpha$numPatterns - 1]]$cardinality
  }else{
    maxChunks <- ddalpha$numPoints
  }
  
  if (is.character(num.chunks) && toupper(num.chunks)=="MAX")
    num.chunks <- maxChunks
  else if (!is.numeric(num.chunks) 
           || is.na(num.chunks) 
           || length(num.chunks) != 1 
           || !.is.wholenumber(num.chunks) 
           || !(num.chunks > 0 && num.chunks <= maxChunks)){
    if (!missing(num.chunks))
      warning("Argument \"num.chunks\" not specified correctly. ", maxChunks, " is used instead")
    num.chunks <- maxChunks
  }
  
  if(!is.numeric(max.degree) 
     || is.na(max.degree) 
     || length(max.degree) != 1 
     || !.is.wholenumber(max.degree) 
     || !(max.degree %in% 1:10)){
    max.degree <- 3
    warning("Argument \"max.degree\" not specified correctly. 3 is used as a default value")
  }
  
  return (list(numChunks = num.chunks, maxDegree = max.degree, debug = (debug == T), methodSeparatorBinary = T))  
}

.polynomial_validate  <- .alpha_validate  # the same  

.knnlm_validate  <- function(ddalpha, knnrange = 10*( (ddalpha$numPoints)^(1/ddalpha$numPatterns) ) + 1,...){
  isnull = missing(knnrange) || is.null(knnrange)
  
  if (is.character(knnrange) && toupper(knnrange)=="MAX")
    knnrange = ceiling(ddalpha$numPoints/2)
  else if(is.null(knnrange)
          || !is.numeric(knnrange) 
          || is.na(knnrange) 
          || length(knnrange) != 1 
          || !.is.wholenumber(knnrange) 
          || !(knnrange >=2 && knnrange <= ceiling(ddalpha$numPoints/2))){    
    knnrange <- 10*( (ddalpha$numPoints)^(1/ddalpha$numPatterns) ) + 1   # Default
    
    knnrange <- min(knnrange, ceiling(ddalpha$numPoints/2))
    knnrange <- max(knnrange, 2)
    
    if (!isnull) warning("Argument \"knnrange\" not specified correctly. ", knnrange, " is used instead")
  }
  return (list(knnrange = knnrange, methodSeparatorBinary = F))
}

.maxD_validate  <- function(ddalpha,...){
  return(list(methodSeparatorBinary = F))
}

.ddplot_validate <- function(ddalpha, ...){
  if(ddalpha$dimension!=ddalpha$numPatterns)
    stop("You must pass a DD-plot, with the number of columns equal to the number of classes as data.")
}

.halfspace_validate  <- function(ddalpha, exact, method, num.directions = 1000,...){
  method = .parse_HSD_pars(exact, method)
  if(method == 0)
    if(!is.numeric(num.directions) 
     || is.na(num.directions) 
     || length(num.directions) != 1 
     || !.is.wholenumber(num.directions) 
     || !(num.directions > 1 && num.directions < 10000000) ){
    num.directions <- 1000
    warning("Argument \"num.directions\" not specified correctly. 1000 is used as a default value")
  }
  return (list(dmethod = method, numDirections = num.directions))
}

.projection_validate <- function(ddalpha, method = "random", num.directions = 1000,...){
  if (!(method %in% c("random","linearize")))
    stop("Wrong method")
  if(method == "random")
  if(!is.numeric(num.directions) 
     || is.na(num.directions) 
     || length(num.directions) != 1 
     || !.is.wholenumber(num.directions) 
     || !(num.directions > 1 && num.directions < 10000000) ){
    num.directions <- 1000
    warning("Argument \"num.directions\" not specified correctly. 1000 is used as a default value")
  }
  return (list(dmethod = method, numDirections = num.directions))
}

.simplicial_validate <- function(ddalpha, exact = F, k = 0.05, ...){
  if (exact)
    return(list(d_exact = exact))
  
  if (k <= 0) stop("k must be positive")
  else if (k < 1) k = choose(ddalpha$numPoints, ddalpha$dimension)*k
  
  return(list(d_exact = exact, d_k = k))
}

.simplicialVolume_validate <- function(ddalpha, exact = F, k = 0.05, mah.estimate = "moment", mah.parMcd = 0.75, ...){
  if (toupper(mah.estimate) == "NONE"){
    useCov <- 0
  } else if (toupper(mah.estimate) == "MOMENT"){
    useCov <- 1
  } else if (toupper(mah.estimate) == "MCD"){
    useCov <- 2
  } else {stop("Wrong argument \"mah.estimate\", should be one of \"moment\", \"MCD\", \"none\"")}
  
  if (exact){
    return(list(d_exact = exact, d_k = k, d_useCov = useCov, 
                d_parMcd = mah.parMcd))
  }
  
  if (k <= 0) stop("k must be positive")
  else if (k < 1) k = choose(ddalpha$numPoints, ddalpha$dimension)*k
  
  return(list(d_exact = exact, d_k = k, d_useCov = useCov, 
              d_parMcd = mah.parMcd))
}

.Mahalanobis_validate  <- function(ddalpha, mah.estimate = "moment", mah.priors = NULL, mah.parMcd = 0.75, ...){ 
  if (!is.character(mah.estimate) 
      || length(mah.estimate) != 1 
      || !(mah.estimate %in% c("moment", "MCD"))){
    mah.estimate <- "moment"
    warning("Argument \"mah.estimate\" not specified correctly. \"moment\" is used as a default value")
  }
  
  if (!is.vector(mah.priors, mode = "double") 
      || is.na(min(mah.priors)) 
      || length(mah.priors) != ddalpha$numPatterns 
      || min(mah.priors) <= 0 
      || max(mah.priors) <= 0){
    if (!is.null(mah.priors)){
      warning("Argument \"mah.priors\" not specified correctly. Defaults in the form of class portions are applied")
    }
    mah.priors <- NULL
  }else{
    mah.priors <- mah.priors/sum(mah.priors)
    
  }
  
  ret <- list(mahEstimate = mah.estimate, mahPriors = mah.priors)
  
  if (mah.estimate == "MCD"){
    if (is.null(mah.parMcd) 
        || !is.vector(mah.parMcd, mode = "double") 
        || is.na(min(mah.parMcd)) 
        || length(mah.parMcd) != 1 
        || mah.parMcd < 0.5 
        || mah.parMcd > 1){
      mah.parMcd <- 0.75
      warning("Argument \"mah.parMcd\" not specified correctly. 0.75 is used as a default value")
    }
    ret$mahParMcd = mah.parMcd
  }
  return (ret)
}

.spatial_validate <- function(ddalpha, mah.estimate = "moment", mah.parMcd = 0.75, ...){ 
  if(mah.estimate == "none")
    return(list(mahEstimate = "none"))
  
  return(.Mahalanobis_validate(ddalpha, mah.estimate = mah.estimate, mah.parMcd = mah.parMcd, ...))
}

.spatialLocal_validate  <- function(ddalpha, kernel = "GKernel", kernel.bandwidth = 1, ...)
{  
  # validate paraameters mah.estimate, mah.parMcd
  spatial_val = .spatial_validate(ddalpha, ...)
  
  if (is.null(kernel)
      || suppressWarnings (
        !((kernel  %in% .potentialKernelTypes) || !is.na(as.numeric(kernel))&&(as.numeric(kernel) %in% c(1:length(.potentialKernelTypes))))
      )){
    stop("Argument \"Kernel\" has invaid format.")
  }  
  if (is.null(kernel.bandwidth) 
      || !is.numeric(kernel.bandwidth)){
    stop("Argument \"kernel.bandwidth\" has invaid format.")
  }
  if (length(kernel.bandwidth) == 1){
    if (length(kernel.bandwidth) || is.na(kernel.bandwidth) || kernel.bandwidth == 0){
      stop("Argument \"kernel.bandwidth\" is Zero or NA.")
    } 
    kernel.bandwidth = rep(kernel.bandwidth, ddalpha$numPatterns)
  }
  else {
    if (sum(!is.na(kernel.bandwidth)) != ddalpha$numPatterns || sum(kernel.bandwidth != 0) != ddalpha$numPatterns){
      stop("Argument \"kernel.bandwidth\" has invaid length, Zero or NA elements.")
    }
     
    
    # order bandwidths the same as the classes
    names = sapply(ddalpha$patterns, FUN=function(X) X$name)
    kernel.bandwidth = kernel.bandwidth[order(names)]
  }
  
  spatial_val$kernel = kernel
  spatial_val$kernel.bandwidth = kernel.bandwidth
  
  return(spatial_val)
}

.potential_validate  <- function(ddalpha, kernel = "GKernel", kernel.bandwidth = NULL, ignoreself = FALSE, ...)
{  
  # if kernel.bandwidth is a vector - the values are given in the alphabetical order of the classes nemes
  if (ddalpha$needtransform == 0)
    stop("'pretransform' must be set for depth = 'potential'")
  
  if (is.null(kernel)
      || suppressWarnings (
        !((kernel  %in% .potentialKernelTypes) || !is.na(as.numeric(kernel))&&(as.numeric(kernel) %in% c(1:length(.potentialKernelTypes))))
        )){
    stop("Argument \"Kernel\" has invaid format.")
  }  
  if (is.null(kernel.bandwidth)) {    # use the rule of thumb
    if (ddalpha$needtransform == 2){
      kernel.bandwidth = sapply(ddalpha$patterns, FUN=function(X) nrow(X$points)) ^ (-2/(ddalpha$dimension+4))
    } else {
      kernel.bandwidth = ddalpha$numPoints ^ (-2/(ddalpha$dimension+4))
    }
  }
  else{
  if (#is.null(kernel.bandwidth) ||
        !is.numeric(kernel.bandwidth)
      ||!(is.vector(kernel.bandwidth) || is.list(kernel.bandwidth))){
    stop("Argument \"kernel.bandwidth\" has invaid format.")
  }
  
  if (ddalpha$needtransform == 2){
    if (length(kernel.bandwidth) == 1)
      kernel.bandwidth = rep(kernel.bandwidth, ddalpha$numPatterns)
    if (sum(!is.na(kernel.bandwidth)) != ddalpha$numPatterns || sum(kernel.bandwidth != 0) != ddalpha$numPatterns){
      stop("Argument \"kernel.bandwidth\" has invaid length, Zero or NA elements.")
    } 
    
    # order bandwidths the same as the classes
    names = sapply(ddalpha$patterns, FUN=function(X) X$name)
    kernel.bandwidth = kernel.bandwidth[order(names)]
  } else if (length(kernel.bandwidth) != 1 || is.na(kernel.bandwidth) || kernel.bandwidth == 0){
    stop("Argument \"kernel.bandwidth\" has invaid length, Zero or NA elements.")
  }  
  }
  
  if (is.null(ignoreself) || !is.logical(ignoreself))
    warning ("Argument \"ignoreself\" has invaid format. FALSE used.")
  
  return(list("kernel" = kernel, "kernel.bandwidth" = kernel.bandwidth, "ignoreself" = ignoreself))
}

lda.train <- function(data){
  new.frm <- data.frame(data)
  z <- lda(formula=as.formula(paste("X", ncol(data), " ~ .", sep="")), data = new.frm, tol = sqrt(.Machine$double.eps))#, prior = rep(1, dimension)/dimension)
  return (z)
}

lda.classify <- function(objects, lda){
  z <- predict(lda, data.frame(objects))$class
  return (z)
}

knnaff.train <- function(data, aggregation.method = "majority", range = -1, k = -1, i = 0){

  knnaff <- knaff.create.structure(data)
  # Checks
  if (!is.character(aggregation.method) 
      || length(aggregation.method) != 1 
      || !(aggregation.method %in% c("majority", "sequent"))){
    warning("In treatment number ", i, ": Argument \"aggregation.method\" not 
            specified correctly. \"majority\" is used as a default value")
    knnaff$methodAggregation <- "majority"
  }else{
    knnaff$methodAggregation <- aggregation.method
  }
  if (!is.numeric(range) 
      || is.na(range) 
      || length(range) != 1 
      || !.is.wholenumber(range) 
      || !(range >= 2 
           && range <= (knnaff$patterns[[knnaff$numPatterns]]$cardinality + 
                  knnaff$patterns[[knnaff$numPatterns - 1]]$cardinality - 1) 
           || range == -1)){
    warning("In treatment number ", i, ": Argument \"range\" not 
            specified correctly. Defaults are applied")
    knnaff$range <- -1
  }else{
    knnaff$range <- range
  }
  if (!is.numeric(k) 
      || is.na(k) 
      || length(k) != 1 
      || !.is.wholenumber(k) 
      || !(k >= 1 
           && k <= (knnaff$patterns[[knnaff$numPatterns]]$cardinality + 
                      knnaff$patterns[[knnaff$numPatterns - 1]]$cardinality) 
           || k == -1)){
    warning("In treatment number ", i, ": Argument \"k\" not specified 
            correctly. Defaults are applied")
    knnaff$k <- -1
  }else{
    knnaff$k <- k
  }
  # Do leave-one-out cross-validation
  knnaff <- knnaff.docv(knnaff)
  
  return (knnaff)
}

knaff.create.structure <- function(data){

  # Elemantary statistics
  dimension <- ncol(data) - 1
  numOfPoints <- nrow(data)
  classNames <- unique(data[,dimension + 1])
  numOfClasses <- length(classNames)
  # Ordering patterns according to their cardinalities
  classCardinalities <- rep(0, numOfClasses)
  for (i in 1:numOfClasses){
    classCardinalities[i] <- nrow(data[data[,dimension + 1] == classNames[i],])
  }
  # Creating pattern templates
  patterns <- as.list("")
  for (i in 1:numOfClasses){
    maxCarIndex <- which.max(classCardinalities)
    # Creating a single template
    pattern.index       <- i
    pattern.points      <- data[data[,dimension + 1] == classNames[maxCarIndex],
                                1:dimension]
    pattern.name        <- classNames[maxCarIndex]
    pattern.cardinality <- classCardinalities[maxCarIndex]
    pattern.votes       <- 0
    pattern <- structure(
      list(index = pattern.index, 
           points = pattern.points, 
           name = pattern.name, 
           cardinality = pattern.cardinality, 
           votes = pattern.votes), 
      .Names = c("index", "points", "name", "cardinality", "votes"))
    # Adding pattern template to the list of patterns
    patterns[[i]] <- pattern
    # Deleting processed pattern
    classCardinalities[maxCarIndex] <- -1    
  }
  # Creating overall structure
  knnaff <- structure(
    list(raw <- data, 
         dimension = dimension, 
         numPatterns = numOfClasses, 
         numPoints = numOfPoints, 
         patterns = patterns, 
         classifiers = list(), 
         numClassifiers = 0, 
         methodAggregation = "majority", 
         range = -1, 
         k = -1), 
    .Names = c("raw", "dimension", "numPatterns", "numPoints", "patterns", 
               "classifiers", "numClassifiers", "methodAggregation", "range", 
               "k"))
  
  return (knnaff)
}

knnaff.docv <- function(knnaff){

  counter <- 1
  # Determining multi-class behaviour
  if (knnaff$methodAggregation == "majority"){
    for (i in 1:(knnaff$numPatterns - 1)){
      for (j in (i + 1):knnaff$numPatterns){
        # Creating a classifier
        classifier.index          <- counter
        classifier.index1         <- i
        classifier.index2         <- j
        classifier.points         <- as.double(t(rbind(knnaff$patterns[[i]]$points, knnaff$patterns[[j]]$points)))
        classifier.cardinalities  <- as.integer(c(knnaff$patterns[[i]]$cardinality, knnaff$patterns[[j]]$cardinality))
        if (knnaff$k < 1 || knnaff$k > (knnaff$patterns[[i]]$cardinality + knnaff$patterns[[j]]$cardinality - 1))
        {
          if (knnaff$range < 2 || knnaff$range > (knnaff$patterns[[i]]$cardinality + knnaff$patterns[[j]]$cardinality - 1)){
            maxk <- 10*( (knnaff$numPoints)^(1/knnaff$dimension) ) + 1
          }else{
            maxk <- knnaff$range
          }
          maxk <- min(maxk, knnaff$patterns[[i]]$cardinality + knnaff$patterns[[j]]$cardinality - 1)
          maxk <- max(maxk, 2)
          classifier.range <- maxk
          classifier.k <- as.integer(.C("KnnAffInvLearnJK", 
                                        classifier.points, 
                                        as.integer(knnaff$dimension), 
                                        classifier.cardinalities, 
                                        as.integer(maxk), 
                                        k=integer(1))$k)
        }else{
          classifier.range <- knnaff$range
          classifier.k <- as.integer(knnaff$k)
        }
        # Adding the classifier to the list of classifiers
        knnaff$classifiers[[counter]] <- 
          list(index = classifier.index, 
               index1 = classifier.index1, 
               index2 = classifier.index2, 
               points = classifier.points, 
               cardinalities = classifier.cardinalities, 
               k = classifier.k, 
               range = classifier.range)
        counter <- counter + 1
      }
    }
  }
  if (knnaff$methodAggregation == "sequent"){
    for (i in 1:knnaff$numPatterns){
      anotherClass <- NULL
      for (j in 1:knnaff$numPatterns){
        if (j != i){
          anotherClass <- rbind(anotherClass, knnaff$patterns[[j]]$points)
        }
      }
      classifier.index          <- counter
      classifier.index1         <- i
      classifier.index2         <- -1
      classifier.points         <- as.double(t(rbind(knnaff$patterns[[i]]$points, anotherClass)))
      classifier.cardinalities  <- as.integer(c(knnaff$patterns[[i]]$cardinality, nrow(anotherClass)))
      if (knnaff$k < 1 || knnaff$k > knnaff$numPoints)
      {
        if (knnaff$range < 2 || knnaff$range > (knnaff$numPoints - 1)){
          maxk <- 10*( (knnaff$numPoints)^(1/knnaff$dimension) ) + 1
        }else{
          maxk <- knnaff$range
        }
        maxk <- min(maxk, knnaff$numPoints - 1)
        maxk <- max(maxk, 2)
        classifier.range <- maxk
        classifier.k <- as.integer(.C("KnnAffInvLearnJK", 
                                      classifier.points, 
                                      as.integer(knnaff$dimension), 
                                      classifier.cardinalities, 
                                      as.integer(maxk), 
                                      k=integer(1))$k)
      }else{
        classifier.range <- knnaff$range
        classifier.k <- as.integer(knnaff$k)
      }
      # Adding the classifier to the list of classifiers
      knnaff$classifiers[[counter]] <- 
        list(index = classifier.index, 
             index1 = classifier.index1, 
             index2 = classifier.index2, 
             points = classifier.points, 
             cardinalities = classifier.cardinalities, 
             k = classifier.k, 
             range = classifier.range)
      counter <- counter + 1
    }
  }
  
  return (knnaff)
}

knnaff.classify <- function(objects, knnaff){

  # Correct input data
  if (!is.matrix(objects)){
    objects <- matrix(objects, nrow=1)
  }
  # Initialization of the vote array
  votes <- matrix(rep(0, nrow(objects)*knnaff$numPatterns), nrow=nrow(objects), ncol=knnaff$numPatterns)
  for (i in 1:length(knnaff$classifiers)){
    res <- .C("KnnAffInvClassify", 
              as.double(t(objects)), 
              as.integer(nrow(objects)), 
              knnaff$classifiers[[i]]$points, 
              as.integer(knnaff$dimension), 
              knnaff$classifiers[[i]]$cardinalities, 
              knnaff$classifiers[[i]]$k, 
              output=integer(nrow(objects)))$output
    for (j in 1:nrow(objects)){
      if (res[j] == 0){
        votes[j,knnaff$classifiers[[i]]$index1] <- votes[j,knnaff$classifiers[[i]]$index1] + 1
      }else{
        votes[j,knnaff$classifiers[[i]]$index2] <- votes[j,knnaff$classifiers[[i]]$index2] + 1
      }
    }
  }
  # Collect results
  results <- list()
  for (i in 1:nrow(objects)){
    results[[i]] <- knnaff$patterns[[which.max(votes[i,])]]$name
  }
  
  return (results)
}

################################################################################
# File:             depth.L2.r
# Created by:       Pavlo Mozharovskyi
# First published:  08.03.2018
# Last revised:     08.03.2018
# 
# Computation of the L2-depth.
################################################################################

.L2_validate <- function(ddalpha, mah.estimate = "moment", mah.parMcd = 0.75, ...) {
  # only validate and stop if anything is wrong
    if( !(toupper(mah.estimate) %in% c("NONE", "MOMENT", "MCD"))){
      stop("Wrong argument \"mah.estimate\", should be one of \"moment\", \"MCD\", \"none\"")
    }

  return (list(mah.estimate = mah.estimate, mah.parMcd = mah.parMcd))
}

.L2_learn <- function(ddalpha) {
  #1. Calculate statistics based on data
  
  if(toupper(ddalpha$mah.estimate) == "NONE"){
    for (i in 1:ddalpha$numPatterns){
      ddalpha$patterns[[i]]$sigma <- diag(ddalpha$dimension)
    }
  } else {
    for (i in 1:ddalpha$numPatterns){
      if(toupper(ddalpha$mah.estimate) == "MOMENT"){
        cov <- cov(as.matrix(ddalpha$patterns[[i]]$points))
      } else if(toupper(ddalpha$mah.estimate) == "MCD"){
        cov <- covMcd(as.matrix(ddalpha$patterns[[i]]$points), ddalpha$mah.parMcd)$cov
      } else {stop("Wrong argument \"mah.estimate\", should be one of \"moment\", \"MCD\", \"none\"")}
      if(sum(is.na(cov)) == 0){
        ddalpha$patterns[[i]]$sigma <- solve(cov)
      } else{
        ddalpha$patterns[[i]]$sigma <- diag(ddalpha$dimension)
        warning("Covariance estimate not found for pattern ", ddalpha$patterns[[i]]$name, ", no affine-invariance-adjustment")
      }
    }
  }
  
  #2. Calculate depths for each pattern
  for (i in 1:ddalpha$numPatterns){
    ddalpha$patterns[[i]]$depths = .L2_depths(ddalpha, ddalpha$patterns[[i]]$points)
  }
  
  return (ddalpha)
}

.L2_depths <- function(ddalpha, objects){
  depths <- NULL
  objects = data.matrix(objects)
  for (j in 1:ddalpha$numPatterns){
    pattern <- as.matrix(ddalpha$patterns[[j]]$points)

    sigma = ddalpha$patterns[[j]]$sigma
    
    ds <- rep(-1, nrow(objects))
    for (i in 1:nrow(objects)){
      tmp1 <- t(objects[i,] - t(pattern))
      tmp2 <- tmp1 %*% sigma
      ds[i] <- 1/(1 + mean(sqrt(rowSums(tmp2 * tmp1))))
    }
    
    depths <- cbind(depths, ds)
  }
  return (depths)
}

depth.L2 <- function(x, data, mah.estimate = "moment", mah.parMcd = 0.75){
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if(is.data.frame(data))
    data = data.matrix(data)
  if (!is.matrix(x)){
    if(is.vector(x))
      x <- matrix(x, nrow=1)
    if(is.data.frame(x))
      x = data.matrix(x)
  }
  if(toupper(mah.estimate) == "NONE"){
    sigma = diag(ncol(data))
  } else {
    if(toupper(mah.estimate) == "MOMENT"){
      cov <- cov(data)
    } else if(toupper(mah.estimate) == "MCD"){
      cov <- covMcd(data, mah.parMcd)$cov
    } else {stop("Wrong argument \"mah.estimate\", should be one of \"moment\", \"MCD\", \"none\"")}
    if(sum(is.na(cov)) == 0){
      sigma <- solve(cov)
    } else{
      sigma = diag(ncol(data))
      warning("Covariance estimate not found, no affine-invariance-adjustment")
    }
  }

  depths <- rep(-1, nrow(x))
  for (i in 1:nrow(x)){
    tmp1 <- t(x[i,] - t(data))
    tmp2 <- tmp1 %*% sigma
    depths[i] <- 1/(1 + mean(sqrt(rowSums(tmp2 * tmp1))))
  }
  return (depths)
}

plot.functional <- function(x, main = "Functional data", xlab = "args", ylab = "vals", colors = c("red", "blue", "green", "black", "orange", "pink"), ...) {
  
  if(main == "Functional data" && !is.null(x$name))
    main = x$name
  if(xlab == "args" && !is.null(x$args))
    xlab = x$args
  if(ylab == "vals" && !is.null(x$vals))
    ylab = x$vals
  
  ylims = matrix(unlist(lapply(x$dataf, function(e) (range(e$vals)))), ncol = 2, byrow = TRUE)
  plot(0, type="n", xlim=range(x$dataf[[1]]$args), ylim=c(min(ylims[,1]), max(ylims[,2])), 
       xlab=xlab, ylab=ylab, 
       main = main, ...)
  grid()
  
  if (!is.null(x$labels))
    labs = sort(unlist(unique(unlist(x$labels)))) # second unlist removes factors, other
  else 
    labs = NULL
  
  for (i in 1:length(x$dataf)){
    if (!is.null(labs))
      ind = match(x$labels[[i]],labs)
    else
      ind = 1
    lineColor <- colors[ind]
    
    lines(x$dataf[[i]]$args, x$dataf[[i]]$vals, col=lineColor)
  }
}

lines.functional <- function(x, colors = c("red", "blue", "green", "black", "orange", "pink"), ...) {
  
  if (!is.null(x$labels))
    labs = sort(unlist(unique(x$labels)))
  else 
    labs = NULL
  
  for (i in 1:length(x$dataf)){
    if (!is.null(labs))
      ind = match(x$labels[[i]],labs)
    else
      ind = 1
    lineColor <- colors[ind]
    
    lines(x$dataf[[i]]$args, x$dataf[[i]]$vals, col=lineColor, ...)
  }
}

points.functional <- function(x, colors = c("red", "blue", "green", "black", "orange", "pink"), ...) {
  
  if (!is.null(x$labels))
    labs = sort(unlist(unique(x$labels)))
  else 
    labs = NULL
  
  for (i in 1:length(x$dataf)){
    if (!is.null(labs))
      ind = match(x$labels[[i]],labs)
    else
      ind = 1
    pointColor <- colors[ind]
    
    points(x$dataf[[i]]$args, x$dataf[[i]]$vals, col=pointColor, ...)
  }
}

################################################################################
# File:             depth.qhpeeling.r
# Created by:       Pavlo Mozharovskyi
# First published:  08.03.2018
# Last revised:     08.03.2018
# 
# Computation of the convex hull peeling depth.
################################################################################

.qhpeeling_validate <- function(ddalpha, ...) {
  return (list())
}

.qhpeeling_learn <- function(ddalpha) {
  # Calculate depths for each pattern
  for (i in 1:ddalpha$numPatterns){
    ddalpha$patterns[[i]]$depths = 
      .qhpeeling_depths(ddalpha, ddalpha$patterns[[i]]$points)
  }
  # Return the updated structure
  return (ddalpha)
}

.qhpeeling_depths <- function(ddalpha, objects){
  depths <- NULL
  objects = data.matrix(objects)
  for (j in 1:ddalpha$numPatterns){
    pattern <- as.matrix(ddalpha$patterns[[j]]$points)
    
    ds <- rep(0, nrow(objects))
    tmpData <- pattern
    for (i in 1:nrow(pattern)){
      if (length(tmpData) < ncol(pattern) * (ncol(pattern) + 1) + 0.5){
        break
      }
      ds <- ds + as.vector(is.in.convex(objects, tmpData, nrow(tmpData)))
      tmpData <- tmpData[-unique(as.vector(convhulln(tmpData))),]
    }
    
    ds = ds / nrow(pattern)
    
    depths <- cbind(depths, ds)
  }
  return (depths)
}

depth.qhpeeling <- function(x, data){
  if (!is.matrix(x) 
      && is.vector(x)){
    x <- matrix(x, nrow=1)
  }
  if (!(is.matrix(data) && is.numeric(data)
        || is.data.frame(data) && prod(sapply(data, is.numeric))) 
      || ncol(data) < 2){
    stop("Argument \"data\" should be a numeric matrix of at least 2-dimensional data")
  }
  if (!is.numeric(x)){
    stop("Argument \"x\" should be numeric")
  }
  
  if (ncol(x) != ncol(data)){
    stop("Dimensions of the arguments \"x\" and \"data\" should coincide")
  }
  if (ncol(data) + 1 > nrow(data)){ #?
    stop("Too few data points")
  }
  if (is.data.frame(x)){
    x <- as.matrix(x)
  }
  if (is.data.frame(data)){
    data <- as.matrix(data)
  }

  depths <- rep(0, nrow(x))
  tmpData <- data
  for (i in 1:nrow(data)){
    if (length(tmpData) < ncol(data) * (ncol(data) + 1) + 0.5){
      break
    }
    depths <- depths + as.vector(is.in.convex(x, tmpData, nrow(tmpData)))
    tmpData <- tmpData[-unique(as.vector(convhulln(tmpData))),]
  }
  
  return (depths / nrow(data))
}

compclassf.train <- function(dataf, labels, subset,
                            to.equalize = TRUE, 
                            to.reduce = TRUE, 
                            classifier.type = c("ddalpha", "maxdepth", "knnaff", "lda", "qda"), 
                            ...){
  # Trains the functional componentwise classifier
  # Args:
  #   dataf:  list containing lists (functions) of two vectors of equal length, 
  #           named "args" and "vals": arguments sorted in ascending order and 
  #           corresponding them values respectively
  #   labels: output labels of the functinal observations
  #   other arguments: TODO
  # Returns:
  #   Functional componentwise clasifier
  
  # Check "dataf"
  if (!is.list(dataf))
    stop("Argument 'dataf' must be a list")
  for (df in dataf)
    if (!(is.list(df) && length(df) == 2 &&
          !is.null(df$args) && !is.null(df$vals) &&
          is.vector(df$args) && is.vector(df$vals) &&
          is.numeric(df$args) && is.numeric(df$vals) &&
          length(df$args) == length(df$vals) &&
          is.sorted(df$args)))
      stop("Argument 'dataf' must be a list containing lists (functions) of two vectors of equal length, named 'args' and 'vals': arguments sorted in ascending order and corresponding them values respectively")
  
  if(!missing(subset)) {
    dataf = dataf[subset]
    labels = labels[subset]
  }
  
  # Check "labels"
  if (!(length(dataf)==length(labels) && length(unique(labels)>=2)))
      stop("Argument 'labels' has wrong format")
  
  # Check classifier.type
  classifier.type = match.arg(classifier.type)
  
  # Bring to finite dimension
  
  # Pointize
  points <- GetPointsDHB12(dataf, labels, to.equalize, to.reduce)
  # CV
  arg.indices <- getBestSpaceDHB12(points$data, classifier.type, num.chunks=10, ...)
  data <- points$data[,c(arg.indices,ncol(points$data))]
  # Apply chosen classifier to train the data
  if (classifier.type == "ddalpha"){
    classifier <- ddalpha.train(data, separator = "alpha", ...)
  }
  if (classifier.type == "maxdepth"){
    classifier <- ddalpha.train(data, separator = "maxD", ...)
  }
  if (classifier.type == "knnaff"){
    classifier <- knnaff.train(data, i = 0, ...)
  }
  if (classifier.type == "lda"){
    classifier <- lda.train(data, ...)
  }
  if (classifier.type == "qda"){
    classifier <- qda.train(data, ...)
  }
  # Create the eventual output structure
  compclassf <- structure(
    list(dataf = points$dataf, 
      labels = points$labels, 
      adc.method = "equalCover", 
      adc.args = list(instance = "val", numFcn = ncol(points$data) - 1, numDer = 0), 
      adc.transmat = points$transmat, 
      the.args = arg.indices, 
      data = points$data, 
      classifier.type = classifier.type, 
      classifier = classifier), 
    .Names = c("dataf", "labels", "adc.method", "adc.args", "adc.transmat", 
               "the.args",  "data", "classifier.type", "classifier"))
  class(compclassf) <- "compclassf"
  
  return (compclassf)
}

compclassf.classify <- function(compclassf, objectsf, subset, ...){
  # Classifies functions
  # Args:
  #   objectsf: sample to classify, a list containing lists (functions) of 
  #             two vectors of equal length, named "args" and "vals": 
  #             arguments sorted in ascending order and corresponding them 
  #             values respectively
  #   compclassf: functional DDalpha-classifier
  # Returns:
  #   List of labels assigned to the functions from "objectsf"

  # Check "objectsf"
  if (!is.list(objectsf))
    stop("Argument 'objectsf' must be a list")
  if (!is.null(objectsf$args)){
    objectsf = list(objectsf) # there was a single element 
  }
  
  if(!missing(subset)) {
    objectsf = objectsf[subset]
  }
  
  for (df in objectsf)
    if (!(is.list(df) && length(df) == 2 &&
            !is.null(df$args) && !is.null(df$vals) &&
            is.vector(df$args) && is.vector(df$vals) &&
            is.numeric(df$args) && is.numeric(df$vals) &&
            length(df$args) == length(df$vals) &&
            is.sorted(df$args)))
      stop("Argument 'objectsf' must be a list containing lists (functions) of two vectors of equal length, named 'args' and 'vals': arguments sorted in ascending order and corresponding them values respectively")

  # Prepare to multivariate classification
  objectsf.equalized <- equalize(objectsf)
  if (compclassf$adc.method == "equalCover"){
    if (compclassf$adc.args$instance == "val"){
      input <- getValGrid(objectsf.equalized, 
                          compclassf$adc.args$numFcn, compclassf$adc.args$numDer)
    }
    if (compclassf$adc.args$instance == "avr"){
      input <- getAvrGrid(objectsf.equalized, 
                          compclassf$adc.args$numFcn, compclassf$adc.args$numDer)
    }
    if (!is.null(compclassf$adc.transmat)){
      input <- input%*%compclassf$adc.transmat
    }
  }
  input <- input[,compclassf$the.args]
  # Classify and assign class labels
  if (compclassf$classifier.type == "ddalpha" || compclassf$classifier.type == "maxdepth"){
    output <- ddalpha.classify(objects = input, ddalpha = compclassf$classifier, ...)
  }
  if (compclassf$classifier.type == "knnaff"){
    output <- knnaff.classify(objects = input, compclassf$classifier, ...)
  }
  if (compclassf$classifier.type == "lda"){
    output <- lda.classify(objects = input, compclassf$classifier, ...)
  }
  if (compclassf$classifier.type == "qda"){
    output <- qda.classify(objects = input, compclassf$classifier, ...)
  }
  classes <- list()
  for (i in 1:length(output)){
#    if (is.numeric(output[[i]])){
      classes[[i]] <- compclassf$labels[[ output[[i]] ]]
#    }else{
#      classes[[i]] <- output[[i]]
#    }
  }

  return (classes)
}

predict.compclassf <- function(object, objectsf, subset, ...){
  compclassf.classify(object, objectsf, subset, ...)
}

print.compclassf <- function(x, ...){
  cat("compclassf:\n")
  cat("\t num.functions = ", length(x$dataf), 
      ", num.patterns = ", length(unique(x$labels)), "\n", sep="")
  #  cat("\t adc.method", x$adc.method, "\"\n", sep="")
  cat("\t adc:", x$adc.args$instance, "; numFcn:", x$adc.args$numFcn, "; numDer:", x$adc.args$numDer, "\"\n", sep="")
  cat("\t adc.transmat", x$adc.transmat, "\"\n", sep="")
  cat("\t classifier.type", x$classifier.type, "\"\n", sep="")
  cat("\t classifier:\n") 
  print(x$classifier)
}

################################################################################
# Functions below are used for intermediate computations                       #
################################################################################

GetPointsDHB12 <- function(dataf, labels, to.equalize=T, to.reduce=F){
  # Numerize labels
  names <- unique(labels)
  output <- rep(0, length(labels))
  for (i in 1:length(labels)){
    for (j in 1:length(names)){
      if (labels[[i]] == names[[j]]){
        output[i] = j
        break
      }
    }
  }
  # Prepare data
  if (to.equalize){
    num.times = length(dataf[[1]]$args)
    dataf.equalized <- equalize(dataf)
    adc.args = list(instance = "val", 
                    numFcn = num.times, 
                    numDer = 0)
    input <- getValGrid(dataf.equalized, adc.args$numFcn, adc.args$numDer)
  }else{
    input <- NULL
    for (i in 1:length(dataf)){
      input <- rbind(input, dataf[[i]]$vals)
    }
  }
  transmat <- NULL
  if (to.reduce){# Reduce dimension if needed
    princomps <- NULL
    newDim <- ncol(input)
    for (i in 1:length(names)){
      classi <- input[output == i,1:ncol(input)]
      princompsi <- prcomp(x=classi, tol=sqrt(.Machine$double.eps))
      newDimi <- sum(princompsi$sdev > sqrt(.Machine$double.eps))
      if (newDimi < newDim){
        newDim <- newDimi
        princomps <- princompsi
      }
    }
    transmat <- NULL
    if (newDim < ncol(input)){
      transmat <- matrix(as.vector(princomps$rotation[,1:newDim]), ncol=newDim)
      input <- input%*%transmat
    }
  }
  # Combine data
  data <- cbind(input, output, deparse.level=0)
  return (list(data = data, dataf = dataf.equalized, labels = names, transmat = transmat))
}

getBestSpaceDHB12 <- function(data, 
                              classifier.type = "ddalpha", 
                              num.chunks = 10, 
                              ...){
  indices.num <- ncol(data) - 1
  indices.avlbl <- rep(TRUE, indices.num)
  indices.best <- c()
  error.last <- nrow(data) + 1
  r <- 0
  while (sum(indices.avlbl) > 0){
    # If this is the first iteration search through all possible pairs
    if (r == 0){
      # Generate all combinations with smallest distance 2
      combinations <- combn((1:indices.num)[indices.avlbl], 2)
      tmp.cmb <- rbind(combinations[-1,], rep(-1000000, ncol(combinations)))
      tmp.cmb <- (tmp.cmb - combinations)==T
      combinations <- combinations[,apply(tmp.cmb, 2, sum)==0]
      # Choose the best combination
      errors <- c()
      for (i in 1:ncol(combinations)){
        cat("r = ", r, ": ", i, "/", ncol(combinations), ".\n", sep="")
        errors <- c(errors, 0)
        # Actually CV
        num.points <- nrow(data)
        indices.off <- num.chunks*(0:(ceiling(num.points/num.chunks) - 1))
        for (j in 1:num.chunks){
          # Determine points to be taken off
          take.off <- (indices.off + j)[(indices.off + j) <= num.points]
          # Apply chosen classifier
          if (classifier.type == "ddalpha"){
            classifier <- ddalpha.train(data[-take.off,c(combinations[,i], indices.num + 1)], separator = "alpha", ...)
            results <- ddalpha.classify(objects = data[take.off,combinations[,i]], ddalpha = classifier)
          }
          if (classifier.type == "maxdepth"){
            classifier <- ddalpha.train(data[-take.off,c(combinations[,i], indices.num + 1)], separator = "maxD", ...)
            results <- ddalpha.classify(objects = data[take.off,combinations[,i]], ddalpha = classifier)
          }
          if (classifier.type == "knnaff"){
            classifier <- knnaff.train(data[-take.off,c(combinations[,i], indices.num + 1)], i = i, ...)
            results <- knnaff.classify(data[take.off,combinations[,i]], classifier)
          }
          if (classifier.type == "lda"){
            classifier <- lda.train(data[-take.off,c(combinations[,i], indices.num + 1)], ...)
            results <- lda.classify(data[take.off,combinations[,i]], classifier)
          }
          if (classifier.type == "qda"){
            classifier <- qda.train(data[-take.off,c(combinations[,i], indices.num + 1)], ...)
            results <- qda.classify(data[take.off,combinations[,i]], classifier)
          }
          # Collect errors
          errors[i] <- errors[i] + sum(unlist(results) != data[take.off,indices.num + 1])
        }
      }
      # Collect results
      error.last <- min(errors)
      indices.best <- combinations[,which.min(errors)]
      indices.avlbl <- rep(TRUE, indices.num)
      indices.to.dsbl <- unique(c(indices.best, indices.best - 1, indices.best + 1))
      indices.to.dsbl <- indices.to.dsbl[indices.to.dsbl >= 1 && indices.to.dsbl <= indices.num]
      indices.avlbl[indices.to.dsbl] <- FALSE
      r <- 2
      next
    }
    # First, sequential approach
    errors <- c()
    variants <- c()
    for (i in 1:indices.num){
      if (indices.avlbl[i]){
        errors <- c(errors, 0)
        variants <- c(variants, i)
        indices.cur <- c(indices.best, i)
        # Actually CV
        num.points <- nrow(data)
        indices.off <- num.chunks*(0:(ceiling(num.points/num.chunks) - 1))
        for (j in 1:num.chunks){
          # Determine points to be taken off
          take.off <- (indices.off + j)[(indices.off + j) <= num.points]
          # Apply chosen classifier
          if (classifier.type == "ddalpha"){
            classifier <- ddalpha.train(data[-take.off,c(indices.cur, indices.num + 1)], separator = "alpha", ...)
            results <- ddalpha.classify(objects = data[take.off,indices.cur], ddalpha = classifier)
          }
          if (classifier.type == "maxdepth"){
            classifier <- ddalpha.train(data[-take.off,c(indices.cur, indices.num + 1)], separator = "maxD", ...)
            results <- ddalpha.classify(objects = data[take.off,indices.cur], ddalpha = classifier)
          }
          if (classifier.type == "knnaff"){
            classifier <- knnaff.train(data[-take.off,c(indices.cur, indices.num + 1)], i = i, ...)
            results <- knnaff.classify(data[take.off,indices.cur], classifier)
          }
          if (classifier.type == "lda"){
            classifier <- lda.train(data[-take.off,c(combinations[,i], indices.num + 1)], ...)
            results <- lda.classify(data[take.off,combinations[,i]], classifier)
          }
          if (classifier.type == "qda"){
            classifier <- qda.train(data[-take.off,c(combinations[,i], indices.num + 1)], ...)
            results <- qda.classify(data[take.off,combinations[,i]], classifier)
          }
          # Collect errors
          errors[i] <- errors[i] + sum(unlist(results) != data[take.off,indices.num + 1])
        }
      }
    }
    error.best <- min(errors)
    best.i <- variants[which.min(errors)[1]]
    indices.new <- c(indices.best, best.i)
    # Refinements for r=2, 3 and 4
    if (r %in% 2:3){
      # Define the grid
      if (r == 2){step <- 10}
      if (r == 3){step <- 5}
      grid.one <- c(-(1:step*2), 0, 1:step*2)
      grid <- c()
      for (i in 1:length(indices.new)){
        grid <- c(grid, indices.new[i] + grid.one)
      }
      grid <- unique(grid)
      grid <- sort(grid[(grid >= 1) & (grid <= indices.num)])
      # Generate all combinations with smallest distance 2
      combinations <- combn(grid, r + 1)
      tmp.cmb <- rbind(combinations[-1,], rep(-1000000, ncol(combinations)))
      tmp.cmb <- (tmp.cmb - combinations)==T
      combinations <- combinations[,apply(tmp.cmb, 2, sum)==0]
      # Choose the best combination
      #indices.grid <- (1:indices.num)[indices.avlbl & ((1:induces.num) %in% grid)]
      # Go through the combinations
      errors <- c()
      #combinations <- combn(indices.grid, r + 1)
      for (i in 1:ncol(combinations)){
        cat("r = ", r, ": ", i, "/", ncol(combinations), ".\n", sep="")
        errors <- c(errors, 0)
        # Actually CV
        num.points <- nrow(data)
        indices.off <- num.chunks*(0:(ceiling(num.points/num.chunks) - 1))
        for (j in 1:num.chunks){
          # Determine points to be taken off
          take.off <- (indices.off + j)[(indices.off + j) <= num.points]
          # Apply chosen classifier
          if (classifier.type == "ddalpha"){
            classifier <- ddalpha.train(data[-take.off,c(combinations[,i], indices.num + 1)], separator = "alpha", ...)
            results <- ddalpha.classify(objects = data[take.off,combinations[,i]], ddalpha = classifier)
          }
          if (classifier.type == "maxdepth"){
            classifier <- ddalpha.train(data[-take.off,c(combinations[,i], indices.num + 1)], separator = "maxD", ...)
            results <- ddalpha.classify(objects = data[take.off,combinations[,i]], ddalpha = classifier)
          }
          if (classifier.type == "knnaff"){
            classifier <- knnaff.train(data[-take.off,c(combinations[,i], indices.num + 1)], i = i, ...)
            results <- knnaff.classify(data[take.off,combinations[,i]], classifier)
          }
          if (classifier.type == "lda"){
            classifier <- lda.train(data[-take.off,c(combinations[,i], indices.num + 1)], ...)
            results <- lda.classify(data[take.off,combinations[,i]], classifier)
          }
          if (classifier.type == "qda"){
            classifier <- qda.train(data[-take.off,c(combinations[,i], indices.num + 1)], ...)
            results <- qda.classify(data[take.off,combinations[,i]], classifier)
          }
          # Collect errors
          errors[i] <- errors[i] + sum(unlist(results) != data[take.off,indices.num + 1])
        }
      }
      error.best <- min(errors)
      indices.cur <- combinations[,which.min(errors)]
    }else{
      indices.cur <- indices.new
    }
    if (error.best < error.last){
      indices.best <- indices.cur
      error.last <- error.best
      indices.avlbl <- rep(TRUE, indices.num)
      indices.to.dsbl <- unique(c(indices.best, indices.best - 1, indices.best + 1))
      indices.to.dsbl <- indices.to.dsbl[indices.to.dsbl >= 1 && indices.to.dsbl <= indices.num]
      indices.avlbl[indices.to.dsbl] <- FALSE
      r <- r + 1
    }else{
      break
    }
  }
  return (indices.best)
}

#all_datasets = c("baby","banknoten","biomed","bloodtransfusion","breast_cancer_wisconsin","bupa","chemdiab_1vs2","chemdiab_1vs3","chemdiab_2vs3","cloud","crabB_MvsF","crabF_BvsO","crabM_BvsO","crabO_MvsF","crab_BvsO","crab_MvsF","cricket_CvsP","diabetes","ecoli_cpvsim","ecoli_cpvspp","ecoli_imvspp","gemsen_MvsF","glass","groessen_MvsF","haberman","heart","hemophilia","indian_liver_patient_1vs2","indian_liver_patient_FvsM","iris_setosavsversicolor","iris_setosavsvirginica","iris_versicolorvsvirginica","irish_ed_MvsF","kidney","pima","plasma_retinol_MvsF","segmentation","socmob_IvsNI","socmob_WvsB","tae","tennis_MvsF","tips_DvsN","tips_MvsF","uscrime_SvsN","vertebral_column","veteran_lung_cancer","vowel_MvsF","wine_1vs2","wine_1vs3","wine_2vs3")

getdata <- function (name){
  data(list = name, envir = environment())
  return(get(name))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/depth.fd.R
\name{depth.sample}
\alias{depth.sample}
\title{Fast Depth Computation for Univariate and Bivariate Random Samples}
\usage{
depth.sample(A, B)
}
\arguments{
\item{A}{Univariate or bivariate points whose depth is computed, represented by a matrix of 
size \code{m*2}. \code{m} stands for the number of points, \code{d} is 1 for univariate and 2 
for bivariate data.}

\item{B}{Random sample points with respect to which the depth of \code{A} is computed. 
\code{B} is represented by a matrix of size \code{n*2}, where \code{n} is the sample size.}
}
\value{
Vector of length \code{m} of depth halfspace depth values is returned.
}
\description{
Faster implementation of the halfspace and the simplicial depth. Computes the depth 
of a whole random sample of a univariate or a bivariate data in one run.
}
\details{
The function returns vectors of sample halfspace and simplicial depth values.
}
\examples{
n = 100
m = 150
A = matrix(rnorm(2*n),ncol=2)
B = matrix(rnorm(2*m),ncol=2)
depth.sample(A,B)
system.time(D1<-depth.halfspace(A,B))
system.time(D2<-depth.sample(A,B))
max(D1-D2$Half)

A = rnorm(100)
B = rnorm(150)
depth.sample(A,B)
# depth.halfspace(matrix(A,ncol=1),matrix(B,ncol=1))

}
\seealso{
\code{\link{depth.halfspace}}

\code{\link{depth.simplicial}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{depth}

\name{depth.zonoid}
\alias{depth.zonoid}
\title{
Calculate Zonoid Depth
}
\description{
Calculates the zonoid depth of points w.r.t. a multivariate data set.
}
\usage{
depth.zonoid(x, data, seed = 0)
}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
  \item{seed}{
the random seed. The default value \code{seed=0} makes no changes.
}
}
\details{
Calculates zonoid depth (Koshevoy and Mosler, 1997; Mosler, 2002) exactly based on the algorithm of Dyckerhoff, Koshevoy and Mosler (1996), implemented in C++ (and provided) by Rainer Dyckerhoff.
}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector.
}
\references{
Dyckerhoff, R., Koshevoy, G., and Mosler, K. (1996). Zonoid data depth: theory and computation. In: Prat A. (ed), \emph{COMPSTAT 1996. Proceedings in computational statistics}, Physica-Verlag (Heidelberg), 235--240.

Koshevoy, G. and Mosler, K. (1997). Zonoid trimming for multivariate distributions \emph{Annals of Statistics} \bold{25} 1998--2017.

Mosler, K. (2002). \emph{Multivariate dispersion, central regions and depth: the lift zonoid approach} Springer (New York).
}
\seealso{
\code{\link{depth.halfspace}} for calculation of the Tukey depth.

\code{\link{depth.Mahalanobis}} for calculation of Mahalanobis depth.

\code{\link{depth.projection}} for calculation of projection depth.

\code{\link{depth.simplicial}} for calculation of simplicial depth.

\code{\link{depth.simplicialVolume}} for calculation of simplicial volume depth.

\code{\link{depth.spatial}} for calculation of spatial depth.

\code{\link{depth.potential}} for calculation of data potential.

}
\examples{
# 5-dimensional normal distribution
data <- mvrnorm(1000, rep(0, 5), 
                matrix(c(1, 0, 0, 0, 0, 
                         0, 2, 0, 0, 0, 
                         0, 0, 3, 0, 0, 
                         0, 0, 0, 2, 0, 
                         0, 0, 0, 0, 1),
                nrow = 5))
x <- mvrnorm(10, rep(1, 5), 
             matrix(c(1, 0, 0, 0, 0, 
                      0, 1, 0, 0, 0, 
                      0, 0, 1, 0, 0, 
                      0, 0, 0, 1, 0, 
                      0, 0, 0, 0, 1),
             nrow = 5))
                
depths <- depth.zonoid(x, data)
cat("Depths: ", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{ddalphaf.train}
\alias{ddalphaf.train}
\title{
Functional DD-Classifier
}
\description{
Trains the functional DD-classifier
}
\usage{
ddalphaf.train(dataf, labels, subset, 
                adc.args = list(instance = "avr", 
                               numFcn = -1, 
                               numDer = -1), 
                classifier.type = c("ddalpha", "maxdepth", "knnaff", "lda", "qda"), 
                cv.complete = FALSE, 
                maxNumIntervals = min(25, ceiling(length(dataf[[1]]$args)/2)),
                seed = 0,
                ...)
}

\arguments{
  \item{dataf}{
list containing lists (functions) of two vectors of equal length, named "args" and "vals": arguments sorted in ascending order and corresponding them values respectively
}
  \item{labels}{
list of output labels of the functional observations
}
  \item{subset}{
an optional vector specifying a subset of observations to be used in training the classifier.
}
  \item{adc.args}{
Represents a function sample as a multidimensional (dimension=\code{"numFcn"+"numDer"}) 
one averaging (\code{instance = "avr"}) or evaluating (\code{instance = "val"}) for that each function and it derivative on \code{"numFcn"} 
(resp. \code{"numDer"}) equal nonoverlapping covering intervals

First two named \code{"args"} and \code{"vals"} are arguments sorted in 
ascending order and having same bounds for all functions and 
corresponding them values respectively

  \describe{
    \item{instance}{
type of discretizing the functions: \cr
"avr" - by averaging over intervals of the same length \cr
"val" - by taking values on equally-spaced grid
    }
    \item{numFcn}{
  number of function intervals
    }  
    \item{numDer}{
  number of first-derivative intervals 
    }
  }
  Set \code{numFcn} and \code{numDer} to -1 to apply cross-validation.
  
  Set  \code{adc.args} to a list of "adc.args" objects to cross-validate only over these values.
}
  \item{classifier.type}{
the classifier which is used on the transformed space. The default value is 'ddalpha'.
}
  \item{cv.complete}{
T: apply complete cross-validation\cr
F: restrict cross-validation by Vapnik-Chervonenkis bound
}
  \item{maxNumIntervals}{
maximal number of intervals for cross-validation ( max(numFcn + numDer) = maxNumIntervals )
}
  \item{seed}{
the random seed. The default value \code{seed=0} makes no changes.
}
  \item{\dots}{
additional parameters, passed to the classifier, selected with parameter \code{classifier.type}.
}
}
\details{
The functional DD-classifier is fast nonparametric procedure for classifying functional data. It consists of
a two-step transformation of the original data plus a classifier operating on a low-dimensional
hypercube. The functional data are first mapped into a finite-dimensional location-slope space
and then transformed by a multivariate depth function into the DD-plot, which is a subset of
the unit hypercube. This transformation yields a new notion of depth for functional data. Three
alternative depth functions are employed for this, as well as two rules for the final classification. 
The resulting classifier is cross-validated over a small range of parameters only,
which is restricted by a Vapnik-Cervonenkis bound. The entire methodology does not involve
smoothing techniques, is completely nonparametric and allows to achieve Bayes optimality under
standard distributional settings. It is robust and efficiently computable.
}
\value{
Trained functional DD-classifier
%%  If it is a LIST, use
%%  \item{comp1 }{Description of 'comp1'}
%%  \item{comp2 }{Description of 'comp2'}
%% ...
}
\references{
Mosler, K. and Mozharovskyi, P. (2017). Fast DD-classification of functional data. \emph{Statistical Papers} \bold{58} 1055--1089.

Mozharovskyi, P. (2015). \emph{Contributions to Depth-based Classification and Computation of the Tukey Depth}. Verlag Dr. Kovac (Hamburg).
}

\seealso{
 \code{\link{ddalphaf.classify}} for classification using functional DD\eqn{\alpha}-classifier,
 \code{\link{compclassf.train}} to train the functional componentwise classifier,
 \code{\link{dataf.*}} for functional data sets included in the package.
}
\examples{

\dontrun{

## load the Growth dataset
dataf = dataf.growth()

learn = c(head(dataf$dataf, 49), tail(dataf$dataf, 34))
labels= c(head(dataf$labels, 49), tail(dataf$labels, 34)) 
test  = tail(head(dataf$dataf, 59), 10)    # elements 50:59. 5 girls, 5 boys

#cross-validate over the whole variants up to dimension 3
c1 = ddalphaf.train (learn, labels, classifier.type = "ddalpha", maxNumIntervals = 3)

classified1 = ddalphaf.classify(c1, test)

print(unlist(classified1))
print(c1$adc.args) 

# cross-validate over these two variants
c2 = ddalphaf.train (learn, labels, classifier.type = "ddalpha", 
                     adc.args = list(
                       list(instance = "avr", 
                            numFcn = 1, 
                            numDer = 2),
                       list(instance = "avr", 
                            numFcn = 0, 
                            numDer = 2)))

classified2 = ddalphaf.classify(c2, test)

print(unlist(classified2))
print(c2$adc.args) 

}
}

\keyword{ functional }
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }
\keyword{ classif }

\name{depth.contours.ddalpha}
\alias{depth.contours.ddalpha}
\title{
Depth Contours
}
\description{
Builds the data depth contours for multiclass 2-dimensional data using the trained classifier.
Also accessible from \code{\link{plot.ddalpha}}.
}
\usage{
depth.contours.ddalpha(ddalpha,
              main = "", xlab="", ylab = "", 
              drawplot = T, frequency=100, levels = 10, drawsep = T, ...)
}

\arguments{
  \item{ddalpha}{
DD\eqn{\alpha}-classifier (obtained by \code{\link{ddalpha.train}}).
}
  \item{main}{
an overall title for the plot: see \code{\link{title}}
}
  \item{xlab, ylab}{
labels of the axes
}
  \item{drawplot}{
if set to false, the contours are built on the existing plot. 
}
  \item{frequency}{
number of points on each direction, x and y. Impacts the smoothness of the contours.
}
  \item{levels}{
numeric vector of levels at which to draw contour lines. 
If the vector contains only ONE element, the levels are generated automatically as \code{seq(0, max(depth), length.out = levels)}.
}
  \item{drawsep}{
draws the separation on the DD-plot (currently for 2 classes and not for knn)
}
  \item{\dots}{
additional parameters passed to the depth functions and to \code{\link{plot}}
}
}

\seealso{
 \code{\link{depth.}}, 
 \code{\link{depth.contours}}, 
 \code{\link{depth.graph}}.
}
\examples{

\dontrun{

par(mfrow = c(2,2))
data(hemophilia)

ddalpha = ddalpha.train(hemophilia, depth = "none")
depth.contours.ddalpha(ddalpha, main = "data")

for (depth in c("zonoid", "Mahalanobis", "projection", "spatial")){
  ddalpha = ddalpha.train(hemophilia, depth = depth)
  depth.contours.ddalpha(ddalpha, main = depth)
}

for (depth in c("halfspace", "simplicial", "simplicialVolume")){
  ddalpha = ddalpha.train(hemophilia, depth = depth, exact = T)
  depth.contours.ddalpha(ddalpha, main = depth)
}


}
}

\keyword{ visualization }

\name{L2metric}
\alias{L2metric}
\title{Fast Computation of the \eqn{L^2} Metric for Sets of Functional Data}
\usage{
L2metric(A, B)
}
\arguments{
\item{A}{Functions of the first set, represented by a matrix of their functional values of 
size \code{m*d}. \code{m} stands for the number of functions, \code{d}
is the number of the equi-distant points {1,...,d} in the domain of the data [1,d] at which the functional
values of the \code{m} functions are evaluated.}

\item{B}{Functions of the second set, represented by a matrix of their functional values of 
size \code{n*d}. \code{n} stands for the number of functions, \code{d}
is the number of the equi-distant points {1,...,d} in the domain of the data [1,d] at which the functional
values of the \code{n} functions are evaluated. The grid of observation points for the 
functions \code{A} and \code{B} must be the same.}
}
\value{
A symmetric matrix of the distances of the functions of size \code{m*n}.
}
\description{
Returns the matrix of \eqn{L^2} distances between two sets of functional data.
}
\details{
For two sets of functional data of sizes \code{m} and \code{n}
represented by matrices of their functional values on the common domain {1,...,d}, 
this function returns the symmetric matrix of size \code{m*n} whose entry in the
\code{i}-th row and \code{j}-th column is the approximated \eqn{L^2} distance of the 
\code{i}-th function from the first set, and the \code{j}-th function from the second set.
This function is utilized in the computation of the h-mode depth.
}
\examples{
datapop = dataf2rawfd(dataf.population()$dataf,range=c(1950,2015),d=66)
A = datapop[1:20,]
B = datapop[21:50,]
L2metric(A,B)

}
\seealso{
\code{\link{depthf.hM}}

\code{\link{dataf2rawfd}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{functional}
\keyword{metric}

\name{ddalphaf.getErrorRateCV}
\alias{ddalphaf.getErrorRateCV}
\title{
Test Functional DD-Classifier
}
\description{
Performs a cross-validation procedure over the given data. 
On each step every \code{numchunks} observation is removed from the data, the functional DD-classifier is trained on these data and tested on the removed observations.
}
\usage{
ddalphaf.getErrorRateCV (dataf, labels, numchunks = 10, disc.type = c("LS", "comp"),  ...)
}
\arguments{
  \item{dataf}{
list containing lists (functions) of two vectors of equal length, named "args" and "vals": arguments sorted in ascending order and corresponding them values respectively
}
  \item{labels}{
list of output labels of the functional observations
}
  \item{numchunks}{
number of subsets of testing data. Equals to the number of times the classifier is trained.
}
  \item{disc.type}{
type of the used discretization scheme. "LS" for \code{\link{ddalphaf.train}}, "comp" for  for \code{\link{compclassf.train}}
}
  \item{\dots}{
additional parameters passed to \code{\link{ddalphaf.train}}
}
}

\value{

  \item{errors}{
  the part of incorrectly classified data
  }
  \item{time}{
  the mean training time
  }
  \item{time_sd}{
  the standard deviation of training time
  }

}


\seealso{
\code{\link{ddalphaf.train}} to train the functional DD\eqn{\alpha}-classifier, 
\code{\link{ddalphaf.classify}} for classification using functional DD\eqn{\alpha}-classifier, 
\code{\link{ddalphaf.test}} to test the functional DD-classifier on particular learning and testing data,
\code{\link{ddalphaf.getErrorRatePart}} to perform a benchmark study of the functional DD-classifier on particular data.
}
\examples{
# load the fdata
df = dataf.growth()

stat <- ddalphaf.getErrorRateCV(dataf = df$dataf, labels = df$labels, 
                                numchunks = 5,
                                adc.args = list(instance = "avr", 
                                                numFcn = 2, 
                                                numDer = 2))
                                                
cat("Classification error rate: ", stat$errors, ".\n", sep = "")

    
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ benchmark }

\name{depth.space.halfspace}
\alias{depth.space.halfspace}
\title{
Calculate Depth Space using Halfspace Depth
}
\description{
Calculates the representation of the training classes in depth space using the halfspace depth.
}
\usage{
depth.space.halfspace(data, cardinalities, exact, method, num.directions = 1000, seed = 0)
}
\arguments{
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{exact}{
The type of the used method. The default is \code{exact=F}, which leads to approximate computation of the halfspace depth. For \code{exact=F}, \code{method="Sunif.1D"} is used by default. If \code{exact=T}, the halfspace depth is computed exactly, with \code{method="recursive"} by default.}
  \item{method}{
For \code{exact=F}, if \code{method="Sunif.1D"} (by default), the halfspace depth is computed approximately by being minimized over univariate projections (see details).

For \code{exact=T}, the halfspace depth is calculated as the minimum over all combinations of \eqn{k} points from \code{data} (see details). In this case parameter \code{method} specifies \eqn{k}, with possible values \eqn{1} for \code{method="recursive"} (by default), \eqn{d-2} for \code{method="plane"}, \eqn{d-1} for \code{method="line"}.

The name of the method may be given as well as just parameter \code{exact}, in which case the default method will be used.
}
  \item{num.directions}{
Number of random directions to be generated. As the same direction set is used for all observations, the algorithmic complexity of calculating the depth of each single point in \code{data} is logarithmic in the number of observations in \code{data}, given the number of directions, see Mozharovskyi et al. (2015), Section 2.3 for discussion.
}
  \item{seed}{
The random seed. The default value \code{seed=0} makes no changes.
}
}
\details{
The depth representation is calculated in the same way as in \code{\link{depth.halfspace}}, see References below for more information and details.
}
\value{
Matrix of objects, each object (row) is represented via its depths (columns) w.r.t. each of the classes of the training sample; order of the classes in columns corresponds to the one in the argument \code{cardinalities}.
}
\references{
Cuesta-Albertos, J.A. and Nieto-Reyes, A. (2008). The random Tukey depth. \emph{Computational Statistics and Data Analysis} \bold{52} 4979--4988.

Dyckerhoff, R. and Mozharovskyi, P. (2016). Exact computation of the halfspace depth. \emph{Computational Statistics and Data Analysis} \bold{98} 19--30.

Mozharovskyi, P., Mosler, K., and Lange, T. (2015). Classifying real-world data with the DD\eqn{\alpha}-procedure. \emph{Advances in Data Analysis and Classification} \bold{9} 287--314.

Rousseeuw, P.J. and Ruts, I. (1996). Algorithm AS 307: Bivariate location depth. \emph{Journal of the Royal Statistical Society. Series C (Applied Statistics)} \bold{45} 516--526.

Tukey, J.W. (1974). Mathematics and the picturing of data. In: \emph{Proceeding of the International Congress of Mathematicians}, Vancouver, 523--531.
}
\seealso{
\code{\link{ddalpha.train}} and \code{\link{ddalpha.classify}} for application, \code{\link{depth.halfspace}} for calculation of the Tukey depth.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 20 training objects
class1 <- mvrnorm(10, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(10, c(1,1), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
data <- rbind(class1, class2)
plot(data, col = c(rep(1,10), rep(2,10)))
# Get depth space using the random Tukey depth
dhA = depth.space.halfspace(data, c(10, 10))
(dhA)

# Get depth space using default exact method - "recursive"
dhE = depth.space.halfspace(data, c(10, 10), exact = TRUE)
(dhE)

data <- getdata("hemophilia")
cardinalities = c(sum(data$gr == "normal"), sum(data$gr == "carrier"))
depth.space.halfspace(data[,1:2], cardinalities)
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{Custom Methods}
\alias{Custom Methods}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Using Custom Depth Functions and Classifiers
}
\description{
To use a custom depth function or classifier one has to implement three functions: parameters validator, learning and calculating functions.

}
%\usage{
%ddalpha.train(depth = "MyDepth", separator = "MyClassifier", ...)
%}
%- maybe also 'usage' for other objects documented here.
%\arguments{
%  \item{x}{
%%     ~~Describe \code{x} here~~
%}
%}
\details{

\bold{To define a depth function:}
\describe{
  \item{.<NAME>_validate}{
  validates parameters passed to \code{\link{ddalpha.train}} and passes them to the \code{ddalpha} object.
  
  \tabular{ll}{IN:\cr
    \code{ddalpha} \tab {the ddalpha object, containing the data and settings (mandatory)}\cr
    \code{<custom parameters>} \tab  {parameters that are passed to the user-defined method}\cr
    \code{...} \tab {other parameters (mandatory)}
    %}  \tabular{ll}{
    \cr
    OUT:\cr 
    \code{list()} \tab {list of output parameters, after the validation is finished, these parameters are stored in the \code{ddalpha} object}
    }
  }

  \item{.<NAME>_learn}{
  trains the depth
  
  \tabular{ll}{IN:\cr
    \code{ddalpha} \tab {the ddalpha object, containing the data and settings}
    %}  \tabular{ll}{
    \cr
  MODIFIES:\cr
  \code{ddalpha} \tab store the calculated statistics in the \code{ddalpha} object \cr
  depths \tab calculate the depths of each pattern, e.g. \cr \tab \code{for (i in 1:ddalpha$numPatterns) ddalpha$patterns[[i]]$depths = .<NAME>_depths(ddalpha, ddalpha$patterns[[i]]$points)}
  %}  \tabular{ll}{
  \cr
  OUT:\cr
    \code{ddalpha} \tab {the updated \code{ddalpha} object}
    }
  
  }
  \item{.<NAME>_depths}{
  calculates the depths
  
    \tabular{ll}{IN:\cr
    \code{ddalpha} \tab {the ddalpha object, containing the data and settings} \cr
    \code{objects} \tab {the objects for which the depths are calculated} \cr
    \cr
    OUT:\cr
      \code{depths} \tab {the calculated depths for each object (rows), with respect to each class (cols)}
      }
  }
}
Usage: \code{ddalpha.train(data, depth = "<NAME>", <custom parameters>, ...)}

\verb{
#### Custom depths ####

.MyDepth_validate  <- function(ddalpha, mydepth.parameter = "value", ...){ 
  print("MyDepth validating")
  # validate the parameters
  if (!is.valid(mydepth.parameter)){
    warning("Argument \"mydepth.parameter\" not specified correctly. Default value is used")
    mydepth.parameter = "value"
    # or stop("Argument \"mydepth.parameter\" not specified correctly.")
  }
  
  # the values from the return list will be stored in the ddalpha object
  return (list(mydepthpar = mydepth.parameter))
}

.MyDepth_learn <- function(ddalpha){
  print("MyDepth learning")
  
  #1. Calculate any statistics based on data that .MyDepth_depths needs
  #   and store them to the ddalpha object: 
  ddalpha$mydepth.statistic = "some value"
  
  #2. Calculate depths for each pattern
  for (i in 1:ddalpha$numPatterns){
    ddalpha$patterns[[i]]$depths = .MyDepth_depths(ddalpha, ddalpha$patterns[[i]]$points)
  }
  
  return(ddalpha)
}

.MyDepth_depths <- function(ddalpha, objects){
  print("MyDepth calculating")
  depths <- NULL
  
  # The depth parameters are accessible in the ddalpha object:
  mydepth.parameter = ddalpha$mydepth.parameter
  mydepth.statistic = ddalpha$mydepth.statistic
  
  #calculate the depths of the objects w.r.t. each pattern
  for (i in 1:ddalpha$numPatterns){
    depth_wrt_i = #calculate depths of the objects, as vector
    
    depths <- cbind(depths, depth_wrt_i)
  }
    
  return (depths)
}

ddalpha.train(data, depth = "MyDepth", ...)
}

\bold{To define a classifier:}
\describe{
  \item{.<NAME>_validate}{
  validates parameters passed to \code{\link{ddalpha.train}} and passes them to the \code{ddalpha} object
    \tabular{ll}{IN:\cr
    \code{ddalpha} \tab {the ddalpha object, containing the data and settings (mandatory)}\cr
    \code{<custom parameters>} \tab  {parameters that are passed to the user-defined method}\cr
    \code{...} \tab {other parameters (mandatory)}
    \cr
    OUT:\cr 
    \code{list()} \tab {list of output parameters, after the validation is finished, these parameters are stored in the \code{ddalpha} object. 
    
    In case of a multiclass classifier the validator must return \code{methodSeparatorBinary = F} and/or pass \code{aggregation.method = "none"} to \code{ddalpha.train}}
    }
  }
  \item{.<NAME>_learn}{
  trains the classifier. Is different for binnary and mylticlass classifiers.
    \tabular{ll}{IN:\cr
    \code{ddalpha} \tab {the ddalpha object, containing the data and settings}\cr
    \code{index1} \tab {(only for binary) index of the first class}\cr
    \code{index2} \tab {(only for binary) index of the second class}\cr
    \code{depths1} \tab {(only for binary) depths of the first class w.r.t. all classes}\cr
    \code{depths2} \tab {(only for binary) depths of the second class w.r.t. all classes}\cr\cr
    \tab depths w.r.t. only given classes are received by \code{depths1[,c(index1, index2)]}\cr\cr
    \tab for the multiclass classifiers the depths are accessible via \code{ddalpha$patterns[[i]]$depths}
    \cr
    OUT:\cr 
    \code{classifier} \tab { the trained \code{classifier} object}
    }
  }
  \item{.<NAME>_classify}{
  classifies the objects
   \tabular{ll}{IN:\cr
    \code{ddalpha} \tab {the ddalpha object, containing the data and global settings} \cr
    \code{classifier} \tab {the previously trained classifier}\cr
    \code{objects} \tab {the objects (depths) that are classified} \cr
    \cr
    OUT:\cr
      \code{result} \tab { a vector with classification results}\cr
        \tab {(binary) the objects from class \code{"classifier$index1"} get positive values} \cr
        \tab {(multiclass) the objects get the numbers of patterns in \code{ddalpha}}
    }
      
  }
}
Usage: \code{ddalpha.train(data, separator = "<NAME>", ...)}

\verb{
#### Custom classifiers ####


.MyClassifier_validate  <- function(ddalpha, my.parameter = "value", ...){
  print("MyClassifier validating")
  # validate the parameters
  ...
  
  # always include methodSeparatorBinary. 
  # TRUE for the binary classifier, FALSE otherwise
  return(list(methodSeparatorBinary = T, 
              my.parameter = my.parameter
              ))
}

# a binary classifier 
# the package takes care of the voting procedures. Just train it as if there are only two classes
.MyClassifier_learn <- function(ddalpha, index1, index2, depths1, depths2){
  print("MyClassifier (binary) learning")
  
  # The parameters are accessible in the ddalpha object:
  my.parameter = ddalpha$my.parameter
  
  #depths w.r.t. only given classes are received by
  depths1[,c(index1, index2)]
  depths2[,c(index1, index2)]
  
  # train the classifier
  classifier <-  ...
  
  #return the needed values in a list, e.g.
  return(list(
    coefficients  = classifier$coefficients,
    ... ))
}

# a multiclass classifier 
.MyClassifier_learn <- function(ddalpha){
  print("MyClassifier (multiclass) learning")
  
  # access the data through the ddalpha object, e.g.
  for (i in 1:ddalpha$numPatterns){
    depth  <- ddalpha$patterns[[i]]$depths
    number <- ddalpha$patterns[[i]]$cardinality
    ...
  }
  
  # train the classifier
  classifier <-  ...
  
  # return the classifier
  return(classifier)
}

# the interface of the classify function is equal for binary and multiclass classifiers
.MyClassifier_classify <- function(ddalpha, classifier, depths){
  print("MyClassifier classifying")
  
  # The global parameters are accessible in the ddalpha object:
  my.parameter = ddalpha$my.parameter
  # The parameters generated by .MyClassifier_learn are accessible in the classifier object:
  classifier$coefficients
  
  # here are the depths w.r.t. the first class
  depths[,classifier$index1]
  # here are the depths w.r.t. the second class
  depths[,classifier$index2]
  
  # fill results in a vector, so that:
  # (binary) the objects from class "classifier$index1" get positive values
  # (multiclass) the objects get the numbers of patterns in ddalpha
  result <- ...
  
  return(result)
}

ddalpha.train(data, separator = "MyClassifier", ...)
}
}
%\value{
%%  ~Describe the value returned
%%  If it is a LIST, use
%%  \item{comp1 }{Description of 'comp1'}
%%  \item{comp2 }{Description of 'comp2'}
%% ...
%}
%\references{
%% ~put references to the literature/web site here ~
%}
%\author{
%%  ~~who you are~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{ddalpha.train}}
}
\examples{

\dontrun{

#### example:  Euclidean depth ####

#.Euclidean_validate is basically not needed

.Euclidean_learn <- function(ddalpha){
  print("Euclidean depth learning")
  
  #1. Calculate any statistics based on data that .MyDepth_depths needs
  #   and store them to the ddalpha object: 
  for (i in 1:ddalpha$numPatterns){
    ddalpha$patterns[[i]]$center <- colMeans(ddalpha$patterns[[i]]$points)
  }
  
  #2. Calculate depths for each pattern
  for (i in 1:ddalpha$numPatterns){
    ddalpha$patterns[[i]]$depths = .Euclidian_depths(ddalpha, ddalpha$patterns[[i]]$points)
  }
  
  return(ddalpha)
}

.Euclidean_depths <- function(ddalpha, objects){
  print("Euclidean depth calculating")
  depths <- NULL
  
  #calculate the depths of the objects w.r.t. each pattern
  for (i in 1:ddalpha$numPatterns){
    # The depth parameters are accessible in the ddalpha object:
    center = ddalpha$patterns[[i]]$center
    
    depth_wrt_i <- 1/(1 + colSums((t(objects) - center)^2))
    depths <- cbind(depths, depth_wrt_i)
  }
  
  return (depths)
}

#### example:  binary decision tree ####

library(rpart)

.tree_validate  <- function(ddalpha, ...){
  print("tree validating")
  return(list(methodSeparatorBinary = T))
}

# a binary classifier 
# the package takes care of the voting procedures. Just train it as if there are only two classes
.tree_learn <- function(ddalpha, index1, index2, depths1, depths2){
  print("tree learning")
  
  # prepare the data
  data = as.data.frame(cbind( (rbind(depths1, depths2)), 
                              c(rep(1, times = nrow(depths1)), rep(-1, times = nrow(depths1)))))
  names(data) <- paste0("V",seq_len(ncol(data)))
  names(data)[ncol(data)] <- "O"
  # train the classifier
  classifier <- rpart(O~., data = data)
  
  #return the needed values in a list, e.g.
  return(classifier)
}

# the interface of the classify function is equal for binary and multiclass classifiers
.tree_classify <- function(ddalpha, classifier, depths){
  print("tree classifying")
  
  # fill results in a vector, so that the objects from class "classifier$index1" get positive values
  data = as.data.frame(depths)
  names(data) <- paste0("V",seq_len(ncol(data)))
  result <- predict(classifier, as.data.frame(depths), type = "vector")
  
  return(result)
}




#### checking ####

library(ddalpha)
data = getdata("hemophilia")

ddalpha = ddalpha.train(data = data, depth = "Euclidean", separator = "tree")
c = ddalpha.classify(ddalpha, data[,1:2])
errors = sum(unlist(c) != data[,3])/nrow(data)
print(paste("Error rate: ",errors))

# building the depth contours using the custom depth
depth.contours.ddalpha(ddalpha, asp = T, levels = seq(0.5, 1, length.out = 10))

}

}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ depth }
\keyword{ classif }
\keyword{ custom }

\name{depth.betaSkeleton}
\alias{depth.betaSkeleton}
\title{
Calculate Beta-Skeleton Depth
}
\description{
Calculates the beta-skeleton depth of points w.r.t. a multivariate data set.
}
\usage{
depth.betaSkeleton(x, data, beta = 2, distance = "Lp", Lp.p = 2, 
                   mah.estimate = "moment", mah.parMcd = 0.75)
}
\arguments{
  \item{x}{Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
  \item{beta}{The paremeter defining the positionning of the balls' centers, see Yang and Modarres (2017) for details. By default (together with other arguments) equals \code{2}, which corresponds to the lens depth, see Liu and Modarres (2011).
}
  \item{distance}{A character string defining the distance to be used for determining inclusion of a point into the lens (influence region), see Yang and Modarres (2017) for details. Possibilities are \code{"Lp"} for the Lp-metric (default) or \code{"Mahalanobis"} for the Mahalanobis distance adjustment.
  }
  \item{Lp.p}{A non-negative number defining the distance's power equal \code{2} by default (Euclidean distance); is used only when \code{distance = "Lp"}.
  }
  \item{mah.estimate}{A character string specifying which estimates to use when calculating sample covariance matrix; can be \code{"none"}, \code{"moment"} or \code{"MCD"}, determining whether traditional moment or Minimum Covariance Determinant (MCD) (see \code{\link{covMcd}}) estimates for mean and covariance are used. By default \code{"moment"} is used. Is used only when \code{distance = "Mahalanobis"}.
}
  \item{mah.parMcd}{The value of the argument \code{alpha} for the function \code{\link{covMcd}}; is used when \code{distance = "Mahalanobis"} and \code{mah.estimate = "MCD"}.
}
}
\details{
Calculates the beta-skeleton depth, see Yang and Modarres (2017). Its particular case, lens depth, see Liu and Modarres (2011), is obtained when \code{beta = 2}, \code{distance = "Lp"} and \code{Lp.p = 2} (default settings). For tne example of the lens depth, the depth of an observation \code{x} is calculated as the portion of lens containing \code{x}, with lens being an intersection of two closed balls centered at two sample's points each having radius equal to the distance between these two points.
}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector.
}
\references{
Liu, Z. and Modarres, R. (2011). Lens data depth and median. \emph{Journal of Nonparametric Statistics} \bold{23}(4) 1063--1074.

Yang, M. and Modarres, R. (2017). \eqn{\beta}-skeleton depth functions and medians. \emph{Commmunications in Statistics - Theory and Methods} to appear.
}
\seealso{
\code{\link{depth.halfspace}} for calculation of the Tukey depth.

\code{\link{depth.Mahalanobis}} for calculation of Mahalanobis depth.

\code{\link{depth.projection}} for calculation of projection depth.

\code{\link{depth.simplicial}} for calculation of simplicial depth.

\code{\link{depth.simplicialVolume}} for calculation of simplicial volume depth.

\code{\link{depth.spatial}} for calculation of spatial depth.

\code{\link{depth.zonoid}} for calculation of zonoid depth.

\code{\link{depth.potential}} for calculation of data potential.

}
\examples{
# 5-dimensional normal distribution
data <- mvrnorm(1000, rep(0, 5), 
                matrix(c(1, 0, 0, 0, 0, 
                         0, 2, 0, 0, 0, 
                         0, 0, 3, 0, 0, 
                         0, 0, 0, 2, 0, 
                         0, 0, 0, 0, 1),
                nrow = 5))
x <- mvrnorm(10, rep(1, 5), 
             matrix(c(1, 0, 0, 0, 0, 
                      0, 1, 0, 0, 0, 
                      0, 0, 1, 0, 0, 
                      0, 0, 0, 1, 0, 
                      0, 0, 0, 0, 1),
             nrow = 5))
                
depths <- depth.betaSkeleton(x, data)
cat("Depths:", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{depth.contours}
\alias{depth.contours}
\title{
Depth Contours
}
\description{
Builds the data depth contours for 2-dimensional data.
}
\usage{
depth.contours(data, depth, 
              main = "", xlab="", ylab = "", 
              drawplot = T, frequency=100, levels = 10,
              col = "red",
              ...)
}

\arguments{
  \item{data}{
2-dimensional numeric data frame or matrix
}
  \item{depth}{
the name of the depth function. The list of the supported depths and described in the topic \code{\link{depth.}}.
}
  \item{main}{
an overall title for the plot: see \code{\link{title}}
}
  \item{xlab, ylab}{
labels of the axes
}
  \item{drawplot}{
if set to false, the contours are built on the existing plot. 
}
  \item{frequency}{
number of points on each direction, x and y. Impacts the smoothness of the contours.
}
  \item{levels}{
numeric vector of levels at which to draw contour lines. 
If the vector contains only ONE element, the levels are generated automatically as \code{seq(0, max(depth), length.out = levels)}.
}
  \item{col}{
color, used to draw points and contours
}
  \item{\dots}{
additional parameters passed to the depth functions and to \code{\link{plot}}
}
}

\seealso{
 \code{\link{depth.}}, 
 \code{\link{depth.contours.ddalpha}}, 
 \code{\link{depth.graph}}.
}
\examples{

\dontrun{

par(mfrow = c(2,2))
data(hemophilia)

depth.contours(hemophilia[,1:2], depth = "none", main = "data")

for (depth in c("zonoid", "Mahalanobis", "projection", "spatial")){
  depth.contours(hemophilia[,1:2], depth = depth, main = depth)
}

for (depth in c("halfspace", "simplicial", "simplicialVolume")){
  depth.contours(hemophilia[,1:2], depth = depth, main = depth, exact = T)
}


}
}

\keyword{ visualization }

\name{draw.ddplot}
\alias{draw.ddplot}
\title{
Draw \emph{DD}-Plot
}
\description{
The function draws the \emph{DD}-plot either of the existing DD\eqn{\alpha}-classifier of the depth space.
Also accessible from \code{\link{plot.ddalpha}}.
}
\usage{
draw.ddplot(ddalpha, depth.space, cardinalities, 
            main = "DD plot", xlab = "C1", ylab = "C2", xlim, ylim,
            classes = c(1, 2), colors = c("red", "blue", "green"), drawsep = T)

}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{ddalpha}{
DD\eqn{\alpha}-classifier (obtained by \code{\link{ddalpha.train}}).
}
  \item{depth.space}{
The ready depth space obtained by \code{\link{depth.space.}}}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{main}{
an overall title for the plot: see \code{\link{title}}
}
  \item{xlab, ylab}{
class labels
}
  \item{xlim, ylim}{
range of axis
}
  \item{classes}{
vector of numbers of two classes used for depth calculation
}
  \item{colors}{
vector of the classes' colors
}
  \item{drawsep}{
draws the separation on the DD-plot (currently for 2 classes and not for knn)
}
}

\seealso{
\code{\link{ddalpha.train}}

\code{\link{depth.space.}}

}
\examples{
  data = getdata("kidney")
  
  #1. using the existing ddalpha classifier
  ddalpha = ddalpha.train(data, depth = "spatial")
  draw.ddplot(ddalpha, main = "DD-plot")
  
  #2. using depth.space.
  # Sort the data w.r.t. classes
  data = rbind(data[data$C == 1,], data[data$C == 2,])
  cardinalities = c(sum(data$C == 1), sum(data$C == 2))
  
  dspace = depth.space.spatial(data[,-6], cardinalities = cardinalities)
  draw.ddplot(depth.space = dspace, cardinalities = cardinalities, 
              main = "DD-plot", xlab = 1, ylab = 2)
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ visualization }

\name{infimalRank}
\alias{infimalRank}
\title{Adjusted Ranking of Functional Data Based on the Infimal Depth}
\usage{
infimalRank(ID, IA, ties.method = "max")
}
\arguments{
\item{ID}{The vector of infimal depths of the curves of length \code{n}.}

\item{IA}{The vector of the infimal areas corresponding to the infimal depths from \code{ID}
of length \code{n}.}

\item{ties.method}{Parameter for breaking ties in infimal area index. By default \code{max}, see 
\code{rank}.}
}
\value{
A vector of length \code{n}. Low depth values mean high ranks, i.e. potential outlyingness. 
If some of the infimal depths are identical, the ranking of these functions is made according to the 
values of the infimal area. There, higher infimal area index means higher rank, i.e. non-centrality.
}
\description{
Returns a vector of adjusted depth-based ranks for infimal depth for functional data.
}
\details{
Infimal depths for functional data tend to give to many functional observations the same 
value of depth. Using this function, the data whose depth is the same is ranked according
to the infimal area indicator. This indicator is provided in functions \code{depthf.fd1} along
the value of the infimal depth.
}
\examples{
datafA = dataf.population()$dataf[1:20]
datafB = dataf.population()$dataf[21:50]
D = depthf.fd1(datafA,datafB)
infimalRank(D$Half_ID,D$Half_IA) 

ID = c(0,1,0,0,0,1,1)
IA = c(2,3,1,0,2,4,1)
infimalRank(ID,IA)
}
\references{
Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
Depth-based recognition of shape outlying functions. 
\emph{Journal of Computational and Graphical Statistics}, \bold{26} (4), 883--893.
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{depth}
\keyword{functional}
\keyword{rank}

\name{depth.space.potential}
\alias{depth.space.potential}
\title{
Calculate Potential Space
}
\description{
Calculates the representation of the training classes in potential space.
}
\usage{
depth.space.potential(data, cardinalities, pretransform = "NMom", 
            kernel = "GKernel", kernel.bandwidth = NULL, mah.parMcd = 0.75)
}
\arguments{
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{pretransform}{
The method of data scaling. 

\code{NULL} to use the original data, 

The data may be scaled jointly or separately:

\code{1Mom} or \code{1MCD} for joint scaling of the classes, 

\code{NMom} or \code{NMCD} for separate scaling of the classes.

You may use traditional moments or Minimum Covariance Determinant (MCD) estimates for mean and covariance:

\code{1Mom} or \code{NMom} for scaling using traditional data moments, 

\code{1MCD} or \code{NMCD} for scaling using robust MCD data moments.
}
  \item{kernel}{
\code{"EDKernel"} for the kernel of type 1/(1+kernel.bandwidth*EuclidianDistance2(x, y)), 

\code{"GKernel"} [default and recommended] for the simple Gaussian kernel, 

\code{"EKernel"} exponential kernel: exp(-kernel.bandwidth*EuclidianDistance(x, y)), 

%\code{"TriangleKernel"}, 
\code{"VarGKernel"} variable Gaussian kernel, where \code{kernel.bandwidth} is proportional to the \code{depth.zonoid} of a point.
}
  \item{kernel.bandwidth}{
  the bandwidth parameter of the kernel. If \code{NULL} - the Scott's rule of thumb is used.
  May be a single value for all classes, or a vector of values for each of the classes.
}
  \item{mah.parMcd}{
is the value of the argument \code{alpha} for the function \code{\link{covMcd}}; is used when \code{pretransform = "*MCD"}.
}
}
\details{
The potential representation is calculated in the same way as in \code{\link{depth.potential}}, see References below for more information and details.
}
\value{
Matrix of objects, each object (row) is represented via its potentials (columns) w.r.t. each of the classes of the training sample; order of the classes in columns corresponds to the one in the argument \code{cardinalities}.
}
\references{
Aizerman, M.A., Braverman, E.M., and Rozonoer, L.I. (1970). \emph{The Method of Potential Functions in the Theory of Machine Learning}. Nauka (Moscow).

Pokotylo, O. and Mosler, K. (2015). Classification with the pot-pot plot. \emph{Mimeo}.
}
\seealso{
\code{\link{ddalpha.train}} and \code{\link{ddalpha.classify}} for application, \code{\link{depth.potential}} for calculation of the potential.
}
\examples{

# Generate a bivariate normal location-shift classification task
# containing 20 training objects
class1 <- mvrnorm(50, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(50, c(1,1), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
data <- rbind(class1, class2)
plot(data, col = c(rep(1,50), rep(2,50)))
# potential with rule of thumb bandwidth
ds = depth.space.potential(data, c(50, 50))
# draw.ddplot(depth.space = ds, cardinalities = c(50, 50))

# potential with bandwidth = 0.5 and joint scaling
ds = depth.space.potential(data, c(50, 50), kernel.bandwidth = 0.5,
                           pretransform = "1Mom")
# draw.ddplot(depth.space = ds, cardinalities = c(50, 50))

# potential with bandwidth = 0.5 and separate scaling
ds = depth.space.potential(data, c(50, 50), kernel.bandwidth = 0.5, 
                           pretransform = "NahMom") # or without pretransform
# draw.ddplot(depth.space = ds, cardinalities = c(50, 50))

data <- getdata("hemophilia")
cardinalities = c(sum(data$gr == "normal"), sum(data$gr == "carrier"))
ds = depth.space.potential(data[,1:2], cardinalities)
# draw.ddplot(depth.space = ds, cardinalities = cardinalities)

}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{depthf.ABD}
\alias{depthf.ABD}
\title{Adjusted Band Depth for Functional Data}
\usage{
depthf.ABD(datafA, datafB, range = NULL, d = 101, norm = c("C", "L2"),
  J = 2, K = 1)
}
\arguments{
\item{datafA}{Functions whose depth is computed, represented by a \code{dataf} object of their arguments
and functional values. \code{m} stands for the number of functions.}

\item{datafB}{Random sample functions with respect to which the depth of \code{datafA} is computed. 
\code{datafB} is represented by a \code{dataf} object of their arguments
 and functional values. \code{n} is the sample size. The grid of observation points for the 
functions \code{datafA} and \code{datafB} may not be the same.}

\item{range}{The common range of the domain where the functions \code{datafA} and \code{datafB} are observed.
Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
\code{datafA} and \code{datafB}.}

\item{d}{Grid size to which all the functional data are transformed. For depth computation, 
all functional observations are first transformed into vectors of their functional values of length \code{d}
corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
points are reconstructed using linear interpolation, and extrapolation, see Nagy et al. (2016).}

\item{norm}{The norm used for the computation of the depth. Two possible 
choices are implemented: \code{C} for the uniform norm of continuous functions, 
and \code{L2} for the \eqn{L^2} norm of integrable functions.}

\item{J}{The order of the adjusted band depth, that is the maximal number of functions
taken in a band. Acceptable values are \code{2}, \code{3},... By default this value is set to \code{2}. 
Note that this is NOT the order as
defined in the order-extended version of adjusted band depths in Nagy et al. (2016), used
for the detection of shape outlying curves.}

\item{K}{Number of sub-samples of the functions from \code{B} taken to speed up the
computation. By default, sub-sampling is not performed. Values of \code{K} larger than \code{1}
result in an approximation of the adjusted band depth.}
}
\value{
A vectors of length \code{m} of the adjusted band depths.
}
\description{
The adjusted band depth 
of functional real-valued data based on either the
\eqn{C} (uniform) norm, or on the \eqn{L^2} norm of functions.
}
\details{
The function returns the vector of the sample adjusted band depth values. The kernel 
used in the evaluation is the function \eqn{K(u) = exp(-u)}.
}
\examples{
datafA = dataf.population()$dataf[1:20]
datafB = dataf.population()$dataf[21:50]
depthf.ABD(datafA,datafB)
depthf.ABD(datafA,datafB,norm="L2")

}
\references{
Gijbels, I., Nagy, S. (2015).
Consistency of non-integrated depths for functional data.
\emph{Journal of Multivariate Analysis} \bold{140}, 259--282.

Nagy, S., Gijbels, I. and Hlubinka, D. (2016). 
Weak convergence of discretely observed functional data with applications. 
\emph{Journal of Multivariate Analysis}, \bold{146}, 46--62.

Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
Depth-based recognition of shape outlying functions. 
\emph{Journal of Computational and Graphical Statistics}, \bold{26} (4), 883--893.
}
\seealso{
\code{\link{depthf.BD}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{depth}
\keyword{functional}

\name{depthf.fd2}
\alias{depthf.fd2}
\title{Bivariate Integrated and Infimal Depth for Functional Data}
\usage{
depthf.fd2(datafA, datafB, range = NULL, d = 101)
}
\arguments{
\item{datafA}{Bivariate functions whose depth is computed, represented by a multivariate \code{dataf} object of 
their arguments (vector), and a matrix with two columns of the corresponding bivariate functional values. 
\code{m} stands for the number of functions.}

\item{datafB}{Bivariate random sample functions with respect to which the depth of \code{datafA} is computed. 
\code{datafB} is represented by a multivariate \code{dataf} object of their arguments
 (vector), and a matrix with two columns of the corresponding bivariate functional values.
\code{n} is the sample size. The grid of observation points for the 
functions \code{datafA} and \code{datafB} may not be the same.}

\item{range}{The common range of the domain where the functions \code{datafA} and \code{datafB} are observed.
Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
\code{datafA} and \code{datafB}.}

\item{d}{Grid size to which all the functional data are transformed. For depth computation, 
all functional observations are first transformed into vectors of their functional values of length \code{d}
corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
points are reconstructed using linear interpolation, and extrapolation.}
}
\value{
Four vectors of length \code{m} are returned:
\itemize{
\item \code{Simpl_FD} the integrated depth based on the bivariate simplicial depth,
\item \code{Half_FD} the integrated depth based on the bivariate halfspace depth,
\item \code{Simpl_ID} the infimal depth based on the bivariate simplicial depth,
\item \code{Half_ID} the infimal depth based on the bivariate halfspace depth.
}
In addition, two vectors of length \code{m} of the relative area of smallest depth values is returned:
\itemize{
\item \code{Simpl_IA} the proportions of points at which the depth \code{Simpl_ID} was attained,
\item \code{Half_IA} the proportions of points at which the depth \code{Half_ID} was attained.
}
The values \code{Simpl_IA} and \code{Half_IA} are always in the interval [0,1]. 
They introduce ranking also among functions having the same
infimal depth value - if two functions have the same infimal depth, the one with larger infimal area
\code{IA} is said to be less central.
}
\description{
Integrated and infimal depths 
of functional bivariate data (that is, data of the form \eqn{X:[a,b] \to R^2},
or \eqn{X:[a,b] \to R} and the derivative of \eqn{X}) based on the
bivariate halfspace and simplicial depths.
}
\details{
The function returns the vectors of sample integrated and infimal depth values.
}
\examples{
datafA = dataf.population()$dataf[1:20]
datafB = dataf.population()$dataf[21:50]

dataf2A = derivatives.est(datafA,deriv=c(0,1))
dataf2B = derivatives.est(datafB,deriv=c(0,1))
depthf.fd2(dataf2A,dataf2B)

}
\references{
Hlubinka, D., Gijbels, I., Omelka, M. and Nagy, S. (2015). 
Integrated data depth for smooth functions and its application in supervised classification. 
\emph{Computational Statistics}, \bold{30} (4), 1011--1031.

Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
Depth-based recognition of shape outlying functions. 
\emph{Journal of Computational and Graphical Statistics}, \bold{26} (4), 883--893.
}
\seealso{
\code{\link{depthf.fd1}}, \code{\link{infimalRank}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{depth}
\keyword{derivatives}
\keyword{functional}

\name{depth.}
\alias{depth.}
\title{
Calculate Depth
}
\description{
Calculates the depth of points w.r.t. a multivariate data set.

The detailed descriptions are found in the corresponding topics.
}
\usage{
depth.(x, data, notion, ...)

## beta-skeleton depth
# depth.betaSkeleton(x, data, beta = 2, distance = "Lp", Lp.p = 2, 
#                   mah.estimate = "moment", mah.parMcd = 0.75)

## Tukey depth
# depth.halfspace(x, data, exact, method, num.directions = 1000, seed = 0)

## L2-depth
# depth.L2(x, data, mah.estimate = "moment", mah.parMcd = 0.75)

## Mahalanobis depth
# depth.Mahalanobis(x, data, mah.estimate = "moment", mah.parMcd = 0.75)

## projection depth
# depth.projection(x, data, method = "random", num.directions = 1000)

## simplicial depth
# depth.simplicial(x, data, exact = F, k = 0.05, seed = 0)

## simplicial volume depth
# depth.simplicialVolume(x, data, exact = F, k = 0.05, seed = 0)

## spatial depth
# depth.spatial(x, data)

## zonoid depth
# depth.zonoid(x, data)

## potential
# depth.potential (x, data, pretransform = "1Mom", 
#            kernel = "GKernel", kernel.bandwidth = NULL, mah.parMcd = 0.75)

## convex hull peeling depth
# depth.qhpeeling(x, data)

}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
  \item{notion}{
The name of the depth notion (shall also work with a user-defined depth function named \code{"depth.<name>"}).
}
  \item{\dots}{
Additional parameters passed to the depth functions.
}
}

\seealso{

\code{\link{depth.betaSkeleton}}

\code{\link{depth.halfspace}}

\code{\link{depth.L2}}

\code{\link{depth.Mahalanobis}}

\code{\link{depth.projection}}

\code{\link{depth.simplicial}}

\code{\link{depth.simplicialVolume}}

\code{\link{depth.spatial}}

\code{\link{depth.zonoid}}

\code{\link{depth.potential}}

\code{\link{depth.qhpeeling}}

\code{\link{depth.graph}} for building the depth surfaces of the two dimensional data.

}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector.
}
\examples{
# 5-dimensional normal distribution
data <- mvrnorm(1000, rep(0, 5), 
                matrix(c(1, 0, 0, 0, 0, 
                         0, 2, 0, 0, 0, 
                         0, 0, 3, 0, 0, 
                         0, 0, 0, 2, 0, 
                         0, 0, 0, 0, 1),
                nrow = 5))
x <- mvrnorm(10, rep(1, 5), 
             matrix(c(1, 0, 0, 0, 0, 
                      0, 1, 0, 0, 0, 
                      0, 0, 1, 0, 0, 
                      0, 0, 0, 1, 0, 
                      0, 0, 0, 0, 1),
             nrow = 5))
                
depths <- depth.(x, data, notion = "zonoid")
cat("Depths: ", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{plot.ddalphaf}
\alias{plot.ddalphaf}
\title{
Plots for the "ddalphaf" Class
}
\description{

\code{\link{plot.functional}} -- plots the functional data used by classifier

\code{\link{depth.contours.ddalpha}} -- builds the data depth contours for multiclass 2-dimensional data using the trained classifier.
\code{\link{draw.ddplot}} -- draws the \emph{DD}-plot of the existing DD\eqn{\alpha}-classifier.
}
\usage{
\method{plot}{ddalphaf}(x, type = c("functional.data", "ddplot", "depth.contours"), ...)
}

\arguments{
  \item{x}{
functional DD\eqn{\alpha}-classifier (obtained by \code{\link{ddalphaf.train}}).
}
  \item{type}{
type of the plot for \code{\link{plot.functional}}, \code{\link{draw.ddplot}} or \code{\link{depth.contours.ddalpha}}
}
  \item{\dots}{
additional parameters passed to the depth functions and to \code{\link{plot}}
}
}

\seealso{
\code{\link{depth.}}

\code{\link{depth.contours}}

\code{\link{depth.graph}}
}
\examples{

\dontrun{

dataf = dataf.growth()
ddalphaf = ddalphaf.train (dataf$dataf, dataf$labels, 
                            classifier.type = "ddalpha", maxNumIntervals = 2)

# plot the functional data
plot(ddalphaf)

# plot depth contours and separation in the transformed space 
# (possible only if maxNumIntervals = 2)
plot(ddalphaf, type = "depth.contours")

# plot the DD-plot
plot(ddalphaf, type = "ddplot")

}
}
\keyword{ functional }
\keyword{ visualization }

\name{depthf.BD}
\alias{depthf.BD}
\title{Band Depth for Functional Data}
\usage{
depthf.BD(datafA, datafB, range = NULL, d = 101)
}
\arguments{
\item{datafA}{Functions whose depth is computed, represented by a \code{dataf} object of their arguments
and functional values. \code{m} stands for the number of functions.}

\item{datafB}{Random sample functions with respect to which the depth of \code{datafA} is computed. 
\code{datafB} is represented by a \code{dataf} object of their arguments
 and functional values. \code{n} is the sample size. The grid of observation points for the 
functions \code{datafA} and \code{datafB} may not be the same.}

\item{range}{The common range of the domain where the functions \code{datafA} and \code{datafB} are observed.
Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
\code{datafA} and \code{datafB}.}

\item{d}{Grid size to which all the functional data are transformed. For depth computation, 
all functional observations are first transformed into vectors of their functional values of length \code{d}
corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
points are reconstructed using linear interpolation, and extrapolation.}
}
\value{
A vector of length \code{m} of the band depth values.
}
\description{
The (unadjusted) band depth 
for functional real-valued data of order \code{J=2}.
}
\details{
The function returns the vector of the sample (unadjusted) band depth values.
}
\examples{
datafA = dataf.population()$dataf[1:20]
datafB = dataf.population()$dataf[21:50]
depthf.BD(datafA,datafB)

}
\references{
Lopez-Pintado, S. and Romo, J. (2009), On the concept of depth for functional data,
\emph{J. Amer. Statist. Assoc.} \bold{104} (486), 718 - 734.
}
\seealso{
\code{\link{depthf.ABD}}, \code{\link{depthf.fd1}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{depth}
\keyword{functional}

\name{is.in.convex}
\alias{is.in.convex}
\title{
Check Outsiderness
}
\description{
Checks the belonging to at least one of class convex hulls of the training sample.
}
\usage{
is.in.convex(x, data, cardinalities, seed = 0)
}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose belonging to convex hulls is to be checked; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects, representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{seed}{
the random seed. The default value \code{seed=0} makes no changes.
}
}
\details{
Checks are conducted w.r.t. each separate class in \code{data} using the simplex algorithm, taken from the C++ implementation of the zonoid depth calculation  by Rainer Dyckerhoff.
}
\value{
Matrix of \code{number of objects} rows and \code{number of classes} columns, containing \code{1} if an object belongs to the convex hull of the corresponding class, and \code{0} otherwise.
}
\references{
Dyckerhoff, R., Koshevoy, G., and Mosler, K. (1996). Zonoid data depth: theory and computation. In: Prat A. (ed), \emph{COMPSTAT 1996. Proceedings in computational statistics}, Physica-Verlag (Heidelberg), 235--240.
}
\author{
Implementation of the simplex algorithm is taken from the algorithm for computation of zonoid depth (Dyckerhoff, Koshevoy and Mosler, 1996) that has been implemented in C++ by Rainer Dyckerhoff.
}
\seealso{
\code{\link{ddalpha.train}} and \code{\link{ddalpha.classify}} for application.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 400 training objects and 1000 to test with
class1 <- mvrnorm(700, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(700, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
trainIndices <- c(1:200)
testIndices <- c(201:700)
propertyVars <- c(1:2)
classVar <- 3
trainData <- rbind(cbind(class1[trainIndices,], rep(1, 200)), 
                   cbind(class2[trainIndices,], rep(2, 200)))
testData <- rbind(cbind(class1[testIndices,], rep(1, 500)), 
                  cbind(class2[testIndices,], rep(2, 500)))
data <- list(train = trainData, test = testData)

# Count outsiders
numOutsiders = sum(rowSums(is.in.convex(data$test[,propertyVars], 
                                data$train[,propertyVars], c(200, 200))) == 0)
cat(numOutsiders, "outsiders found in the testing sample.\n")
}
\keyword{ multivariate }

\name{depth.simplicialVolume}
\alias{depth.simplicialVolume}
\title{
Calculate Simplicial Volume Depth
}
\description{
Calculates the simpicial volume depth of points w.r.t. a multivariate data set.
}
\usage{
depth.simplicialVolume(x, data, exact = F, k = 0.05, mah.estimate = "moment", 
                       mah.parMcd = 0.75, seed = 0)
}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
  \item{exact}{
\code{exact=F} (by default) implies the approximative algorithm, considering \code{k} simplices, \code{exact=T} implies the exact algorithm.
}
  \item{k}{
Number (\eqn{k>1}) or portion (if \eqn{0<k<1}) of simplices that are considered if \code{exact=F}. If \eqn{k>1}, then the algorithmic complexity is polynomial in \eqn{d} but is independent of the number of observations in \code{data}, given \eqn{k}. If \eqn{0<k<1}, then the algorithmic complexity is exponential in the number of observations in \code{data}, but the calculation precision stays approximately the same.
}
  \item{mah.estimate}{
A character string specifying affine-invariance adjustment; can be \code{"none"}, \code{"moment"} or \code{"MCD"}, determining whether no affine-invariance adjustemt or moment or Minimum Covariance Determinant (MCD) (see \code{\link{covMcd}}) estimates of the covariance are used. By default \code{"moment"} is used.
}
  \item{mah.parMcd}{
The value of the argument \code{alpha} for the function \code{\link{covMcd}}; is used when \code{mah.estimate =} \code{"MCD"}.
}
  \item{seed}{
The random seed. The default value \code{seed=0} makes no changes.
}
}
\details{
Calculates Oja depth (also: Simplicial volume depth). 
At first the Oja outlyingness function \code{O(x,data)} is calculated as the average of the volumes of simplices built on \eqn{d} data points and the measurement point \code{x} (Oja, 1983).

Zuo and Serfling (2000) proposed Oja depth based on the Oja outlyingness function as \code{1/(1 + O(x,data)/S)}, where S is a square root of the determinant of \code{cov(data)}, which makes the depth function affine-invariant.
}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector.
}
\references{
Oja, H. (1983). Descriptive statistics for multivariate distributions. \emph{Statistics & Probability Letters} \bold{1} 327--332.

Zuo, Y.J. and Serfling, R. (2000). General notions of statistical depth function. \emph{The Annals of Statistics} \bold{28} 461--482.
}
\seealso{
\code{\link{depth.halfspace}} for calculation of the Tukey depth.

\code{\link{depth.Mahalanobis}} for calculation of Mahalanobis depth.

\code{\link{depth.projection}} for calculation of projection depth.

\code{\link{depth.simplicial}} for calculation of simplicial depth.

\code{\link{depth.spatial}} for calculation of spatial depth.

\code{\link{depth.zonoid}} for calculation of zonoid depth.

\code{\link{depth.potential}} for calculation of data potential.

}
\examples{
# 3-dimensional normal distribution
data <- mvrnorm(20, rep(0, 3), 
                matrix(c(1, 0, 0,
                         0, 2, 0,
                         0, 0, 1),
                       nrow = 3))
x <- mvrnorm(10, rep(1, 3), 
             matrix(c(1, 0, 0,
                      0, 1, 0,
                      0, 0, 1),
                    nrow = 3))

#exact
depths <- depth.simplicialVolume(x, data, exact = TRUE)
cat("Depths: ", depths, "\n")

#approximative
depths <- depth.simplicialVolume(x, data, exact = FALSE, k = 0.2)
cat("Depths: ", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{depthf.hM2}
\alias{depthf.hM2}
\title{Bivariate h-Mode Depth for Functional Data Based on the \eqn{L^2} Metric}
\usage{
depthf.hM2(datafA, datafB, range = NULL, d = 101, q = 0.2)
}
\arguments{
\item{datafA}{Bivariate functions whose depth is computed, represented by a multivariate \code{dataf} object of 
their arguments (vector), and a matrix with two columns of the corresponding bivariate functional values. 
\code{m} stands for the number of functions.}

\item{datafB}{Bivariate random sample functions with respect to which the depth of \code{datafA} is computed. 
\code{datafB} is represented by a multivariate \code{dataf} object of their arguments
 (vector), and a matrix with two columns of the corresponding bivariate functional values.
\code{n} is the sample size. The grid of observation points for the 
functions \code{datafA} and \code{datafB} may not be the same.}

\item{range}{The common range of the domain where the functions \code{datafA} and \code{datafB} are observed.
Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
\code{datafA} and \code{datafB}.}

\item{d}{Grid size to which all the functional data are transformed. For depth computation, 
all functional observations are first transformed into vectors of their functional values of length \code{d}
corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
points are reconstructed using linear interpolation, and extrapolation.}

\item{q}{The quantile used to determine the value of the bandwidth \eqn{h}
in the computation of the h-mode depth. \eqn{h} is taken as the \code{q}-quantile of
all non-zero distances between the functions \code{B}. By default, this value is set
to \code{q=0.2}, in accordance with the choice of Cuevas et al. (2007).}
}
\value{
Three vectors of length \code{m} of h-mode depth values are returned:
\itemize{
\item \code{hM} the unscaled h-mode depth,
\item \code{hM_norm} the h-mode depth \code{hM} linearly transformed so that its range is [0,1],
\item \code{hM_norm2} the h-mode depth \code{FD} linearly transformed by a transformation such that 
   the range of the h-mode depth of \code{B} with respect to \code{B} is [0,1]. This depth may give negative values.
}
}
\description{
The h-mode depth 
of functional bivariate data (that is, data of the form \eqn{X:[a,b] \to R^2},
or \eqn{X:[a,b] \to R} and the derivative of \eqn{X}) based on the
\eqn{L^2} metric of functions.
}
\details{
The function returns the vectors of sample h-mode depth values. The kernel 
used in the evaluation is the standard Gaussian kernel, the bandwidth value is chosen
as a quantile of the non-zero distances between the random sample curves.
}
\examples{
datafA = dataf.population()$dataf[1:20]
datafB = dataf.population()$dataf[21:50]

datafA2 = derivatives.est(datafA,deriv=c(0,1))
datafB2 = derivatives.est(datafB,deriv=c(0,1))

depthf.hM2(datafA2,datafB2)

depthf.hM2(datafA2,datafB2)$hM
# depthf.hM2(cbind(A2[,,1],A2[,,2]),cbind(B2[,,1],B2[,,2]))$hM
# the two expressions above should give the same result

}
\references{
Cuevas, A., Febrero, M. and Fraiman, R.  (2007). 
Robust estimation and classification for functional data via projection-based depth notions. 
\emph{Computational Statistics} \bold{22} (3), 481--496.
}
\seealso{
\code{\link{depthf.hM}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{depth}
\keyword{derivatives}
\keyword{functional}

\name{depth.space.zonoid}
\alias{depth.space.zonoid}
\title{
Calculate Depth Space using Zonoid Depth
}
\description{
Calculates the representation of the training classes in depth space using zonoid depth.
}
\usage{
depth.space.zonoid(data, cardinalities, seed = 0)
}
\arguments{
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{seed}{
the random seed. The default value \code{seed=0} makes no changes.
}
}
\details{
The depth representation is calculated in the same way as in \code{\link{depth.zonoid}}, see 'References' for more information and details.
}
\value{
Matrix of objects, each object (row) is represented via its depths (columns) w.r.t. each of the classes of the training sample; order of the classes in columns corresponds to the one in the argument \code{cardinalities}.
}
\references{
Dyckerhoff, R., Koshevoy, G., and Mosler, K. (1996). Zonoid data depth: theory and computation. In: Prat A. (ed), \emph{COMPSTAT 1996. Proceedings in computational statistics}, Physica-Verlag (Heidelberg), 235--240.

Koshevoy, G. and Mosler, K. (1997). Zonoid trimming for multivariate distributions \emph{Annals of Statistics} \bold{25} 1998--2017.

Mosler, K. (2002). \emph{Multivariate dispersion, central regions and depth: the lift zonoid approach} Springer (New York).
}
\seealso{
\code{\link{ddalpha.train}} and \code{\link{ddalpha.classify}} for application, \code{\link{depth.zonoid}} for calculation of zonoid depth.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 20 training objects
class1 <- mvrnorm(10, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(10, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
data <- rbind(class1, class2)
# Get depth space using zonoid depth
depth.space.zonoid(data, c(10, 10))

data <- getdata("hemophilia")
cardinalities = c(sum(data$gr == "normal"), sum(data$gr == "carrier"))
depth.space.zonoid(data[,1:2], cardinalities)

}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{ddalpha.test}
\alias{ddalpha.test}
\title{
Test DD-Classifier
}
\description{
Trains DD-classifier on the learning sequence of the data and tests it on the testing sequence.
}
\usage{
ddalpha.test(learn, test, ...)
}
\arguments{
  \item{learn}{
the learning sequence of the data. Matrix containing training sample where each of \eqn{n} rows is one object of the training sample where first \eqn{d} entries are inputs and the last entry is output (class label).
}
  \item{test}{
the testing sequence. Has the same format as \code{learn}
}
  \item{\dots}{
additional parameters passed to \code{\link{ddalpha.train}}
}
}

\value{

  \item{error}{
  the part of incorrectly classified data
  }
  \item{correct}{
  the number of correctly classified objects
  }
  \item{incorrect}{
  the number of incorrectly classified objects
  }
  \item{total}{
  the number of classified objects
  }
  \item{ignored}{
  the number of ignored objects (outside the convex hull of the learning data)
  }
  \item{n}{
  the number of objects in the testing sequence
  }
  \item{time}{
  training time
  }

}


\seealso{
\code{\link{ddalpha.train}} to train the DD-classifier, 
\code{\link{ddalpha.classify}} for classification using DD-classifier, 
\code{\link{ddalpha.getErrorRateCV}} and \code{\link{ddalpha.getErrorRatePart}} to get error rate of the DD-classifier on particular data.
}
\examples{

# Generate a bivariate normal location-shift classification task
# containing 200 training objects and 200 to test with
class1 <- mvrnorm(200, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(200, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
trainIndices <- c(1:100)
testIndices <- c(101:200)
propertyVars <- c(1:2)
classVar <- 3
trainData <- rbind(cbind(class1[trainIndices,], rep(1, 100)), 
                   cbind(class2[trainIndices,], rep(2, 100)))
testData <- rbind(cbind(class1[testIndices,], rep(1, 100)), 
                  cbind(class2[testIndices,], rep(2, 100)))
data <- list(train = trainData, test = testData)

# Train 1st DDalpha-classifier (default settings) 
# and get the classification error rate
stat <- ddalpha.test(data$train, data$test)
cat("1. Classification error rate (defaults): ", 
    stat$error, ".\n", sep = "")

# Train 2nd DDalpha-classifier (zonoid depth, maximum Mahalanobis 
# depth classifier with defaults as outsider treatment) 
# and get the classification error rate
stat2 <- ddalpha.test(data$train, data$test, depth = "zonoid", 
                          outsider.methods = "depth.Mahalanobis")
cat("2. Classification error rate (depth.Mahalanobis): ", 
    stat2$error, ".\n", sep = "")

}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ benchmark }

\name{depthf.hM}
\alias{depthf.hM}
\title{h-Mode Depth for Functional Data}
\usage{
depthf.hM(datafA, datafB, range = NULL, d = 101, norm = c("C", "L2"),
  q = 0.2)
}
\arguments{
\item{datafA}{Functions whose depth is computed, represented by a \code{dataf} object of their arguments
and functional values. \code{m} stands for the number of functions.}

\item{datafB}{Random sample functions with respect to which the depth of \code{datafA} is computed. 
\code{datafB} is represented by a \code{dataf} object of their arguments
 and functional values. \code{n} is the sample size. The grid of observation points for the 
functions \code{datafA} and \code{datafB} may not be the same.}

\item{range}{The common range of the domain where the functions \code{datafA} and \code{datafB} are observed.
Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
\code{datafA} and \code{datafB}.}

\item{d}{Grid size to which all the functional data are transformed. For depth computation, 
all functional observations are first transformed into vectors of their functional values of length \code{d}
corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
points are reconstructed using linear interpolation, and extrapolation.}

\item{norm}{The norm used for the computation of the depth. Two possible 
choices are implemented: \code{C} for the uniform norm of continuous functions, 
and \code{L2} for the \eqn{L^2} norm of integrable functions.}

\item{q}{The quantile used to determine the value of the bandwidth \eqn{h}
in the computation of the h-mode depth. \eqn{h} is taken as the \code{q}-quantile of
all non-zero distances between the functions \code{B}. By default, this value is set
to \code{q=0.2}, in accordance with the choice of Cuevas et al. (2007).}
}
\value{
A vector of length \code{m} of the h-mode depth values.
}
\description{
The h-mode depth of functional real-valued data.
}
\details{
The function returns the vectors of the sample h-mode depth values. The kernel 
used in the evaluation is the standard Gaussian kernel, the bandwidth value is chosen
as a quantile of the non-zero distances between the random sample curves.
}
\examples{
datafA = dataf.population()$dataf[1:20]
datafB = dataf.population()$dataf[21:50]
depthf.hM(datafA,datafB)
depthf.hM(datafA,datafB,norm="L2")
}
\references{
Cuevas, A., Febrero, M. and Fraiman, R.  (2007).
Robust estimation and classification for functional data via projection-based depth notions. 
\emph{Computational Statistics} \bold{22} (3), 481--496.

Nagy, S., Gijbels, I. and Hlubinka, D. (2016). 
Weak convergence of discretely observed functional data with applications. 
\emph{Journal of Multivariate Analysis}, \bold{146}, 46--62.
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{depth}
\keyword{functional}

\name{depth.space.spatial}
\alias{depth.space.spatial}
\title{
Calculate Depth Space using Spatial Depth
}
\description{
Calculates the representation of the training classes in depth space using spatial depth.
}
\usage{
depth.space.spatial(data, cardinalities, mah.estimate = "moment", mah.parMcd = 0.75)
}
\arguments{
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{mah.estimate}{ is a character string specifying which estimates to use when calculating sample covariance matrix; can be \code{"none"}, \code{"moment"} or \code{"MCD"}, determining whether traditional moment or Minimum Covariance Determinant (MCD) (see \code{\link{covMcd}}) estimates for mean and covariance are used. By default \code{"moment"} is used. With \code{"none"} the non-affine invariant version of Spatial depth is calculated
}
  \item{mah.parMcd}{
is the value of the argument \code{alpha} for the function \code{\link{covMcd}}; is used when \code{mah.estimate =} \code{"MCD"}.
}
}
\details{
The depth representation is calculated in the same way as in \code{\link{depth.spatial}}, see 'References' for more information and details.
}
\value{
Matrix of objects, each object (row) is represented via its depths (columns) w.r.t. each of the classes of the training sample; order of the classes in columns corresponds to the one in the argument \code{cardinalities}.
}
\references{
Chaudhuri, P. (1996). On a geometric notion of quantiles for multivariate data. \emph{Journal of the Americal Statistical Association} \bold{91} 862--872.

Koltchinskii, V.I. (1997). M-estimation, convexity and quantiles. \emph{The Annals of Statistics} \bold{25} 435--477.

Serfling, R. (2006). Depth functions in nonparametric multivariate inference. In: Liu, R., Serfling, R., Souvaine, D. (eds.), \emph{Data Depth: Robust Multivariate Analysis, Computational Geometry and Applications}, American Mathematical Society, 1--16.

Vardi, Y. and Zhang, C.H. (2000). The multivariate L1-median and associated data depth. \emph{Proceedings of the National Academy of Sciences, U.S.A.} \bold{97} 1423--1426.
}
\seealso{
\code{\link{ddalpha.train}} and \code{\link{ddalpha.classify}} for application, \code{\link{depth.spatial}} for calculation of spatial depth.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 20 training objects
class1 <- mvrnorm(10, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(10, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
data <- rbind(class1, class2)
# Get depth space using spatial depth
depth.space.spatial(data, c(10, 10))

data <- getdata("hemophilia")
cardinalities = c(sum(data$gr == "normal"), sum(data$gr == "carrier"))
depth.space.spatial(data[,1:2], cardinalities)
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{compclassf.train}
\alias{compclassf.train}
\title{
Functional Componentwise Classifier
}
\description{
Trains the functional componentwise classifier
}
\usage{
compclassf.train (dataf, labels, subset,
                  to.equalize = TRUE, 
                  to.reduce = TRUE, 
                  classifier.type = c("ddalpha", "maxdepth", "knnaff", "lda", "qda"), 
                  ...)
}

\arguments{
  \item{dataf}{
list containing lists (functions) of two vectors of equal length, named "args" and "vals": arguments sorted in ascending order and corresponding them values respectively
}
  \item{labels}{
list of output labels of the functional observations
}
  \item{subset}{
an optional vector specifying a subset of observations to be used in training the classifier.
}
  \item{to.equalize}{
Adjust the data to have equal (the largest) argument interval.
}
  \item{to.reduce}{
If the data spans a subspace only, project on it (by PCA).
}
  \item{classifier.type}{
the classifier which is used on the transformed space. The default value is 'ddalpha'.
}
  \item{\dots}{
additional parameters, passed to the classifier, selected with parameter \code{classifier.type}.
}
}
\details{
The finite-dimensional space is directly constructed from the observed values. 
Delaigle, Hall and Bathia (2012) consider (almost) all sets of discretization points
that have a given cardinality. 

The usual classifiers are then trained on the constructed finite-dimensional space.
}
\value{
Trained functional componentwise classifier
%%  If it is a LIST, use
%%  \item{comp1 }{Description of 'comp1'}
%%  \item{comp2 }{Description of 'comp2'}
%% ...
}
\references{
Delaigle, A., Hall, P., and Bathia, N. (2012). Componentwise classification and clustering of functional data. \emph{Biometrika} \bold{99} 299--313.
}

\seealso{
 \code{\link{compclassf.classify}} for classification using functional componentwise classifier,
 
 \code{\link{ddalphaf.train}} to train the functional DD-classifier,
 
 \code{\link{dataf.*}} for functional data sets included in the package.
 
}
\examples{

\dontrun{
## load the Growth dataset
dataf = dataf.growth()

learn = c(head(dataf$dataf, 49), tail(dataf$dataf, 34))
labels =c(head(dataf$labels, 49), tail(dataf$labels, 34)) 
test = tail(head(dataf$dataf, 59), 10)    # elements 50:59. 5 girls, 5 boys

c = compclassf.train (learn, labels, classifier.type = "ddalpha")

classified = compclassf.classify(c, test)

print(unlist(classified))

}
}

\keyword{ functional }
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }
\keyword{ classif }

\name{depth.graph}
\alias{depth.graph}
\title{
Depth Graph
}
\description{
Builds the data depth graphs for 2-dimensional data. The graph is built using \code{\link{persp}}.
}
\usage{
depth.graph(data, 
  depth_f = c("halfspace", "Mahalanobis", "projection", "simplicial", 
              "simplicialVolume", "spatial", "zonoid", "none"), 
  apoint = NULL, 
  main = depth_f,  
  xlim = c(min(data[, 1]), max(data[, 1])), 
  ylim = c(min(data[, 2]), max(data[, 2])), 
  zlim = c(0, max(z)), 
  xnum = 250, 
  ynum = 250, 
  theta=15, phi=60,
  bold = F,
  ...)
}

\arguments{
  \item{data}{
2-dimensional numeric data frame or matrix
}
  \item{depth_f}{
the name of the depth function. The list of the supported depths and described in the topic \code{\link{depth.}}.
}
  \item{apoint}{
a 2-dimensional point which is shown in black color.
}
  \item{main}{
an overall title for the plot: see \code{\link{title}}
}
  \item{xlim, ylim, zlim}{
numeric vectors of length 2, giving the x, y and z coordinates ranges: see \code{\link{plot.window}}
}
  \item{xnum, ynum}{
number of points on each direction, x and y. Impacts the smoothness of the surface. 
}
  \item{theta, phi}{
rotation angles
}
  \item{bold}{
draws bold points
}
  \item{\dots}{
additional parameters passed to \code{\link{persp}}
}
}

\seealso{
\code{\link{depth.}}

\code{\link{persp}}
}
\examples{

\dontrun{

par(mfrow = c(2,3), mar = c(0,0,0,0), mai = c(0,0,0.2,0))
data(hemophilia)
depth.graph(hemophilia, "none", xnum = 100, ynum = 100)
depth.graph(hemophilia, "Mahalanobis", xnum = 100, ynum = 100)
depth.graph(hemophilia, "halfspace", xnum = 100, ynum = 100)
depth.graph(hemophilia, "projection", xnum = 100, ynum = 100)
depth.graph(hemophilia, "zonoid", xnum = 100, ynum = 100)
depth.graph(hemophilia, "spatial", xnum = 100, ynum = 100)

}
}

\keyword{ visualization }

\name{dknn.classify}
\alias{dknn.classify}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Depth-Based kNN
}
\description{
The implementation of the affine-invariant depth-based kNN of Paindaveine and Van Bever (2015). 
}
\usage{
dknn.classify(objects, data, k, depth = "halfspace", seed = 0)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{objects}{
Matrix containing objects to be classified; each row is one \eqn{d}-dimensional object.
}
  \item{data}{
Matrix containing training sample where each of \eqn{n} rows is one object of the training sample where first \eqn{d} entries are inputs and the last entry is output (class label).
}
  \item{k}{
the number of neighbours
}
  \item{depth}{
Character string determining which depth notion to use; the default value is \code{"halfspace"}. 
Currently the method supports the following depths: "halfspace", "Mahalanobis", "simplicial".
}
  \item{seed}{
the random seed. The default value \code{seed=0} makes no changes.
}
}
%\details{
%%  ~~ If necessary, more details than the description above ~~
%}
\value{
List containing class labels, or character string "Ignored" for the outsiders if "Ignore" was specified as the outsider treating method.
}
\references{
Paindaveine, D. and Van Bever, G. (2015). Nonparametrically consistent depth-based classifiers. \emph{Bernoulli} \bold{21} 62--82.
}
%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{dknn.train}} to train the Dknn-classifier.

\code{\link{dknn.classify.trained}} to classify with the Dknn-classifier.

\code{\link{ddalpha.train}} to train the DD\eqn{\alpha}-classifier.

\code{\link{ddalpha.getErrorRateCV}} and \code{\link{ddalpha.getErrorRatePart}} to get error rate of the Dknn-classifier on particular data (set \code{separator = "Dknn"}).
}
\examples{

# Generate a bivariate normal location-shift classification task
# containing 200 training objects and 200 to test with
class1 <- mvrnorm(200, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(200, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
trainIndices <- c(1:100)
testIndices <- c(101:200)
propertyVars <- c(1:2)
classVar <- 3
trainData <- rbind(cbind(class1[trainIndices,], rep(1, 100)), 
                   cbind(class2[trainIndices,], rep(2, 100)))
testData <- rbind(cbind(class1[testIndices,], rep(1, 100)), 
                  cbind(class2[testIndices,], rep(2, 100)))
data <- list(train = trainData, test = testData)

# Train the classifier
# and get the classification error rate
cls <- dknn.train(data$train, kMax = 20, depth = "Mahalanobis")
cls$k
classes1 <- dknn.classify.trained(data$test[,propertyVars], cls)
cat("Classification error rate: ", 
    sum(unlist(classes1) != data$test[,classVar])/200)

# Classify the new data based on the old ones in one step
classes2 <- dknn.classify(data$test[,propertyVars], data$train, k = cls$k, depth = "Mahalanobis")
cat("Classification error rate: ", 
    sum(unlist(classes2) != data$test[,classVar])/200)

}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ multivariate }
\keyword{ nonparametric }
\keyword{ classif }

\name{depth.simplicial}
\alias{depth.simplicial}
\title{
Calculate Simplicial Depth
}
\description{
Calculates the simplicial depth of points w.r.t. a multivariate data set.
}
\usage{
depth.simplicial(x, data, exact = F, k = 0.05, seed = 0)
}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
  \item{exact}{
\code{exact=F} (by default) implies the approximative algorithm, considering \code{k} simplices, \code{exact=T} implies the exact algorithm.
}
  \item{k}{
Number (\eqn{k>1}) or portion (if \eqn{0<k<1}) of simplices that are considered if \code{exact=F}. If \eqn{k>1}, then the algorithmic complexity is polynomial in \eqn{d} but is independent of the number of observations in \code{data}, given \eqn{k}. If \eqn{0<k<1}, then the algorithmic complexity is exponential in the number of observations in \code{data}, but the calculation precision stays approximately the same.
}
  \item{seed}{
the random seed. The default value \code{seed=0} makes no changes.
}
}
\details{
Calculates simplicial depth. Simplicial depth is counted as a probability that a point lies in a simplex, built on \eqn{d+1} data points.
}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector.
}
\references{
Chaudhuri, P. (1996). On a geometric notion of quantiles for multivariate data. \emph{Journal of the American Statistical Association} \bold{91} 862--872.

Liu, R. Y. (1990). On a notion of data depth based on random simplices. \emph{The Annals of Statistics} \bold{18} 405--414.

Rousseeuw, P.J. and Ruts, I. (1996). Algorithm AS 307: Bivariate location depth. \emph{Journal of the Royal Statistical Society. Seriec C (Applied Statistics)} \bold{45} 516--526.
}
\seealso{
\code{\link{depth.halfspace}} for calculation of the Tukey depth.

\code{\link{depth.Mahalanobis}} for calculation of Mahalanobis depth.

\code{\link{depth.projection}} for calculation of projection depth.

\code{\link{depth.simplicialVolume}} for calculation of simplicial volume depth.

\code{\link{depth.spatial}} for calculation of spatial depth.

\code{\link{depth.zonoid}} for calculation of zonoid depth.

\code{\link{depth.potential}} for calculation of data potential.

}
\examples{
# 3-dimensional normal distribution
data <- mvrnorm(20, rep(0, 3), 
                matrix(c(1, 0, 0,
                         0, 2, 0,
                         0, 0, 1),
                       nrow = 3))
x <- mvrnorm(10, rep(1, 3), 
             matrix(c(1, 0, 0,
                      0, 1, 0,
                      0, 0, 1),
                    nrow = 3))

#exact
depths <- depth.simplicial(x, data, exact = TRUE)
cat("Depths: ", depths, "\n")

#approximative
depths <- depth.simplicial(x, data, exact = FALSE, k = 0.2)
cat("Depths: ", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{dataf.tecator}
\alias{dataf.tecator}
\alias{tecator}
\docType{data}
\title{
Functional Data Set Spectrometric Data (Tecator)
}
\description{
This dataset is a part of the original one which can be found at
\url{https://www.stat.cmu.edu/}. For each peace of finely chopped meat we observe one spectrometric curve which
corresponds to the absorbance measured at 100 wavelengths. 
The peaces are split according to Ferraty and Vieu (2006) into two classes: with small (<20) and large fat
content obtained by an analytical chemical processing.
}
\usage{
dataf.tecator()
}
\format{
  The functional data as a data structure.
  \describe{
    \item{\code{dataf}}{
      The functional data as a list of objects. Each object is characterized by two coordinates.
      \describe{
        \item{\code{args}}{\bold{wavelength} - a numeric vector of discretization points from 850 to 1050mm }
        \item{\code{vals}}{\bold{absorbance} - a numeric vector of absorbance values}
       }
    }
    \item{\code{labels}}{The classes of the objects: "small" (<20) and "large" fat content}
  }
}
\author{

Febrero-Bande, M and Oviedo de la Fuente, Manuel
}
\source{

\url{https://www.stat.cmu.edu/}

}
\references{
Ferraty, F. and Vieu, P. (2006). \emph{Nonparametric functional data analysis: theory and practice}. Springer.
}
\seealso{
\code{\link{dataf.*}} for other functional data sets

\code{\link{plot.functional}} for building plots of functional data
}
\examples{
## load the dataset
dataf = dataf.tecator()

## view the classes
unique(dataf$labels)

## access the 5th point of the 2nd object
dataf$dataf[[2]]$args[5]
dataf$dataf[[2]]$vals[5]

## plot the data
\dontrun{
labels = unlist(dataf$labels)
plot(dataf, 
  xlab="Wavelengths", ylab="Absorbances", 
  main=paste("Tecator: < 20 red (", sum(labels == "small"), "),", 
            " >= 20 blue (", sum(labels == "large"), ")", sep=""),
    colors = c("blue", "red"))
}
}
\keyword{datasets}
\keyword{functional}

\name{ddalphaf.classify}
\alias{ddalphaf.classify}
\alias{predict.ddalphaf}
\title{
Classify using Functional DD-Classifier
}
\description{
Classifies data using the functional DD-classifier.
}
\usage{
ddalphaf.classify(ddalphaf, objectsf, subset, ...)

\method{predict}{ddalphaf}(object, objectsf, subset, ...)
}
\arguments{
  \item{ddalphaf, object}{
Functional DD-classifier (obtained by \code{\link{ddalphaf.train}}).
}
  \item{objectsf}{list containing lists (functions) of two vectors of equal length, named "args" and "vals": arguments sorted in ascending order and corresponding them values respectively
}
  \item{subset}{
an optional vector specifying a subset of observations to be classified.
}
  \item{\dots}{
additional parameters, passed to the classifier, selected with parameter \code{classifier.type} in \code{\link{ddalphaf.train}}.
}
}



\value{
List containing class labels.
}
\references{
Mosler, K. and Mozharovskyi, P. (2017). Fast DD-classification of functional data. \emph{Statistical Papers} \bold{58} 1055--1089.

Mozharovskyi, P. (2015). \emph{Contributions to Depth-based Classification and Computation of the Tukey Depth}. Verlag Dr. Kovac (Hamburg).
}
\seealso{
\code{\link{ddalphaf.train}} to train the functional DD\eqn{\alpha}-classifier.
}
\examples{

\dontrun{
## load the Growth dataset
dataf = dataf.growth()

learn = c(head(dataf$dataf, 49), tail(dataf$dataf, 34))
labels= c(head(dataf$labels, 49), tail(dataf$labels, 34)) 
test  = tail(head(dataf$dataf, 59), 10)    # elements 50:59. 5 girls, 5 boys

c = ddalphaf.train (learn, labels, classifier.type = "ddalpha")

classified = ddalphaf.classify(c, test)

print(unlist(classified))

}
}

\keyword{ functional }
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }
\keyword{ classif }

\name{plot.ddalpha}
\alias{plot.ddalpha}
\title{
Plots for the "ddalpha" Class
}
\description{
\code{\link{depth.contours.ddalpha}} -- builds the data depth contours for multiclass 2-dimensional data using the trained classifier.
\code{\link{draw.ddplot}} -- draws the \emph{DD}-plot of the existing DD\eqn{\alpha}-classifier.
}
\usage{
\method{plot}{ddalpha}(x, type = c("ddplot", "depth.contours"), ...)
}

\arguments{
  \item{x}{
DD\eqn{\alpha}-classifier (obtained by \code{\link{ddalpha.train}}).
}
  \item{type}{
type of the plot for \code{\link{draw.ddplot}} or \code{\link{depth.contours.ddalpha}}
}
  \item{\dots}{
additional parameters passed to the depth functions and to \code{\link{plot}}
}
}

\seealso{
\code{\link{depth.}}

\code{\link{depth.contours}}

\code{\link{depth.graph}}
}
\examples{

\dontrun{


par(mfrow = c(2,2))
data(hemophilia)

ddalpha = ddalpha.train(hemophilia, depth = "none")
plot(ddalpha, type = "depth.contours", main = "data")
plot(ddalpha, type = "ddplot", main = "data", drawsep = F)

for (depth in c("zonoid", "Mahalanobis", "projection", "spatial")){
  ddalpha = ddalpha.train(hemophilia, depth = depth)
  plot(ddalpha, type = "depth.contours", main = depth, drawsep = T)
  plot(ddalpha, type = "ddplot", main = depth)
}

}
}

\keyword{ visualization }

\name{depth.qhpeeling}
\alias{depth.qhpeeling}
\title{
Calculate Convex Hull Peeling Depth
}
\description{
Calculates the convex hull peeling depth of points w.r.t. a multivariate data set.
}
\usage{
depth.qhpeeling(x, data)
}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
}
\details{
Calculates the convex hull peeling depth (Eddy, 1982; see also Cascos, 2009).
}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector. Each depth value equals the number of the convex hulls to be peeled from \code{data} so that (the corresponding row of) \code{x} is not contained in the convex hull of the rest of the data; the depths are normalized by the number of points in \code{data}.
}
\references{
Eddy, W.F. (1982). Convex hull peeling. In: Caussinus, H., Ettinger, P. and Tomassone, R. (eds), \emph{COMPSTAT 1982. Proceedings in computational statistics}, Physica-Verlag (Vienna), 42--47.

Cascos, I. (2009). Data depth: multivariate statistics and geometry. In: Kendall, W.S. and Molchanov, I. (eds) \emph{New Perspectives in Stochastic Geometry}, Clarendon/Oxford University Press (Oxford).
}
\seealso{
\code{\link{depth.halfspace}} for calculation of the Tukey depth.

\code{\link{depth.L2}} for calculation of L2-depth.

\code{\link{depth.Mahalanobis}} for calculation of Mahalanobis depth.

\code{\link{depth.projection}} for calculation of projection depth.

\code{\link{depth.simplicial}} for calculation of simplicial depth.

\code{\link{depth.simplicialVolume}} for calculation of simplicial volume depth.

\code{\link{depth.spatial}} for calculation of spatial depth.

\code{\link{depth.potential}} for calculation of data potential.

\code{\link{depth.zonoid}} for calculation of zonoid depth.

}
\examples{
# Mixture of 3-variate normal distributions
data <- mvrnorm(25, rep(0, 3), diag(3))
x <- rbind(mvrnorm(10, rep(1, 3), diag(3)), data)
depths <- depth.qhpeeling(x, data)
cat("Depths:", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{depth.space.projection}
\alias{depth.space.projection}
\title{
Calculate Depth Space using Projection Depth
}
\description{
Calculates the representation of the training classes in depth space using projection depth.
}
\usage{
depth.space.projection(data, cardinalities, 
                       method = "random", num.directions = 1000, seed = 0)
}
\arguments{
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{method}{
  to be used in calculations.
  
  \code{"random"} Here the depth is determined as the minimum univariate depth of the data projected on lines in several directions. The directions are distributed uniformly on the \eqn{(d-1)}-sphere; the same direction set is used for all points.
  
  \code{"linearize"} The Nelder-Mead method for function minimization, taken from Olsson, Journal of Quality Technology, 1974, 6, 56. R-codes of this function were written by Subhajit Dutta.
  }
  \item{num.directions}{
Number of random directions to be generated for \code{method = "random"}. With the growth of n the complexity grows linearly for the same number of directions. 
}
  \item{seed}{
the random seed. The default value \code{seed=0} makes no changes.
}
}
\details{
The depth representation is calculated in the same way as in \code{\link{depth.projection}}, see 'References' for more information and details.
}
\value{
Matrix of objects, each object (row) is represented via its depths (columns) w.r.t. each of the classes of the training sample; order of the classes in columns corresponds to the one in the argument \code{cardinalities}.
}
\references{
Donoho, D.L. (1982). \emph{Breakdown properties of multivariate location estimators}. Ph.D. qualifying paper. Department of Statistics, Harvard University.

Liu, R.Y. (1992). Data depth and multivariate rank tests. In: Dodge, Y. (ed.), L1-Statistics and Related Methods, North-Holland (Amsterdam), 279--294.

Liu, X. and Zuo, Y. (2014). Computing projection depth and its associated estimators. \emph{Statistics and Computing} \bold{24} 51--63.

Stahel, W.A. (1981). \emph{Robust estimation: infinitesimal optimality and covariance matrix estimators}. Ph.D. thesis (in German). Eidgenossische Technische Hochschule Zurich.

Zuo, Y.J. and Lai, S.Y. (2011). Exact computation of bivariate projection depth and the Stahel-Donoho estimator. \emph{Computational Statistics and Data Analysis} \bold{55} 1173--1179.
}
\seealso{
\code{\link{ddalpha.train}} and \code{\link{ddalpha.classify}} for application, \code{\link{depth.projection}} for calculation of projection depth.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 20 training objects
class1 <- mvrnorm(10, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(10, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
data <- rbind(class1, class2)
# Get depth space using projection depth
depth.space.projection(data, c(10, 10), method = "random", num.directions = 1000)
depth.space.projection(data, c(10, 10), method = "linearize")

data <- getdata("hemophilia")
cardinalities = c(sum(data$gr == "normal"), sum(data$gr == "carrier"))
depth.space.projection(data[,1:2], cardinalities)
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{ddalpha.getErrorRatePart}
\alias{ddalpha.getErrorRatePart}
\title{
Test DD-Classifier
}
\description{
Performs a benchmark procedure by partitioning the given data. 
On each of \code{times} steps \code{size} observations are removed from the data, the DD-classifier is trained on these data and tested on the removed observations.
}
\usage{
ddalpha.getErrorRatePart(data, size = 0.3, times = 10,  ...)
}
\arguments{
  \item{data}{
Matrix containing training sample where each of \eqn{n} rows is one object of the training sample where first \eqn{d} entries are inputs and the last entry is output (class label).
}
  \item{size}{
  the excluded sequences size. Either an integer between \eqn{1} and \eqn{n}, or a fraction of data between \eqn{0} and \eqn{1}.
}
  \item{times}{
  the number of times the classifier is trained.
}
  \item{\dots}{
additional parameters passed to \code{\link{ddalpha.train}}
}
}

\value{

  \item{errors}{
  the part of incorrectly classified data (mean)
  }
  \item{errors_sd}{
  the standard deviation of errors
  }
  \item{errors_vec}{
  vector of errors
  }
  \item{time}{
  the mean training time
  }
  \item{time_sd}{
  the standard deviation of training time
  }

}


\seealso{
\code{\link{ddalpha.train}} to train the DD\eqn{\alpha}-classifier, 
\code{\link{ddalpha.classify}} for classification using DD\eqn{\alpha}-classifier, 
\code{\link{ddalpha.test}} to test the DD-classifier on particular learning and testing data,
\code{\link{ddalpha.getErrorRateCV}} to get error rate of the DD-classifier on particular data.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 200 objects
class1 <- mvrnorm(100, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(100, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
propertyVars <- c(1:2)
classVar <- 3
data <- rbind(cbind(class1, rep(1, 100)), cbind(class2, rep(2, 100)))

# Train 1st DDalpha-classifier (default settings) 
# and get the classification error rate
stat <- ddalpha.getErrorRatePart(data, size = 10, times = 10)
cat("1. Classification error rate (defaults): ", 
    stat$error, ".\n", sep = "")

# Train 2nd DDalpha-classifier (zonoid depth, maximum Mahalanobis 
# depth classifier with defaults as outsider treatment) 
# and get the classification error rate
stat2 <- ddalpha.getErrorRatePart(data, depth = "zonoid", 
                                outsider.methods = "depth.Mahalanobis", size = 0.2, times = 10)
cat("2. Classification error rate (depth.Mahalanobis): ", 
    stat2$error, ".\n", sep = "")



}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ benchmark }

\name{dataf2rawfd}
\alias{dataf2rawfd}
\title{Transform a \code{dataf} Object to Raw Functional Data}
\usage{
dataf2rawfd(dataf, range = NULL, d = 101)
}
\arguments{
\item{dataf}{Functions to be transformed, represented by a (possibly multivariate) \code{dataf} object of their arguments
and functional values. \code{m} stands for the number of functions. The grid of observation points for the 
functions in \code{dataf} may not be the same.}

\item{range}{The common range of the domain where the functions \code{dataf} are observed.
Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
\code{dataf}. If the range is not provided, the smallest interval in which all the arguments from the data functions
are contained is chosen as the domain.}

\item{d}{Grid size to which all the functional data are transformed. All functional observations are 
transformed into vectors of their functional values of length \code{d}
corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
points are reconstructed using linear interpolation, and extrapolation, see Nagy et al. (2016).}
}
\value{
If the functional data are univariate (scalar-valued), a matrix of size \code{m*d} is given, with each row
corresponding to one function. If the functional data are \code{k}-variate with k>1, an array of size \code{m*d*k}
of the functional values is given.
}
\description{
From a (possibly multivariate) functional data object \code{dataf} constructs an array of the functional values
evaluated at an equi-distant grid of points.
}
\examples{
## transform a matrix into a functional data set and back
n = 5
d = 21
X = matrix(rnorm(n*d),ncol=d)
R = rawfd2dataf(X,range=c(0,1))
R2 = dataf2rawfd(R,range=c(0,1),d=d)
all.equal(X,R2)

## transform a functional dataset into a raw matrix of functional values
dataf = dataf.population()$dataf
dataf2rawfd(dataf,range=c(1950,2015),d=66)

## transform an array into a multivariate functional data set and back
k = 3
X = array(rnorm(n*d*k),dim=c(n,d,k))
R = rawfd2dataf(X,range=c(-1,1))
dataf2rawfd(R,range=c(-1,1),d=50)

}
\seealso{
\code{\link{rawfd2dataf}}

\code{\link{depthf.fd1}}

\code{\link{depthf.fd2}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{functional}

\name{depth.spatial}
\alias{depth.spatial}
\title{
Calculate Spatial Depth
}
\description{
Calculates the spatial depth of points w.r.t. a multivariate data set.
}
\usage{
depth.spatial(x, data, mah.estimate = "moment", mah.parMcd = 0.75)
}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
  \item{mah.estimate}{ is a character string specifying which estimates to use when calculating sample covariance matrix; can be \code{"none"}, \code{"moment"} or \code{"MCD"}, determining whether traditional moment or Minimum Covariance Determinant (MCD) (see \code{\link{covMcd}}) estimates for mean and covariance are used. By default \code{"moment"} is used. With \code{"none"} the non-affine invariant version of Spatial depth is calculated
}
  \item{mah.parMcd}{
is the value of the argument \code{alpha} for the function \code{\link{covMcd}}; is used when \code{mah.estimate =} \code{"MCD"}.
}
}
\details{
Calculates spatial depth. Spatial depth (also L1-depth) is a distance-based depth exploiting the idea of spatial quantiles of Chaudhuri (1996) and Koltchinskii (1997), formulated by Vardi & Zhang (2000) and Serfling (2002).
}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector.
}
\references{
Chaudhuri, P. (1996). On a geometric notion of quantiles for multivariate data. \emph{Journal of the Americal Statistical Association} \bold{91} 862--872.

Koltchinskii, V.I. (1997). M-estimation, convexity and quantiles. \emph{The Annals of Statistics} \bold{25} 435--477.

Serfling, R. (2006). Depth functions in nonparametric multivariate inference. In: Liu, R., Serfling, R., Souvaine, D. (eds.), \emph{Data Depth: Robust Multivariate Analysis, Computational Geometry and Applications}, American Mathematical Society, 1--16.

Vardi, Y. and Zhang, C.H. (2000). The multivariate L1-median and associated data depth. \emph{Proceedings of the National Academy of Sciences, U.S.A.} \bold{97} 1423--1426.
}
\seealso{
\code{\link{depth.halfspace}} for calculation of the Tukey depth.

\code{\link{depth.Mahalanobis}} for calculation of Mahalanobis depth.

\code{\link{depth.projection}} for calculation of projection depth.

\code{\link{depth.simplicial}} for calculation of simplicial depth.

\code{\link{depth.simplicialVolume}} for calculation of simplicial volume depth.

\code{\link{depth.zonoid}} for calculation of zonoid depth.

\code{\link{depth.potential}} for calculation of data potential.

}
\examples{
# 5-dimensional normal distribution
data <- mvrnorm(1000, rep(0, 5), 
                matrix(c(1, 0, 0, 0, 0, 
                         0, 2, 0, 0, 0, 
                         0, 0, 3, 0, 0, 
                         0, 0, 0, 2, 0, 
                         0, 0, 0, 0, 1),
                nrow = 5))
x <- mvrnorm(10, rep(1, 5), 
             matrix(c(1, 0, 0, 0, 0, 
                      0, 1, 0, 0, 0, 
                      0, 0, 1, 0, 0, 
                      0, 0, 0, 1, 0, 
                      0, 0, 0, 0, 1),
             nrow = 5))
                
depths <- depth.spatial(x, data)
cat("Depths: ", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{depth.space.simplicialVolume}
\alias{depth.space.simplicialVolume}
\title{
Calculate Depth Space using Simplicial Volume Depth
}
\description{
Calculates the representation of the training classes in depth space using simplicial volume depth.
}
\usage{
depth.space.simplicialVolume(data, cardinalities, exact = F, k = 0.05, 
                             mah.estimate = "moment", mah.parMcd = 0.75, seed = 0)
}
\arguments{
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{exact}{
\code{exact=F} (by default) implies the approximative algorithm, considering \code{k} simplices, \code{exact=T} implies the exact algorithm.
}
  \item{k}{
Number (\eqn{k>1}) or portion (if \eqn{0<k<1}) of simplices that are considered if \code{exact=F}. If \eqn{k>1}, then the algorithmic complexity is polynomial in \eqn{d} but is independent of the number of observations in \code{data}, given \eqn{k}. If \eqn{0<k<1}, then the algorithmic complexity is exponential in the number of observations in \code{data}, but the calculation precision stays approximately the same.
}
  \item{mah.estimate}{
A character string specifying affine-invariance adjustment; can be \code{"none"}, \code{"moment"} or \code{"MCD"}, determining whether no affine-invariance adjustemt or moment or Minimum Covariance Determinant (MCD) (see \code{\link{covMcd}}) estimates of the covariance are used. By default \code{"moment"} is used.
}
  \item{mah.parMcd}{
The value of the argument \code{alpha} for the function \code{\link{covMcd}}; is used when \code{mah.estimate =} \code{"MCD"}.
}
  \item{seed}{
The random seed. The default value \code{seed=0} makes no changes.
}
}
\details{
The depth representation is calculated in the same way as in \code{\link{depth.simplicialVolume}}, see References below for more information and details.
}
\value{
Matrix of objects, each object (row) is represented via its depths (columns) w.r.t. each of the classes of the training sample; order of the classes in columns corresponds to the one in the argument \code{cardinalities}.
}
\references{
Oja, H. (1983). Descriptive statistics for multivariate distributions. \emph{Statistics & Probability Letters} \bold{1} 327--332.

Zuo, Y.J. and Serfling, R. (2000). General notions of statistical depth function. \emph{The Annals of Statistics} \bold{28} 461--482.
}
\seealso{
\code{\link{ddalpha.train}} and \code{\link{ddalpha.classify}} for application, \code{\link{depth.simplicialVolume}} for calculation of simplicial depth.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 20 training objects
class1 <- mvrnorm(10, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(10, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
data <- rbind(class1, class2)
# Get depth space using Oja depth
depth.space.simplicialVolume(data, c(10, 10))

data <- getdata("hemophilia")
cardinalities = c(sum(data$gr == "normal"), sum(data$gr == "carrier"))
depth.space.simplicialVolume(data[,1:2], cardinalities)
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{rawfd2dataf}
\alias{rawfd2dataf}
\title{Transform Raw Functional Data to a \code{dataf} Object}
\usage{
rawfd2dataf(X, range)
}
\arguments{
\item{X}{Either a matrix of size \code{n*d}, or an array of dimension \code{n*d*k} of functional values. Here \code{n}
stands for the number of functions, \code{d} is the number of equi-distant points in the domain where the functional
values are evaluated, and if applicable, \code{k} is the dimensionality of the (vector-valued) functional data.}

\item{range}{A vector of size two that represents the endpoints of the common domain of all functions \code{X}.}
}
\value{
A (possibly multivariate) \code{dataf} object corresponding to the functional data \code{X} evaluated at an
equi-distant grid of points.
}
\description{
Constructs a (possibly multivariate) functional data object given by an array of its functional values
evaluated at an equi-distant grid of points, and transforms it into a \code{dataf} object more suitable 
for work in the \code{ddalpha} package.
}
\examples{
## transform a matrix into a functional data set
n = 5
d = 21
X = matrix(rnorm(n*d),ncol=d)
rawfd2dataf(X,range=c(0,1))

## transform an array into a multivariate functional data set
k = 3
X = array(rnorm(n*d*k),dim=c(n,d,k))
rawfd2dataf(X,range=c(-1,1))

}
\seealso{
\code{\link{dataf2rawfd}}

\code{\link{depthf.fd1}}

\code{\link{depthf.fd2}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{functional}

\name{depth.Mahalanobis}
\alias{depth.Mahalanobis}
\title{
Calculate Mahalanobis Depth
}
\description{
Calculates the Mahalanobis depth of points w.r.t. a multivariate data set.
}
\usage{
depth.Mahalanobis(x, data, mah.estimate = "moment", mah.parMcd = 0.75)
}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
  \item{mah.estimate}{ is a character string specifying which estimates to use when calculating the Mahalanobis depth; can be \code{"moment"} or \code{"MCD"}, determining whether traditional moment or Minimum Covariance Determinant (MCD) (see \code{\link{covMcd}}) estimates for mean and covariance are used. By default \code{"moment"} is used.
}
  \item{mah.parMcd}{
is the value of the argument \code{alpha} for the function \code{\link{covMcd}}; is used when \code{mah.estimate =} \code{"MCD"}.
}
}
\details{
Calculates Mahalanobis depth. Mahalanobis depth is based on an outlyingness measure (Zuo & Serfling, 2000), \emph{viz.} the Mahalanobis distance between the given point and the center of the data (Mahalanobis, 1936).

 \emph{Moment estimates} may be used i.e. traditional \emph{mean} and \emph{covariance matrix}, the corresponding depth may be sensitive to
outliers. A more robust depth is obtained with \emph{minimum volume ellipsoid} (MVE) or \emph{minimum
covariance determinant} (MCD) estimators, see Rousseeuw & Leroy (1987) and Lopuhaa &
Rousseeuw (1991).
}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector.
}
\references{
Mahalanobis, P. (1936). On the generalized distance in statistics. \emph{Proceedings of the National
Academy India} \bold{12} 49--55.

Liu, R.Y. (1992). Data depth and multivariate rank tests. In: Dodge, Y. (ed.), \emph{L1-Statistics and Related Methods}, North-Holland (Amsterdam), 279--294.

Lopuhaa, H.P. and Rousseeuw, P.J. (1991). Breakdown points of affine equivariant estimators of multivariate location and covariance matrices. \emph{The Annals of Statistics} \bold{19} 229--248.

Rousseeuw, P.J. and Leroy, A.M. (1987). Robust Regression and Outlier Detection. John Wiley & Sons (New York).

Zuo, Y.J. and Serfling, R. (2000). General notions of statistical depth function. \emph{The Annals of Statistics} \bold{28} 461--482.
}
\seealso{
\code{\link{depth.halfspace}} for calculation of the Tukey depth.

\code{\link{depth.projection}} for calculation of projection depth.

\code{\link{depth.simplicial}} for calculation of simplicial depth.

\code{\link{depth.simplicialVolume}} for calculation of simplicial volume depth.

\code{\link{depth.spatial}} for calculation of spatial depth.

\code{\link{depth.zonoid}} for calculation of zonoid depth.

\code{\link{depth.potential}} for calculation of data potential.
}
\examples{
# 5-dimensional normal distribution
data <- mvrnorm(1000, rep(0, 5), 
                matrix(c(1, 0, 0, 0, 0, 
                         0, 2, 0, 0, 0, 
                         0, 0, 3, 0, 0, 
                         0, 0, 0, 2, 0, 
                         0, 0, 0, 0, 1),
                nrow = 5))
x <- mvrnorm(10, rep(1, 5), 
             matrix(c(1, 0, 0, 0, 0, 
                      0, 1, 0, 0, 0, 
                      0, 0, 1, 0, 0, 
                      0, 0, 0, 1, 0, 
                      0, 0, 0, 0, 1),
             nrow = 5))
                
depths <- depth.Mahalanobis(x, data)
cat("Depths moment: ", depths, "\n")
depths <- depth.Mahalanobis(x, data, mah.estimate = "MCD", mah.parMcd = 0.75)
cat("Depths MCD: ", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{depth.space.}
\alias{depth.space.}

\title{
Calculate Depth Space using the Given Depth
}
\description{
Calculates the representation of the training classes in depth space.

The detailed descriptions are found in the corresponding topics.
}
\usage{
depth.space.(data, cardinalities, notion, ...)

## Mahalanobis depth
# depth.space.Mahalanobis(data, cardinalities, mah.estimate = "moment", mah.parMcd = 0.75)

## projection depth
# depth.space.projection(data, cardinalities, method = "random", num.directions = 1000)

## Tukey depth
# depth.space.halfspace(data, cardinalities, exact, alg, num.directions = 1000)

## spatial depth
# depth.space.spatial(data, cardinalities)

## zonoid depth
# depth.space.zonoid(data, cardinalities)

# Potential
# depth.space.potential(data, cardinalities, pretransform = "NMom", 
#            kernel = "GKernel", kernel.bandwidth = NULL, mah.parMcd = 0.75)
}

\arguments{
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{notion}{
The name of the depth notion (shall also work with \code{\link{Custom Methods}}).
}
  \item{\dots}{
Additional parameters passed to the depth functions.
}
}

\value{
Matrix of objects, each object (row) is represented via its depths (columns) w.r.t. each of the classes of the training sample; order of the classes in columns corresponds to the one in the argument \code{cardinalities}.
}

\seealso{

\code{\link{depth.space.Mahalanobis}}

\code{\link{depth.space.projection}}

\code{\link{depth.space.halfspace}}

\code{\link{depth.space.spatial}}

\code{\link{depth.space.zonoid}}

}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 20 training objects
class1 <- mvrnorm(10, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(10, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
data <- rbind(class1, class2)
# Get depth space using zonoid depth
depth.space.(data, c(10, 10), notion = "zonoid")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{ddalphaf.test}
\alias{ddalphaf.test}
\title{
Test Functional DD-Classifier
}
\description{
Trains functional DD-classifier on the learning sequence of the data and tests it on the testing sequence.
}
\usage{
ddalphaf.test(learn, learnlabels, test, testlabels, disc.type = c("LS", "comp"), ...)
}
\arguments{
  \item{learn}{
list containing lists (functions) of two vectors of equal length, named "args" and "vals": arguments sorted in ascending order and corresponding them values respectively
}
  \item{learnlabels}{
list of output labels of the functional observations
}
  \item{test}{
the testing sequence. Has the same format as \code{learn}
}
  \item{disc.type}{
type of the used discretization scheme. "LS" for \code{\link{ddalphaf.train}}, "comp" for  for \code{\link{compclassf.train}}
}
  \item{testlabels}{
list of output labels of the functinal observations
}
  \item{\dots}{
additional parameters passed to \code{\link{ddalphaf.train}}
}
}

\value{

  \item{error}{
  the part of incorrectly classified data
  }
  \item{correct}{
  the number of correctly classified objects
  }
  \item{incorrect}{
  the number of incorrectly classified objects
  }
  \item{total}{
  the number of classified objects
  }
  \item{ignored}{
  the number of ignored objects (outside the convex hull of the learning data)
  }
  \item{n}{
  the number of objects in the testing sequence
  }
  \item{time}{
  training time
  }

}


\seealso{
\code{\link{ddalphaf.train}} to train the functional DD\eqn{\alpha}-classifier, 
\code{\link{ddalphaf.classify}} for classification using functonal DD\eqn{\alpha}-classifier, 
\code{\link{ddalphaf.getErrorRateCV}} and \code{\link{ddalphaf.getErrorRatePart}} to get error rate of the functional DD-classifier on particular data.
}
\examples{

# load the fdata
df = dataf.growth()

samp = c(35:70)

ddalphaf.test(learn = df$dataf[-samp], learnlabels = df$labels[-samp], 
              test =  df$dataf[samp],  testlabels =  df$labels[samp], 
              adc.args = list(instance = "avr", 
                              numFcn = 2, 
                              numDer = 2))

}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ benchmark }

\name{ddalpha.classify}
\alias{ddalpha.classify}
\alias{predict.ddalpha}
\title{
Classify using DD-Classifier
}
\description{
Classifies data using the DD-classifier and a specified outsider treatment.
}
\usage{
ddalpha.classify(ddalpha, objects, subset, outsider.method = NULL, use.convex = NULL)

\method{predict}{ddalpha}(object, objects, subset, outsider.method = NULL, use.convex = NULL, ...)
}
\arguments{
  \item{ddalpha, object}{
DD\eqn{\alpha}-classifier (obtained by \code{\link{ddalpha.train}}).
}
  \item{objects}{
Matrix containing objects to be classified; each row is one \eqn{d}-dimensional object.
}
  \item{subset}{
an optional vector specifying a subset of observations to be classified.
}
  \item{outsider.method}{
Character string, name of a treatment to be used for outsiders; one of those trained by \code{\link{ddalpha.train}}. If the treatment was specified using the argument \code{outsider.methods} then use the name of the method.
}
  \item{use.convex}{
Logical variable indicating whether outsiders should be determined as the points not contained in any of the convex hulls of the classes from the training sample (\code{TRUE}) or those having zero depth w.r.t. each class from the training sample (\code{FALSE}). For \code{depth =} \code{"zonoid"} both values give the same result. If \code{NULL} the value specified in DD\eqn{\alpha}-classifier (in \code{\link{ddalpha.train}}) is used.
}
  \item{\dots}{
  additional parameters are ignored
}
}
\details{
Only one outsider treatment can be specified.

See Lange, Mosler and Mozharovskyi (2014) for details and additional information.
}
\value{
List containing class labels, or character string "Ignored" for the outsiders if "Ignore" was specified as the outsider treating method.
}
\references{
Dyckerhoff, R., Koshevoy, G., and Mosler, K. (1996). Zonoid data depth: theory and computation. In: Prat A. (ed), \emph{COMPSTAT 1996. Proceedings in computational statistics}, Physica-Verlag (Heidelberg), 235--240.

Lange, T., Mosler, K., and Mozharovskyi, P. (2014). Fast nonparametric classification based on data depth. \emph{Statistical Papers} \bold{55} 49--69.

Li, J., Cuesta-Albertos, J.A., and Liu, R.Y. (2012). DD-classifier: Nonparametric classification procedure based on DD-plot. \emph{Journal of the American Statistical Association} \bold{107} 737--753.

Mozharovskyi, P. (2015). \emph{Contributions to Depth-based Classification and Computation of the Tukey Depth}. Verlag Dr. Kovac (Hamburg).

Mozharovskyi, P., Mosler, K., and Lange, T. (2015). Classifying real-world data with the DD\eqn{\alpha}-procedure. \emph{Advances in Data Analysis and Classification} \bold{9} 287--314.

Vasil'ev, V.I. (2003). The reduction principle in problems of revealing regularities I. \emph{Cybernetics and Systems Analysis} \bold{39} 686--694.
}
\seealso{
\code{\link{ddalpha.train}} to train the DD-classifier.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 200 training objects and 200 to test with
class1 <- mvrnorm(200, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(200, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
trainIndices <- c(1:100)
testIndices <- c(101:200)
propertyVars <- c(1:2)
classVar <- 3
trainData <- rbind(cbind(class1[trainIndices,], rep(1, 100)), 
                   cbind(class2[trainIndices,], rep(2, 100)))
testData <- rbind(cbind(class1[testIndices,], rep(1, 100)), 
                  cbind(class2[testIndices,], rep(2, 100)))
data <- list(train = trainData, test = testData)

# Train the DDalpha-Classifier (zonoid depth, maximum Mahalanobis depth 
# classifier with defaults as outsider treatment)
ddalpha <- ddalpha.train(data$train, 
                         depth = "zonoid", 
                         outsider.methods = "depth.Mahalanobis")
# Get the classification error rate
classes <- ddalpha.classify(data$test[,propertyVars], ddalpha, 
                            outsider.method = "depth.Mahalanobis")
cat("Classification error rate: ", 
    sum(unlist(classes) != data$test[,classVar])/200, ".\n", sep="")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }
\keyword{ classif }

\name{dataf.*}
\alias{dataf.*}
\docType{data}
\title{
Functional Data Sets
}
\description{
The functions generate data sets of functional two-dimensional data of two or more classes.
}
\usage{
# dataf.[name]()

# load the data set by name
# data(list = "name")

# load the data set by name to a variable
# getdata("name")
}
\format{
  The functional data as a data structure.
  \describe{
    \item{\code{dataf}}{
      The functional data as a list of objects. Each object is characterized by two coordinates 
      \describe{
        \item{\code{args}}{The arguments vector}
        \item{\code{vals}}{The values vector}
       }
    }
    \item{\code{labels}}{The classes of the objects}
  }
}
\details{
More details about the datasets in the topics:

\code{\link{dataf.geneexp}}

\code{\link{dataf.growth}}

\code{\link{dataf.medflies}}

\code{\link{dataf.population}}

\code{\link{dataf.population2010}}

\code{\link{dataf.tecator}}

\code{\link{dataf.tecator}}

The following datasets provide simulated data:

\code{\link{dataf.sim.1.CFF07}}

\code{\link{dataf.sim.2.CFF07}}

}
\seealso{
\code{\link{plot.functional}} for building plots of functional data
}
\examples{
## load the Growth dataset
dataf = dataf.growth()

## view the classes
unique(dataf$labels)

## access the 5th point of the 2nd object
dataf$dataf[[2]]$args[5]
dataf$dataf[[2]]$vals[5]

\dontrun{plot.functional(dataf)}
}
\keyword{datasets}
\keyword{functional}

\name{shape.fd.analysis}
\alias{shape.fd.analysis}
\title{Diagnostic Plot for First and Second Order Integrated and Infimal Depths}
\usage{
shape.fd.analysis(datafA, datafB, range = NULL, d = 101, order = 1,
  method = c("halfspace", "simplicial"), approx = 0, title = "",
  nfun = 10, plot = TRUE)
}
\arguments{
\item{datafA}{A single function whose depth is computed, represented by a 
\code{dataf} object of arguments and functional values.}

\item{datafB}{Functional dataset with respect to which the depth of \code{datafA} is computed. 
\code{datafB} is represented by a \code{dataf} object of arguments and functional values. 
\code{n} stands for the number of functions. The grid of observation points for the 
functions in \code{datafA} and \code{datafB} may not be the same.}

\item{range}{The common range of the domain where the functions \code{datafA} and \code{datafB} are observed.
Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
\code{datafA} and \code{datafB}.}

\item{d}{Grid size to which all the functional data are transformed. For depth computation, 
all functional observations are first transformed into vectors of their functional values of length \code{d}
corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
points are reconstructed using linear interpolation, and extrapolation.}

\item{order}{The order of the depth to be used in the plot, for \code{order=1} produces
the plot of univariate marginal depth of \code{A} and \code{nfun} functions from \code{B} 
over the domain of the functions. For \code{order=2} produces the bivariate contour plot 
of the bivariate depths of \code{A} at couples of points from the domain.}

\item{method}{The depth that is used in the diagnostic plot. possible values are \code{halfspace} for 
the halfspace depth, or \code{simplicial} for the simplicial depth.}

\item{approx}{For \code{order=2}, the number of approximations used in the computation of the order extended depth. By default
this is set to \code{0}, meaning that the depth is computed at all possible \code{d^2}
combinations of the points in the domain. When set to a positive integer, \code{approx}
bivariate points are randomly sampled in unit square, and at these points the bivariate depths of the
corresponding functional values are computed.}

\item{title}{The title of the diagnostic plot.}

\item{nfun}{For \code{order=1}, the number of functions from \code{B} whose coordinate-wise
univariate depths of functional values should be displayed with the depth of \code{A}.
The depth of \code{A} is displayed in solid red line, the depths of the functions from \code{B}
in dashed black.}

\item{plot}{Logical: should the function by plotted?}
}
\value{
For \code{order=1} two depth values, and two vectors of pointwise depths:
\itemize{
\item \code{Simpl_FD} the first order integrated depth based on the simplicial depth,
\item \code{Half_FD} the first order integrated depth based on the halfspace depth,
\item \code{Simpl_ID} the first order infimal depth based on the simplicial depth,
\item \code{Half_ID} the first order infimal depth based on the halfspace depth,
   \item \code{PSD} the vector of length \code{d} containing the computed 
   pointwise univariate simplicial depths used for the computation of \code{Simpl_FD} and \code{Simpl_ID},
   \item \code{PHD} the vector of length \code{d} containing the computed 
   pointwise univariate halfspace depths used for the computation of \code{Half_FD} and \code{Half_ID}.
}
   In addition, the first order integrated / infimal depth diagnostic plot of the function \code{A} with respect to
   the random sample given by the functions corresponding to the rows of the matrix \code{B} is produced.

   For \code{order=2} four depth values, and two matrices of pointwise depths:
\itemize{
\item \code{Simpl_FD} the second order integrated depth based on the simplicial depth,
\item \code{Half_FD} the second order integrated depth based on the halfspace depth,
\item \code{Simpl_ID} the second order infimal depth based on the simplicial depth,
\item \code{Half_ID} the second order infimal depth based on the halfspace depth,
   \item \code{PSD} the matrix of size \code{d*d} containing the computed 
   pointwise bivariate simplicial depths used for the computation of \code{Simpl_FD} and \code{Simpl_ID},
   \item \code{PHD} the matrix of size \code{d*d} containing the computed 
   pointwise bivariate halfspace depths used for the computation of \code{Half_FD} and \code{Half_ID}.
}
   In addition, the second order integrated / infimal depth diagnostic plot of the function \code{A} with respect to
   the random sample given by the functions corresponding to the rows of the matrix \code{B} is produced.
}
\description{
Produce the diagnostic plot based on the fist or second order extended integrated / infimal depths.
}
\details{
Plots a diagnostic plot of pointwise univariate (or bivariate) depths for all possible points (or couples of points) from the domain of the 
functional data. From such a plot it is possible to infer into the first order (or second order) properties of a single function \emph{x} with respect 
to the given set of functional data. For \code{order=1}, the integral of the displayed function is the integrated depth of \emph{x}, 
the smallest value of the function is the infimal depth of \emph{x}. 
For \code{order=2}, the bivariate integral of the displayed surface gives the second order extended 
integrated depth of \emph{x}, the infimum of this bivariate function gives the second order infimal depth of \emph{x}. 
For details see Nagy et al. (2016) and \code{\link{depthf.fd1}}.
}
\examples{
datafA = dataf.population()$dataf[1]
dataf = dataf.population()$dataf[2:20]
shape.fd.analysis(datafA,dataf,order=1)
shape.fd.analysis(datafA,dataf,order=2,approx=0)

}
\references{
Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
Depth-based recognition of shape outlying functions. 
\emph{Journal of Computational and Graphical Statistics}, \bold{26} (4), 883--893.
}
\seealso{
\code{\link{depthf.fd1}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{depth}
\keyword{functional}
\keyword{outlier}
\keyword{plot}
\keyword{shape}

\name{depth.L2}
\alias{depth.L2}
\title{
Calculate L2-Depth
}
\description{
Calculates the L2-depth of points w.r.t. a multivariate data set.
}
\usage{
depth.L2(x, data, mah.estimate = "moment", mah.parMcd = 0.75)
}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
  \item{mah.estimate}{ is a character string specifying which estimates to use when calculating sample covariance matrix; can be \code{"none"}, \code{"moment"} or \code{"MCD"}, determining whether traditional moment or Minimum Covariance Determinant (MCD) (see \code{\link{covMcd}}) estimates for mean and covariance are used. By default \code{"moment"} is used. With \code{"none"} the non-affine invariant version of the L2-depth is calculated
}
  \item{mah.parMcd}{
is the value of the argument \code{alpha} for the function \code{\link{covMcd}}; is used when \code{mah.estimate =} \code{"MCD"}.
}
}
\details{
Calculates L2-depth (Mosler, 2013). L2-depth is based on the oultyingness distance calculated as the average L2-distance from (a row of) \code{x} to each point in \code{data}.
}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector.
}
\references{
Mosler, K. (2013). Depth statistics. In: Becker, C., Fried, R. and Kuhnt, S. (eds), \emph{Robustness and Complex Data Structures: Festschrift in Honour of Ursula Gather}, Springer-Verlag (Berlin, Heidelberg), 17--34.
}
\seealso{
\code{\link{depth.halfspace}} for calculation of the Tukey depth.

\code{\link{depth.Mahalanobis}} for calculation of Mahalanobis depth.

\code{\link{depth.projection}} for calculation of projection depth.

\code{\link{depth.qhpeeling}} for calculation of convex hull peeling depth.

\code{\link{depth.simplicial}} for calculation of simplicial depth.

\code{\link{depth.simplicialVolume}} for calculation of simplicial volume depth.

\code{\link{depth.spatial}} for calculation of spatial depth.

\code{\link{depth.potential}} for calculation of data potential.

\code{\link{depth.zonoid}} for calculation of zonoid depth.

}
\examples{
# 5-dimensional normal distribution
data <- mvrnorm(1000, rep(0, 5), 
                matrix(c(1, 0, 0, 0, 0, 
                         0, 2, 0, 0, 0, 
                         0, 0, 3, 0, 0, 
                         0, 0, 0, 2, 0, 
                         0, 0, 0, 0, 1),
                nrow = 5))
x <- mvrnorm(10, rep(1, 5), 
             matrix(c(1, 0, 0, 0, 0, 
                      0, 1, 0, 0, 0, 
                      0, 0, 1, 0, 0, 
                      0, 0, 0, 1, 0, 
                      0, 0, 0, 0, 1),
             nrow = 5))
                
depths <- depth.spatial(x, data)
cat("Depths:", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{depth.space.simplicial}
\alias{depth.space.simplicial}
\title{
Calculate Depth Space using Simplicial Depth
}
\description{
Calculates the representation of the training classes in depth space using simplicial depth.
}
\usage{
depth.space.simplicial(data, cardinalities, exact = F, k = 0.05, seed = 0)
}
\arguments{
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{exact}{
\code{exact=F} (by default) implies the approximative algorithm, considering \code{k} simplices, \code{exact=T} implies the exact algorithm.
}
  \item{k}{
Number (\eqn{k>1}) or portion (if \eqn{0<k<1}) of simplices that are considered if \code{exact=F}. If \eqn{k>1}, then the algorithmic complexity is polynomial in \eqn{d} but is independent of the number of observations in \code{data}, given \eqn{k}. If \eqn{0<k<1}, then the algorithmic complexity is exponential in the number of observations in \code{data}, but the calculation precision stays approximately the same.
}
  \item{seed}{
The random seed. The default value \code{seed=0} makes no changes.
}
}
\details{
The depth representation is calculated in the same way as in \code{\link{depth.simplicial}}, see 'References' for more information and details.
}
\value{
Matrix of objects, each object (row) is represented via its depths (columns) w.r.t. each of the classes of the training sample; order of the classes in columns corresponds to the one in the argument \code{cardinalities}.
}
\references{
Chaudhuri, P. (1996). On a geometric notion of quantiles for multivariate data. \emph{Journal of the American Statistical Association} \bold{91} 862--872.

Liu, R. Y. (1990). On a notion of data depth based on random simplices. \emph{The Annals of Statistics} \bold{18} 405--414.

Rousseeuw, P.J. and Ruts, I. (1996). Algorithm AS 307: Bivariate location depth. \emph{Journal of the Royal Statistical Society. Seriec C (Applied Statistics)} \bold{45} 516--526.
}
\seealso{
\code{\link{ddalpha.train}} and \code{\link{ddalpha.classify}} for application, \code{\link{depth.simplicial}} for calculation of simplicial depth.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 20 training objects
class1 <- mvrnorm(10, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(10, c(1,1), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
data <- rbind(class1, class2)
# Get depth space using simplicial depth
depth.space.simplicial(data, c(10, 10))

data <- getdata("hemophilia")
cardinalities = c(sum(data$gr == "normal"), sum(data$gr == "carrier"))
depth.space.simplicial(data[,1:2], cardinalities)
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\docType{data}
\name{dataf.population}
\alias{dataf.population}
\alias{population}
\title{World Historical Population-by-Country Dataset}
\format{
  The functional data as a data structure.
  \describe{
    \item{\code{dataf}}{
      The functional data as a list of objects. Each object is characterized by two coordinates 
      \describe{
        \item{\code{args}}{\bold{year} - a numeric vector of years 1950-2015 (66 years)}
        \item{\code{vals}}{\bold{population} - a numeric vector of the estimated total population in thousands in 233 countries and regions}
       }
    }
    \item{\code{labels}}{The geographic region of the country: Africa, Asia, Europe, Latin America, North America, Oceania}
    \item{\code{identifier}}{The name of country or region}
  }
}
\source{
United Nations, Department of Economic and Social Affairs,
Population Division,
\url{https://esa.un.org/unpd/wpp/Download/Standard/Population/}, 
file \code{Total population - Both sexes}
}
\usage{
dataf.population()
}
\description{
Historical world population data by countries.
}
\details{
World population data by a country, area or region as of 1 July 
of the year indicated. Figures are presented in thousands.
}
\examples{
## load the Population dataset
dataf = dataf.population()

## view the classes
unique(dataf$labels)

## access the 5th point of the 2nd object
dataf$dataf[[2]]$args[5]
dataf$dataf[[2]]$vals[5]

## plot the data
## Not run:
labels = unlist(dataf$labels)
plot(dataf,
  main = "World population data",
  xlab="Year", ylab="Population (in thousands)"
  )
## End(Not run)

## compute the integrated and infimal depths of the data curves
## with respect to the same set of curves
depthf.fd1(dataf$dataf, dataf$dataf)
}
\seealso{
\code{\link{dataf.population2010}}

\code{\link{dataf.*}} for other functional data sets

\code{\link{plot.functional}} for building plots of functional data
}
\keyword{datasets}
\keyword{functional}

\name{depth.space.Mahalanobis}
\alias{depth.space.Mahalanobis}
\title{
Calculate Depth Space using Mahalanobis Depth
}
\description{
Calculates the representation of the training classes in depth space using Mahalanobis depth.
}
\usage{
depth.space.Mahalanobis(data, cardinalities, mah.estimate = "moment", mah.parMcd = 0.75)
}
\arguments{
  \item{data}{
Matrix containing training sample where each row is a \eqn{d}-dimensional object, and objects of each class are kept together so that the matrix can be thought of as containing blocks of objects representing classes.
}
  \item{cardinalities}{
Numerical vector of cardinalities of each class in \code{data}, each entry corresponds to one class.
}
  \item{mah.estimate}{ is a character string specifying which estimates to use when calculating the Mahalanobis depth; can be \code{"moment"} or \code{"MCD"}, determining whether traditional moment or Minimum Covariance Determinant (MCD) (see \code{\link{covMcd}}) estimates for mean and covariance are used. By default \code{"moment"} is used.
}
  \item{mah.parMcd}{
is the value of the argument \code{alpha} for the function \code{\link{covMcd}}; is used when \code{mah.estimate =} \code{"MCD"}.
}
}
\details{
The depth representation is calculated in the same way as in \code{\link{depth.Mahalanobis}}, see 'References' for more information and details.
}
\value{
Matrix of objects, each object (row) is represented via its depths (columns) w.r.t. each of the classes of the training sample; order of the classes in columns corresponds to the one in the argument \code{cardinalities}.
}
\references{
Mahalanobis, P. (1936). On the generalized distance in statistics. \emph{Proceedings of the National
Academy India} \bold{12} 49--55.

Liu, R.Y. (1992). Data depth and multivariate rank tests. In: Dodge, Y. (ed.), \emph{L1-Statistics and Related Methods}, North-Holland (Amsterdam), 279--294.

Lopuhaa, H.P. and Rousseeuw, P.J. (1991). Breakdown points of affine equivariant estimators of multivariate location and covariance matrices. \emph{The Annals of Statistics} \bold{19} 229--248.

Rousseeuw, P.J. and Leroy, A.M. (1987). Robust Regression and Outlier Detection. John Wiley & Sons (New York).

Zuo, Y.J. and Serfling, R. (2000). General notions of statistical depth function. \emph{The Annals of Statistics} \bold{28} 461--482.
}
\seealso{
\code{\link{ddalpha.train}} and \code{\link{ddalpha.classify}} for application, \code{\link{depth.Mahalanobis}} for calculation of Mahalanobis depth.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 20 training objects
class1 <- mvrnorm(10, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(10, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
data <- rbind(class1, class2)
# Get depth space using Mahalanobis depth
depth.space.Mahalanobis(data, c(10, 10))
depth.space.Mahalanobis(data, c(10, 10), mah.estimate = "MCD", mah.parMcd = 0.75)

data <- getdata("hemophilia")
cardinalities = c(sum(data$gr == "normal"), sum(data$gr == "carrier"))
depth.space.Mahalanobis(data[,1:2], cardinalities)
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{ddalphaf.getErrorRatePart}
\alias{ddalphaf.getErrorRatePart}
\title{
Test Functional DD-Classifier
}
\description{
Performs a benchmark procedure by partitioning the given data. 
On each of \code{times} steps \code{size} observations are removed from the data, the functional DD-classifier is trained on these data and tested on the removed observations.
}
\usage{
ddalphaf.getErrorRatePart(dataf, labels, size = 0.3, times = 10, 
                          disc.type = c("LS", "comp"),  ...)
}
\arguments{
  \item{dataf}{
list containing lists (functions) of two vectors of equal length, named "args" and "vals": arguments sorted in ascending order and corresponding them values respectively
}
  \item{labels}{
list of output labels of the functional observations
}
  \item{size}{
  the excluded sequences size. Either an integer between \eqn{1} and \eqn{n}, or a fraction of data between \eqn{0} and \eqn{1}.
}
  \item{times}{
  the number of times the classifier is trained.
}
  \item{disc.type}{
type of the used discretization scheme. "LS" for \code{\link{ddalphaf.train}}, "comp" for  for \code{\link{compclassf.train}}
}
  \item{\dots}{
additional parameters passed to \code{\link{ddalphaf.train}}
}
}

\value{

  \item{errors}{
  the part of incorrectly classified data (mean)
  }
  \item{errors_sd}{
  the standard deviation of errors
  }
  \item{errors_vec}{
  vector of errors
  }
  \item{time}{
  the mean training time
  }
  \item{time_sd}{
  the standard deviation of training time
  }

}


\seealso{
\code{\link{ddalphaf.train}} to train the functional DD\eqn{\alpha}-classifier, 
\code{\link{ddalphaf.classify}} for classification using functional DD\eqn{\alpha}-classifier, 
\code{\link{ddalphaf.test}} to test the functional DD-classifier on particular learning and testing data,
\code{\link{ddalphaf.getErrorRateCV}} to get error rate of the functional DD-classifier on particular data.
}
\examples{
# load the fdata
df = dataf.growth()

stat <- ddalphaf.getErrorRatePart(dataf = df$dataf, labels = df$labels, 
                          size = 0.3, times = 5,
                          adc.args = list(instance = "avr", 
                                         numFcn = 2, 
                                         numDer = 2))

cat("Classification error rate: ", stat$errors, ".\n", sep = "")


}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ benchmark }

\name{depthf.fd1}
\alias{depthf.fd1}
\title{Univariate Integrated and Infimal Depth for Functional Data}
\usage{
depthf.fd1(datafA, datafB, range = NULL, d = 101, order = 1, approx = 0)
}
\arguments{
\item{datafA}{Functions whose depth is computed, represented by a \code{dataf} object of their arguments
and functional values. \code{m} stands for the number of functions.}

\item{datafB}{Random sample functions with respect to which the depth of \code{datafA} is computed. 
\code{datafB} is represented by a \code{dataf} object of their arguments
 and functional values. \code{n} is the sample size. The grid of observation points for the 
functions \code{datafA} and \code{datafB} may not be the same.}

\item{range}{The common range of the domain where the functions \code{datafA} and \code{datafB} are observed.
Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
\code{datafA} and \code{datafB}.}

\item{d}{Grid size to which all the functional data are transformed. For depth computation, 
all functional observations are first transformed into vectors of their functional values of length \code{d}
corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
points are reconstructed using linear interpolation, and extrapolation, see Nagy et al. (2016).}

\item{order}{The order of the order extended integrated and infimal depths.
By default, this is set to \code{1}, meaning that the usual univariate depths of 
the functional values are computed. For \code{order=2} or \code{3}, the second
and the third order extended integrated and infimal depths are computed, 
respectively.}

\item{approx}{Number of approximations used in the computation of the order extended depth
for \code{order} greater than \code{1}. For \code{order=2}, the default
value is set to \code{0}, meaning that the depth is computed at all possible \code{d^order}
combinations of the points in the domain. For \code{order=3}, 
the default value is set to \code{101}. When \code{approx} is a positive integer, \code{approx}
points are randomly sampled in \code{[0,1]^order} and at these points the \code{order}-variate depths of the
corresponding functional values are computed.}
}
\value{
Four vectors of length \code{m} of depth values are returned:
\itemize{
\item \code{Simpl_FD} the integrated depth based on the simplicial depth,
\item \code{Half_FD} the integrated depth based on the halfspace depth,
\item \code{Simpl_ID} the infimal depth based on the simplicial depth,
\item \code{Half_ID} the infimal depth based on the halfspace depth.
}
In addition, two vectors of length \code{m} of the relative area of smallest depth values is returned:
\itemize{
\item \code{Simpl_IA} the proportions of points at which the depth \code{Simpl_ID} was attained,
\item \code{Half_IA} the proportions of points at which the depth \code{Half_ID} was attained.
}
The values \code{Simpl_IA} and \code{Half_IA} are always in the interval [0,1]. 
They introduce ranking also among functions having the same
infimal depth value - if two functions have the same infimal depth, the one with larger infimal area
\code{IA} is said to be less central. 	
For \code{order=2} and \code{m=1}, two additional matrices of pointwise depths are also returned:
\itemize{
   \item \code{PSD} the matrix of size \code{d*d} containing the computed 
   pointwise bivariate simplicial depths used for the computation of \code{Simpl_FD} and \code{Simpl_ID},
   \item \code{PHD} the matrix of size \code{d*d} containing the computed 
   pointwise bivariate halfspace depths used for the computation of \code{Half_FD} and \code{Half_ID}.
}
For \code{order=3}, only \code{Half_FD} and \code{Half_ID} are provided.
}
\description{
Usual, and order extended integrated and infimal depths for real-valued functional data based on the
halfspace and simplicial depth.
}
\details{
The function returns vectors of sample integrated and infimal depth values.
}
\examples{
datafA = dataf.population()$dataf[1:20]
datafB = dataf.population()$dataf[21:50]
depthf.fd1(datafA,datafB)
depthf.fd1(datafA,datafB,order=2)
depthf.fd1(datafA,datafB,order=3,approx=51)

}
\references{
Nagy, S., Gijbels, I. and Hlubinka, D. (2016). 
Weak convergence of discretely observed functional data with applications. 
\emph{Journal of Multivariate Analysis}, \bold{146}, 46--62.

Nagy, S., Gijbels, I. and Hlubinka, D.  (2017).
Depth-based recognition of shape outlying functions. 
\emph{Journal of Computational and Graphical Statistics}, \bold{26} (4), 883--893.
}
\seealso{
\code{\link{depthf.fd2}}, \code{\link{infimalRank}}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{depth}
\keyword{functional}

\name{depthf.simplicialBand}
\alias{depthf.simplicialBand}

\title{
Calculate Simplicial Band Depth
}
\description{
Calculate the simplicial band depth defined by Lopez-Pintado, Sun, Lin, Genton (2014).
}
\usage{
depthf.simplicialBand(objectsf, dataf, modified = TRUE, J = NULL, 
                      range = NULL, d = 101)
}
\arguments{
  \item{objectsf}{
Functoins for which the depth should be computed; a list containing lists (functions) of two vectors of equal length, named \code{args} and \code{vals}: arguments sorted in ascending order and corresponding them values respectively.
}
  \item{dataf}{
Data sample of functoins w.r.t. which the depth should be computed; structure as for \code{objectsf}.
}
  \item{modified}{
Whether modified simplicial band depth should be computed; logical, \code{TRUE} by default.
}
  \item{J}{
How many functions to consider in each tuple of the U-statistics; integer, \code{d+1} by default.
}
  \item{range}{
The common range of the domain where the functions of objectsf and dataf are observed. Vector of length \code{2} with the left and the right end of the interval. Must contain all arguments given in objectsf and dataf.
}
  \item{d}{
Grid size to which all the functional data are transformed. For depth computation, all functional observations are first transformed into vectors of their functional values of length d corresponding to equi-spaced points in the domain given by the interval range. Functional values in these points are reconstructed using linear interpolation, and extrapolation.
}
}
\value{
A vector of depths of each of \code{objectsf} w.r.t. \code{dataf}.
}
\references{
Lopez-Pintado, Sun, Lin, Genton (2014). Simplicial band depth for multivariate data. \emph{Advances in Data Analysis and Classification} \bold{8}(3) 321--338.
}
\seealso{
\code{\link{depthf.BD}}, \code{\link{depthf.ABD}}.
}
\examples{
datafA = dataf.population()$dataf[1:20]
datafB = dataf.population()$dataf[21:50]

dataf2A = derivatives.est(datafA,deriv=c(0,1))
dataf2B = derivatives.est(datafB,deriv=c(0,1))
depthf.simplicialBand(dataf2A,dataf2B)
}
\keyword{ robust }
\keyword{ functional }
\keyword{ nonparametric }

\name{resetPar}
\alias{resetPar}
\title{
Reset Graphical Parameters
}
\description{
The function returns the default graphical parameters for \code{\link{par}}.
}
\usage{
resetPar()
}
\details{
The returned parameters are used as input parameters for \code{\link{par}}.
}
\value{
The list of graphical parameters described in the 'Graphical Parameters' section of \code{\link{par}}.
}
\examples{

par(mfrow = c(1,2), mar = c(2,2,2,2))
plot(sin, -pi, 2*pi)
plot(cos, -pi, 2*pi)

par(resetPar())
plot(sin, -pi, 2*pi)
plot(cos, -pi, 2*pi)

}

\keyword{ visualization }

\name{ddalpha.getErrorRateCV}
\alias{ddalpha.getErrorRateCV}
\title{
Test DD-Classifier
}
\description{
Performs a cross-validation procedure over the given data. 
On each step every \code{numchunks} observation is removed from the data, the DD-classifier is trained on these data and tested on the removed observations.
}
\usage{
ddalpha.getErrorRateCV (data, numchunks = 10,  ...)
}
\arguments{
  \item{data}{
Matrix containing training sample where each of \eqn{n} rows is one object of the training sample where first \eqn{d} entries are inputs and the last entry is output (class label).
}
  \item{numchunks}{
number of subsets of testing data. Equals to the number of times the classifier is trained.
}
  \item{\dots}{
additional parameters passed to \code{\link{ddalpha.train}}
}
}

\value{

  \item{errors}{
  the part of incorrectly classified data
  }
  \item{time}{
  the mean training time
  }
  \item{time_sd}{
  the standard deviation of training time
  }

}


\seealso{
\code{\link{ddalpha.train}} to train the DD\eqn{\alpha}-classifier, 
\code{\link{ddalpha.classify}} for classification using DD\eqn{\alpha}-classifier, 
\code{\link{ddalpha.test}} to test the DD-classifier on particular learning and testing data,
\code{\link{ddalpha.getErrorRatePart}} to perform a benchmark study of the DD-classifier on particular data.
}
\examples{
# Generate a bivariate normal location-shift classification task
# containing 200 training objects and 200 to test with
class1 <- mvrnorm(150, c(0,0), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
class2 <- mvrnorm(150, c(2,2), 
                  matrix(c(1,1,1,4), nrow = 2, ncol = 2, byrow = TRUE))
propertyVars <- c(1:2)
classVar <- 3
data <- rbind(cbind(class1, rep(1, 150)), cbind(class2, rep(2, 150)))

# Train 1st DDalpha-classifier (default settings) 
# and get the classification error rate
stat <- ddalpha.getErrorRateCV(data, numchunks = 5)
cat("1. Classification error rate (defaults): ", 
    stat$error, ".\n", sep = "")

# Train 2nd DDalpha-classifier (zonoid depth, maximum Mahalanobis 
# depth classifier with defaults as outsider treatment) 
# and get the classification error rate
stat2 <- ddalpha.getErrorRateCV(data, depth = "zonoid", 
                          outsider.methods = "depth.Mahalanobis")
cat("2. Classification error rate (depth.Mahalanobis): ", 
    stat2$error, ".\n", sep = "")


}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ benchmark }

\name{dataf}
\alias{dataf}
\docType{data}
\title{
Converts data from fdata class to the functional class.
}
\description{
\pkg{fda.usc} contains a handy function \code{\link[fda.usc:fdata]{fdata}} that converts varios types of functional data to the \code{fdata} class. 
To use these data in \code{\link{ddalphaf.train}} it must first be converted with \code{dataf}.

The function may be used either to convert a fdata object that contains multiple classes, or to convert multiple fdata objects, each of which contains one class.

Note that \code{fdata$fdata2d = TRUE} is not supported.
}
\usage{

dataf(fdatas, labels) 

}
\arguments{
  \item{fdatas}{ 
  an \code{fdata} object with curves belonging to multiple classes, or a list of \code{fdata} objects, each of which contains curves of the same class
}
  \item{labels}{
  a list of labels of the functional observations.
  If \code{fdatas} is a single \code{fdata} object, the list contains labels for each curve. 
  If \code{fdatas} is a list of \code{fdata} objects, the list labels for each of these \code{fdata} objects. 
}
}
\format{
  The functional data as a data structure (see \code{\link{dataf.*}}).
  \describe{
    \item{\code{dataf}}{
      The functional data as a list of objects. Each object is characterized by two coordinates 
      \describe{
        \item{\code{args}}{The arguments vector}
        \item{\code{vals}}{The values vector}
       }
    }
    \item{\code{labels}}{The classes of the objects}
  }
}
\seealso{
\code{\link{dataf.*}} for the functional data format.


 \code{\link{ddalphaf.train}} to train the functional DD\eqn{\alpha}-classifier
 
 \code{\link{compclassf.train}} to train the functional componentwise classifier


\code{\link{plot.functional}} for building plots of functional data
}
\examples{
\dontrun{
library(fda.usc)
data(phoneme)

# 1. convert a fdata object that contains multiple classes.
#    labels are defined for each curve
converted = dataf(phoneme$learn, phoneme$classlearn)
plot.functional(converted)

# 2. convert multiple fdata objects, each of which contains one class
#    the same label is applied to all curves of each fdata object
converted = dataf(list(phoneme$learn, phoneme$test), c("1 red", "2 blue"))
converted$name = "Phoneme learn (red) and test (blue)"
plot.functional(converted)
}
}
\keyword{fdata}
\keyword{functional}

\name{derivatives.est}
\alias{derivatives.est}
\title{Estimation of the First Two Derivatives for Functional Data}
\usage{
derivatives.est(dataf, range = NULL, d = 101, spar = NULL, deriv = c(0,1))
}
\arguments{
\item{dataf}{Functional dataset, represented by a \code{dataf} object of their arguments
and functional values. \code{m} stands for the number of functions.}

\item{range}{The common range of the domain where the functions \code{dataf} are observed.
Vector of length 2 with the left and the right end of the interval. Must contain all arguments given in 
\code{dataf}.}

\item{d}{Grid size to which all the functional data are transformed. For computation, 
all functional observations are first transformed into vectors of their functional values of length \code{d}
corresponding to equi-spaced points in the domain given by the interval \code{range}. Functional values in these
points are reconstructed using linear interpolation, and extrapolation.}

\item{spar}{If provided, this parameter is passed to functions \code{D1ss} and \code{D2ss} from package \code{sfsmisc}
as the value of the smoothing spline parameter in order to numerically approximate
the derivatives of \code{dataf}.}

\item{deriv}{A vector composed of \code{0}, \code{1}, and \code{2} of the demanded 
functional values / derivatives of the functions in the rows of \code{dataf}.
\code{0} stands for the functional values, \code{1} for the first derivatives, 
\code{2} for the second derivatives.}
}
\value{
A multivariate \code{dataf} object of the functional values and / or the derivatives of \code{dataf}. 
The dimensionality of the vector-valued functional data is \code{nd}. The arguments of the data are all equal to 
an equi-distant grid of \code{d} points in the domain given by \code{range}. \code{nd} is the demanded number 
of derivatives at the output, i.e. the length of the vector \code{deriv}.
}
\description{
Returns the estimated values of derivatives of functional data.
}
\details{
If the input \code{dataf} is a functional random sample of size \code{m}, 
the function returns a \code{dataf} object of \code{nd}-dimensional functional data, where 
in the elements of the vector-valued functional data represent the estimated values of the 
derivatives of \code{dataf}. All derivatives are evaluated at an equi-distant grid of \code{d}
points in the domain given by \code{range}. \code{nd} here stands for \code{1}, \code{2} or \code{3}, 
depending on how many derivatives of \code{dataf} are
requested to be computed. For the estimation, functions \code{D1ss} and \code{D2ss} from the package
\code{sfsmisc} are utilized.
}
\examples{
dataf = dataf.population()$dataf
derivatives.est(dataf,deriv=c(0,1,2))
}
\seealso{
\code{\link[sfsmisc:D2ss]{D1ss}} in package \pkg{sfsmisc}

\code{\link[sfsmisc:D2ss]{D2ss}} in package \pkg{sfsmisc}
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{derivatives}
\keyword{functional}
\keyword{kernel}

\name{depth.halfspace}
\alias{depth.halfspace}
\title{
Calculate Halfspace Depth
}
\description{
Calculates the exact or random Tukey (=halfspace, location) depth  (Tukey, 1975) of points w.r.t. a multivariate data set.
}
\usage{
depth.halfspace(x, data, exact, method, num.directions = 1000, seed = 0)
}
\arguments{
  \item{x}{
Matrix of objects (numerical vector as one object) whose depth is to be calculated; each row contains a \eqn{d}-variate point. Should have the same dimension as \code{data}.
}
  \item{data}{
Matrix of data where each row contains a \eqn{d}-variate point, w.r.t. which the depth is to be calculated.
}
  \item{exact}{
The type of the used method. The default is \code{exact=F}, which leads to approximate computation of the Tukey depth. For \code{exact=F}, \code{method="Sunif.1D"} is used by default. If \code{exact=T}, the Tukey depth is computed exactly, with \code{method="recursive"} by default.}
  \item{method}{
For \code{exact=F}, if \code{method="Sunif.1D"} (by default), the Tukey depth is computed approximately by being minimized over univariate projections (see Details below).

For \code{exact=T}, the Tukey depth is calculated as the minimum over all combinations of \eqn{k} points from \code{data} (see Details below). In this case parameter \code{method} specifies \eqn{k}, with possible values \eqn{1} for \code{method="recursive"} (by default), \eqn{d-2} for \code{method="plane"}, \eqn{d-1} for \code{method="line"}.

The name of the method may be given as well as just parameter \code{exact}, in which case the default method will be used.
}
  \item{num.directions}{
Number of random directions to be generated (for \code{method="Sunif.1D"}). The algorithmic complexity is linear in the number of observations in \code{data}, given the number of directions.
}
  \item{seed}{
The random seed. The default value \code{seed=0} makes no changes (for \code{method="Sunif.1D"}).
}
}
\details{
For \code{exact=F}, if \code{method="Sunif.1D"}, the Tukey depth is computed approximately using the random Tukey depth method proposed by Cuesta-Albertos and Nieto-Reyes (2008). Here the depth is determined as the minimum univariate Tukey depth of the - on lines in several directions - projected data. The directions are distributed uniformly on the \eqn{(d-1)}-sphere; the same direction set is used for all points.

For \code{exact=T}, the Tukey depth is computed exactly as the minimum of the sum of the depths in two orthogonal complementary affine subspaces, which dimensions add to \eqn{d}: one of the subspaces (combinatorial) is the \eqn{k}-dimensional hyperplane through (a point from) \code{x} and \eqn{k} points from \code{data}, another one is its orthogonal complement (see Dyckerhoff and Mozharovskyi, 2016 for the detailed description of the algorithmic framework). The algorithm then minimizes the depth over all combinations of \eqn{k} points, in which the depth in the orthogonal complements is computed using an exact algorithm. In this case, parameter \code{method} specifies the dimensionality \eqn{k} of the combinatorial space. The implemented (reasonable) algorithms (and corresponding names) are: \eqn{k=1} (or \code{method="recursive"}), \eqn{k=d-2} (or \code{method="plane"}), and \eqn{k=d-1} (or \code{method="line"}).
}
\value{
Numerical vector of depths, one for each row in \code{x}; or one depth value if \code{x} is a numerical vector.
}
\references{
Cuesta-Albertos, J.A. and Nieto-Reyes, A. (2008). The random Tukey depth. \emph{Computational Statistics and Data Analysis} \bold{52} 4979--4988.

Dyckerhoff, R. and Mozharovskyi, P. (2016). Exact computation of the halfspace depth. \emph{Computational Statistics and Data Analysis} \bold{98} 19--30.

Rousseeuw, P.J. and Ruts, I. (1996). Algorithm AS 307: Bivariate location depth. \emph{Journal of the Royal Statistical Society. Seriec C (Applied Statistics)} \bold{45} 516--526.

Tukey, J.W. (1974). Mathematics and the picturing of data. In: \emph{Proceeding of the International Congress of Mathematicians}, Vancouver, 523--531.
}
\seealso{
\code{\link{depth.Mahalanobis}} for calculation of Mahalanobis depth.

\code{\link{depth.projection}} for calculation of projection depth.

\code{\link{depth.simplicial}} for calculation of simplicial depth.

\code{\link{depth.simplicialVolume}} for calculation of simplicial volume depth.

\code{\link{depth.spatial}} for calculation of spatial depth.

\code{\link{depth.zonoid}} for calculation of zonoid depth.

\code{\link{depth.potential}} for calculation of data potential.

}
\examples{
# 3-dimensional normal distribution
data <- mvrnorm(200, rep(0, 3), 
                matrix(c(1, 0, 0,
                         0, 2, 0, 
                         0, 0, 1),
                nrow = 3))
x <- mvrnorm(10, rep(1, 3), 
             matrix(c(1, 0, 0,
                      0, 1, 0, 
                      0, 0, 1),
             nrow = 3))
              
# default - random Tukey depth
depths <- depth.halfspace(x, data)
cat("Depths: ", depths, "\n")

# default exact method - "recursive"
depths <- depth.halfspace(x, data, exact = TRUE)
cat("Depths: ", depths, "\n")

# method "line"
depths <- depth.halfspace(x, data, method = "line")
cat("Depths: ", depths, "\n")
}
\keyword{ robust }
\keyword{ multivariate }
\keyword{ nonparametric }

\name{plot.functional}
\alias{plot.functional}
\alias{lines.functional}
\alias{points.functional}
\title{
Plot functions for the Functional Data
}
\description{
Plots the functional data given in the form which is described in the topic \code{\link{dataf.*}}.
}
\usage{
\method{plot}{functional}(x, 
      main = "Functional data", xlab = "args", ylab = "vals", 
      colors = c("red", "blue", "green", "black", "orange", "pink"), ...)
      
\method{lines}{functional}(x, 
      colors = c("red", "blue", "green", "black", "orange", "pink"), ...)
      
\method{points}{functional}(x, 
      colors = c("red", "blue", "green", "black", "orange", "pink"), ...)
}

\arguments{
  \item{x}{
The functional data as in the topic \code{\link{dataf.*}}. 
Note, that the in order to use s3 methods the data must be of class "functional". 
}
  \item{main}{
an overall title for the plot: see \code{\link{title}}
}
  \item{xlab}{
a title for the x axis: see \code{\link{title}}
}
  \item{ylab}{
a title for the y axis: see \code{\link{title}}
}
  \item{colors}{
the colors for the classes of the data. The colors are applied to the classes sorted in alphabetical order. Use the same set of classes to ensure that the same colours are selected in \code{lines} and \code{points} as in \code{plot} (do not remove entire classes). 
}
  \item{\dots}{
  additional parameters
}
}

\seealso{
\code{\link{dataf.*}} for functional data description
}
\examples{

\dontrun{
  ## load the Growth dataset
  dataf = dataf.growth()
  
  labels = unlist(dataf$labels)
  plot(dataf, 
       main = paste("Growth: girls red (", sum(labels == "girl"), "),", 
                    " boys blue (", sum(labels == "boy"), ")", sep=""),
       xlab="Year", ylab="Height, cm",
       colors = c("blue", "red")   # in alphabetical order of class labels   
  )
  
  # plot an observation as a line
  observation = structure(list(dataf = list(dataf$dataf[[1]])), class = "functional")
  lines(observation, colors = "green", lwd = 3)
  
  # plot hight at the age of 14 
  indexAge14 = which(observation$dataf[[1]]$args == 14)
  hightAge14 = observation$dataf[[1]]$vals[indexAge14]
  atAge14 = structure(list(
                      dataf = list(dataf = list(args = 14, vals = hightAge14))
                      ), class = "functional")
  points(atAge14, colors = "yellow", pch = 18)
}

}
\keyword{ visualization }
\keyword{ functional }

\name{FKS}
\alias{FKS}
\title{Fast Kernel Smoothing}
\usage{
FKS(dataf, Tout, kernel = c("uniform", "triangular", "Epanechnikov",
  "biweight", "triweight", "Gaussian"), m = 51, K = 20)
}
\arguments{
\item{dataf}{A set of functional data given by a \code{dataf} object that are to be smoothed.}

\item{Tout}{vector of values in the domain of the functions at which the
resulting smoothed function is evaluated}

\item{kernel}{Kernel used for smoothing. Admissible values are \code{uniform}, 
\code{triangular}, \code{Epanechnikov}, \code{biweight}, \code{triweight} and \code{Gaussian}.
By default, \code{uniform} is used.}

\item{m}{Number of points in the grid for choosing the cross-validated bandwidth.}

\item{K}{Performs \code{K}-fold cross-validation based on randomly shuffled data.}
}
\value{
A \code{dataf} object corresponding to \code{Tout} of smoothed functional values.
}
\description{
Produces a kernel smoothed version of a function based on
the vectors given in the input. Bandwidth is selected using cross-validation.
}
\details{
A vector of the same length as \code{Tout}
corresponding to the values of the
function produced using kernel smoothing, is provided. Bandwidth is selected using the
\code{K}-fold cross-validation of randomly shuffled input values.
}
\examples{
d = 10
T = sort(runif(d))
X = T^2+ rnorm(d,sd=.1)
Tout = seq(0,1,length=101)

plot(T,X)
dataf = list(list(args=T,vals=X))
data.sm = FKS(dataf,Tout,kernel="Epan")
lines(data.sm[[1]]$args,data.sm[[1]]$vals,col=2)

datafs = structure(list(dataf=dataf,labels=1:length(dataf)),class="functional")
plot(datafs)
points(T,X)
data.sms = structure(list(dataf=data.sm,labels=1:length(data.sm)),class="functional")
plot(data.sms)

n = 6
dataf = list()
for(i in 1:n) dataf[[i]] = list(args = T<-sort(runif(d)), vals = T^2 + rnorm(d,sd=.1))
data.sm = FKS(dataf,Tout,kernel="triweight")
data.sms = structure(list(dataf=data.sm,labels=1:length(data.sm)),class="functional")
plot(data.sms)
}
\author{
Stanislav Nagy, \email{nagy@karlin.mff.cuni.cz}
}
\keyword{functional}
\keyword{kernel}
\keyword{smoothing}

\name{depthf.}
\alias{depthf.}
\title{
Calculate Functional Depth
}
\description{
Calculates the depth of functions w.r.t. a functional data set.

The detailed descriptions are found in the corresponding topics.
}
\usage{
depthf.(datafA, datafB, notion, ...)

## Adjusted band depth
# depthf.ABD(datafA, datafB, range = NULL, d = 101, norm = c("C", "L2"), 
# J = 2, K = 1)

## Band depth
# depthf.BD(datafA, datafB, range = NULL, d = 101)

## Univariate integrated and infimal depth
# depthf.fd1(datafA, datafB, range = NULL, d = 101, order = 1, approx = 0)

## Bivariate integrated and infimal depth
# depthf.fd2(datafA, datafB, range = NULL, d = 101)

## h-mode depth
# depthf.hM(datafA, datafB, range = NULL, d = 101, norm = c("C", "L2"),
#  q = 0.2)

## Bivariate h-mode depth
# depthf.hM2(datafA, datafB, range = NULL, d = 101, q = 0.2)

## Half-region depth
# depthf.HR(datafA, datafB, range = NULL, d = 101)

## Univariate random projection depths
# depthf.RP1(datafA, datafB, range = NULL, d = 101, nproj = 50, nproj2 = 5)

# Bivariate random projection depths
# depthf.RP2(datafA, datafB, range = NULL, d = 101, nproj = 51)
}
\arguments{
  \item{datafA}{
Functions whose depth is computed, represented by a \code{dataf} object of their arguments
and functional values.
}
  \item{datafB}{
Random sample functions with respect to which the depth of \code{datafA} is computed. 
\code{datafB} is represented by a \code{dataf} object of their arguments
 and functional values.
}
  \item{notion}{
The name of the depth notion (shall also work with a user-defined depth function named \code{"depthf.<name>"}).
}
  \item{\dots}{
Additional parameters passed to the depth functions.
}
}

\seealso{

\code{\link{depthf.ABD}}

\code{\link{depthf.BD}}

\code{\link{depthf.fd1}}

\code{\link{depthf.fd2}}

\code{\link{depthf.hM}}

\code{\link{depthf.hM2}}

\code{\link{depthf.HR}}

\code{\link{depthf.RP1}}

\code{\link{depthf.RP2}}

}
\value{
Numerical vector of depths, one for each function in \code{datafA}; or one depth value if \code{datafA} is a single function.
}
\examples{
# real data example
datafA = dataf.population()$dataf[1:20]
datafB = dataf.population()$dataf[21:50]
                
depthf.(datafA, datafB, notion = "HR")

dataf2A = derivatives.est(datafA,deriv=c(0,1))
dataf2B = derivatives.est(datafB,deriv=c(0,1))

depthf.(dataf2A, dataf2B, notion = "fd2")
}
\keyword{ robust }
\keyword{ functional }
\keyword{ nonparametric }

\name{dataf.sim.2.CFF07}
\alias{dataf.sim.2.CFF07}
\title{
Model 2 from Cuevas et al. (2007)
}
\description{
Model 2 from Cuevas et al. (2007)

Processes: \cr
X(t) = m_0(t) + e(t), m_0(t) = 30*(1-t)*t^2 + 0.5*abs(sin(20*pi*t)) \cr
Y(t) = an 8-knot spline approximation of X \cr
e(t): Gaussian with mean = 0, cov(X(s), X(t)) = 0.2*exp(-abs(s - t)/0.3)\cr
the processes are discretized at \code{numDiscrets} equally distant points on [0, 1].
}
\usage{
dataf.sim.2.CFF07(numTrain = 100, numTest = 50, numDiscrets = 51, plot = FALSE)
}

\arguments{
  \item{numTrain}{
number of objects in the training sample
  }
  \item{numTest}{
number of objects in the test sample
  }
  \item{numDiscrets}{
number of points for each object
  }
  \item{plot}{
if TRUE the training sample is plotted
  }
}

\format{
  A data strusture containing \code{$learn} and \code{$test} functional data.
  The functional data are given as data structures.
  \describe{
    \item{\code{dataf}}{
      The functional data as a list of objects. Each object is characterized by two coordinates.
      \describe{
        \item{\code{args}}{a numeric vector}
        \item{\code{vals}}{a numeric vector}
       }
    }
    \item{\code{labels}}{The classes of the objects: 0 for X(t), 1 for Y(t)}
  }
}

\source{

Cuevas, A., Febrero, M. and Fraiman, R. (2007). Robust estimation and classification for functional data via projection-based depth notions. Computational Statistics 22 481-496.

}
\seealso{
\code{\link{dataf.*}} for other functional data sets

\code{\link{plot.functional}} for building plots of functional data
}
\examples{
## load the dataset
dataf = dataf.sim.2.CFF07(numTrain = 100, numTest = 50, numDiscrets = 51)
learn = dataf$learn
test = dataf$test

## view the classes
unique(learn$labels)

## access the 5th point of the 2nd object
learn$dataf[[2]]$args[5]
learn$dataf[[2]]$vals[5]

\dontrun{
## plot the data
plot(learn)
plot(test)

## or
dataf = dataf.sim.2.CFF07(numTrain = 100, numTest = 50, numDiscrets = 51, plot = TRUE)
}

}
\keyword{datasets}
\keyword{functional}
